#include <stdio.h>
#include "ImagesTest.h"
#include "Resolution.h"
#include "groundtruth.h"
#include "iniconfig.h"
#include "resize.h"
#include <stdlib.h>
#define snprintf _snprintf

#include "bp-vision/stereo.h"

#define RECT_CENTER 0

#ifndef M_PI
# define M_PI               3.14159265358979323846
#endif

#define __BEGIN__       {

#define __END__         goto exit; exit: ; }



static inline double square(double x) {
	return x*x;
}


int main( int argc, char **argv )
{
	double scale = 0;
	double h = 1.;

	//tali: the projection matrix of the second camera
	CvMat* P = 0;
	CvMat* R = 0;
	CvMat* C = 0;
	CvMat* T = 0;

	//The input images are turned into grey scale images
	CvMat** Images_0 = 0;
	CvMat** Images_t = 0;

	//Ground Truth
	CvMat *True_Z0 = 0;
	CvMat *True_U = 0;
	CvMat *True_V = 0;
	CvMat *True_W = 0;

	FILE *logfile;
	// tali: temporal txt file for debug
	char temp_filename[256];
	FILE *temp_file;


	if (argc == 1) {
		fprintf(stderr,"Usage: %s <primary.ini> <secondary.ini> ...\n", argv[0]);
		fprintf(stderr,
			"Each ini file may contain the following values (default values are indicated).\n"
			"The last file on the command-line overrides the previous ones, and you can put\n"
			"different settings in different files.\n");
		ReadConfig(NULL);
		PrintConfig(stderr);
		exit(3);
	}
	for (int i=1; i<argc; i++)
		ReadConfig(argv[i]);

	logfile = fopen(LogFile, "w");
	PrintConfig(logfile);


	//Reading the images in the sequence
	int view = 0;
	Images_0 = new CvMat*[NumOfViews];
	Images_t = new CvMat*[NumOfViews];
	for( view = 0; view < NumOfViews; view++) {
		if(Img_0_FileName) {
			sprintf(temp_filename, "%s%d.%s", Img_0_FileName, view, ImgPrefix);  
			(CvMat*)(Images_0[view]) = cvLoadImageM((const char *)&temp_filename, CV_LOAD_IMAGE_GRAYSCALE);
			if(!Images_0[view]) { 
				fprintf(stderr,"Error: could not load Images0=%s\n", Img_0_FileName);
				exit(1);
			}
		}
		if(Img_t_FileName) {
			sprintf(temp_filename, "%s%d.%s", Img_t_FileName, view, ImgPrefix);  
			Images_t[view] = cvLoadImageM(temp_filename, CV_LOAD_IMAGE_GRAYSCALE);
			if(!Images_t[view]) { 
				fprintf(stderr,"Error: could not load Imagest=%s\n", Img_t_FileName);
				exit(1);
			}
		}
	}    

	CvSize final_size = cvGetSize(Images_0[0]);
	int N_x = final_size.width;
	int N_y = final_size.height;

	// handle default values
	if (AlphaSF == -1)
		AlphaSF = Alpha;
	if (AlphaW == -1)
		AlphaW = Alpha;
	if (AlphaZ == -1)
		AlphaZ = Alpha;
	if (u0 == -1)
		u0 = N_x/2;
	if (v0 == -1)
		v0 = N_y/2;

	//might consider some more effective implementaion of file reading
	double ref_params[4] = {f0ku,f0kv,u0,v0};
	CvMat** C_Mat = new CvMat*[NumOfViews];
	CvMat** R_Mat = new CvMat*[NumOfViews];
	CvMat** T_Mat = new CvMat*[NumOfViews];

	for( view = 0; view < NumOfViews; view++) {
		sprintf(temp_filename,"%s%d.txt",CMat_FileName,view);
		C_Mat[view] = cvCreateMat(3, 3, CV_64FC1);
		ReadMat(temp_filename,(CvMat *)(C_Mat[view])); 

		sprintf(temp_filename,"%s%d.txt",RMat_FileName,view);
		R_Mat[view] = cvCreateMat(3, 3, CV_64FC1);
		ReadMat(temp_filename,(CvMat *)(R_Mat[view])); 

		sprintf(temp_filename,"%s%d.txt",TMat_FileName,view);
		T_Mat[view] = cvCreateMat(3, 1, CV_64FC1);
		ReadMat(temp_filename,(CvMat *)(T_Mat[view])); 
	}

	//Multiresolution images computation
	if (PyramidFloatingPoint) {
		CvMat *floatim;
		if (Images_0) {
			for(view = 0; view < NumOfViews; view++){
				floatim = cvCreateMat(N_y, N_x, CV_32FC1);
				cvConvert(Images_0[view],floatim);
				cvReleaseMat(&(Images_0[view]));
				(CvMat *)(Images_0[view]) = floatim;
			}
		}

		if (Images_t) {
			floatim = cvCreateMat(N_y, N_x, CV_32FC1);
			for(view = 0; view < NumOfViews; view++){
				floatim = cvCreateMat(N_y, N_x, CV_32FC1);
				cvConvert(Images_t[view],floatim);
				cvReleaseMat(&(Images_t[view]));
				Images_t[view] = floatim;
			}
		}
	}
	double eta_total = 1;
	int level;
	for(level = 1; level < PyramidLevels; level++){
		eta_total *= PyramidEta;
	}
	if (cvCeil(N_x* eta_total) < 4 || cvCeil(N_y* eta_total) < 4) {
		fprintf(stderr, "Too many pyramid levels, resulting images are smaller than 4x4, please try with less pyramid levels\n");
		exit(1);
	}
	// tali: creating a pyramid for each of the input images at each time step
	CvMat ***Images_0_pyramid = 0;
	if(Images_0) {
		Images_0_pyramid = new CvMat**[NumOfViews];
		for(view = 0; view < NumOfViews; view++){
			Images_0_pyramid[view] = new CvMat*[PyramidLevels];
			PyramidBuild(Images_0[view], PyramidEta, PyramidLevels, (CvMat **)(Images_0_pyramid[view]));

		}
		// Il0_pyramid is pre-smoothed
		PyramidSmooth(PyramidLevels, (CvMat**)(Images_0_pyramid[0]));
	}
	CvMat ***Images_t_pyramid = 0;
	if(Images_t) {
		Images_t_pyramid = new CvMat**[NumOfViews];
		for(view = 0; view < NumOfViews; view++){
			Images_t_pyramid[view] = new CvMat*[PyramidLevels];
			PyramidBuild(Images_t[view], PyramidEta, PyramidLevels, Images_t_pyramid[view]);
		}
	}
	CvSize size_start = cvGetSize(Images_0_pyramid[0][PyramidLevels - 1]);
	CvSize size_final = cvGetSize(Images_0_pyramid[0][PyramidLevelFinal - 1]);


	//////////////////////////////////
	//      Z initilization         //
	//////////////////////////////////
	CvMat *Z0_start =  cvCreateMat(size_start.height, size_start.width, CV_32FC1);
	CvMat *Z0_end = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *img = 0;
	CvMat *D0_start = 0;
	CvMat *D0temp = 0;

	if(InitStereo){
		D0_start = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
		D0temp = cvCreateMat(N_y,N_x,CV_32FC1);

		if(InitStereoBP){
			CvMat* Il0 = 0;
			CvMat* Ir0 = 0;

			Il0 = cvLoadImageM(Il0_FileName, CV_LOAD_IMAGE_GRAYSCALE);
			if (!Il0) {
				fprintf(stderr,"Error: could not load Il0=%s\n", Il0_FileName);
				exit(1);
			}
			Ir0 = cvLoadImageM(Ir0_FileName, CV_LOAD_IMAGE_GRAYSCALE);
			if (!Ir0) {
				fprintf(stderr,"Error: could not load Il0=%s\n", Ir0_FileName);
				exit(1);
			}
			char Command[1024];
			char *ST_Il0 = Il0_FileName;
			char *ST_Ir0 = Ir0_FileName;
            

			snprintf(Command, sizeof(Command), "%s %s %s %s", StereoExecutable, ST_Il0, ST_Ir0,  ST_D0_Init_Middlebury_FileName);
			//Belief Propagation for d
			if (logfile) {
				fprintf(logfile, "** Initialization: external command \"%s\"\n", Command);
				fflush(logfile);
			}
			int status = system(Command);
			if (status != 0) {
				fprintf(stderr,"Error: command \"%s\" failed\n", Command);
				exit(1);
			}
			cvReleaseMat(&Il0);
			cvReleaseMat(&Ir0);

			//Taking the result
			img = cvLoadImageM(ST_D0_Init_Middlebury_FileName,0);
			if (!img) {
				fprintf(stderr,"Error: could not load external stereo result %s\n", ST_D0_Init_Middlebury_FileName);
				exit(1);
			}
			D0temp = cvCreateMat(N_y,N_x,CV_32FC1);//cvCreateMat(N_y, N_x, img->type);
			cvZero(D0temp);
			cvConvertScale(img,D0temp,-1*(1/(double)BP_SCALE));
		} else {
			if (ST_D0_initialFileName) {
				img = cvLoadImageM(ST_D0_initialFileName,0);
				if(D_SCALE){
					cvConvertScale(img,D0temp,(-1/(double)D_SCALE));
					cvSubS(D0temp, cvScalar(RECT_CENTER), D0temp);
				}
				else
					cvConvertScale(img,D0temp,-1*(1/(double)BP_SCALE));

				if (logfile) {
					fprintf(logfile, "** Initialization: D is read from file\n");
					fflush(logfile);
				}
			}else {
				fprintf(stderr,"Error: initial disparity file is not supplied!\n");
				exit(1);
			}
		}

	}else {
		if(SF_Z0_start_init_FileName){
			CvMat *img = cvLoadImageM(SF_Z0_start_init_FileName,0);
			CvMat *Z0 = cvCreateMat(N_y,N_x,CV_32FC1);
			cvConvertScale(img,Z0,Z0_init_Scale,0);
			cvResize(Z0, Z0_start,CV_INTER_AREA);
			cvReleaseMat(&img);
			cvReleaseMat(&Z0);
			if (logfile) {
				fprintf(logfile, "** Initialization: Z is read from file\n");
				fflush(logfile);
			}
		} else if(SF_Z0_start_Txt_FileName){
			CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
			ReadMat(SF_Z0_start_Txt_FileName,temp);
			cvResize(temp, Z0_start,CV_INTER_AREA);
			cvReleaseMat(&temp);
		}else {
			if(Z_CONST) {
				cvSet(Z0_start, cvScalarAll(Z_CONST),0);
				if (logfile) {
					fprintf(logfile, "** Initialization: Z to const\n");
					fflush(logfile);
				}
			}
			else {					
				cvSetZero(Z0_start);
				if (logfile) {
					fprintf(logfile, "** Initialization: Z to null\n");
					fflush(logfile);
				}
			}
		}
	}


	// Adjust border values by replication
	if(InitStereo){
		CvMat *D0 = cvCreateMat(N_y,N_x,CV_32FC1);
		cvConvert(D0temp,D0);
		cvReleaseMat(&img);
		cvReleaseMat(&D0temp);
		int x, y;
		for(y = 1; y < N_y-1; y++) {
			cvSetReal2D(D0, y,0, cvGetReal2D(D0, y, 1));
			cvSetReal2D(D0, y,N_x-1, cvGetReal2D(D0, y, N_x-2));
		}
		for(x = 0; x < N_x; x++) {
			cvSetReal2D(D0, 0,x, cvGetReal2D(D0, 1, x));
		}
		for(x = 0; x < N_x; x++) {
			cvSetReal2D(D0, N_y-1,x, cvGetReal2D(D0, N_y-2,x));
		}
		cvResize(D0, D0_start, CV_INTER_AREA);
		//cvConvertScale(D0_start, D0_start, (double)size_stereo_start.width / N_x);

		if(ST_initD0_FileName)
			cvSaveImage(ST_initD0_FileName,D0_start);
		// converting the disparity to Z
		CvMat *Z0 = cvCreateMat(N_y,N_x,CV_32FC1);
		cvZero(Z0);
		scale = Tx*f0ku;
		ConvertDZ(D0,Z0,scale,maxZ);
		if(ST_initZ0_FileName)
			cvSaveImage(ST_initZ0_FileName,Z0);
		cvResize(Z0, Z0_start, CV_INTER_AREA);
		cvReleaseMat(&D0);
		cvReleaseMat(&D0_start);
	}

	//////////////////////////////////////
	//      Scene flow computation      //
	//////////////////////////////////////
	CvMat *U_start = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
	CvMat *V_start = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
	CvMat *W_start = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
	CvMat *U_end = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *V_end = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *W_end = cvCreateMat(N_y, N_x, CV_32FC1);

	cvSet(W_start, cvScalarAll(0),0);
	cvSet(U_start, cvScalarAll(0),0);
	cvSet(V_start, cvScalarAll(0),0);

	if(SF_U_start_init_FileName){
		CvMat *img = cvLoadImageM(SF_U_start_init_FileName,0);
		CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
		cvConvertScale(img,temp,-0.25,0);
		cvSubS(temp, cvScalar(-50), temp);
		cvResize(temp, U_start,CV_INTER_AREA);
		cvReleaseMat(&img);
		cvReleaseMat(&temp);
		if (logfile) {
			fprintf(logfile, "** Initialization: U is read from file\n");
			fflush(logfile);
		}
	}else if(SF_U_start_Txt_FileName){
		CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
		ReadMat(SF_U_start_Txt_FileName,temp);
		cvResize(temp, U_start,CV_INTER_AREA);
		cvReleaseMat(&temp);
	}else
	{
		if(U_CONST){
			cvSet(U_start, cvScalarAll(U_CONST),0);
			if (logfile) {
				fprintf(logfile, "** Initialization: U to %d\n",U_CONST);
				fflush(logfile);
			}
		}else{
			if (logfile) {
				fprintf(logfile, "** Initialization: U to zero\n");
				fflush(logfile);
			}
		}
	}
	if(SF_V_start_init_FileName){
		CvMat *img = cvLoadImageM(SF_V_start_init_FileName,0);
		CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
		cvConvertScale(img,temp,-0.25,0);
		cvSubS(temp, cvScalar(-50), temp);
		cvResize(temp, V_start,CV_INTER_AREA);
		cvReleaseMat(&img);
		cvReleaseMat(&temp);
	}else if(SF_V_start_Txt_FileName){
		CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
		ReadMat(SF_V_start_Txt_FileName,temp);
		cvResize(temp, V_start,CV_INTER_AREA);
		cvReleaseMat(&temp);
	}else {
		if(V_CONST){
			cvSet(V_start, cvScalarAll(V_CONST),0);
			if (logfile) {
				fprintf(logfile, "** Initialization: V to %d\n",V_CONST);
				fflush(logfile);
			}
		}else{
			if (logfile) {
				fprintf(logfile, "** Initialization: V to zero\n");
				fflush(logfile);
			}
		}
	}
	if(SF_W_start_init_FileName){
		CvMat *img = cvLoadImageM(SF_W_start_init_FileName,0);
		CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
		cvConvertScale(img,temp,-0.25,0);
		cvSubS(temp, cvScalar(-50), temp);
		cvResize(temp, W_start,CV_INTER_AREA);
		cvReleaseMat(&img);
		cvReleaseMat(&temp);
	}else if(SF_W_start_Txt_FileName){
		CvMat *temp = cvCreateMat(N_y,N_x,CV_32FC1);
		ReadMat(SF_W_start_Txt_FileName,temp);
		cvResize(temp, W_start,CV_INTER_AREA);
		cvReleaseMat(&temp);
	}else {
		if(W_CONST){
			cvSet(W_start, cvScalarAll(W_CONST),0);
			if (logfile) {
				fprintf(logfile, "** Initialization: W to %d\n",W_CONST);
				fflush(logfile);
			}
		}else{
			if (logfile) {
				fprintf(logfile, "** Initialization: W to zero\n");
				fflush(logfile);
			}
		}
	}

	sprintf(temp_filename, "%s/Z0_start.txt", TxtResultsDir);
	temp_file = fopen(temp_filename,"wb");
	WriteRes(temp_file,Z0_start);
	fclose(temp_file);
	sprintf(temp_filename, "%s/W_start.txt", TxtResultsDir);
	temp_file = fopen(temp_filename,"wb");
	WriteRes(temp_file,W_start);
	fclose(temp_file);
	sprintf(temp_filename, "%s/U_start.txt", TxtResultsDir);
	temp_file = fopen(temp_filename,"wb");
	WriteRes(temp_file,U_start);
	fclose(temp_file);
	sprintf(temp_filename, "%s/V_start.txt", TxtResultsDir);
	temp_file = fopen(temp_filename,"wb");
	WriteRes(temp_file,V_start);
	fclose(temp_file);

	U_end = cvCreateMat(size_final.height, size_final.width, CV_32FC1);
	V_end = cvCreateMat(size_final.height, size_final.width, CV_32FC1);
	W_end = cvCreateMat(size_final.height, size_final.width, CV_32FC1);
	Z0_end = cvCreateMat(size_final.height, size_final.width, CV_32FC1);

	SceneFlow_MVZ(logfile, 
		PyramidEta, 
		PyramidLevels,
		Images_0_pyramid,
		Images_t_pyramid,
		PyramidLevels,
		U_start,
		V_start,
		W_start,
		Z0_start,
		PyramidLevelFinal,
		U_end,
		V_end,
		W_end,
		Z0_end, 
		h, 
		EpsilonOuter,
		EpsilonInner,
		EpsilonSOR,
		MaxIterOuter,
		MaxIterInner,
		MaxIterSOR,
		OmegaSOR,
		AlphaSF,
		AlphaW,
		AlphaZ,
		MuZ,
		MuU,
		MuV,
		MuW,
		f0ku,
		f0kv, 
		u0, 
		v0,
		R_Mat,
		T_Mat,
		C_Mat,
		SF_Steps);

	sprintf(temp_filename, "%s/Z0_final.txt", TxtResultsDir);
	FILE *f = fopen(temp_filename,"w");
	WriteRes(f, Z0_end);
	fclose(f);
	sprintf(temp_filename, "%s/W_final.txt", TxtResultsDir);
	f = fopen(temp_filename,"w");
	WriteRes(f, W_end);
	fclose(f);
	sprintf(temp_filename, "%s/U_final.txt", TxtResultsDir);
	f = fopen(temp_filename,"w");
	WriteRes(f, U_end);
	fclose(f);
	sprintf(temp_filename, "%s/V_final.txt", TxtResultsDir);
	f = fopen(temp_filename,"w");
	WriteRes(f, V_end);
	fclose(f);

	if (logfile) {
		fprintf(logfile, "** Saving results\n");
		fflush(logfile);
	}
	if (SF_U_Final_FileName)
		SaveResults(SF_U_Final_FileName, U_end);

	if (logfile) {
		fprintf(logfile, "** Saving results : U\n");
		fflush(logfile);
	}
	if (SF_V_Final_FileName)
		SaveResults(SF_V_Final_FileName, V_end);

	if (logfile) {
		fprintf(logfile, "** Saving results : V\n");
		fflush(logfile);
	}
	if (SF_W_Final_FileName)
		SaveResults(SF_W_Final_FileName, W_end);

	if (logfile) {
		fprintf(logfile, "** Saving results : W\n");
		fflush(logfile);
	}



	///////////////////////////////
	//        Ground truth       //
	///////////////////////////////
	if (GT_Z0_FileName) {
		CvMat *img = cvLoadImageM(GT_Z0_FileName,0);
		if (!img) {
			fprintf(stderr,"Error: could not load Z0 ground truth %s\n", GT_Z0_FileName);
			exit(1);
		}
		True_Z0 = cvCreateMat(N_y, N_x, CV_32FC1);
		cvConvertScale(img, True_Z0, GT_Z0_Scale);
		sprintf(temp_filename, "%s/Z0_GT.txt", TxtResultsDir);
		temp_file = fopen(temp_filename,"wb");
		WriteRes(temp_file,True_Z0);
		fclose(temp_file);
		cvReleaseMat(&img);
	}
	if (GT_U_FileName) {
		CvMat *img = cvLoadImageM(GT_U_FileName,0);
		if (!img) {
			fprintf(stderr,"Error: could not load U ground truth %s\n", GT_U_FileName);
			exit(1);
		}
		True_U = cvCreateMat(N_y, N_x, CV_32FC1);
		cvConvertScale(img, True_U, GT_U_Scale);
		sprintf(temp_filename, "%s/U_GT.txt", TxtResultsDir);
		temp_file = fopen(temp_filename,"wb");
		WriteRes(temp_file,True_U);
		fclose(temp_file);
		cvReleaseMat(&img);
	}
	if (GT_V_FileName) {
		CvMat *img = cvLoadImageM(GT_V_FileName,0);
		if (!img) {
			fprintf(stderr,"Error: could not load U ground truth %s\n", GT_V_FileName);
			exit(1);
		}
		True_V = cvCreateMat(N_y, N_x, CV_32FC1);
		cvConvertScale(img, True_V, GT_V_Scale);
		sprintf(temp_filename, "%s/V_GT.txt", TxtResultsDir);
		temp_file = fopen(temp_filename,"wb");
		WriteRes(temp_file,True_V);
		fclose(temp_file);
		cvReleaseMat(&img);
	}

	if (GT_W_FileName) {
		CvMat *img = cvLoadImageM(GT_W_FileName,0);
		if (!img) {
			fprintf(stderr,"Error: could not load V ground truth %s\n", GT_W_FileName);
			exit(1);
		}
		True_W = cvCreateMat(N_y, N_x, CV_32FC1);
		cvConvertScale(img, True_W, GT_W_Scale);
		sprintf(temp_filename, "%s/W_GT.txt", TxtResultsDir);
		temp_file = fopen(temp_filename,"wb");
		WriteRes(temp_file,True_W);
		fclose(temp_file);
		cvReleaseMat(&img);
	}




	if (True_Z0)
		cvReleaseMat(&True_Z0);

	if (Images_0_pyramid) {
		for(view = 0; view < NumOfViews; view++){
			for(level = 0; level < PyramidLevels; level++){
				cvReleaseMat(&Images_0_pyramid[view][level]);
			}
		}
		delete [] Images_0_pyramid;
	}

	if (Images_t_pyramid) {
		for(view = 0; view < NumOfViews; view++){
			for(level = 0; level < PyramidLevels; level++){
				cvReleaseMat(&Images_t_pyramid[view][level]);
			}
		}
		delete [] Images_t_pyramid;
	}

	if (P)
		cvReleaseMat(&P);        
	if (R)
		cvReleaseMat(&R);        
	if (T)
		cvReleaseMat(&T);        
	if (C)
		cvReleaseMat(&C);        
	
	printf("Scene Flow finished!");

	return(0);

	
}


