#ifndef RESOLUTION_H
#define RESOLUTION_H
#include <math.h>
#include <stdlib.h>
#include <cv.h>
#include <cvaux.h>
#include <cvtypes.h>
#include <highgui.h>
#include <stdio.h>
#include <cxcore.h>


void SaveImage(const char *filename, const CvArr* arr);
void SaveResults(const char *filename, const CvArr* Results, int scale CV_DEFAULT(1), double min CV_DEFAULT(0.), double max CV_DEFAULT(-1.));

void PyramidBuild(CvMat *I,
                  double eta,
                  int pyramid_levels,
                  CvMat **I_pyramid);

void PyramidSmooth(int pyramid_levels,
                   CvMat **I_pyramid);

void PyramidCopy(int pyramid_levels,
                 const CvMat * const * const In_pyramid,
                 CvMat ** Out_pyramid);


#endif

void ComputeXY(CvMat* X,CvMat* Y,CvMat* Z0,
			   CvMat* MapIdentityX,CvMat* MapIdentityY,
			   int N_x,int N_y,
			   const double* params,
			   double scale);

void ConvertDZ(CvMat* D,CvMat* Z,double scale,double maxZ);

void WriteRes(FILE* filename,CvMat* M);

void ReadMat(char* filename,CvMat* M);

void WriteRes8u(FILE* filename,const CvMat* M);

void ComputeProjMat(CvMat *P, CvMat *R, CvMat *T, CvMat *C,double eta_curr);

void ComputeProjMaps(CvMat* X,CvMat* Y,
					 CvMat* Z,CvMat *R, 
					 CvMat* T, CvMat* C,
					 CvMat* Proj_MapX_Ir0,
					 CvMat* Proj_MapY_Ir0,
					 double eta_curr);
void ComputeCOP(CvMat **COP, CvMat **R_Mat, CvMat **T_Mat);
void ComputeOcclusion(CvMat *MapX, CvMat *MapY, CvMat *X, CvMat *Y, CvMat *Z0,  CvMat *disactivated_pixels_map, CvMat *cop CV_DEFAULT(NULL));
void ProjectPoints2(   const CvMat* object_points, const CvMat* rotation_vector,
                              const CvMat* translation_vector, const CvMat* camera_matrix,
                              const CvMat* distortion_coeffs, CvMat* image_points,
							  double  scale CV_DEFAULT(1),
                              CvMat* dpdrot CV_DEFAULT(NULL), CvMat* dpdt CV_DEFAULT(NULL),
                              CvMat* dpdf CV_DEFAULT(NULL), CvMat* dpdc CV_DEFAULT(NULL),
                              CvMat* dpddist CV_DEFAULT(NULL),
                              double aspect_ratio CV_DEFAULT(0));

void GradientProjCoor(CvMat* X, CvMat* Y, CvMat* Z, CvMat *X_driv, CvMat *Y_driv,CvMat* Z_driv, 
					  CvMat* M, int grad_y, CvMat* Proj_driv CV_DEFAULT(NULL));



void SceneFlow_MVZ(FILE* logfile, //the log file
	       const double eta, //reduction factor
	       int pyramid_levels, // total number of pyramid levels
	       const CvMat * const * const * const Images_0_pyramid, //time t Images
	       const CvMat * const * const * const Images_t_pyramid, //time t+1 Images
           int pyramid_level_start, // starting level
	       CvMat* U_start, //horizontal left scene flow initialization
	       CvMat* V_start, //vertical scene flow initialization
		   CvMat* W_start, //depth scene flow initialization
           CvMat* Z0_start, // Z(x,y) initialization
           int pyramid_level_end, // end pyramid level
	       CvMat* U_end, //final U
	       CvMat* V_end, //final V
		   CvMat* W_end, //final V
		   CvMat* Z0_end, //t Z 
		   const double h, //grid step for finite elements
	       const double epsilon_outer, //stopping criteria for outer iterations 
	       const double epsilon_inner, //stopping criteria for inner iterations
	       const double epsilon_SOR, //stopping criteria for the linear system resolution
	       const int max_iter_outer, //maximal number of outer iterations
	       const int max_iter_inner, //maximal number of inner iterations
	       const int max_iter_SOR, //maximal number of linear system resolution iterations
	       const double omega_SOR, //SOR method weight
           double alpha, // smooth term weight
		   double alpha_w, // smooth term weight
		   double alpha_z, // smooth term weight
		   const double mu_z, // smooth term: disparity weight
		   const double mu_u, // smooth term: disparity weight
		   const double mu_v, // smooth term: disparity weight
           const double mu_w, // smooth term: disparity weight
		   const double f0ku, // x focal length of the reference camera
		   const double f0kv, // y focal length of the refernce camera
		   const double u0, 
		   const double v0,
		   CvMat** R_Mat, //the rotation matrixes
		   CvMat** T_Mat, //the translation matrixes
		   CvMat** C_Mat, //the calibration matrixes
		   const int methode);