#include <math.h>
#include <assert.h>
#include "Resolution.h"
#include "iniconfig.h"
#include "resize.h"
#include <stdlib.h>
#define snprintf _snprintf

#define MAX_HUE 150

#define DISACTIVATION_THRESHOLD 1.7
#define DISACTIVATION_PRECENT  0.2

#define WARP_DETERMINANT_THRESHOLD 0.01// should be >= 0. and < 1.

#define EPSILON_PSI_SMOOTH 0.001           // epsilon for the psi function
#define EPSILON_PSI_SMOOTH_W 0.001           // epsilon for the psi function
#define EPSILON_PSI_DATA 1.               // epsilon for the psi function

#define MAX_INCR  20                  // maximum increment between two iterates in pixels
#define MAX_INCR_SF 5
#define MAX_NUM_OF_VIEWS 9


//tali 
#define ENABLE_ALPHA 1

#define ASSERT  0
//#define DEBUG  //uncomment if you want to run in DEBUG mode
#define __BEGIN__       {

#define __END__         goto exit; exit: ; }

// Pixels that go beyond this border (in pixels) are disactivated. Must be at
// least 3, since gradient estimation for warped images is
// wrong 2 pixels from the border, and warping is wrong 1 pixel from the border
// but 4 (more than 2*sqrt(2)+1) is safer in
// case the transformed image border is at 45deg
#define DISACTIVATE_BORDER 4

#ifndef M_PI
# define M_PI               3.14159265358979323846
#endif


static inline double square(double x) {
    return x*x;
}

static inline double fmax(double a, double b)
{
	return (a > b ? a : b);
}


CV_INLINE  unsigned char cvmGet8u( const CvMat* mat, int row, int col )
{
    return ((unsigned char*)(mat->data.ptr + (size_t)mat->step*row))[col];
}

CV_INLINE  static void  cvmSet8u( CvMat* mat, int row, int col, unsigned char value )
{
    ((unsigned char*)(mat->data.ptr + (size_t)mat->step*row))[col] = value;
}

// used to mark border pixels as disactivated (although they are not used anyway)
CV_INLINE  static void  cvmSet8u_borders( CvMat* mat, unsigned char value )
{
    int rows  = mat->rows;
    int cols = mat->cols;
    int i, j;

    // top border
    for(j = 0; j < cols; j++)
        cvmSet8u(mat,0,j,value);
    // bottom border
    for(j = 0; j < cols; j++)
        cvmSet8u(mat,rows-1,j,value);
    // left border
    for(i = 1; i < rows-1; i++)
        cvmSet8u(mat,i,0,value);
    // right border
    for(i = 1; i < rows-1; i++)
        cvmSet8u(mat,i,cols-1,value);
}

CV_INLINE  static void  cvmSet8u_col_borders( CvMat* mat, unsigned char value, int left, int col )
{
    int rows  = mat->rows;
    int cols = mat->cols;
    int i, j;

	if(left){
    // left border
		for(j=0; j < col; j++){
			for(i = 0; i < rows; i++)
				cvmSet8u(mat,i,j,value);
		}
	}else
	{
		for(j=cols; j > col; j--){
			for(i = 0; i < rows; i++)
				cvmSet8u(mat,i,j,value);
		}
	}
}



static inline double sign(double a)
{
    return (a > 0 ? 1.0 : -1.0);
}
static inline float sign(float a)
{
    return (a > 0.0 ? 1.0f : -1.0f);
}

/***************************************************Horizontal linear Interpolation**************************************************/
static inline void Interpolate_X(const CvMat* data, //data to be interpolated
				 const int x, //pixel x
				 const int y, //pixel y
				 const double alpha, //Interpolation coefficient
				 double *Interpolate, //Result
				 const int Nx){
    if(x < Nx - 1)
	*Interpolate = alpha * cvmGet(data, y, x + 1) + (1 - alpha) * cvmGet(data, y, x);
    else
	*Interpolate = alpha * cvmGet(data, y, x) + (1 - alpha) * cvmGet(data, y, x - 1); 
}

static inline void SmoothImage(CvArr *img) {
    cvSmooth(img,img, CV_GAUSSIAN, 3,3);
}


static inline double& Crop(double &val, double threshold) {
#if 1
    if (val > threshold)
        val = threshold;
    else if (val < -threshold)
        val = -threshold;
#endif
    return val;
}

static inline void CropArr(CvArr *arr, double threshold) {
#if 1
    cvMinS(arr, threshold, arr);
    cvMaxS(arr,-threshold, arr);
#endif
}



/* lightweight convolution with 3x3 kernel */
static void GradientFast_32f( const CvMat* srcmat, CvMat* dstxmat, CvMat* dstymat )
{
    int  x, y;

    int src_type, dstx_type, dsty_type;
    src_type = CV_MAT_TYPE( srcmat->type );
    dstx_type = CV_MAT_TYPE( dstxmat->type );
    dsty_type = CV_MAT_TYPE( dstymat->type );
  
	int src_step = srcmat->step;
    int dst_step = dstxmat->step;
  
    float* src = (float*)srcmat->data.ptr;
    float* dstx = (float*)dstxmat->data.ptr;
    float* dsty = (float*)dstymat->data.ptr;

    src_step /= sizeof(src[0]);
    dst_step /= sizeof(dstx[0]);

    for( y = 0; y < srcmat->height; y++, src += src_step, dstx += dst_step, dsty += dst_step)
    {
        float* srcprevline;
        float* srcnextline;
        float muly = 0.5;
        if (y == 0) {
            srcprevline = src;
            muly = 1;
        }
        else {
            srcprevline = src - src_step;
        }
        if (y == srcmat->height - 1) {
            srcnextline = src;
            muly = 1;
        }
        else {
            srcnextline = src + src_step;
        }
        for( x = 0; x < srcmat->width; x++ ) {
            float srcvalprevx;
            float srcvalnextx;
            float mulx = 0.5;
            if (x == 0) {
                srcvalprevx = src[x];
                mulx = 1;
            }
            else {
                srcvalprevx = src[x-1];
            }
            if (x == srcmat->width - 1) {
                srcvalnextx = src[x];
                mulx = 1;
            }
            else {
                srcvalnextx = src[x+1];
            }
            dstx[x] = (srcvalnextx-srcvalprevx)*mulx;
            dsty[x] = (srcnextline[x]-srcprevline[x])*muly;
        }
    }
}

/* lightweight convolution with 3x3 kernel */
static void ZBufMap_32f( const CvMat* src, CvMat* dst, const CvMat* mapx, const CvMat* mapy, float fillval )
{
    int  x, y;

    int src_type, dst_type, mapx_type, mapy_type;
    src_type = CV_MAT_TYPE( src->type );
    dst_type = CV_MAT_TYPE( dst->type );
    mapx_type = CV_MAT_TYPE( mapx->type );
    mapy_type = CV_MAT_TYPE( mapy->type );

	cvSet(dst, cvScalar(fillval));
    for( y = 0; y < src->height; y++)
    {
        for( x = 0; x < src->width; x++ ) {
            const double dstx = cvmGet(mapx,y,x);
            const double dsty = cvmGet(mapy,y,x);
            const double srcval = cvmGet(src,y,x);
            
            if (dstx < 0 || dstx > (dst->width-1) ||
                dsty < 0 || dsty > (dst->height-1))
                continue;
                
            for(int i=cvFloor(dsty); i <= cvCeil(dsty); i++) {
                for(int j=cvFloor(dstx); j <= cvCeil(dstx); j++) {
                    const double dstval = cvmGet(dst,i,j);
                    if (srcval < dstval)
                        cvmSet(dst,i,j,srcval);
                }
            }
        }
    }
}



/************** creates a distances map. for each of the pixels in the reference image it computes the distance of P from COP*****************/
static void Dist_map(CvMat* dst,  CvMat *X, CvMat *Y, CvMat *Z,CvMat *COP)
{
    int  x, y;
	double dist = 0.0;
	int dst_type;
    dst_type = CV_MAT_TYPE( dst->type );
 
	double copx = cvmGet(COP, 0, 0);
	double copy = cvmGet(COP, 1, 0);
	double copz = cvmGet(COP, 2, 0);


    for( y = 0; y < dst->height; y++)
    {
        for( x = 0; x < dst->width; x++ ) {
            const double X_val = cvmGet(X,y,x);
            const double Y_val = cvmGet(Y,y,x);
			const double Z_val = cvmGet(Z,y,x);
			dist  = sqrt((X_val - copx)*(X_val - copx) + (Y_val - copy)*(Y_val - copy) + (Z_val - copz)*(Z_val - copz));
            cvmSet(dst,y,x,dist);
        }
    }
}


/***************************Saving results in false color images********************/
void SaveImage(const char *filename, const CvArr* arr)
{
    int type = cvGetElemType(arr);
    if(type == CV_8UC1 || type == CV_8UC3) {
        cvSaveImage(filename, arr);
    }
    else {
        CvMat *charim = cvCreateMat(cvGetDimSize(arr,0), cvGetDimSize(arr,1), CV_8UC1);
        cvConvert(arr, charim);
        cvSaveImage(filename, charim);
        cvReleaseMat(&charim);
    }
}

/***************************Saving results in false color images********************/
void SaveResults(const char *filename, const CvArr* arr, int addscale, double min, double max)
{  
    double values[MAX_HUE+1];
    int i;
    char chaine[256];
    CvFont police;
    int ydiv = 1;
    CvSize size;

    if (CV_IS_SPARSE_MAT( arr ) && ((CvSparseMat*)arr)->dims == 2) {
        size.height = ((CvSparseMat*)arr)->size[0];
        size.width = ((CvSparseMat*)arr)->size[1];
    }
    else {
        size = cvGetSize(arr);
    }
    
    CvMat *Temp = cvCreateMat(size.height, size.width, CV_8UC1);

    if (min == 0. && max == -1.) {
        cvMinMaxLoc(arr, &min, &max);
    }
    //Data scaling
    if(max <= min) {
	cvSetZero(Temp);
    }
    else {
        double scale = (double)MAX_HUE/(max-min);
        double shift =  -(double)MAX_HUE*min/(max - min);
        
        if (!CV_IS_SPARSE_MAT( arr )) {
            cvConvertScale(arr, Temp, scale, shift);
        }
        else {
            cvSetZero(Temp);
            CvSparseMatIterator mat_iterator;
            const CvSparseMat *array = (CvSparseMat*) arr;
            CvSparseNode* node = cvInitSparseMatIterator( array, &mat_iterator );
            for( ; node != 0; node = cvGetNextSparseNode( &mat_iterator ))
            {
                int* idx = CV_NODE_IDX( array, node ); /* get pointer to the element indices */
                double val;
                switch(array->type) {
                    case CV_8UC1:
                        val = *(unsigned char*)CV_NODE_VAL( array, node );
                        break;
                    case CV_32FC1:
                        val = *(float*)CV_NODE_VAL( array, node );
                        break;
                    case CV_64FC1:
                        val = *(double*)CV_NODE_VAL( array, node );
                        break;
                    default:
                        val = 1.;
                }
                cvSetReal2D(Temp, idx[0], idx[1], val*scale+shift);
            }
        }
    }
  
    for(i = 0; i <= MAX_HUE; i++)
	values[i] = min + i*(max - min)/(double)MAX_HUE;
  
    //We use HSV images 
    IplImage *HSV_arr = cvCreateImage(size, IPL_DEPTH_8U, 3);
    //H channel update
    cvSetImageCOI(HSV_arr, 1);
    cvCopy(Temp, HSV_arr);
    //S and V channels update

    //HSV space use
    cvSet(Temp, cvScalar(255));
    cvSetImageCOI(HSV_arr, 2);
    cvCopy(Temp, HSV_arr);
    cvSetImageCOI(HSV_arr, 3);
    cvCopy(Temp, HSV_arr);
    cvSetImageCOI(HSV_arr, 0);
    cvReleaseMat(&Temp);
 
    //We put a value scale on the images
    if (size.width < 50 || size.height < 32)
        addscale = 0;
    if (size.height < MAX_HUE)
        ydiv = cvCeil((double)MAX_HUE/size.height);
  
    if(addscale){
	for(i = 0; i <= MAX_HUE; i++){
	    cvLine(HSV_arr, cvPoint(0,i/ydiv), cvPoint(10,i/ydiv), cvScalar(i,255,255));
	}
    }

    //HSV => RGB
    IplImage *RGB_arr = cvCreateImage(size, IPL_DEPTH_8U, 3);
    cvCvtColor(HSV_arr, RGB_arr, 55);
    cvReleaseImage(&HSV_arr);

    if(addscale){
	cvInitFont(&police, CV_FONT_HERSHEY_SIMPLEX, 0.3, 0.3, 0.);
	snprintf(chaine, 256,"%g", values[0]);
	cvPutText(RGB_arr, chaine, cvPoint(11,10), &police, cvScalar(0,0,0));
	snprintf(chaine, 256,"%g", values[MAX_HUE/2]);
	cvPutText(RGB_arr, chaine, cvPoint(11,5+MAX_HUE/(2*ydiv)), &police, cvScalar(0,0,0));
	snprintf(chaine, 256,"%g", values[MAX_HUE]);
	cvPutText(RGB_arr, chaine, cvPoint(11,MAX_HUE/ydiv), &police, cvScalar(0,0,0));
    }
  
    //Writing the values scale 
    cvSaveImage(filename, RGB_arr);
    cvReleaseImage(&RGB_arr);
}

/************************************* Psi values estimation ********************/
static inline double PsiData(const double x){
    return sqrt(x + EPSILON_PSI_DATA*EPSILON_PSI_DATA);
}
static inline double PsiSmooth(const double x){
    return sqrt(x + EPSILON_PSI_SMOOTH*EPSILON_PSI_SMOOTH);
}
static inline double PsiSmoothW(const double x){
    return sqrt(x + EPSILON_PSI_SMOOTH_W*EPSILON_PSI_SMOOTH_W);
}

//**************************************Divergence operator coefficients computation***********************************************************
static inline double DPsiSmooth(const double grad2)
{
    //phi(s^2) = 2sqrt(1+s^2)-2
    return 1./sqrt((EPSILON_PSI_SMOOTH*EPSILON_PSI_SMOOTH + grad2));
}

static inline double DPsiSmoothW(const double grad2)
{
    //phi(s^2) = 2sqrt(1+s^2)-2
    return 1./sqrt((EPSILON_PSI_SMOOTH_W*EPSILON_PSI_SMOOTH_W + grad2));
}

static inline double DPsiData(const double intensity2)
{
    //phi(s^2) = 2sqrt(1+s^2)-2
    return 1./sqrt((EPSILON_PSI_DATA*EPSILON_PSI_DATA + intensity2));
}

//**************************************Half pixel interpolation for the divergence coefficients in the principal directions *****************
static inline void divergence_coefficients_interpolate(CvMat * coeffs, const int i, const int j, double *b1, double *b2, double *b3, double *b4)
{
    double valcentre = cvmGet(coeffs, i, j);
    //The function is called only on the inside points so no condition over the image side
    *b1 = (valcentre + cvmGet(coeffs, i, j + 1)) / 2;	
    *b2 = (valcentre + cvmGet(coeffs, i, j - 1)) / 2;
    *b3 = (valcentre + cvmGet(coeffs, i + 1, j)) / 2;	
    *b4 = (valcentre + cvmGet(coeffs, i - 1, j)) / 2;
}

//**************************************Half pixel interpolation for the divergence coefficients in the diagonal directions *****************
static inline void divergence_coefficients_interpolate_diag(CvMat * coeffs, const int i, const int j, double *b5, double *b6, double *b7, double *b8)
{
    double valcentre = cvmGet(coeffs, i, j);
    //The function is called only on the inside points so no condition over the image side
    *b5 = (valcentre + cvmGet(coeffs, i+1, j + 1)) / 2;	//en X
    *b6 = (valcentre + cvmGet(coeffs, i+1, j - 1)) / 2;
    *b7 = (valcentre + cvmGet(coeffs, i - 1, j+1)) / 2;	//en Y
    *b8 = (valcentre + cvmGet(coeffs, i - 1, j-1)) / 2;
}


//**************************************Images gradient computation********************************
static inline void Derivative(const CvArr * image,  int dx, int dy, CvArr * derivative_int, CvArr * derivative)
{
    if (derivative_int == 0)
        derivative_int = derivative;
    if ((dx == 1 && dy == 0) || (dx == 0 && dy == 1)) {
        cvSobel(image, derivative_int, dx, dy, CV_SCHARR);
        cvConvertScale(derivative_int, derivative, 1.0/32.0);
    }
    else {
      
        cvSobel(image, derivative_int, dx, dy, 3);
        cvConvertScale(derivative_int, derivative, 1.0/4.0);
    }
}



void EnergyComputation3D(const CvMat *I0,   //Images at time t
					   const CvMat **I0w,   //Wraped images at time t 
					   const CvMat **Itw,   //Wraped images at time t+1
					   const CvMat* U    ,  //Horizontal scene flow
					   const CvMat* V    ,  //Vertical scene flow
					   const CvMat* W    ,  //Z scene flow t+1
					   const CvMat* Z0   ,  //surface at time t
                       const CvMat *disactivated_pixels_ref_flow,
                       const CvMat **disactivated_pixels_flow,
                       const CvMat **disactivated_pixels_Z0,
                       const CvMat **disactivated_pixels_Zt,
					   const int N_x, //Horizontal pixels number 
					   const int N_y, //Vertical pixels number
		               double *data, //Data term
		               double *smooth, //Smoothing term
		               const double mu_u, //Smooth parameter for t disparity 
					   const double mu_v, //Smooth parameter for t disparity 
					   const double mu_w, //Smooth parameter for t disparity 
					   const double mu_z, //Smooth parameter for t disparity 
					   const double alpha, //Smooth parameter for t disparity 
					   const double alpha_w, //Smooth parameter for t disparity 
					   const double alpha_z, //Smooth parameter for t disparity 
		               const int opt_Z, 
		               double *E_Ref_flow, //Reference optical flow contribution
		               double *E_Flow, //Right optical flow contribution
		               double *E_Stereo_0, //t stereo contribution
		               double *E_Stereo_t, //t+1 stereo contribution
		               double *Es_Motion, 
		               double *Es_Z,
					   double *Eintensityp,
					   int level){

						   double Es_motion, Es_structure, Esmoothtmp, Esmooth;
						   double E_ref_flow = 0;
						   double E_flow[MAX_NUM_OF_VIEWS] = {0}, E_stereo_0[MAX_NUM_OF_VIEWS] = {0}, E_stereo_t[MAX_NUM_OF_VIEWS] = {0};
						   double Es, Esz0, ps;
						   double Ei, Eintensity, Edata_pixel, Esmooth_pixel;
						   int view = 1;
						   CvMat *EnergyMapData = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *EnergyMapSmooth = cvCreateMat(N_y, N_x, CV_32FC1);
						   cvZero(EnergyMapData);
						   cvZero(EnergyMapSmooth);

						   CvMat *U_x = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *U_y = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *V_x = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *V_y = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *W_x = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *W_y = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *Z0_x = cvCreateMat(N_y, N_x, CV_32FC1);
						   CvMat *Z0_y   = cvCreateMat(N_y, N_x, CV_32FC1);


						   GradientFast_32f(U, U_x, U_y);
						   GradientFast_32f(V, V_x, V_y);
						   GradientFast_32f(W, W_x, W_y);

						   if (opt_Z) {
							   GradientFast_32f(Z0, Z0_x, Z0_y);
						   }

						   // energy
						   Es_motion = 0.;
						   Es_structure = 0.;
						   Esz0 = 0.;
						   Eintensity = 0.;
						   Esmooth = 0.;

						   for(int yl0=0; yl0 < N_y; yl0++){
							   for(int xl0=0; xl0 < N_x; xl0++){
								   Edata_pixel = 0.;
								   Esmooth_pixel = 0.;
								   double it = 0., itw = 0., itw_ref = 0.;
								   double i0 = 0., i0w = 0.;
								   int pixel_do_ref_flow = 1;
								   int pixel_do_flow = 1;
								   int pixel_doZ0 = 1;
								   int pixel_doZt = 1;
								   for(view = 1; view < NumOfViews; view++){
									   if (disactivated_pixels_ref_flow)
										   pixel_do_ref_flow = !cvmGet8u(disactivated_pixels_ref_flow,yl0,xl0);
									   if (disactivated_pixels_flow)
										   pixel_do_flow = !cvmGet8u(disactivated_pixels_flow[view],yl0,xl0);
									   if (disactivated_pixels_Z0)
										   pixel_doZ0 = !cvmGet8u(disactivated_pixels_Z0[view],yl0,xl0);
									   if (disactivated_pixels_Zt)
										   pixel_doZt = !cvmGet8u(disactivated_pixels_Zt[view],yl0,xl0);

									   i0 = cvGetReal2D(I0,yl0,xl0);
									   if (I0w) {
										   i0w = cvGetReal2D(I0w[view],yl0,xl0);
									   }
									   if (Itw) {
										   itw_ref = cvGetReal2D(Itw[0],yl0,xl0);
										   itw = cvGetReal2D(Itw[view],yl0,xl0);
									   }

									   // data terms
									   if (pixel_do_ref_flow) {
										   Ei = square(itw_ref - i0);
										   Edata_pixel += PsiData(Ei);
										   E_ref_flow += PsiData(Ei);
										   Eintensity += Ei;
									   }
									   if (pixel_do_flow) {
										   Ei = square(itw - i0w);
										   Edata_pixel += PsiData(Ei);
										   E_flow[view] += PsiData(Ei);
										   Eintensity += Ei;
									   }
									   if (pixel_doZ0 && opt_Z) {
										   Ei = square(i0w - i0);
										   Edata_pixel += PsiData(Ei);
										   E_stereo_0[view] += PsiData(Ei);
										   Eintensity += Ei;
									   }
									   if (pixel_doZt) {
										   Ei = square(itw - itw_ref);
										   Edata_pixel += PsiData(Ei);
										   E_stereo_t[view] += PsiData(Ei);
										   Eintensity += Ei;
									   }
								   }
								   // regularization terms
								   Esmoothtmp = 0.;
								   double u_x = cvmGet(U_x,yl0,xl0);
								   double u_y = cvmGet(U_y,yl0,xl0);
								   double v_x = cvmGet(V_x,yl0,xl0);
								   double v_y = cvmGet(V_y,yl0,xl0);
								   double w_x = cvmGet(W_x,yl0,xl0);
								   double w_y = cvmGet(W_y,yl0,xl0);
								   double z0_x = 0;
								   double z0_y = 0;

								   if (opt_Z) {
									   z0_x = cvmGet(Z0_x,yl0,xl0);
									   z0_y = cvmGet(Z0_y,yl0,xl0);
								   }

								   if(DIV_UVWZ & opt_Z){
									    Es = mu_u*(square(u_x) + square(u_y)) + mu_v*(square(v_x) + square(v_y)) + mu_w*(square(w_x) + square(w_y)) + mu_z*(square(z0_x) + square(z0_y));
										ps = PsiSmooth(Es);
										Esmooth_pixel += alpha*ps;
										Es_motion += ps;
								   }else
									   if(DIV_UVW){
										   Es = mu_u*(square(u_x) + square(u_y)) + mu_v*(square(v_x) + square(v_y)) + mu_w*(square(w_x) + square(w_y));
										   ps = PsiSmooth(Es);
										   Esmooth_pixel += alpha*ps;
										   Es_motion += ps;
									   }
									   else{
										   Es = mu_u*(square(u_x) + square(u_y)) + mu_v*(square(v_x) + square(v_y));
										   ps = PsiSmooth(Es);
										   Esmooth_pixel += alpha*ps;
										   Es_motion += ps;

										   if(~DIV_UV_ZW & opt_Z | ~opt_Z){
											   Es = mu_w*(square(w_x) + square(w_y)) ;
											   ps = PsiSmooth(Es);
											   Esmooth_pixel += alpha_w*ps;
											   Es_motion += ps;
										   }
									   }

								   if (opt_Z & ~DIV_UVWZ) {
									   if(DIV_UV_ZW)
											Es = mu_w*(square(w_x) + square(w_y)) + mu_z*(square(z0_x) + square(z0_y));
									   else
										   Es =mu_z*(square(z0_x) + square(z0_y));
									    
									   ps = PsiSmooth(Es);
									   Esmooth_pixel += alpha_z*ps;   
									   Es_structure += ps;
								   }

								   Esmooth += Esmooth_pixel;
								   cvmSet(EnergyMapData, yl0, xl0, Edata_pixel);
								   cvmSet(EnergyMapSmooth, yl0, xl0, Esmooth_pixel);
							   }
						   }
#ifdef DEBUG
						   FILE *f;
						   char file[256];
						   sprintf(file,"%s/SmoothEnergyMap%d.txt", TxtResultsDir, level);
						   f = fopen(file,"w");
						   WriteRes(f,EnergyMapSmooth);
						   fclose(f);
						   sprintf(file,"%s/DataEnergyMap%d.txt", TxtResultsDir, level);
						   f = fopen(file,"w");
						   WriteRes(f,EnergyMapData);
						   fclose(f);
#endif
						   cvReleaseMat(&EnergyMapData);
						   cvReleaseMat(&EnergyMapSmooth);
						   *data = E_ref_flow;
						   for(view = 1; view < NumOfViews; view++) {
							   *data += E_flow[view] + E_stereo_0[view] + E_stereo_t[view];
							   if (E_Flow)
								   E_Flow[view] = E_flow[view];
							   if (E_Stereo_0)
								   E_Stereo_0[view] = E_stereo_0[view];
							   if (E_Stereo_t)
								   E_Stereo_t[view] = E_stereo_t[view];
						   }

						   *smooth = Esmooth;
						   if(E_Ref_flow)
							   *E_Ref_flow = E_ref_flow;
						   if (Es_Motion)
							   *Es_Motion = Es_motion;
						   if (Es_Z)
							   *Es_Z = Es_structure;
						   if (Eintensityp)
							   *Eintensityp = Eintensity;

						   cvReleaseMat(&U_x);
						   cvReleaseMat(&U_y);
						   cvReleaseMat(&V_x);
						   cvReleaseMat(&V_y);
						   cvReleaseMat(&W_x);
						   cvReleaseMat(&W_y);
						   cvReleaseMat(&Z0_x);
						   cvReleaseMat(&Z0_y);

}



//************************3 elements maxima computation*****************
static inline double max_3_arg(const double x, const double y, const double z)
{
    return fmax(fmax(x,y),z);
}

//**************************************Vector euclidean norm computation*******************************************************
static inline double norm_vect(const double x, const double y)
{
    return sqrt(x * x + y * y);
}

//**************************************Matrix euclidean norm computation*******************************************************
static inline double norm_mat(const double a, const double b, const double c, const double d)
{
    return sqrt(a * a + b * b + c * c + d * d);
}

//***************************************Gaussian data pyramid computation **********
static inline void GaussianPyramid(CvArr * Data, CvArr * Undersampled_Data)
{
    cvPyrDown(Data, Undersampled_Data);
}

//***************************************Angle of a vector computation with the X axis*********************
static inline double teta_vect(const double x, const double y){
    return acos(x/norm_vect(x,y));
}

//***********************************************Underpolate data to the k-ith resolution level with the reduction factor eta ***************
#if 1
static inline void Underpolate(const CvMat* data, //Data to be undersampled
			       CvMat* data_underpolated, //Undersampled data
			       const int k, //Resolution level
			       const double eta, //Reduction factor
			       const double total_scale)
{
	(void)eta;
	(void)k;
    cvResize(data, data_underpolated, CV_INTER_AREA);
    cvConvertScale(data_underpolated,data_underpolated, total_scale);
}
#else
static void Underpolate(CvMat* data, CvMat* data_unterpolated, const int k, const double eta, const double total_scale){
   
    int N_x = cvGetSize(data).width;
    int N_y = cvGetSize(data).height;
    CvMat* data_cur = cvCreateMat(N_y, N_x, CV_32FC1);
    cvCopy(data, data_cur);
    double eta_total = 1.;
    for(int i = 0; i < k; i++){
        eta_total *= eta;
        CvMat* data_next =  cvCreateMat(cvCeil(N_y*eta_total), cvCeil(N_x*eta_total) , CV_32FC1);
        cvResize(data_cur, data_next, CV_INTER_AREA);
        //SmoothData(data_next);
	cvReleaseMat(&data_cur);
	data_cur = data_next;
    }
    assert(CV_ARE_SIZES_EQ(data_cur,data_underpolated));
    cvConvertScale(data_cur,data_underpolated, total_scale);
    cvReleaseMat(&data_cur);
}
#endif

/***************************************************euclidean norm in R^n*****************************/
double n_norm(CvMat *x, const int Nx){
    double n = 0.;
    for(int k=0; k < Nx; k++)
        n += square(cvGetReal2D(x,k,0));
    return sqrt(n);
}

/***************************************************QR Factorization*********************************/
void QR(CvMat *M, CvMat *Q, CvMat *R, const int Nx, const int Ny){
    cvSetZero(Q);
    cvSetZero(R);
  
}

/**************************************************Matrix Transposition*****************************/
void Transpose(CvMat *Q1, CvMat *Q2, const int N){
    double tmp;
    cvSetZero(Q2);
    for(int i = 0; i < N; i++){
        for(int j = 0; j <= i; j++){
            if(i != j){
                tmp = cvGetReal2D(Q1,i,j);
                cvSetReal2D(Q2,j,i,tmp);
                tmp = cvGetReal2D(Q1,j,i);
                cvSetReal2D(Q2,i,j,tmp);
            }
            else{
                tmp = cvGetReal2D(Q1,i,j);
                cvSetReal2D(Q2,i,j,tmp);
            }
        }
    }
}

/********************************************Scene flow matrix with vector multiplication to spare memory******/
void SFMatrixProduct(CvMat *SFMatrix, CvMat *x, CvMat *x_end, const int Nx, const int Ny, const int dimension){
  
    int i;
    int N = dimension * Nx * Ny;
    for (i = 0; i < N; i++) {
		    
	if(i < dimension*Nx){
	    double x1 = cvGetReal2D(x,i,0);
	    double x2 = cvGetReal2D(x,i+dimension*Nx,0);
	    double m1 = cvGetReal2D(SFMatrix,i,0);
	    double m2 = cvGetReal2D(SFMatrix,i,1);
	    cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
	}
	else if(i >= dimension*Nx*(Ny-1)){
	    double x1 = cvGetReal2D(x,i,0);
	    double x2 = cvGetReal2D(x,i-dimension*Nx,0);
	    double m1 = cvGetReal2D(SFMatrix,i,0);
	    double m2 = cvGetReal2D(SFMatrix,i,1);
	    cvSetReal2D(x_end, i, 0,  m1*x1 + m2*x2);
	}
	else{
	    if(i%(dimension*Nx) == 0){
		double x1 = cvGetReal2D(x,i,0);
		double x2 = cvGetReal2D(x,i+dimension,0);
		double m1 = cvGetReal2D(SFMatrix,i,0);
		double m2 = cvGetReal2D(SFMatrix,i,1);
		cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
	    }
	    else if((i+dimension)%(dimension*Nx)==0){
		double x1 = cvGetReal2D(x,i,0);
		double x2 = cvGetReal2D(x,i-dimension,0);
		double m1 = cvGetReal2D(SFMatrix,i,0);
		double m2 = cvGetReal2D(SFMatrix,i,1);
		cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
	    }
	    else{

		int rest1 = i%(dimension*Nx);
		if(rest1 == 1 || rest1 == 2 || rest1 == 3){
                    double x1 = cvGetReal2D(x,i,0);
                    double x2 = cvGetReal2D(x,i+dimension,0);
                    double m1 = cvGetReal2D(SFMatrix,i,0);
                    double m2 = cvGetReal2D(SFMatrix,i,1);
                    cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2); 
		}
		else if(rest1 == 4*Nx - 3 || rest1 == 4*Nx - 2 || rest1 == 4*Nx - 1){
		    double x1 = cvGetReal2D(x,i,0);
		    double x2 = cvGetReal2D(x,i-dimension,0);
		    double m1 = cvGetReal2D(SFMatrix,i,0);
		    double m2 = cvGetReal2D(SFMatrix,i,1);
		    cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
		}
		else{
		    int rest2 = rest1 % dimension;
		    
		    if(rest2 == 0){

                        double x1 = cvGetReal2D(x,i,0);
                        double x2 = cvGetReal2D(x,i-dimension*Nx,0);
                        double x3 = cvGetReal2D(x,i-dimension,0);
                        double x4 = cvGetReal2D(x,i+1,0);
                        double x5 = cvGetReal2D(x,i+2,0);
                        double x6 = cvGetReal2D(x,i+3,0);
                        double x7 = cvGetReal2D(x,i+dimension*Nx,0);
                        double x8 = cvGetReal2D(x,i+dimension,0);
			
                        double m1 = cvGetReal2D(SFMatrix,i,0);
                        double m2 = cvGetReal2D(SFMatrix,i,4);
                        double m3 = cvGetReal2D(SFMatrix,i,6);
                        double m4 = cvGetReal2D(SFMatrix,i,1);
                        double m5 = cvGetReal2D(SFMatrix,i,2);
                        double m6 = cvGetReal2D(SFMatrix,i,3);
                        double m7 = cvGetReal2D(SFMatrix,i,5);
                        double m8 = cvGetReal2D(SFMatrix,i,7);
			 
                        cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);
					
                    }
		    else if(rest2 == 1){
			//v equation
                        double x1 = cvGetReal2D(x,i,0);
                        double x2 = cvGetReal2D(x,i-dimension*Nx,0);
                        double x3 = cvGetReal2D(x,i-dimension,0);
                        double x4 = cvGetReal2D(x,i-1,0);
                        double x5 = cvGetReal2D(x,i+1,0);
                        double x6 = cvGetReal2D(x,i+2,0);
                        double x7 = cvGetReal2D(x,i+dimension*Nx,0);
                        double x8 = cvGetReal2D(x,i+dimension,0);
			
                        double m1 = cvGetReal2D(SFMatrix,i,1);
                        double m2 = cvGetReal2D(SFMatrix,i,4);
                        double m3 = cvGetReal2D(SFMatrix,i,6);
                        double m4 = cvGetReal2D(SFMatrix,i,0);
                        double m5 = cvGetReal2D(SFMatrix,i,2);
                        double m6 = cvGetReal2D(SFMatrix,i,3);
                        double m7 = cvGetReal2D(SFMatrix,i,5);
                        double m8 = cvGetReal2D(SFMatrix,i,7);
			 
                        cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);

		    }

		    else if(rest2 == 2){
			//dprime equation
                        double x1 = cvGetReal2D(x,i,0);
                        double x2 = cvGetReal2D(x,i-dimension*Nx,0);
                        double x3 = cvGetReal2D(x,i-dimension,0);
                        double x4 = cvGetReal2D(x,i-2,0);
                        double x5 = cvGetReal2D(x,i-1,0);
                        double x6 = cvGetReal2D(x,i+1,0);
                        double x7 = cvGetReal2D(x,i+dimension*Nx,0);
                        double x8 = cvGetReal2D(x,i+dimension,0);
			
                        double m1 = cvGetReal2D(SFMatrix,i,2);
                        double m2 = cvGetReal2D(SFMatrix,i,4);
                        double m3 = cvGetReal2D(SFMatrix,i,6);
                        double m4 = cvGetReal2D(SFMatrix,i,0);
                        double m5 = cvGetReal2D(SFMatrix,i,1);
                        double m6 = cvGetReal2D(SFMatrix,i,3);
                        double m7 = cvGetReal2D(SFMatrix,i,5);
                        double m8 = cvGetReal2D(SFMatrix,i,7);
			 
                        cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);
			  
					    
		    }
		    else{
			//d equation
			double x1 = cvGetReal2D(x,i,0);
			double x2 = cvGetReal2D(x,i-dimension*Nx,0);
			double x3 = cvGetReal2D(x,i-dimension,0);
			double x4 = cvGetReal2D(x,i-3,0);
			double x5 = cvGetReal2D(x,i-2,0);
			double x6 = cvGetReal2D(x,i-1,0);
			double x7 = cvGetReal2D(x,i+dimension*Nx,0);
			double x8 = cvGetReal2D(x,i+dimension,0);
			
			double m1 = cvGetReal2D(SFMatrix,i,3);
			double m2 = cvGetReal2D(SFMatrix,i,4);
			double m3 = cvGetReal2D(SFMatrix,i,6);
			double m4 = cvGetReal2D(SFMatrix,i,0);
			double m5 = cvGetReal2D(SFMatrix,i,1);
			double m6 = cvGetReal2D(SFMatrix,i,2);
			double m7 = cvGetReal2D(SFMatrix,i,5);
			double m8 = cvGetReal2D(SFMatrix,i,7);
			 
			cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);
					
		    }
		    
		}
	    }
	}
    }

}






/*******************************************Scalar Product in R^k****************************/
double scalar_prod(CvMat *x, CvMat *y, const int k){
    double scalar = 0.;
    for(int p = 0; p < k; p++)
        scalar += cvGetReal2D(x,p,0)*cvGetReal2D(y,p,0);
    return scalar;
}

/*************************************Extract a matrix column ****************/
void ColumnExtract(CvMat* Q, CvMat* q1, const int j, const int N){
    for(int p = 0; p < N; p++)
        cvSetReal2D(q1,p,0,cvGetReal2D(Q,p,j));
}


void PyramidBuild(CvMat *I,
                  double eta,
                  int pyramid_levels,
                  CvMat **I_pyramid)//in: array[pyramid_levels] of non-allocated pointers. out: pyramid
{
    //Images Gaussian Pyramid Computation
    double eta_total = 1.;
    int N_x = cvGetSize(I).width;
    int N_y = cvGetSize(I).height;

    I_pyramid[0] = cvCloneMat(I);
    
    for (int level = 1; level < pyramid_levels; level++) {
        eta_total *= eta;
        CvSize size_i = cvSize(cvCeil(N_x*eta_total),cvCeil(N_y*eta_total));
	//Images memory allocation
	I_pyramid[level] = cvCreateMat(size_i.height, size_i.width, CV_MAT_TYPE(I->type));       
	cvResize(I_pyramid[level-1], I_pyramid[level], CV_INTER_AREA);
    }
}

void PyramidSmooth(int pyramid_levels,
                   CvMat **I_pyramid)
{
    for (int level = 0; level < pyramid_levels; level++)
        SmoothImage(I_pyramid[level]);  // Il0_pyramid is pre-smoothed
}

void PyramidCopy(int pyramid_levels,
                 const CvMat * const * const In_pyramid,
                 CvMat ** Out_pyramid)
{
    for (int level = 0; level < pyramid_levels; level++) {
        Out_pyramid[level] = cvCloneMat(In_pyramid[level]);
    }
}


/**************************************************Scene Flow Estimation*****************/


/***************Computes 3D points with respect to the reference image******************/
/* tali: this function computes the 3D points given the reference image coordinates and 
   the depth. it scale the projection matrix to fit the current pyramid level*/

void ComputeXY(CvMat* X,CvMat* Y,CvMat* Z0,
			   CvMat* MapIdentityX,CvMat* MapIdentityY,
			   int N_x,int N_y,
			   const double* params,
			   double scale)
{
	CvMat* temp = cvCreateMat(X->rows, X->cols, CV_32FC1);

	// x-scale*u0
    cvSubS(MapIdentityX,cvScalar(params[2]*scale),temp);
	// Z*(x-u0)/(scale*fku)
	cvMul(temp,Z0,X,double(1/(params[0]*scale)));
	// y-scale*v0
	cvSubS(MapIdentityY,cvScalar(params[3]*scale),temp);
	// Z*(y-v0)/(scale*fkv)
	cvMul(temp,Z0,Y,double(1/(params[1]*scale)));

	cvReleaseMat(&temp);

}
/***************Converts between disparity and depth (rectified images)*****************/
void ConvertDZ(CvMat* D,CvMat* Z,double scale, double maxZ){
	int i=0;
	int j=0;
	cvZero(Z);
	double Val = 0;
	for (int i = 0; i < D->rows; ++i){
	  for (int j = 0; j < D->cols; ++j){
		  Val  = cvmGet(D,i,j);
		  if(abs(Val)==0){
			  cvmSet(Z,i,j,1.05*maxZ);
			  continue;
		  }


		  if(abs(Val) > abs(scale/maxZ))
			  cvmSet(Z,i,j,scale/Val);
		  else
			cvmSet(Z,i,j,maxZ);
	  }
	}
}
/********Copy a specific row from the src mat to a specific row in the dst mat ***********/
void SetRow(CvMat* src, CvMat* dst, int src_row, int dst_row){
	int i=0;
	switch(CV_MAT_TYPE(src->type)){
		case CV_8UC1:  
			for(i=0; i<src->cols; i++){
				((unsigned char *)(dst->data.ptr + SIZE_T(dst_row*dst->step)))[i] = ((unsigned char *)(src->data.ptr + SIZE_T(src_row*src->step)))[i];				
			}
			break;
		case CV_32FC1:
			for(i=0; i<src->cols; i++){
				((float *)(dst->data.ptr + SIZE_T(dst_row*dst->step)))[i] = ((float *)(src->data.ptr + SIZE_T(src_row*src->step)))[i];				
			}
			
			break;
		case CV_64FC1:
			for(i=0; i<src->cols; i++){
				((double *)(dst->data.ptr + SIZE_T(dst_row*dst->step)))[i] = ((double *)(src->data.ptr + SIZE_T(src_row*src->step)))[i];				
			}
			
			break;
		default:
			break;
	}
}
/*****Computes projection matrix from the intrinsic and extrinsic parameters***********/
void ComputeProjMat(CvMat *P, CvMat *R, CvMat *T, CvMat *C,double eta_curr) {
	CvMat *RT = cvCreateMat(3, 4, CV_64FC1);
	cvSetZero(RT);
	double R11 = cvmGet(R,0,0);
	double R12 = cvmGet(R,0,1);
	double R13 = cvmGet(R,0,2);
	double R21 = cvmGet(R,1,0);
	double R22 = cvmGet(R,1,1);
	double R23 = cvmGet(R,1,2);
	double R31 = cvmGet(R,2,0);
	double R32 = cvmGet(R,2,1);
	double R33 = cvmGet(R,2,2);
	double T1 = cvmGet(T,0,0);
	double T2 = cvmGet(T,1,0);
	double T3 = cvmGet(T,2,0);
    // setting the elements in the matrix [R | T]
	cvmSet(RT,0,0,R11);
	cvmSet(RT,0,1,R12);
	cvmSet(RT,0,2,R13);
	cvmSet(RT,1,0,R21);
	cvmSet(RT,1,1,R22);
	cvmSet(RT,1,2,R23);
	cvmSet(RT,2,0,R31);
	cvmSet(RT,2,1,R32);
	cvmSet(RT,2,2,R33);
	cvmSet(RT,0,3,T1);
	cvmSet(RT,1,3,T2);
	cvmSet(RT,2,3,T3);
	// scale the matrix according to the pyramid level
	cvmSet(C,2,2,1/eta_curr);
	cvMatMul(C,RT,P);
	cvReleaseMat(&RT);
}

void ComputeRefProjMat(CvMat *P, CvMat *R, CvMat *T, CvMat *C,double eta_curr) {
	CvMat *RT = cvCreateMat(3, 4, CV_64FC1);
	cvSetZero(RT);
	cvSetZero(C);

	 // setting the elements in the matrix [R | T]
	cvmSet(RT,0,0,1);
	cvmSet(RT,0,1,0);
	cvmSet(RT,0,2,0);
	cvmSet(RT,1,0,0);
	cvmSet(RT,1,1,1);
	cvmSet(RT,1,2,0);
	cvmSet(RT,2,0,0);
	cvmSet(RT,2,1,0);
	cvmSet(RT,2,2,1);
	cvmSet(RT,0,3,0);
	cvmSet(RT,1,3,0);
	cvmSet(RT,2,3,0);

	cvmSet(T,0,0,0);
	cvmSet(T,1,0,0);
	cvmSet(T,2,0,0);
	cvmSet(R,0,0,1);
	cvmSet(R,0,1,0);
	cvmSet(R,0,2,0);
	cvmSet(R,1,0,0);
	cvmSet(R,1,1,1);
	cvmSet(R,1,2,0);
	cvmSet(R,2,0,0);
	cvmSet(R,2,1,0);
	cvmSet(R,2,2,1);
    // setting the elements in the calibriation matrix
	cvmSet(C,0,0,f0ku);
	cvmSet(C,1,1,f0kv);
	//scaling the projection matrix to fit to the current pyramid size
	cvmSet(C,2,2,1/eta_curr);
	cvmSet(C,0,2,u0);
	cvmSet(C,1,2,v0);

	cvMatMul(C,RT,P);
    cvReleaseMat(&RT);

}


void ProjectPoints2( const CvMat* objectPoints,
                  const CvMat* r_vec,
                  const CvMat* t_vec,
                  const CvMat* A,
                  const CvMat* distCoeffs,
                  CvMat* imagePoints,
				  double scale,
				  CvMat* dpdr,
                  CvMat* dpdt, CvMat* dpdf,
                  CvMat* dpdc, CvMat* dpdk,
                  double aspectRatio )
{
    CvMat *_M = 0, *_m = 0;
    CvMat *_dpdr = 0, *_dpdt = 0, *_dpdc = 0, *_dpdf = 0, *_dpdk = 0;

    CV_FUNCNAME( "cvProjectPoints2" );

    __BEGIN__;

    int i, j, count;
    int calc_derivatives;
    const CvPoint3D64f* M;
    CvPoint2D64f* m;
    double r[3], R[9], dRdr[27], t[3], a[9], k[5] = {0,0,0,0,0}, fx, fy, cx, cy;
    CvMat _r, _t, _a = cvMat( 3, 3, CV_64F, a ), _k;
    CvMat _R = cvMat( 3, 3, CV_64F, R ), _dRdr = cvMat( 3, 9, CV_64F, dRdr );
    double *dpdr_p = 0, *dpdt_p = 0, *dpdk_p = 0, *dpdf_p = 0, *dpdc_p = 0;
    int dpdr_step = 0, dpdt_step = 0, dpdk_step = 0, dpdf_step = 0, dpdc_step = 0;
    bool fixedAspectRatio = aspectRatio > FLT_EPSILON;

    if( !CV_IS_MAT(objectPoints) || !CV_IS_MAT(r_vec) ||
        !CV_IS_MAT(t_vec) || !CV_IS_MAT(A) ||
        /*!CV_IS_MAT(distCoeffs) ||*/ !CV_IS_MAT(imagePoints) )
        CV_ERROR( CV_StsBadArg, "One of required arguments is not a valid matrix" );

    count = MAX(objectPoints->rows, objectPoints->cols);

    if( CV_IS_CONT_MAT(objectPoints->type) && CV_MAT_DEPTH(objectPoints->type) == CV_64F &&
        ((objectPoints->rows == 1 && CV_MAT_CN(objectPoints->type) == 3) ||
        (objectPoints->rows == count && CV_MAT_CN(objectPoints->type)*objectPoints->cols == 3)))
        _M = (CvMat*)objectPoints;
    else
    {
        CV_CALL( _M = cvCreateMat( 1, count, CV_64FC3 ));
        CV_CALL( cvConvertPointsHomogeneous( objectPoints, _M ));
    }

    if( CV_IS_CONT_MAT(imagePoints->type) && CV_MAT_DEPTH(imagePoints->type) == CV_64F &&
        ((imagePoints->rows == 1 && CV_MAT_CN(imagePoints->type) == 2) ||
        (imagePoints->rows == count && CV_MAT_CN(imagePoints->type)*imagePoints->cols == 2)))
        _m = imagePoints;
    else
        CV_CALL( _m = cvCreateMat( 1, count, CV_64FC2 ));

    M = (CvPoint3D64f*)_M->data.db;
    m = (CvPoint2D64f*)_m->data.db;

    if( (CV_MAT_DEPTH(r_vec->type) != CV_64F && CV_MAT_DEPTH(r_vec->type) != CV_32F) ||
        (((r_vec->rows != 1 && r_vec->cols != 1) ||
        r_vec->rows*r_vec->cols*CV_MAT_CN(r_vec->type) != 3) &&
        ((r_vec->rows != 3 && r_vec->cols != 3) || CV_MAT_CN(r_vec->type) != 1)))
        CV_ERROR( CV_StsBadArg, "Rotation must be represented by 1x3 or 3x1 "
                  "floating-point rotation vector, or 3x3 rotation matrix" );

    if( r_vec->rows == 3 && r_vec->cols == 3 )
    {
        _r = cvMat( 3, 1, CV_64FC1, r );
        CV_CALL( cvRodrigues2( r_vec, &_r ));
        CV_CALL( cvRodrigues2( &_r, &_R, &_dRdr ));
        cvCopy( r_vec, &_R );
    }
    else
    {
        _r = cvMat( r_vec->rows, r_vec->cols, CV_MAKETYPE(CV_64F,CV_MAT_CN(r_vec->type)), r );
        CV_CALL( cvConvert( r_vec, &_r ));
        CV_CALL( cvRodrigues2( &_r, &_R, &_dRdr ) );
    }

    if( (CV_MAT_DEPTH(t_vec->type) != CV_64F && CV_MAT_DEPTH(t_vec->type) != CV_32F) ||
        (t_vec->rows != 1 && t_vec->cols != 1) ||
        t_vec->rows*t_vec->cols*CV_MAT_CN(t_vec->type) != 3 )
        CV_ERROR( CV_StsBadArg,
            "Translation vector must be 1x3 or 3x1 floating-point vector" );

    _t = cvMat( t_vec->rows, t_vec->cols, CV_MAKETYPE(CV_64F,CV_MAT_CN(t_vec->type)), t );
    CV_CALL( cvConvert( t_vec, &_t ));

    if( (CV_MAT_TYPE(A->type) != CV_64FC1 && CV_MAT_TYPE(A->type) != CV_32FC1) ||
        A->rows != 3 || A->cols != 3 )
        CV_ERROR( CV_StsBadArg, "Instrinsic parameters must be 3x3 floating-point matrix" );

    CV_CALL( cvConvert( A, &_a ));
    fx = a[0]; fy = a[4];
    cx = a[2]; cy = a[5];

    if( fixedAspectRatio )
        fx = fy*aspectRatio;

    if( distCoeffs )
    {
        if( !CV_IS_MAT(distCoeffs) ||
            (CV_MAT_DEPTH(distCoeffs->type) != CV_64F &&
            CV_MAT_DEPTH(distCoeffs->type) != CV_32F) ||
            (distCoeffs->rows != 1 && distCoeffs->cols != 1) ||
            (distCoeffs->rows*distCoeffs->cols*CV_MAT_CN(distCoeffs->type) != 4 &&
            distCoeffs->rows*distCoeffs->cols*CV_MAT_CN(distCoeffs->type) != 5) )
            CV_ERROR( CV_StsBadArg,
                "Distortion coefficients must be 1x4, 4x1, 1x5 or 5x1 floating-point vector" );

        _k = cvMat( distCoeffs->rows, distCoeffs->cols,
                    CV_MAKETYPE(CV_64F,CV_MAT_CN(distCoeffs->type)), k );
        CV_CALL( cvConvert( distCoeffs, &_k ));
    }

    if( dpdr )
    {
        if( !CV_IS_MAT(dpdr) ||
            (CV_MAT_TYPE(dpdr->type) != CV_32FC1 &&
            CV_MAT_TYPE(dpdr->type) != CV_64FC1) ||
            dpdr->rows != count*2 || dpdr->cols != 3 )
            CV_ERROR( CV_StsBadArg, "dp/drot must be 2Nx3 floating-point matrix" );

        if( CV_MAT_TYPE(dpdr->type) == CV_64FC1 )
            _dpdr = dpdr;
        else
            CV_CALL( _dpdr = cvCreateMat( 2*count, 3, CV_64FC1 ));
        dpdr_p = _dpdr->data.db;
        dpdr_step = _dpdr->step/sizeof(dpdr_p[0]);
    }

    if( dpdt )
    {
        if( !CV_IS_MAT(dpdt) ||
            (CV_MAT_TYPE(dpdt->type) != CV_32FC1 &&
            CV_MAT_TYPE(dpdt->type) != CV_64FC1) ||
            dpdt->rows != count*2 || dpdt->cols != 3 )
            CV_ERROR( CV_StsBadArg, "dp/dT must be 2Nx3 floating-point matrix" );

        if( CV_MAT_TYPE(dpdt->type) == CV_64FC1 )
            _dpdt = dpdt;
        else
            CV_CALL( _dpdt = cvCreateMat( 2*count, 3, CV_64FC1 ));
        dpdt_p = _dpdt->data.db;
        dpdt_step = _dpdt->step/sizeof(dpdt_p[0]);
    }

    if( dpdf )
    {
        if( !CV_IS_MAT(dpdf) ||
            (CV_MAT_TYPE(dpdf->type) != CV_32FC1 && CV_MAT_TYPE(dpdf->type) != CV_64FC1) ||
            dpdf->rows != count*2 || dpdf->cols != 2 )
            CV_ERROR( CV_StsBadArg, "dp/df must be 2Nx2 floating-point matrix" );

        if( CV_MAT_TYPE(dpdf->type) == CV_64FC1 )
            _dpdf = dpdf;
        else
            CV_CALL( _dpdf = cvCreateMat( 2*count, 2, CV_64FC1 ));
        dpdf_p = _dpdf->data.db;
        dpdf_step = _dpdf->step/sizeof(dpdf_p[0]);
    }

    if( dpdc )
    {
        if( !CV_IS_MAT(dpdc) ||
            (CV_MAT_TYPE(dpdc->type) != CV_32FC1 && CV_MAT_TYPE(dpdc->type) != CV_64FC1) ||
            dpdc->rows != count*2 || dpdc->cols != 2 )
            CV_ERROR( CV_StsBadArg, "dp/dc must be 2Nx2 floating-point matrix" );

        if( CV_MAT_TYPE(dpdc->type) == CV_64FC1 )
            _dpdc = dpdc;
        else
            CV_CALL( _dpdc = cvCreateMat( 2*count, 2, CV_64FC1 ));
        dpdc_p = _dpdc->data.db;
        dpdc_step = _dpdc->step/sizeof(dpdc_p[0]);
    }

    if( dpdk )
    {
        if( !CV_IS_MAT(dpdk) ||
            (CV_MAT_TYPE(dpdk->type) != CV_32FC1 && CV_MAT_TYPE(dpdk->type) != CV_64FC1) ||
            dpdk->rows != count*2 || (dpdk->cols != 5 && dpdk->cols != 4 && dpdk->cols != 2) )
            CV_ERROR( CV_StsBadArg, "dp/df must be 2Nx5, 2Nx4 or 2Nx2 floating-point matrix" );

        if( !distCoeffs )
            CV_ERROR( CV_StsNullPtr, "distCoeffs is NULL while dpdk is not" );

        if( CV_MAT_TYPE(dpdk->type) == CV_64FC1 )
            _dpdk = dpdk;
        else
            CV_CALL( _dpdk = cvCreateMat( dpdk->rows, dpdk->cols, CV_64FC1 ));
        dpdk_p = _dpdk->data.db;
        dpdk_step = _dpdk->step/sizeof(dpdk_p[0]);
    }

    calc_derivatives = dpdr || dpdt || dpdf || dpdc || dpdk;

    for( i = 0; i < count; i++ )
    {
        double X = M[i].x, Y = M[i].y, Z = M[i].z;
        double x = R[0]*X + R[1]*Y + R[2]*Z + t[0];
        double y = R[3]*X + R[4]*Y + R[5]*Z + t[1];
        double z = R[6]*X + R[7]*Y + R[8]*Z + t[2];
        double r2, r4, r6, a1, a2, a3, cdist;
        double xd, yd;

        z = z ? scale/z : scale/1;
        x *= z; y *= z;

        r2 = x*x + y*y;
        r4 = r2*r2;
        r6 = r4*r2;
        a1 = 2*x*y;
        a2 = r2 + 2*x*x;
        a3 = r2 + 2*y*y;
        cdist = 1 + k[0]*r2 + k[1]*r4 + k[4]*r6;
        xd = x*cdist + k[2]*a1 + k[3]*a2;
        yd = y*cdist + k[2]*a3 + k[3]*a1;

        m[i].x = xd*fx + cx*scale;
        m[i].y = yd*fy + cy*scale;

        if( calc_derivatives )
        {
            if( dpdc_p )
            {
                dpdc_p[0] = 1; dpdc_p[1] = 0;
                dpdc_p[dpdc_step] = 0;
                dpdc_p[dpdc_step+1] = 1;
                dpdc_p += dpdc_step*2;
            }

            if( dpdf_p )
            {
                if( fixedAspectRatio )
                {
                    dpdf_p[0] = 0; dpdf_p[1] = xd*aspectRatio;
                    dpdf_p[dpdf_step] = 0;
                    dpdf_p[dpdf_step+1] = yd;
                }
                else
                {
                    dpdf_p[0] = xd; dpdf_p[1] = 0;
                    dpdf_p[dpdf_step] = 0;
                    dpdf_p[dpdf_step+1] = yd;
                }
                dpdf_p += dpdf_step*2;
            }

            if( dpdk_p )
            {
                dpdk_p[0] = fx*x*r2;
                dpdk_p[1] = fx*x*r4;
                dpdk_p[dpdk_step] = fy*y*r2;
                dpdk_p[dpdk_step+1] = fy*y*r4;
                if( _dpdk->cols > 2 )
                {
                    dpdk_p[2] = fx*a1;
                    dpdk_p[3] = fx*a2;
                    dpdk_p[dpdk_step+2] = fy*a3;
                    dpdk_p[dpdk_step+3] = fy*a1;
                    if( _dpdk->cols > 4 )
                    {
                        dpdk_p[4] = fx*x*r6;
                        dpdk_p[dpdk_step+4] = fy*y*r6;
                    }
                }
                dpdk_p += dpdk_step*2;
            }

            if( dpdt_p )
            {
                double dxdt[] = { z, 0, -x*z }, dydt[] = { 0, z, -y*z };
                for( j = 0; j < 3; j++ )
                {
                    double dr2dt = 2*x*dxdt[j] + 2*y*dydt[j];
                    double dcdist_dt = k[0]*dr2dt + 2*k[1]*r2*dr2dt + 3*k[4]*r4*dr2dt;
                    double da1dt = 2*(x*dydt[j] + y*dxdt[j]);
                    double dmxdt = fx*(dxdt[j]*cdist + x*dcdist_dt +
                                k[2]*da1dt + k[3]*(dr2dt + 2*x*dxdt[j]));
                    double dmydt = fy*(dydt[j]*cdist + y*dcdist_dt +
                                k[2]*(dr2dt + 2*y*dydt[j]) + k[3]*da1dt);
                    dpdt_p[j] = dmxdt;
                    dpdt_p[dpdt_step+j] = dmydt;
                }
                dpdt_p += dpdt_step*2;
            }

            if( dpdr_p )
            {
                double dx0dr[] =
                {
                    X*dRdr[0] + Y*dRdr[1] + Z*dRdr[2],
                    X*dRdr[9] + Y*dRdr[10] + Z*dRdr[11],
                    X*dRdr[18] + Y*dRdr[19] + Z*dRdr[20]
                };
                double dy0dr[] =
                {
                    X*dRdr[3] + Y*dRdr[4] + Z*dRdr[5],
                    X*dRdr[12] + Y*dRdr[13] + Z*dRdr[14],
                    X*dRdr[21] + Y*dRdr[22] + Z*dRdr[23]
                };
                double dz0dr[] =
                {
                    X*dRdr[6] + Y*dRdr[7] + Z*dRdr[8],
                    X*dRdr[15] + Y*dRdr[16] + Z*dRdr[17],
                    X*dRdr[24] + Y*dRdr[25] + Z*dRdr[26]
                };
                for( j = 0; j < 3; j++ )
                {
                    double dxdr = z*(dx0dr[j] - x*dz0dr[j]);
                    double dydr = z*(dy0dr[j] - y*dz0dr[j]);
                    double dr2dr = 2*x*dxdr + 2*y*dydr;
                    double dcdist_dr = k[0]*dr2dr + 2*k[1]*r2*dr2dr + 3*k[4]*r4*dr2dr;
                    double da1dr = 2*(x*dydr + y*dxdr);
                    double dmxdr = fx*(dxdr*cdist + x*dcdist_dr +
                                k[2]*da1dr + k[3]*(dr2dr + 2*x*dxdr));
                    double dmydr = fy*(dydr*cdist + y*dcdist_dr +
                                k[2]*(dr2dr + 2*y*dydr) + k[3]*da1dr);
                    dpdr_p[j] = dmxdr;
                    dpdr_p[dpdr_step+j] = dmydr;
                }
                dpdr_p += dpdr_step*2;
            }
        }
    }

    if( _m != imagePoints )
        cvConvertPointsHomogeneous( _m, imagePoints );
    if( _dpdr != dpdr )
        cvConvert( _dpdr, dpdr );
    if( _dpdt != dpdt )
        cvConvert( _dpdt, dpdt );
    if( _dpdf != dpdf )
        cvConvert( _dpdf, dpdf );
    if( _dpdc != dpdc )
        cvConvert( _dpdc, dpdc );
    if( _dpdk != dpdk )
        cvConvert( _dpdk, dpdk );

    __END__;

    if( _M != objectPoints )
        cvReleaseMat( &_M );
    if( _m != imagePoints )
        cvReleaseMat( &_m );
    if( _dpdr != dpdr )
        cvReleaseMat( &_dpdr );
    if( _dpdt != dpdt )
        cvReleaseMat( &_dpdt );
    if( _dpdf != dpdf )
        cvReleaseMat( &_dpdf );
    if( _dpdc != dpdc )
        cvReleaseMat( &_dpdc );
    if( _dpdk != dpdk )
        cvReleaseMat( &_dpdk );
}

void ComputeCOP(CvMat **COP, CvMat **R_Mat, CvMat **T_Mat) {
	CvMat *temp = cvCreateMat(3, 1, CV_64F);
	CvMat *R_inv = cvCreateMat(3, 3, CV_64F);
	cvZero(temp);
	COP[0] = cvCreateMat(3, 1, CV_64F);
	cvZero(COP[0]);
	int view = 1;
	for( view = 1; view < NumOfViews; view++) {
		cvmSet(temp, 0, 0, COPX); 
		cvmSet(temp, 1, 0, COPY);
		cvmSet(temp, 2, 0, COPZ);
		cvmSub(temp, T_Mat[view], temp);
		cvInvert(R_Mat[view], R_inv);
		COP[view] = cvCreateMat(3, 1, CV_64F);
		cvMatMul(R_inv, temp, COP[view]);
	}
	cvReleaseMat(&temp);

}

void ComputeOcclusion(CvMat *MapX,CvMat *MapY, CvMat *X, CvMat *Y, CvMat *Z0, CvMat *disactivated_pixels_map, CvMat *cop) {
	int n_x = MapX->cols;
	int n_y = MapX->rows;
	double xk = 0; double yk = 0;
	int rxk = 0; int ryk = 0;
	int x_ref = 0; int y_ref = 0;
	double X_curr = 0; double Y_curr = 0; double Z_curr = 0;
	double X_prev = 0; double Y_prev = 0; double Z_prev = 0;
	double dist = 0;
	double curr_dist = 0;
	double prev_dist = 0;
	double copx = 0; double copy = 0; double copz = 0;
	BOOL is_occluded = 0;
	int disactive = 0;
	CvMat *RefMap_X = cvCreateMat(n_y,n_x,CV_32FC1);
	CvMat *RefMap_Y = cvCreateMat(n_y,n_x,CV_32FC1);
	cvZero(RefMap_X);
	cvZero(RefMap_Y);
	
	if(cop) {
		copx = cvmGet(cop, 0, 0);
		copy = cvmGet(cop, 1, 0);
		copz = cvmGet(cop, 2, 0);
	}
	//loops over all pixels 
	for(int i = 0; i < n_x; i++) {
		for(int j = 0; j < n_y; j++) {
			is_occluded = 0;

			Z_curr = cvmGet(Z0, j, i);
			if(Occl_Z){
				if(Z_curr >= 0.95*double(maxZ)){
					cvmSet8u(disactivated_pixels_map, j, i, 150);
					continue;
				}
			}
			//the coordinates in the second image which correspond to (j,i)
			xk = cvmGet(MapX, j, i);
			yk = cvmGet(MapY, j, i);
			for(int ryk=cvRound(yk); ryk <= cvRound(yk); ryk++) {
				for(int rxk=cvRound(xk); rxk <= cvRound(xk); rxk++) {
			
			if (rxk < 0 || rxk >= n_x || ryk < 0 || ryk >= n_y)
				continue;
			
			// the remap reference coordinates 
			x_ref = (int)cvmGet(RefMap_X, ryk, rxk);
			y_ref = (int)cvmGet(RefMap_Y, ryk, rxk);
            
			
			// if xk and yk where previously then ref_x and ref_y are non zero
			if(x_ref == 0 && y_ref == 0) {
				// its the first time there is a map of reference image pixel to (ryk, rxk)
				cvmSet(RefMap_X, ryk, rxk, i);
				cvmSet(RefMap_Y, ryk, rxk, j);
				continue;
			}
			is_occluded = (x_ref - i)*(x_ref - i) + (y_ref - j)*(y_ref - j) >= 2 ? 1 : 0;
			X_curr = cvmGet(X, j, i);
			Y_curr = cvmGet(Y, j, i);
			

			

			X_prev = cvmGet(X, y_ref, x_ref);
			Y_prev = cvmGet(Y, y_ref, x_ref);
			Z_prev = cvmGet(Z0, y_ref, x_ref);
			disactive = cvmGet8u(disactivated_pixels_map, y_ref, x_ref);
				// computes the distances 
			curr_dist = (X_curr-copx)*(X_curr-copx) + (Y_curr-copy)*(Y_curr-copy) + (Z_curr-copz)*(Z_curr-copz);
			prev_dist = (X_prev-copx)*(X_prev-copx) + (Y_prev-copy)*(Y_prev-copy) + (Z_prev-copz)*(Z_prev-copz);

			if(is_occluded)
				//is_occluded = (X_curr - X_prev)*(X_curr - X_prev) + (Y_curr - Y_prev)*(Y_curr - Y_prev) + (Z_curr - Z_prev)*(Z_curr - Z_prev) > DISACTIVATION_THRESHOLD_Z ? 1 : 0;

				// in case the previous point is occluded
				if(curr_dist < 0.9*prev_dist) {
					cvmSet(RefMap_X, ryk, rxk, i);
					cvmSet(RefMap_Y, ryk, rxk, j);
					if(disactive==0 && is_occluded)
						cvmSet8u(disactivated_pixels_map, y_ref, x_ref, 100);
				}else {
					// the current point is occluded
					if(is_occluded)
						cvmSet8u(disactivated_pixels_map, j, i, 100);
				}
			}
		}
			}
		}

			
	cvReleaseMat(&RefMap_X);
	cvReleaseMat(&RefMap_Y);
}
/***************computes the projection of world points to the image*******************/
void ComputeProjMaps(CvMat* X,CvMat* Y,
					 CvMat* Z,CvMat *R, 
					 CvMat* T, CvMat* C,
					 CvMat* Proj_MapX_Ir0,
					 CvMat* Proj_MapY_Ir0,
					 double eta_curr){
	int i=0;
	int n_x = X->cols;
	int n_y = X->rows;

	CvMat *X_row = cvCreateMat(1,n_y, CV_32FC1); 
	CvMat *Y_row = cvCreateMat(1,n_y, CV_32FC1); 
	CvMat *Z_row = cvCreateMat(1,n_y, CV_32FC1); 
	CvMat *proj_points = cvCreateMat(2,n_x, CV_32FC1); 

	CvMat *Point_row = cvCreateMat(3,n_x, CV_32FC1); 
    
	for(i=0; i< X->rows; i++){
		SetRow(X,Point_row,i,0);
		SetRow(Y,Point_row,i,1);
		SetRow(Z,Point_row,i,2);
		
		ProjectPoints2(Point_row, R, T, C, NULL,proj_points,eta_curr);
		SetRow(proj_points, Proj_MapX_Ir0, 0, i);
		SetRow(proj_points, Proj_MapY_Ir0, 1, i);
		
	}
	
    cvReleaseMat(&X_row);
	cvReleaseMat(&Y_row);
	cvReleaseMat(&Z_row);
	cvReleaseMat(&Point_row);
	cvReleaseMat(&proj_points);
}


/************************* Writes a matrix to txt file **********************************/
void WriteRes(FILE* filename,CvMat* M){
	int i=0;
	int j=0;
	double Val = 0;
	if(M == NULL){
		i=1;
		return;
	}
	
	if (filename){
		  for (int i = 0; i < M->rows; ++i){
			  for (int j = 0; j < M->cols; ++j){
				  Val  = cvmGet(M,i,j);
				  fprintf (filename, "%f\t",Val );
			  }
			  fprintf (filename, "\n");
		  }
	  }
}
/************************* Reads a matrix to txt file **********************************/
void ReadMat(char* filename,CvMat* M){
	int i=0;
	int j=0;
	double Val = 0;
	if(M == NULL){
		i=1;
		return;
	}
	
	if (filename){
		FILE *f = fopen(filename,"r");
		  for (int i = 0; i < M->rows; ++i){
			  for (int j = 0; j < M->cols; ++j){
				  fscanf(f,"%lf",&Val);
				  cvmSet(M,i,j,Val);	  
			  }
		  }
		 fclose(f);
	  }

}

void WriteRes8u(FILE* filename,const CvMat* M){
	int i=0;
	int j=0;
	double Val = 0;
	if(M == NULL){
		i=1;
		return;
	}
	
	if (filename){
		  for (int i = 0; i < M->rows; ++i){
			  for (int j = 0; j < M->cols; ++j){
				  Val  = cvmGet8u(M,i,j);
				  fprintf (filename, "%f\t",Val );
			  }
			  fprintf (filename, "\n");
		  }
	  }
}

/********************** Computes the world coordinates gradients ************************/
void Gradient3DPoints(CvMat* MapIdentityX,CvMat* MapIdentityY, CvMat* Z, CvMat* Z_X, CvMat* Z_Y, 
					  double eta_curr,
					  CvMat* Xx, CvMat* Xy,CvMat* Xz, CvMat* Yx,CvMat* Yy,CvMat* Yz)
{	
	CvMat *temp1 = cvCreateMat(Z->rows, Z->cols, CV_32FC1);
	CvMat *temp2 = cvCreateMat(Z->rows, Z->cols, CV_32FC1);
    //---------------- X derivatives -----------------------
	// temp1<- x - u0*eta_curr
	cvAddS(MapIdentityX,cvScalar(-u0*eta_curr),temp1);
	cvCvtScale(temp1,Xz,1/(f0ku*eta_curr),0);
	
	//computes Xx: [Z_X*temp1 + Z]/[f0ku*eta_curr]
	cvMul(temp1,Z_X,temp2,1/(f0ku*eta_curr));
	cvAddWeighted(temp2,1,Z,1/(f0ku*eta_curr),0,Xx);

	// computes Xy: [Z_Y*temp1]/[f0ku*eta_curr]
    cvMul(temp1,Z_Y,Xy,1/(f0ku*eta_curr));
	//------------------------------------------------------
	cvZero(temp1);
	cvZero(temp2);
	//---------------- Y derivatives -----------------------
	// temp1<- y - v0*eta_curr
	cvAddS(MapIdentityY,cvScalar(-v0*eta_curr),temp1);
	cvCvtScale(temp1,Yz,1/(f0kv*eta_curr),0);
	
	// computes Yx
	cvMul(temp1,Z_X,Yx,1/(f0kv*eta_curr));

	// computes Yy
	cvMul(temp1,Z_Y,temp2,1/(f0kv*eta_curr));
	cvAddWeighted(temp2,1,Z,1/(f0kv*eta_curr),0,Yy);

	cvReleaseMat(&temp1);
	cvReleaseMat(&temp2);
}


/******************** Computes the projection coordinates gradients ************************/
void GradientProjCoor(CvMat* X, CvMat* Y, CvMat* Z, CvMat *X_driv, CvMat *Y_driv,CvMat* Z_driv, 
					  CvMat* M, CvMat* Proj_driv, int grad_y){
    
	CvMat *denom = cvCreateMat(X->rows, X->cols, CV_32FC1);
	CvMat *numer = cvCreateMat(X->rows, X->cols, CV_32FC1);
	CvMat *temp1 = cvCreateMat(X->rows, X->cols, CV_32FC1);
	CvMat *temp2 = cvCreateMat(X->rows, X->cols, CV_32FC1);
	double Mn1; double Mn2; double Mn3; double Mn4;
	double Md1; double Md2; double Md3; double Md4;
		
	// the last line of the projection matrix
	Md1 = cvmGet(M, 2, 0);
	Md2 = cvmGet(M, 2, 1);
	Md3 = cvmGet(M, 2, 2);
	Md4 = cvmGet(M, 2, 3);
	
	// denom = M31*X + M32*Y + M33*Z + M34
	cvAddWeighted(X, Md1, Y, Md2, 0, temp1);
	cvAddWeighted(temp1, 1, Z ,Md3, Md4,denom);
	cvZero(temp1);
	
	// if computing the gradient of the horizintal projected coordinate
	if(grad_y==0){
		Mn1 = cvmGet(M, 0, 0);
		Mn2 = cvmGet(M, 0, 1);
		Mn3 = cvmGet(M, 0, 2);
		Mn4 = cvmGet(M, 0, 3);
	}
	else 
	{
	    Mn1 = cvmGet(M, 1, 0);
		Mn2 = cvmGet(M, 1, 1);
		Mn3 = cvmGet(M, 1, 2);
		Mn4 = cvmGet(M, 1, 3);
	}

	//numer = M11*X + M12*Y + M13*Z + M14
    cvAddWeighted(X, Mn1, Y, Mn2, 0, temp1);
	cvAddWeighted(temp1, 1, Z, Mn3, Mn4,numer);
	

	// Mn1*X_driv + Mn2*Y_driv + Mn3*Z_driv
	cvAddWeighted(X_driv, Mn1, Y_driv, Mn2, 0, temp1);
	cvAddWeighted(temp1,1, Z_driv, Mn3, 0, temp1);

	// Md1*X_driv + Md2*Y_driv + Md3*Z_driv
	cvAddWeighted(X_driv, Md1, Y_driv, Md2, 0,temp2);
	cvAddWeighted(temp2, 1, Z_driv, Md3, 0,temp2);
	//Proj_driv <- temp1*denom
	cvMul(temp1,denom,Proj_driv);
	cvZero(temp1);
	// temp1 <-- temp2*numer
	cvMul(temp2,numer,temp1);
	// Proj_driv <--denom(...)-numer(...)
	cvSub(Proj_driv,temp1, Proj_driv);
    cvZero(temp1);
	cvZero(temp2);
	
	cvMul(denom,denom,temp1);
	cvDiv(Proj_driv, temp1, Proj_driv);
	
    cvReleaseMat(&denom);
	cvReleaseMat(&numer);
	cvReleaseMat(&temp1);
	cvReleaseMat(&temp2);
}

void SceneFlow_MVZ(FILE* logfile, //the log file
	       const double eta, //reduction factor
	       int pyramid_levels, // total number of pyramid levels
	       const CvMat * const * const * const Images_0_pyramid, //time t Images
	       const CvMat * const * const * const Images_t_pyramid, //time t+1 Images
           int pyramid_level_start, // starting level
	       CvMat* U_start, //horizontal left scene flow initialization
	       CvMat* V_start, //vertical scene flow initialization
		   CvMat* W_start, //depth scene flow initialization
           CvMat* Z0_start, // Z(x,y) initialization
           int pyramid_level_end, // end pyramid level
	       CvMat* U_end, //final U
	       CvMat* V_end, //final V
		   CvMat* W_end, //final V
		   CvMat* Z0_end, //t Z 
		   const double h, //grid step for finite elements
	       const double epsilon_outer, //stopping criteria for outer iterations 
	       const double epsilon_inner, //stopping criteria for inner iterations
	       const double epsilon_SOR, //stopping criteria for the linear system resolution
	       const int max_iter_outer, //maximal number of outer iterations
	       const int max_iter_inner, //maximal number of inner iterations
	       const int max_iter_SOR, //maximal number of linear system resolution iterations
	       const double omega_SOR, //SOR method weight
           double alpha, // smooth term weight
		   double alpha_w, // smooth term weight
		   double alpha_z, // smooth term weight
           const double mu_z,
		   const double mu_u,
		   const double mu_v,
		   const double mu_w,
		   const double f0ku, // x focal length of the reference camera
		   const double f0kv, // y focal length of the refernce camera
		   const double u0, 
		   const double v0,
		   CvMat** R_Mat, //the rotation matrixes
		   CvMat** T_Mat, //the translation matrixes
		   CvMat** C_Mat, //the calibration matrixes
		   const int methode) //the chosen method 
{
    int N_x,N_y;
    char file[256];
	FILE *f = 0;
    CvMat *Z0=0, *U=0, *V=0, *W=0;
	CvMat *X=0,  *Y=0, *Xt=0, *Yt=0, *Zt=0;
	CvMat **M_curr = 0; 
	
	int dimension = 0;
	int opt_Z = 0;

    int xl0, yl0;

	//intrinisic parameters of the reference camera
	const double ref_params[4] = {f0ku, f0kv, u0, v0};
	
	double eta_curr = 1;
	for (int i = 1; i < pyramid_level_start; i++) {
		eta_curr *= PyramidEta;
	}
	if(ENABLE_ALPHA) {
		alpha = alpha*eta_curr;
		alpha_z = alpha_z*eta_curr;

		if(DIV_UVWZ){
			alpha_w = alpha;
			alpha_z = alpha;
		}else
			if(DIV_UVW)
				alpha_w = alpha;
			else
				if(DIV_UV_ZW)
					alpha_w = alpha_z;
				else
					alpha_w = alpha_w*eta_curr;
	}

	// computes the initial reference projection matrix
	int view = 0;
	M_curr = new CvMat*[NumOfViews];
	for(view = 0; view < NumOfViews; view++) {
		M_curr[view] = cvCreateMat(3,4,CV_64FC1);
		ComputeProjMat(M_curr[view],R_Mat[view], T_Mat[view], C_Mat[view], eta_curr);
		#ifdef DEBUG
			sprintf(file, "%s/ProjMat%d.txt", TxtResultsDir,view);
			f = fopen(file,"w");
			WriteRes(f,M_curr[view]);
			fclose(f);
		#endif
	}

	//Computes the center of projection for all views
	CvMat **COP = 0; 
	COP = new CvMat*[NumOfViews];
	ComputeCOP(COP, R_Mat, T_Mat);

    if(methode == 0){
        // computing all 4 data terms. U,V,W,Z optimizatio
		opt_Z = 1;
		dimension = 4;

       
    }
	if(methode == 1) {
		// U,V,W optimization 
		opt_Z = 0;
		dimension = 3;
	}
	

	int pyramidtype = CV_MAT_TYPE(Images_0_pyramid[0][pyramid_level_start-1]->type);

    //intial scene flow is set to the lowest resolution
    {
        CvSize size_start = cvGetSize(Images_0_pyramid[0][pyramid_level_start-1]);
		Z0 = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
		cvCopy(Z0_start, Z0);

		U = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
		V = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
		W = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
		cvCopy(U_start, U);
		cvCopy(V_start, V);
		cvCopy(W_start, W);

	}
	
    
    //Loop over the resolution levels
	for(int level = pyramid_level_start - 1; level >= pyramid_level_end - 1; level--){

        CvSize level_size = cvGetSize(Images_0_pyramid[0][level]);
		//Size of the current images
		N_x = level_size.width;
		N_y = level_size.height;

		//Maps used for the images warping : Initialization
		CvMat *MapIdentityX = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *MapIdentityY = cvCreateMat(N_y, N_x, CV_32FC1);

		//3D points at t+1
		Xt = cvCreateMat(N_y, N_x, CV_32FC1);
		Yt = cvCreateMat(N_y, N_x, CV_32FC1);
		Zt = cvCreateMat(N_y, N_x, CV_32FC1);
		
		//MeshGrid 
		for(yl0 = 0; yl0 < N_y; yl0++){
			for(xl0 = 0; xl0 < N_x; xl0++){
				cvmSet(MapIdentityX, yl0, xl0, xl0);
				cvmSet(MapIdentityY, yl0, xl0, yl0);
			}
		}

		////////////////////////////////////////////////////////
		//Warped Images
		CvMat **Images_0w = new CvMat*[(size_t)NumOfViews];
		CvMat **Images_tw = new CvMat*[NumOfViews];
		//Maps for warping
		CvMat **MapX_Img_0 = new CvMat*[NumOfViews];
		CvMat **MapY_Img_0 = new CvMat*[NumOfViews];
		CvMat **MapX_Img_0_inner = new CvMat*[NumOfViews];
		CvMat **MapY_Img_0_inner = new CvMat*[NumOfViews];
		CvMat **MapX_Img_t = new CvMat*[NumOfViews];
		CvMat **MapY_Img_t = new CvMat*[NumOfViews];
		CvMat **MapX_Img_t_inner = new CvMat*[NumOfViews];
		CvMat **MapY_Img_t_inner = new CvMat*[NumOfViews];

		for(view = 0; view < NumOfViews; view++) {
			Images_0w[view] = cvCreateMat(N_y, N_x, pyramidtype);
			MapX_Img_0[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_0[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapX_Img_0_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_0_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			Images_tw[view] = cvCreateMat(N_y, N_x, pyramidtype);
			MapX_Img_t[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_t[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapX_Img_t_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_t_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
		}
		CvMat *X = cvCreateMat(N_y, N_x, CV_32FC1);
        CvMat *Y = cvCreateMat(N_y, N_x, CV_32FC1);

		//Images gradients computation
		CvMat *derivative_tmp = 0;
		if (pyramidtype == CV_8UC1)
			derivative_tmp = cvCreateMat(N_y, N_x, CV_16SC1);
		CvMat **Images_0_X  = new CvMat*[NumOfViews];
		CvMat **Images_0_Y  = new CvMat*[NumOfViews];
		CvMat **Images_0w_X = new CvMat*[NumOfViews];
		CvMat **Images_0w_Y = new CvMat*[NumOfViews];
		CvMat **Images_t_X  = new CvMat*[NumOfViews];
		CvMat **Images_t_Y  = new CvMat*[NumOfViews];
		CvMat **Images_tw_X = new CvMat*[NumOfViews];
		CvMat **Images_tw_Y = new CvMat*[NumOfViews];
		CvMat **Proj_x0_x = 0, **Proj_x0_y = 0, **Proj_x0_z = 0;
		
		CvMat **Proj_y0_x = 0, **Proj_y0_y = 0, **Proj_y0_z = 0;
		CvMat **Proj_y0_u = 0, **Proj_y0_v = 0, **Proj_y0_w = 0;
		CvMat **Proj_xt_x = 0, **Proj_xt_y = 0, **Proj_xt_z = 0;
		CvMat **Proj_xt_u = 0, **Proj_xt_v = 0, **Proj_xt_w = 0;
		CvMat **Proj_yt_x = 0, **Proj_yt_y = 0, **Proj_yt_z = 0;
		CvMat **Proj_yt_u = 0, **Proj_yt_v = 0, **Proj_yt_w = 0;

		CvMat *X_x = 0, *X_y = 0, *X_z = 0;
		CvMat *Y_x = 0, *Y_y = 0, *Y_z = 0;
		CvMat *Xt_x = 0, *Xt_y = 0, *Xt_z = 0;
		CvMat *Yt_x = 0, *Yt_y = 0, *Yt_z = 0;
		CvMat *Zt_X = 0, *Zt_Y = 0;

		// world points X,Y derivatives with respect to the ref camera coordinates
		X_x = cvCreateMat(N_y, N_x, CV_32FC1);
		X_y = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_x = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_y = cvCreateMat(N_y, N_x, CV_32FC1);
        // world points Xt,Yt derivatives with respect to the ref camera coordinates
		Xt_x = cvCreateMat(N_y, N_x, CV_32FC1);
		Xt_y = cvCreateMat(N_y, N_x, CV_32FC1);
		Yt_x = cvCreateMat(N_y, N_x, CV_32FC1);
		Yt_y = cvCreateMat(N_y, N_x, CV_32FC1);
		Zt_X = cvCreateMat(N_y, N_x, CV_32FC1);
		Zt_Y = cvCreateMat(N_y, N_x, CV_32FC1);

		Proj_xt_x = new CvMat*[NumOfViews];
		Proj_xt_y = new CvMat*[NumOfViews];
		Proj_yt_x = new CvMat*[NumOfViews];
		Proj_yt_y = new CvMat*[NumOfViews];

		Proj_xt_u = new CvMat*[NumOfViews]; 
		Proj_xt_v = new CvMat*[NumOfViews];
		Proj_xt_w = new CvMat*[NumOfViews];
		
		Proj_yt_u = new CvMat*[NumOfViews]; 
		Proj_yt_v = new CvMat*[NumOfViews];
		Proj_yt_w = new CvMat*[NumOfViews];

		Proj_x0_x = new CvMat*[NumOfViews];
		Proj_x0_y = new CvMat*[NumOfViews];
		Proj_y0_x = new CvMat*[NumOfViews];
		Proj_y0_y = new CvMat*[NumOfViews];
		Proj_x0_z = new CvMat*[NumOfViews];
		Proj_y0_z = new CvMat*[NumOfViews];
		X_z = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_z = cvCreateMat(N_y, N_x, CV_32FC1);
			
		// if Z optimization 
		if(opt_Z){
			
			Xt_z = cvCreateMat(N_y, N_x, CV_32FC1);
		    Yt_z = cvCreateMat(N_y, N_x, CV_32FC1);
			
			Proj_xt_z = new CvMat*[NumOfViews];
			Proj_yt_z = new CvMat*[NumOfViews];
		}

		for(view = 0; view < NumOfViews; view++){
			Images_0_X[view] = cvCreateMat(N_y, N_x, CV_32FC1);
			Images_0_Y[view] = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_x0_x[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_x0_y[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_y0_x[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_y0_y[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
			Images_t_X[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
			Images_t_Y[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_xt_x[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_xt_y[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_yt_x[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_yt_y[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_xt_u[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_xt_v[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_xt_w[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_yt_u[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_yt_v[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
			Proj_yt_w[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
            //Compute gradients of the only non-warped image
			Derivative(Images_0_pyramid[view][level], 1, 0, derivative_tmp, Images_0_X[view]);
			Derivative(Images_0_pyramid[view][level], 0, 1, derivative_tmp, Images_0_Y[view]);

			if(opt_Z){
				Proj_x0_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
			    Proj_y0_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_xt_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_yt_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
			}
		}
	
      
        
		//M is the matrix of the final linear system and B the second member of this system
		CvMat *M = cvCreateMat(dimension * N_x * N_y, dimension + 4, CV_32FC1); 
	    CvMat *B = cvCreateMat(dimension * N_x * N_y,1,CV_32FC1);
		
        CvMat *disactivated_pixels_ref_flow = 0;
		CvMat **disactivated_pixels_flow = 0;
        CvMat **disactivated_pixels_Z0 = 0;
		CvMat **disactivated_pixels_Zt = 0;
		
		// occlution map for Images_tw[0]: between the reference images at time t and t+1
        disactivated_pixels_ref_flow = cvCreateMat(N_y, N_x, CV_8UC1);
        // occlution map for Itw: between the reference image and other wraped images at time t+1
		disactivated_pixels_flow = new CvMat*[NumOfViews];
		// occlution map for Itw: between the reference image and other wraped images at time t+1
		disactivated_pixels_Zt = new CvMat*[NumOfViews];
		// occlution map for I0w: between the reference image and other wraped images at time t
		if (opt_Z) {
            disactivated_pixels_Z0 = new CvMat*[NumOfViews];
        }
		
        //divergence coefficients
		CvMat *divergence_coefficients = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *divergence_coefficients_Z = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *divergence_coefficients_W = cvCreateMat(N_y, N_x, CV_32FC1);

        CvMat *Z0_X,*Z0_Y,*U_X, *U_Y, *V_X, *V_Y, *W_X, *W_Y;
        CvMat *dZ0, *dU, *dW, *dV;
        CvMat *dZ0_prec,  *dU_prec,  *dV_prec, *dW_prec;
        CvMat *dZ0_X,*dZ0_Y, *dU_X, *dU_Y, *dV_X, *dV_Y,*dW_X, *dW_Y;
        CvMat *Z0_prec,*U_prec, *V_prec, *W_prec;
		CvMat *Z_inner=0, *X_inner=0, *Y_inner=0;
		CvMat *Zt_inner=0, *Xt_inner=0, *Yt_inner=0;
		
		//flow
#ifdef DEBUG
            //saving the initial flow images
            if (SF_U_FileName) {
                SaveResults(SF_U_FileName, U);
            }
            if (SF_V_FileName) {
                SaveResults(SF_V_FileName, V);
            }
			if (SF_W_FileName) {
                SaveResults(SF_W_FileName, W);
            }
#endif

        U_prec = cvCreateMat(N_y, N_x, CV_32FC1);
        V_prec = cvCreateMat(N_y, N_x, CV_32FC1);
		W_prec = cvCreateMat(N_y, N_x, CV_32FC1);
   
        dU = cvCreateMat(N_y, N_x, CV_32FC1);
        dV = cvCreateMat(N_y, N_x, CV_32FC1);
		dW = cvCreateMat(N_y, N_x, CV_32FC1);
        dU_X = cvCreateMat(N_y, N_x, CV_32FC1);
        dU_Y = cvCreateMat(N_y, N_x, CV_32FC1);
        dV_X = cvCreateMat(N_y, N_x, CV_32FC1);
        dV_Y = cvCreateMat(N_y, N_x, CV_32FC1);
		dW_X = cvCreateMat(N_y, N_x, CV_32FC1);
        dW_Y = cvCreateMat(N_y, N_x, CV_32FC1);
     
        U_X = cvCreateMat(N_y, N_x, CV_32FC1);
        U_Y = cvCreateMat(N_y, N_x, CV_32FC1);
        V_X = cvCreateMat(N_y, N_x, CV_32FC1);
        V_Y = cvCreateMat(N_y, N_x, CV_32FC1);
		W_X = cvCreateMat(N_y, N_x, CV_32FC1);
        W_Y = cvCreateMat(N_y, N_x, CV_32FC1);
		Z0_X = cvCreateMat(N_y, N_x, CV_32FC1);
		Z0_Y = cvCreateMat(N_y, N_x, CV_32FC1);

        dU_prec = cvCreateMat(N_y, N_x, CV_32FC1);
        dV_prec = cvCreateMat(N_y, N_x, CV_32FC1);
		dW_prec = cvCreateMat(N_y, N_x, CV_32FC1);

		
		X_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		Zt_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		Xt_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		Yt_inner = cvCreateMat(N_y, N_x, CV_32FC1);

	    Z_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		if(opt_Z) {
			
			Z0_prec = cvCreateMat(N_y, N_x, CV_32FC1);
			if (SF_D0_FileName) {
				SaveResults(SF_D0_FileName, Z0);
			}	   
			dZ0 = cvCreateMat(N_y, N_x, CV_32FC1);
			dZ0_X = cvCreateMat(N_y, N_x, CV_32FC1);
			dZ0_Y = cvCreateMat(N_y, N_x, CV_32FC1);
			dZ0_prec = cvCreateMat(N_y, N_x, CV_32FC1);
		}

		//For the final linear system and SOR iterations
        CvMat *SOR = cvCreateMat(dimension*N_x*N_y,1, CV_32FC1);	
        CvMat *SOR_prec = cvCreateMat(dimension*N_x*N_y,1,CV_32FC1);

		//Main fixed point iterations, loop over optical flow/disparity values
        double norm = 0., norm_inner = 0., norm_SOR = 0.;
        int iter_SOR = 0, iter_inner = 0, iter_outer = 0;
        int first_iteration = 1;
		
		while((norm > epsilon_outer && iter_outer <= max_iter_outer) || iter_outer < 2 || first_iteration == 1){
			first_iteration = 0;
		    
			//Maps update
		
				ComputeXY(X, Y, Z0, MapIdentityX, MapIdentityY, N_x, N_y, ref_params, eta_curr);
				for(view = 1; view < NumOfViews; view++) {
					ComputeProjMaps(X, Y, Z0, R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_0[view], MapY_Img_0[view], eta_curr);
				}
	
			cvAdd(X, U, Xt);
			cvAdd(Y, V, Yt);
			cvAdd(Z0,W, Zt);
		    
			ComputeProjMaps(Xt, Yt, Zt, R_Mat[0], T_Mat[0], C_Mat[0], MapX_Img_t[0], MapY_Img_t[0], eta_curr);
			for(view = 1; view < NumOfViews; view++) {
					ComputeProjMaps(Xt, Yt, Zt, R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_t[view], MapY_Img_t[view], eta_curr);
			}
			
			cvZero(disactivated_pixels_ref_flow);
			ComputeOcclusion(MapX_Img_t[0], MapY_Img_t[0], Xt, Yt, Zt, disactivated_pixels_ref_flow);
			cvmSet8u_borders(disactivated_pixels_ref_flow, 127);
			
			if(opt_Z){
				for(view = 1; view < NumOfViews; view++){
					disactivated_pixels_Z0[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					cvZero(disactivated_pixels_Z0[view]);
					ComputeOcclusion(MapX_Img_0[view], MapY_Img_0[view], X, Y, Z0, disactivated_pixels_Z0[view], COP[view]);
					cvmSet8u_borders(disactivated_pixels_Z0[view], 127);
				}
			}

			// not so sure about how to compute right flow occlusion...
			//CHECK ME: notice that the computation is done with respect reference coordinates:
			// Namely : the occloution is computed between the left and right image (Irtw). 
			
			for(view = 1; view < NumOfViews; view++){
				disactivated_pixels_flow[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
				disactivated_pixels_Zt[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
				cvZero(disactivated_pixels_flow[view]);
				if (BLACK_LINE && view == 1){
					CvMat *img = cvLoadImageM(Zt_initFile, CV_LOAD_IMAGE_GRAYSCALE);
					cvResize(img, disactivated_pixels_flow[view],CV_INTER_AREA);
					cvReleaseMat(&img);
				}
				ComputeOcclusion(MapX_Img_t[view], MapY_Img_t[view], Xt, Yt, Zt, disactivated_pixels_flow[view], COP[view]);
				cvmSet8u_borders(disactivated_pixels_flow[view], 127);
				cvCopy(disactivated_pixels_flow[view],disactivated_pixels_Zt[view]);
			}
			

#ifdef DEBUG
			for(view = 1; view < NumOfViews; view++) {
				if(opt_Z && SF_disactivated_Surf_1_FileName){
				sprintf(file, "%sv%d_l%d.txt", SF_disactivated_Surf_1_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_Z0[view]);
				fclose(f);
				}
				
				if(SF_disactivated_flow_FileName){
					sprintf(file, "%sv%d_l%d.txt",SF_disactivated_flow_FileName,view,level);
					f = fopen(file,"w");
					WriteRes8u(f,disactivated_pixels_flow[view]);
					fclose(f);
				}

				if(SF_disactivated_Surf_2_FileName){
					sprintf(file, "%sv%d_l%d.txt",SF_disactivated_Surf_2_FileName,view,level);
					f = fopen(file,"w");
					WriteRes8u(f,disactivated_pixels_Zt[view]);
					fclose(f);
				}

			}
			if(SF_disactivated_ref_flow_FileName){
				sprintf(file, "%s_l%d.txt",SF_disactivated_ref_flow_FileName,level);
				f = fopen(file,"w");
				WriteRes8u	(f,disactivated_pixels_ref_flow);
				fclose(f);
			}
#endif

			cvRemap(Images_t_pyramid[0][level], Images_tw[0], MapX_Img_t[0], MapY_Img_t[0]);
			SmoothImage(Images_tw[0]);

			Images_tw_X[0] = cvCreateMat(N_y, N_x, CV_32FC1);
			Images_tw_Y[0] = cvCreateMat(N_y, N_x, CV_32FC1);
			Derivative(Images_tw[0] , 1, 0, derivative_tmp, Images_tw_X[0]);
			Derivative(Images_tw[0] , 0, 1, derivative_tmp, Images_tw_Y[0]);
			
			for(view = 1; view < NumOfViews; view++) {
				cvRemap(Images_0_pyramid[view][level], Images_0w[view], MapX_Img_0[view], MapY_Img_0[view]);
				SmoothImage(Images_0w[view]);
				if (SF_I0_Warped_FileName) {
					sprintf(file, "%s%d.%s", SF_I0_Warped_FileName, view, IMG_TYPE);
					SaveImage(file, Images_0w[view]);
				}

				Images_0w_X[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Images_0w_Y[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Derivative(Images_0w[view], 1, 0, derivative_tmp, Images_0w_X[view]);
				Derivative(Images_0w[view], 0, 1, derivative_tmp, Images_0w_Y[view]);
			}
			
			
			for(view = 1; view < NumOfViews; view++) {
				cvRemap(Images_t_pyramid[view][level], Images_tw[view], MapX_Img_t[view], MapY_Img_t[view]);
				SmoothImage(Images_tw[view]);
				if (SF_It_Warped_FileName) {
					sprintf(file, "%s%d.%s", SF_It_Warped_FileName, view, IMG_TYPE);
					SaveImage(file, Images_tw[view]);
				}
		
				Images_tw_X[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Images_tw_Y[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Derivative(Images_tw[view], 1, 0, derivative_tmp, Images_tw_X[view]);
				Derivative(Images_tw[view], 0, 1, derivative_tmp, Images_tw_Y[view]);
			}
			
			//Flow gradients computation
		
				GradientFast_32f(Z0, Z0_X, Z0_Y);
				Gradient3DPoints(MapIdentityX, MapIdentityY, Z0, Z0_X, Z0_Y, eta_curr, X_x, X_y, X_z, Y_x, Y_y, Y_z);
			
			
			GradientFast_32f(U, U_X, U_Y);
			GradientFast_32f(V, V_X, V_Y);
			GradientFast_32f(W, W_X, W_Y);
			cvAdd(U_X,X_x,Xt_x);
			cvAdd(U_Y,X_y,Xt_y);
			cvAdd(V_X,Y_x,Yt_x);
			cvAdd(V_Y,Y_y,Yt_y);
			cvAdd(W_X,Z0_X,Zt_X);
			cvAdd(W_Y,Z0_Y,Zt_Y);
		
			
			//computes the projected coordinates derivatives at time t
			for(view = 1; view < NumOfViews; view++) {
				GradientProjCoor(X, Y, Z0, X_x, Y_x, Z0_X, M_curr[view], Proj_x0_x[view], 0);
				GradientProjCoor(X, Y, Z0, X_y, Y_y, Z0_Y, M_curr[view], Proj_x0_y[view], 0);
				GradientProjCoor(X, Y, Z0, X_x, Y_x, Z0_X, M_curr[view], Proj_y0_x[view], 1);
				GradientProjCoor(X, Y, Z0, X_y, Y_y, Z0_Y, M_curr[view], Proj_y0_y[view], 1);
				if(opt_Z){
					CvMat *temp = cvCreateMat(N_y, N_x, CV_32FC1); 
					cvSet(temp, cvScalarAll(1),0);
					GradientProjCoor(X, Y, Z0, X_z, Y_z, temp, M_curr[view], Proj_x0_z[view], 0);
					GradientProjCoor(X, Y, Z0, X_z, Y_z, temp, M_curr[view], Proj_y0_z[view], 1);
					cvReleaseMat(&temp);
				}
			}

			CvMat *ZerosDerive = cvCreateMat(N_y, N_x, CV_32FC1); 
			CvMat *OnesDerive = cvCreateMat(N_y, N_x, CV_32FC1); 
			cvSet(OnesDerive, cvScalarAll(1),0);
			cvZero(ZerosDerive);

			GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[0], Proj_xt_u[0], 0);
			GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[0], Proj_xt_v[0], 0);
			GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[0], Proj_xt_w[0], 0);

			GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[0], Proj_yt_u[0], 1);
			GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[0], Proj_yt_v[0], 1);
			GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[0], Proj_yt_w[0], 1);
			
		    GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X,  M_curr[0], Proj_xt_x[0], 0);
			GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y,  M_curr[0], Proj_xt_y[0], 0);
			GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X,  M_curr[0], Proj_yt_x[0], 1);
			GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y,  M_curr[0], Proj_yt_y[0], 1);

			for(view = 1; view < NumOfViews; view++) {
				GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[view], Proj_xt_u[view], 0);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[view], Proj_xt_v[view], 0);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[view], Proj_xt_w[view], 0);
				
				GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[view], Proj_yt_u[view], 1);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[view], Proj_yt_v[view], 1);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[view], Proj_yt_w[view], 1);

				GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X, M_curr[view], Proj_xt_x[view], 0);
				GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y, M_curr[view], Proj_xt_y[view], 0);
				GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X, M_curr[view], Proj_yt_x[view], 1);
				GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y, M_curr[view], Proj_yt_y[view], 1);

				if(opt_Z) {
					CvMat *temp = cvCreateMat(N_y, N_x, CV_32FC1); 
					cvSet(temp, cvScalarAll(1),0);
					GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp, M_curr[view], Proj_xt_z[view], 0);
					GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp, M_curr[view], Proj_yt_z[view], 1);

					cvSet(temp, cvScalarAll(1),0);
					GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp,  M_curr[0], Proj_xt_z[0], 0);
					GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp,  M_curr[0], Proj_yt_z[0], 1);

					cvReleaseMat(&temp);
				}
			}

			cvReleaseMat(&ZerosDerive);
			cvReleaseMat(&OnesDerive);
			
			//Increments initialization
			cvSetZero(dU);
			cvSetZero(dV);
			cvSetZero(dW);
			
			if(opt_Z){
				cvSetZero(dZ0);
			}

		  	norm_inner = 0.;
			first_iteration = 1;
			iter_inner = 0;
			//secondary fixed point iterations
			while((norm_inner > epsilon_inner && iter_inner <= max_iter_inner) || first_iteration == 1){
				first_iteration = 0;
	                
				//Coordinates maps update
				cvCopy(X, X_inner);
				cvCopy(Y, Y_inner);
				cvCopy(Z0,Z_inner);
				// need to update Z0 and re-compute the world coordinates at time t
				if(opt_Z){
					cvAdd(Z_inner, dZ0, Z_inner);}
				    ComputeXY(X_inner, Y_inner, Z_inner, MapIdentityX, MapIdentityY, N_x, N_y, ref_params, eta_curr);
					for(view = 1; view < NumOfViews; view++){
						ComputeProjMaps(X_inner, Y_inner, Z_inner, R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_0_inner[view], MapY_Img_0_inner[view], eta_curr);
					}
				/*}else{
					MapX_Img_0 =  MapX_Img_0_inner;
					MapY_Img_0 =  MapY_Img_0_inner ;
				}*/

				
				// need to update the world coordinates at time t+1
				// Xt_inner = X_inner + U + dU
				cvAdd(X_inner, U, Xt_inner);
				cvAdd(Xt_inner, dU, Xt_inner);
				// Yt_inner = Y_inner + V + dV
				cvAdd(Y_inner, V, Yt_inner);
				cvAdd(Yt_inner, dV, Yt_inner);
				// Zt_inner = (Z0+dZ0)+W+dW
				cvAdd(Z_inner, W, Zt_inner);
				cvAdd(Zt_inner, dW, Zt_inner);
				
				ComputeProjMaps(Xt_inner,Yt_inner,Zt_inner,R_Mat[0], T_Mat[0], C_Mat[0], MapX_Img_t_inner[0], MapY_Img_t_inner[0], eta_curr);
				for(view = 1; view < NumOfViews; view++) {
					ComputeProjMaps(Xt_inner,Yt_inner,Zt_inner,R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_t_inner[view], MapY_Img_t_inner[view], eta_curr);
				}
				
								
				//Flow increments gradients computation
				GradientFast_32f(dU, dU_X, dU_Y);
				GradientFast_32f(dV, dV_X, dV_Y);
				GradientFast_32f(dW, dW_X, dW_Y);
				
				if(opt_Z){
					GradientFast_32f(dZ0, dZ0_X, dZ0_Y);
				}

				for(yl0 = 0; yl0 < N_y; yl0++){
					for(xl0 = 0; xl0 < N_x; xl0++){
				    double z0 = 0.;
					double z0_X = 0.;
					double z0_Y = 0.;
					double dz0_X = 0.;
					double dz0_Y = 0.;
					double s2 = 0.;
					double s3 = 0.;
					double s4 = 0.;
					double sZ = 0.;
					
					double du_X = cvmGet(dU_X, yl0, xl0);
					double du_Y = cvmGet(dU_Y, yl0, xl0);
					double u_X = cvmGet(U_X, yl0, xl0);
					double u_Y = cvmGet(U_Y, yl0, xl0);
					double dv_X = cvmGet(dV_X, yl0, xl0);
					double dv_Y = cvmGet(dV_Y, yl0, xl0);
					double v_X = cvmGet(V_X, yl0, xl0);
					double v_Y = cvmGet(V_Y, yl0, xl0);
					double dw_X = cvmGet(dW_X, yl0, xl0);
					double dw_Y = cvmGet(dW_Y, yl0, xl0);
					double w_X = cvmGet(W_X, yl0, xl0);
					double w_Y = cvmGet(W_Y, yl0, xl0);

					if(opt_Z){
						z0_X = cvmGet(Z0_X, yl0, xl0);
						z0_Y = cvmGet(Z0_Y, yl0, xl0);
						dz0_X = cvmGet(dZ0_X, yl0, xl0);
						dz0_Y = cvmGet(dZ0_Y, yl0, xl0);
					}
					
					if(DIV_UVWZ & opt_Z){
						s2 = mu_u*(square(u_X + du_X) + square(u_Y + du_Y)) +
							mu_v*(square(v_X + dv_X) + square(v_Y + dv_Y)) + 
							mu_w*(square(w_X + dw_X) + square(w_Y + dw_Y)) + 
							mu_z*(square(z0_X + dz0_X) + square(z0_Y + dz0_Y)) ;
						s3 = s2;
						s4 = s2;
					}else
						if(DIV_UVW){
							s2 = mu_u*(square(u_X + du_X) + square(u_Y + du_Y)) +
								mu_v*(square(v_X + dv_X) + square(v_Y + dv_Y)) + 
								mu_w*(square(w_X + dw_X) + square(w_Y + dw_Y));
							s3 = s2;
							if(opt_Z){
								s4 = mu_z*(square(z0_X + dz0_X) + square(z0_Y + dz0_Y));
							}
						}
						else{
							s2 = (square(u_X + du_X) + square(u_Y + du_Y) +
								square(v_X + dv_X) + square(v_Y + dv_Y)) ;
							if(opt_Z){

								if(DIV_UV_ZW){
									s3 =  mu_w*(square(w_X + dw_X) + square(w_Y + dw_Y)) + 
										mu_z*(square(z0_Y + dz0_Y) + square(z0_X + dz0_X));
									s4 = s3;
								}else{
									s3 =  mu_w*(square(w_X + dw_X) + square(w_Y + dw_Y)) ;
									s4 = mu_z*(square(z0_X + dz0_X) + square(z0_Y + dz0_Y));
								}
							}else{
								s3 =  mu_w*(square(w_X + dw_X) + square(w_Y + dw_Y));
							}
						}
						
					double a = DPsiSmooth(s2);
					cvmSet(divergence_coefficients, yl0, xl0, a);
					
					a = DPsiSmoothW(s3);
					cvmSet(divergence_coefficients_W, yl0, xl0, a);

					a = DPsiSmoothW(s4);
					cvmSet(divergence_coefficients_Z, yl0, xl0, a);
					}
				
				}
				//Computing M and B for the final linear system 
				cvSetZero(M);
				cvSetZero(B);
				
				int i;
				
				//The dimension*N_x first lines in M correspond to the points on the y = 0 on the image
				//tali: The matrix indexes are a row-stack indexes
				int inf = 0;
				int sup = dimension*N_x;
				for (i = inf; i < sup; i++) {
					cvmSet(M, i, 0, -1./h); 
					cvmSet(M, i, 1,  1./h);	
				}
				//Points on the lower horizontal side of the image      
				inf = dimension*N_x*(N_y-1);
				sup = dimension*N_x*N_y;
			
				for (i = inf; i < sup; i++) {
					cvmSet(M, i, 0,  1./h); 
					cvmSet(M, i, 1, -1./h);

				}

				inf = dimension*N_x;
				sup = dimension*N_x*(N_y-1)-1;
			
				//We fill in the lines of the matrix M corresponding to images pixels except points on the upper and lower horizontal side
				for(i = inf; i <= sup; i+=dimension){
					yl0 = i/(dimension*N_x);
					xl0 = (i % (inf))/dimension;
		           
					if(xl0 == 0){ //if point on the left side of image
						for(int d=0; d<dimension; d++) {
							cvmSet(M,i+d,0,-1./h);
							cvmSet(M,i+d,1, 1./h);
							
						}
					}
					else if(xl0 == (N_x-1)){//if point on the right side
						for(int d=0; d<dimension; d++) {
							cvmSet(M,i+d,0, 1./h);
							cvmSet(M,i+d,1,-1./h);  
						}
					}
					else{
						//Pixel inside the integration domain 
						if(methode == 0){
							double il0 = 0.;
							double il0_X = 0., il0_Y = 0., il0w_Z = 0.;
							double i0 = 0.;
							double i0_X = 0., i0_Y = 0.;
							double i0w_X = 0., i0w_Y = 0., i0w_Z = 0.; 
							double proj_x0_x = 0., proj_x0_y = 0;
							double proj_y0_x = 0., proj_y0_y = 0;
							double proj_x0_z = 0., proj_y0_z = 0;
							double ilt = 0.;
							double ilt_X = 0., ilt_Y = 0.;
							double iltw_X = 0., iltw_Y = 0., iltw_Z = 0.;
							double iltw_U = 0., iltw_V = 0., iltw_W = 0.;
							double proj_xlt_x = 0., proj_xlt_y = 0., proj_xlt_z = 0.;
							double proj_xlt_u = 0., proj_xlt_v = 0., proj_xlt_w = 0.;
							double proj_ylt_x = 0., proj_ylt_y = 0., proj_ylt_z = 0.;
							double proj_ylt_u = 0., proj_ylt_v = 0., proj_ylt_w = 0.;
							double it = 0.;
							double it_X = 0., it_Y = 0.;
							double itw_X = 0., itw_Y = 0., itw_Z = 0.;
							double itw_U = 0., itw_V = 0., itw_W = 0.;
							double proj_xt_x = 0., proj_xt_y = 0., proj_xt_z = 0.;
							double proj_xt_u = 0., proj_xt_v = 0., proj_xt_w = 0.;
							double proj_yt_x = 0., proj_yt_y = 0., proj_yt_z = 0.;
							double proj_yt_u = 0., proj_yt_v = 0., proj_yt_w = 0.;
							double du = 0, dv = 0, dw = 0, dz = 0;

							//interpolated divergence coefficients
							double b_f_1, b_f_2, b_f_3, b_f_4;
							double b_fz_1, b_fz_2, b_fz_3, b_fz_4;
							double b_fw_1, b_fw_2, b_fw_3, b_fw_4;
						
							// Half pixel interpolation for the divergence coefficients 
							divergence_coefficients_interpolate(divergence_coefficients, yl0, xl0, &b_f_1, &b_f_2, &b_f_3, &b_f_4);
							divergence_coefficients_interpolate(divergence_coefficients_W, yl0, xl0, &b_fw_1, &b_fw_2, &b_fw_3, &b_fw_4);
							divergence_coefficients_interpolate(divergence_coefficients_Z, yl0, xl0, &b_fz_1, &b_fz_2, &b_fz_3, &b_fz_4);
							
	                        //Matrix terms
							double a_u_1, a_u_2, a_u_3, a_u_4, a_v_1, a_v_2, a_v_3, a_v_4; 
							double a_w_1, a_w_2, a_w_3, a_w_4,a_z_1, a_z_2, a_z_3, a_z_4;
		    
							a_u_1 = (alpha*mu_u/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_v_1 = 0;
							a_w_1 = 0;	  
							a_z_1 = 0;
		    
							a_u_2 = 0;
							a_v_2 = (alpha*mu_v/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_w_2 = 0;	  
							a_z_2 = 0;
		    
							a_u_3 = 0;
							a_v_3 = 0;

							a_w_3 = (alpha_w*mu_w/(4.*h*h))*(b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4);	  
							a_z_3 = 0; 

							a_u_4 = 0;
							a_v_4 = 0;
							a_w_4 = 0; 
							a_z_4 = (alpha_z*(mu_z)/(4.*h*h))*(b_fz_1 + b_fz_2 + b_fz_3 + b_fz_4);
	
							//second member for the first equation
							double bu0 = cvmGet(U, yl0, xl0);
							double bu1 = cvmGet(U, yl0, xl0+1);
							double bu2 = cvmGet(U, yl0, xl0-1);
							double bu3 = cvmGet(U,yl0+1,xl0);
							double bu4 = cvmGet(U,yl0-1,xl0);
							//Crop(bu0, N_x);
							double bv0 = cvmGet(V, yl0, xl0);
							double bv1 = cvmGet(V, yl0, xl0+1);
							double bv2 = cvmGet(V, yl0, xl0-1);
							double bv3 = cvmGet(V,yl0+1,xl0);
							double bv4 = cvmGet(V,yl0-1,xl0);
							//Crop(bw0, N_x);
							double bw0 = cvmGet(W, yl0, xl0);
							double bw1 = cvmGet(W, yl0, xl0+1);
							double bw2 = cvmGet(W, yl0, xl0-1);
							double bw3 = cvmGet(W,yl0+1,xl0);
							double bw4 = cvmGet(W,yl0-1,xl0);

							double bzi0 = cvmGet(Z0, yl0, xl0);
							double bzi1 = cvmGet(Z0, yl0, xl0+1);
							double bzi2 = cvmGet(Z0, yl0, xl0-1);
							double bzi3 = cvmGet(Z0,yl0+1,xl0);
							double bzi4 = cvmGet(Z0,yl0-1,xl0);

							//tali: bu is the discretization of div(Psi()*grad(u))
							double bu = (alpha*mu_u/(4*h*h)) *
								(b_f_1 * bu1 +
								 b_f_2 * bu2 +
								 b_f_3 * bu3 +
								 b_f_4 * bu4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bu0);

							//Crop(bv0, N_y);
							double bv = (alpha*mu_v/(4*h*h)) *
								(b_f_1 * bv1 +
								 b_f_2 * bv2 +
								 b_f_3 * bv3 +
								 b_f_4 * bv4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bv0);
							//Crop(bv0, N_y);
							double bw = (alpha_w*mu_w/(4*h*h)) *
								(b_fw_1 * bw1 +
								 b_fw_2 * bw2 +
								 b_fw_3 * bw3 +
								 b_fw_4 * bw4
								 - (b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4) * bw0);
							double bz = (alpha_z*mu_z/(4*h*h)) *
								(b_fz_1 * bzi1 +
								 b_fz_2 * bzi2 +
								 b_fz_3 * bzi3 +
								 b_fz_4 * bzi4
								 - (b_fz_1 + b_fz_2 + b_fz_3 + b_fz_4)* bzi0);
							
						    //Pixels out must not be taken into account in the energy data term
							int pixel_do_ref_flow = 0;
							int pixel_do_flow_views[MAX_NUM_OF_VIEWS] = {0};
							int pixel_doZ0_views[MAX_NUM_OF_VIEWS] = {0};
							int pixel_doZt_views[MAX_NUM_OF_VIEWS] = {0};
							
							pixel_do_ref_flow = !cvmGet8u(disactivated_pixels_ref_flow,yl0,xl0);
							
							for(view=1; view < NumOfViews; view++) {
								pixel_do_flow_views[view] = !cvmGet8u(disactivated_pixels_flow[view],yl0,xl0);
								pixel_doZt_views[view] = !cvmGet8u(disactivated_pixels_Zt[view],yl0,xl0);
								pixel_doZ0_views[view] = !cvmGet8u(disactivated_pixels_Z0[view],yl0,xl0);
																
								//is the point visible at time t+1 ?
								if (pixel_do_ref_flow || pixel_doZt_views[view]) {
									double xlt = cvmGet(MapX_Img_t_inner[0], yl0, xl0);
									double ylt = cvmGet(MapY_Img_t_inner[0], yl0, xl0);
									if(xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h ||
									   ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h){
										   if(pixel_do_ref_flow){
												cvmSet8u(disactivated_pixels_ref_flow,yl0,xl0,127);
												pixel_do_ref_flow = 0;
										   }
										   if(pixel_doZt_views[view]){
												cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,127);
												pixel_doZt_views[view] = 0;
										   }
									}
								}
								//On the image at time t
								if (pixel_doZ0_views[view] || pixel_do_flow_views[view]) {
									double x0 = cvmGet(MapX_Img_0_inner[view], yl0, xl0);
									double y0 = cvmGet(MapY_Img_0_inner[view], yl0, xl0);
									if(x0 >= (N_x - 1 - DISACTIVATE_BORDER)*h || x0 <= DISACTIVATE_BORDER*h ||
									   y0 >= (N_y - 1 - DISACTIVATE_BORDER)*h || y0 <= DISACTIVATE_BORDER*h) {
										if (pixel_doZ0_views[view]) {
											cvmSet8u(disactivated_pixels_Z0[view],yl0,xl0,127);
											pixel_doZ0_views[view] = 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,127);
											pixel_do_flow_views[view] = 0;
										}
									}
								}

								//On the right image at time t+1
								if (pixel_doZt_views[view] || pixel_do_flow_views[view]) {
									double xt = cvmGet(MapX_Img_t_inner[view], yl0, xl0);
									double yt = cvmGet(MapY_Img_t_inner[view], yl0, xl0);
									if(xt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xt <= DISACTIVATE_BORDER*h ||
									   yt >= (N_y - 1 - DISACTIVATE_BORDER)*h || yt < DISACTIVATE_BORDER*h) {
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,127);
											pixel_doZt_views[view] = 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,127);
											pixel_do_flow_views[view] = 0;
										}
									}
								}
							}
							int pixel_doZ0 = 0;
							int pixel_doZt = 0;
							int pixel_do_flow = 0;
							for(view = 1; view < NumOfViews; view++){
								pixel_doZ0 += pixel_doZ0_views[view];
								pixel_doZt += pixel_doZt_views[view];
								pixel_do_flow += pixel_do_flow_views[view];
							}
							
							du = cvmGet(dU, yl0, xl0);
							dv = cvmGet(dV, yl0, xl0);
							dw = cvmGet(dW, yl0, xl0);
							
							if (pixel_doZ0) {
								dz = cvmGet(dZ0, yl0, xl0);
							}
							//Data term estimation
							if (pixel_do_ref_flow || pixel_doZ0) {
								il0 = cvGetReal2D(Images_0_pyramid[0][level], yl0, xl0); 
								il0_X = cvmGet(Images_0_X[0],  yl0, xl0);
								il0_Y = cvmGet(Images_0_Y[0],  yl0, xl0);
							}
							for(view = 1; view < NumOfViews; view++){
								if (pixel_do_flow_views[view] || pixel_doZ0_views[view]) {
									i0 = cvGetReal2D(Images_0w[view], yl0, xl0);
									i0w_X = cvmGet(Images_0w_X[view],  yl0, xl0);
									i0w_Y = cvmGet(Images_0w_Y[view],  yl0, xl0);
									proj_x0_x = cvmGet(Proj_x0_x[view], yl0, xl0);
									proj_x0_y = cvmGet(Proj_x0_y[view], yl0, xl0);
									proj_y0_x = cvmGet(Proj_y0_x[view], yl0, xl0);
									proj_y0_y = cvmGet(Proj_y0_y[view], yl0, xl0);
									if(pixel_doZ0) {
										proj_x0_z = cvmGet(Proj_x0_z[view], yl0, xl0);
										proj_y0_z = cvmGet(Proj_y0_z[view], yl0, xl0);
									}
									double det = proj_x0_x*proj_y0_y - proj_x0_y*proj_y0_x;
									if (det <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_doZ0_views[view]) {
											cvmSet8u(disactivated_pixels_Z0[view],yl0,xl0,191);
											pixel_doZ0_views[view]= 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,191);
											pixel_do_flow_views[view] = 0;
										}
									}
									else {
										i0_X = (i0w_X*proj_y0_y - i0w_Y*proj_y0_x)/det;
										i0_Y = (i0w_Y*proj_x0_x - i0w_X*proj_x0_y)/det;
										if(opt_Z) {
											i0w_Z = i0_X*proj_x0_z + i0_Y*proj_y0_z;
										}
									}
								}
								if (pixel_do_ref_flow || pixel_doZt_views[view]) {
									// Wraped image values
									ilt = cvGetReal2D(Images_tw[0], yl0, xl0);
									iltw_X = cvmGet(Images_tw_X[0], yl0, xl0);
									iltw_Y = cvmGet(Images_tw_Y[0],  yl0, xl0);
									// projected coordinated derivatives
									proj_xlt_x = cvmGet(Proj_xt_x[0], yl0, xl0);
									proj_xlt_y = cvmGet(Proj_xt_y[0], yl0, xl0);
									proj_ylt_x = cvmGet(Proj_yt_x[0], yl0, xl0);
									proj_ylt_y = cvmGet(Proj_yt_y[0], yl0, xl0);
									proj_xlt_z = cvmGet(Proj_xt_z[0], yl0, xl0);
									proj_ylt_z = cvmGet(Proj_yt_z[0], yl0, xl0);
									proj_xlt_u = cvmGet(Proj_xt_u[0], yl0, xl0);
									proj_ylt_u = cvmGet(Proj_yt_u[0], yl0, xl0);
									proj_xlt_v = cvmGet(Proj_xt_v[0], yl0, xl0);
									proj_ylt_v = cvmGet(Proj_yt_v[0], yl0, xl0);
									proj_xlt_w = cvmGet(Proj_xt_w[0], yl0, xl0);
									proj_ylt_w = cvmGet(Proj_yt_w[0], yl0, xl0);
								
									double det = proj_xlt_x*proj_ylt_y - proj_xlt_y*proj_ylt_x;
									if ((det) <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_do_ref_flow) {
											cvmSet8u(disactivated_pixels_ref_flow,yl0,xl0,191);
											pixel_do_ref_flow = 0;
										}
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,191);
											pixel_doZt_views[view] = 0;
										}
									}
									else {
										ilt_X = (iltw_X*proj_ylt_y - iltw_Y*proj_ylt_x)/det ;
										ilt_Y = (iltw_Y*proj_xlt_x - iltw_X*proj_xlt_y)/det ;
										iltw_Z = ilt_X*proj_xlt_z + ilt_Y*proj_ylt_z ;
										iltw_U = ilt_X*proj_xlt_u + ilt_Y*proj_ylt_u ; 
										iltw_V = ilt_X*proj_xlt_v + ilt_Y*proj_ylt_v ; 
										iltw_W = ilt_X*proj_xlt_w + ilt_Y*proj_ylt_w ;
									}
								}
								if (pixel_doZt_views[view] || pixel_doZt_views[view]) {
									it = cvGetReal2D(Images_tw[view],yl0,xl0);
									itw_X = cvmGet(Images_tw_X[view],  yl0, xl0);
									itw_Y = cvmGet(Images_tw_Y[view],  yl0, xl0);
									
									proj_xt_x = cvmGet(Proj_xt_x[view], yl0, xl0);
									proj_xt_y = cvmGet(Proj_xt_y[view], yl0, xl0);
									proj_yt_x = cvmGet(Proj_yt_x[view], yl0, xl0);
									proj_yt_y = cvmGet(Proj_yt_y[view], yl0, xl0);
									
									proj_xt_u = cvmGet(Proj_xt_u[view], yl0, xl0);
									proj_xt_v = cvmGet(Proj_xt_v[view], yl0, xl0);
									proj_xt_w = cvmGet(Proj_xt_w[view], yl0, xl0);
									proj_yt_u = cvmGet(Proj_yt_u[view], yl0, xl0);
									proj_yt_v = cvmGet(Proj_yt_v[view], yl0, xl0);
									proj_yt_w = cvmGet(Proj_yt_w[view], yl0, xl0);
									
									if(pixel_doZ0_views[view]) {
										proj_xt_z = cvmGet(Proj_xt_z[view], yl0, xl0);
										proj_yt_z = cvmGet(Proj_yt_z[view], yl0, xl0);
									}
									double det = proj_xt_x*proj_yt_y - proj_xt_y*proj_yt_x;
									if ((det) <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,191);
											pixel_doZt_views[view] = 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,191);
											pixel_do_flow_views[view] = 0;
										}
									}
									else {
										it_X = (itw_X*proj_yt_y - itw_Y*proj_yt_x)/det;
										it_Y = (itw_Y*proj_xt_x - itw_X*proj_xt_y)/det;
										if(pixel_doZ0_views[view]) {
											itw_Z = it_X*proj_xt_z + it_Y*proj_yt_z;
										}
										itw_U = it_X*proj_xt_u + it_Y*proj_yt_u ; 
										itw_V = it_X*proj_xt_v + it_Y*proj_yt_v ; 
										itw_W = it_X*proj_xt_w + it_Y*proj_yt_w ;
									}
								}
								if (pixel_do_flow_views[view]) {
									double a = square(it + itw_U*du + itw_V*dv + itw_W*dw + itw_Z*dz - 
													  (i0 + i0w_Z*dz));
									double psi_flow = DPsiData(a);
									//u equation
									a_u_1 += psi_flow*(itw_U*itw_U);
									a_v_1 += psi_flow*(itw_V*itw_U);
									a_w_1 += psi_flow*(itw_W*itw_U);
									a_z_1 += psi_flow*((itw_Z - i0w_Z)*itw_U);
									//v equation
									a_u_2 += psi_flow*(itw_U*itw_V);
									a_v_2 += psi_flow*(itw_V*itw_V);
									a_w_2 += psi_flow*(itw_W*itw_V);
									a_z_2 += psi_flow*((itw_Z - i0w_Z)*itw_V);
									//w equation
									a_u_3 += psi_flow*(itw_U*itw_W);
									a_v_3 += psi_flow*(itw_V*itw_W);
									a_w_3 += psi_flow*(itw_W*itw_W);
									a_z_3 += psi_flow*((itw_Z - i0w_Z)*itw_W);
									//z equation
									a_u_4 += psi_flow*(itw_U*(itw_Z - i0w_Z));
									a_v_4 += psi_flow*(itw_V*(itw_Z - i0w_Z));
									a_w_4 += psi_flow*(itw_W*(itw_Z - i0w_Z));
									a_z_4 += psi_flow*((itw_Z - i0w_Z)*(itw_Z - i0w_Z));
									// second term
									bu += - psi_flow*(itw_U*(it - i0));		   
									bv += - psi_flow*(itw_V*(it - i0));
									bw += - psi_flow*(itw_W*(it - i0));
									bz += - psi_flow*((itw_Z - i0w_Z)*(i0 - it));
								}
								if (pixel_doZ0_views[view]) {
									double a = square(i0 + i0w_Z*dz - il0);
									double psi_stereo_init = DPsiData(a);
			 
									//z equation
									a_z_4 += psi_stereo_init*(i0w_Z*i0w_Z); 

									// second term
									bz += -psi_stereo_init*((i0 - il0)*i0w_Z);
								}
								if (pixel_doZt_views[view]) {
									double a = square(it + itw_U*du + itw_V*dv + itw_W*dw +itw_Z*dz - 
													 (ilt + iltw_U*du + iltw_V*dv + iltw_W*dw +iltw_Z*dz));
									double psi_stereo = DPsiData(a);
									//u equation
									a_u_1  += psi_stereo*((itw_U - iltw_U)*(itw_U - iltw_U));
									a_v_1  += psi_stereo*((itw_V - iltw_V)*(itw_U - iltw_U));
									a_w_1  += psi_stereo*((itw_W - iltw_W)*(itw_U - iltw_U));
									a_z_1  += psi_stereo*((itw_Z - iltw_Z)*(itw_U - iltw_U));
									//v equation
									a_u_2  += psi_stereo*((itw_U - iltw_U)*(itw_V - iltw_V));
									a_v_2  += psi_stereo*((itw_V - iltw_V)*(itw_V - iltw_V));
									a_w_2  += psi_stereo*((itw_W - iltw_W)*(itw_V - iltw_V));
									a_z_2  += psi_stereo*((itw_Z - iltw_Z)*(itw_V - iltw_V));
									//w equation
									a_u_3  += psi_stereo*((itw_U - iltw_U)*(itw_W - iltw_W));
									a_v_3  += psi_stereo*((itw_V - iltw_V)*(itw_W - iltw_W));
									a_w_3  += psi_stereo*((itw_W - iltw_W)*(itw_W - iltw_W));
									a_z_3  += psi_stereo*((itw_Z - iltw_Z)*(itw_W - iltw_W));
									//z equation
									a_u_4  += psi_stereo*((itw_U - iltw_U)*(itw_Z - iltw_Z));
									a_v_4  += psi_stereo*((itw_V - iltw_V)*(itw_Z - iltw_Z));
									a_w_4  += psi_stereo*((itw_W - iltw_W)*(itw_Z - iltw_Z));
									a_z_4  += psi_stereo*((itw_Z - iltw_Z)*(itw_Z - iltw_Z));

									// second term
									bu += - psi_stereo*((it - ilt)*(itw_U - iltw_U));
									bv += - psi_stereo*((it - ilt)*(itw_V - iltw_V));
									bw += - psi_stereo*((it - ilt)*(itw_W - iltw_W));
									bz += - psi_stereo*((it - ilt)*(itw_Z - iltw_Z));
								}
							} // for(view = 1; view < NumOfViews; view++){
							
							
							if (pixel_do_ref_flow) {
								//psi' coefficients computation
								//tali: the data term of the left camera OF
								double a = square(ilt + iltw_U*du + iltw_V*dv + iltw_W*dw + iltw_Z*dz - il0);
								//divergence coefficients  
								double psi_flowL = DPsiData(a);

								// tali: the _1 refers to the the u equation, the _2 to the v equation the _3 to w and _4 z 
								//u equation								
								a_u_1 += psi_flowL*(iltw_U*iltw_U);
								a_v_1 += psi_flowL*(iltw_V*iltw_U);
								a_w_1 += psi_flowL*(iltw_W*iltw_U);
								a_z_1 += psi_flowL*(iltw_Z*iltw_U);
								//v equation
								a_u_2 += psi_flowL*(iltw_U*iltw_V);
								a_v_2 += psi_flowL*(iltw_V*iltw_V);
								a_w_2 += psi_flowL*(iltw_W*iltw_V);
								a_z_2 += psi_flowL*(iltw_Z*iltw_V);
								//w equation
								a_u_3 += psi_flowL*(iltw_U*iltw_W);
								a_v_3 += psi_flowL*(iltw_V*iltw_W);
								a_w_3 += psi_flowL*(iltw_W*iltw_W);
								a_z_3 += psi_flowL*(iltw_Z*iltw_W);
								// second term
								bu += - psi_flowL*(iltw_U*(ilt - il0));		   
								bv += - psi_flowL*(iltw_V*(ilt - il0));
								bw += - psi_flowL*(iltw_W*(ilt - il0));
								bz += - psi_flowL*(iltw_Z*(ilt - il0));
							}
							// 3 lines corresponding to the 3 equations at (yl0,xl0) 
							cvmSet(M,i  ,0,a_u_1);
							cvmSet(M,i+1,0,a_u_2);
							cvmSet(M,i+2,0,a_u_3);
							cvmSet(M,i+3,0,a_u_4);
			   
							cvmSet(M,i  ,1,a_v_1);
							cvmSet(M,i+1,1,a_v_2);
							cvmSet(M,i+2,1,a_v_3);
							cvmSet(M,i+3,1,a_v_4);
			
							cvmSet(M,i  ,2,a_w_1);
							cvmSet(M,i+1,2,a_w_2);
							cvmSet(M,i+2,2,a_w_3);
							cvmSet(M,i+3,2,a_w_4);
			   
							cvmSet(M,i  ,3,a_z_1);
							cvmSet(M,i+1,3,a_z_2);
							cvmSet(M,i+2,3,a_z_3);
							cvmSet(M,i+3,3,a_z_4);

							cvmSet(M,i  ,4,(-alpha*mu_u/(4*h*h))*b_f_4);
							cvmSet(M,i+1,4,(-alpha*mu_v/(4*h*h))*b_f_4);
							cvmSet(M,i+2,4,(-alpha_w*mu_w/(4*h*h))*b_fw_4);
							cvmSet(M,i+3,4,(-alpha_z*mu_z/(4*h*h))*b_fz_4);
							
		 
							cvmSet(M,i  ,5,(-alpha*mu_u/(4*h*h))*b_f_3);
							cvmSet(M,i+1,5,(-alpha*mu_v/(4*h*h))*b_f_3);
							cvmSet(M,i+2,5,(-alpha_w*mu_w/(4*h*h))*b_fw_3);
							cvmSet(M,i+3,5,(-alpha_z*mu_z/(4*h*h))*b_fz_3);
									 
							cvmSet(M,i  ,6,(-alpha*mu_u/(4*h*h))*b_f_2);
							cvmSet(M,i+1,6,(-alpha*mu_v/(4*h*h))*b_f_2);
							cvmSet(M,i+2,6,(-alpha_w*mu_w/(4*h*h))*b_fw_2);
							cvmSet(M,i+3,6,(-alpha_z*mu_z/(4*h*h))*b_fz_2);
							
							cvmSet(M,i  ,7,(-alpha*mu_u/(4*h*h))*b_f_1);
							cvmSet(M,i+1,7,(-alpha*mu_v/(4*h*h))*b_f_1);
							cvmSet(M,i+2,7,(-alpha_w*mu_w/(4*h*h))*b_fw_1);
							cvmSet(M,i+3,7,(-alpha_z*mu_z/(4*h*h))*b_fz_1);
							
							cvmSet(B,i,0,bu);
							cvmSet(B,i+1,0,bv);
							cvmSet(B,i+2,0,bw);
							cvmSet(B,i+3,0,bz);
				}
				else if(methode == 1) {
							double il0 = 0.;
							double il0_X = 0., il0_Y = 0.;
							double i0 = 0.;
							double i0_X = 0., i0_Y = 0.;
							double i0w_X = 0., i0w_Y = 0.;
							double proj_x0_x = 0., proj_x0_y = 0;
							double proj_y0_x = 0., proj_y0_y = 0;
							
							double ilt = 0.;
							double ilt_X = 0., ilt_Y = 0.;
							double iltw_X = 0., iltw_Y = 0., iltw_Z = 0.;
							double iltw_U = 0., iltw_V = 0., iltw_W = 0.;
							double proj_xlt_x = 0., proj_xlt_y = 0.;
							double proj_xlt_u = 0., proj_xlt_v = 0., proj_xlt_w = 0.;
							double proj_ylt_x = 0., proj_ylt_y = 0.;
							double proj_ylt_u = 0., proj_ylt_v = 0., proj_ylt_w = 0.;
							double it = 0.;
							double it_X = 0., it_Y = 0.;
							double itw_X = 0., itw_Y = 0.;;
							double itw_U = 0., itw_V = 0., itw_W = 0.;
							double proj_xt_x = 0., proj_xt_y = 0.;
							double proj_xt_u = 0., proj_xt_v = 0., proj_xt_w = 0.;
							double proj_yt_x = 0., proj_yt_y = 0.;
							double proj_yt_u = 0., proj_yt_v = 0., proj_yt_w = 0.;
							double du = 0, dv = 0, dw = 0;

							//interpolated divergence coefficients
							double b_f_1, b_f_2, b_f_3, b_f_4;
							double b_fw_1, b_fw_2, b_fw_3, b_fw_4;
						
							// Half pixel interpolation for the divergence coefficients 
							divergence_coefficients_interpolate(divergence_coefficients, yl0, xl0, &b_f_1, &b_f_2, &b_f_3, &b_f_4);
							divergence_coefficients_interpolate(divergence_coefficients_W, yl0, xl0, &b_fw_1, &b_fw_2, &b_fw_3, &b_fw_4);
							
	                        //Matrix terms
							double a_u_1, a_u_2, a_u_3, a_v_1, a_v_2, a_v_3, a_w_1, a_w_2, a_w_3;
		    
							a_u_1 = (alpha*mu_u/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_v_1 = 0;
							a_w_1 = 0;	  
							
		    
							a_u_2 = 0;
							a_v_2 = (alpha*mu_v/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_w_2 = 0;	  
							
		    
							a_u_3 = 0;
							a_v_3 = 0;
							a_w_3 = (alpha_w*mu_w/(4.*h*h))*(b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4);	  
							
	
							//second member for the first equation
							double bu0 = cvmGet(U, yl0, xl0);
							double bu1 = cvmGet(U, yl0, xl0+1);
							double bu2 = cvmGet(U, yl0, xl0-1);
							double bu3 = cvmGet(U,yl0+1,xl0);
							double bu4 = cvmGet(U,yl0-1,xl0);
							//Crop(bu0, N_x);
							double bv0 = cvmGet(V, yl0, xl0);
							double bv1 = cvmGet(V, yl0, xl0+1);
							double bv2 = cvmGet(V, yl0, xl0-1);
							double bv3 = cvmGet(V,yl0+1,xl0);
							double bv4 = cvmGet(V,yl0-1,xl0);
							//Crop(bw0, N_x);
							double bw0 = cvmGet(W, yl0, xl0);
							double bw1 = cvmGet(W, yl0, xl0+1);
							double bw2 = cvmGet(W, yl0, xl0-1);
							double bw3 = cvmGet(W,yl0+1,xl0);
							double bw4 = cvmGet(W,yl0-1,xl0);

							//tali: bu is the discretization of div(Psi()*grad(u))
							double bu = (alpha*mu_u/(4*h*h)) *
								(b_f_1 * bu1 +
								 b_f_2 * bu2 +
								 b_f_3 * bu3 +
								 b_f_4 * bu4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bu0);

							//Crop(bv0, N_y);
							double bv = (alpha*mu_v/(4*h*h)) *
								(b_f_1 * bv1 +
								 b_f_2 * bv2 +
								 b_f_3 * bv3 +
								 b_f_4 * bv4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bv0);
							//Crop(bv0, N_y);
							double bw = (alpha_w*mu_w/(4*h*h)) *
								(b_fw_1 * bw1 +
								 b_fw_2 * bw2 +
								 b_fw_3 * bw3 +
								 b_fw_4 * bw4
								 - (b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4) * bw0);

							//Pixels out must not be taken into account in the energy data term
							int pixel_do_ref_flow = 0;
							int pixel_do_flow_views[MAX_NUM_OF_VIEWS] = {0};
							int pixel_doZt_views[MAX_NUM_OF_VIEWS] = {0};
							
							pixel_do_ref_flow = !cvmGet8u(disactivated_pixels_ref_flow,yl0,xl0);
							if(pixel_do_ref_flow){
								double xlt = cvmGet(MapX_Img_t_inner[0], yl0, xl0);
								double ylt = cvmGet(MapY_Img_t_inner[0], yl0, xl0);
								if(xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h ||
									ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h){

										cvmSet8u(disactivated_pixels_ref_flow,yl0,xl0,127);
										pixel_doZt_views[view] = 0;
										pixel_do_ref_flow = 0;
								}
							}
							
							for(view=1; view < NumOfViews; view++) {
								pixel_do_flow_views[view] = !cvmGet8u(disactivated_pixels_flow[view],yl0,xl0);
								pixel_doZt_views[view] = !cvmGet8u(disactivated_pixels_Zt[view],yl0,xl0);
																

								//On the image at time t
								if (pixel_do_flow_views[view] ) {
									double x0 = cvmGet(MapX_Img_0_inner[view], yl0, xl0);
									double y0 = cvmGet(MapY_Img_0_inner[view], yl0, xl0);
									double xt = cvmGet(MapX_Img_t_inner[view], yl0, xl0);
									double yt = cvmGet(MapY_Img_t_inner[view], yl0, xl0);

									if(x0 >= (N_x - 1 - DISACTIVATE_BORDER)*h || x0 <= DISACTIVATE_BORDER*h ||
									   y0 >= (N_y - 1 - DISACTIVATE_BORDER)*h || y0 <= DISACTIVATE_BORDER*h )
									{
										cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,127);
										pixel_doZt_views[view] = 0;
										pixel_do_flow_views[view] = 0;
										
									}
									if(xt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xt <= DISACTIVATE_BORDER*h ||
									   yt >= (N_y - 1 - DISACTIVATE_BORDER)*h || yt <= DISACTIVATE_BORDER*h )
									{
										cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,127);
										pixel_doZt_views[view] = 0;
										pixel_do_flow_views[view] = 0;
									}
								
								}
							}
							
							int pixel_doZt = 0;
							int pixel_do_flow = 0;
							for(view = 1; view < NumOfViews; view++){
								pixel_doZt += pixel_doZt_views[view];
								pixel_do_flow += pixel_do_flow_views[view];
							}
							
							du = cvmGet(dU, yl0, xl0);
							dv = cvmGet(dV, yl0, xl0);
							dw = cvmGet(dW, yl0, xl0);
							
							//Data term estimation
							if (pixel_do_ref_flow) {
								il0 = cvGetReal2D(Images_0_pyramid[0][level], yl0, xl0); 
								il0_X = cvmGet(Images_0_X[0],  yl0, xl0);
								il0_Y = cvmGet(Images_0_Y[0],  yl0, xl0);
							}
							for(view = 1; view < NumOfViews; view++){
								if (pixel_do_flow_views[view]) {
									i0 = cvGetReal2D(Images_0w[view], yl0, xl0);
									i0w_X = cvmGet(Images_0w_X[view],  yl0, xl0);
									i0w_Y = cvmGet(Images_0w_Y[view],  yl0, xl0);
									proj_x0_x = cvmGet(Proj_x0_x[view], yl0, xl0);
									proj_x0_y = cvmGet(Proj_x0_y[view], yl0, xl0);
									proj_y0_x = cvmGet(Proj_y0_x[view], yl0, xl0);
									proj_y0_y = cvmGet(Proj_y0_y[view], yl0, xl0);

									double det = proj_x0_x*proj_y0_y - proj_x0_y*proj_y0_x;
									if (det <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,191);
										pixel_do_flow_views[view] = 0;

									}
									else {
										i0_X = (i0w_X*proj_y0_y - i0w_Y*proj_y0_x)/det;
										i0_Y = (i0w_Y*proj_x0_x - i0w_X*proj_x0_y)/det;
									}
								}

								if (pixel_do_ref_flow || pixel_doZt_views[view]) {
									// Wraped image values
									ilt = cvGetReal2D(Images_tw[0], yl0, xl0);
									iltw_X = cvmGet(Images_tw_X[0], yl0, xl0);
									iltw_Y = cvmGet(Images_tw_Y[0],  yl0, xl0);
									// projected coordinated derivatives
									proj_xlt_x = cvmGet(Proj_xt_x[0], yl0, xl0);
									proj_xlt_y = cvmGet(Proj_xt_y[0], yl0, xl0);
									proj_ylt_x = cvmGet(Proj_yt_x[0], yl0, xl0);
									proj_ylt_y = cvmGet(Proj_yt_y[0], yl0, xl0);
									proj_xlt_u = cvmGet(Proj_xt_u[0], yl0, xl0);
									proj_ylt_u = cvmGet(Proj_yt_u[0], yl0, xl0);
									proj_xlt_v = cvmGet(Proj_xt_v[0], yl0, xl0);
									proj_ylt_v = cvmGet(Proj_yt_v[0], yl0, xl0);
									proj_xlt_w = cvmGet(Proj_xt_w[0], yl0, xl0);
									proj_ylt_w = cvmGet(Proj_yt_w[0], yl0, xl0);
								
									double det = proj_xlt_x*proj_ylt_y - proj_xlt_y*proj_ylt_x;
									if ((det) <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_do_ref_flow) {
											cvmSet8u(disactivated_pixels_ref_flow,yl0,xl0,191);
											pixel_do_ref_flow = 0;
										}
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,191);
											pixel_doZt_views[view] = 0;
										}
									}
									else {
										ilt_X = (iltw_X*proj_ylt_y - iltw_Y*proj_ylt_x)/det ;
										ilt_Y = (iltw_Y*proj_xlt_x - iltw_X*proj_xlt_y)/det ;
										iltw_U = ilt_X*proj_xlt_u + ilt_Y*proj_ylt_u ; 
										iltw_V = ilt_X*proj_xlt_v + ilt_Y*proj_ylt_v ; 
										iltw_W = ilt_X*proj_xlt_w + ilt_Y*proj_ylt_w ;
									}
								}
								if(pixel_doZt_views[view]){
									it = cvGetReal2D(Images_tw[view],yl0,xl0);
									itw_X = cvmGet(Images_tw_X[view],  yl0, xl0);
									itw_Y = cvmGet(Images_tw_Y[view],  yl0, xl0);
									
									proj_xt_x = cvmGet(Proj_xt_x[view], yl0, xl0);
									proj_xt_y = cvmGet(Proj_xt_y[view], yl0, xl0);
									proj_yt_x = cvmGet(Proj_yt_x[view], yl0, xl0);
									proj_yt_y = cvmGet(Proj_yt_y[view], yl0, xl0);
									
									proj_xt_u = cvmGet(Proj_xt_u[view], yl0, xl0);
									proj_xt_v = cvmGet(Proj_xt_v[view], yl0, xl0);
									proj_xt_w = cvmGet(Proj_xt_w[view], yl0, xl0);
									proj_yt_u = cvmGet(Proj_yt_u[view], yl0, xl0);
									proj_yt_v = cvmGet(Proj_yt_v[view], yl0, xl0);
									proj_yt_w = cvmGet(Proj_yt_w[view], yl0, xl0);
									
									double det = proj_xt_x*proj_yt_y - proj_xt_y*proj_yt_x;
									if ((det) <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,191);
										pixel_doZt_views[view] = 0;
									
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,191);
											pixel_do_flow_views[view] = 0;
										}
									}
									else {
										it_X = (itw_X*proj_yt_y - itw_Y*proj_yt_x)/det;
										it_Y = (itw_Y*proj_xt_x - itw_X*proj_xt_y)/det;
									
										itw_U = it_X*proj_xt_u + it_Y*proj_yt_u ; 
										itw_V = it_X*proj_xt_v + it_Y*proj_yt_v ; 
										itw_W = it_X*proj_xt_w + it_Y*proj_yt_w ;
									}
								}

								if (pixel_do_flow_views[view]) {
									double a = square(it + itw_U*du + itw_V*dv + itw_W*dw  -i0 );
									double psi_flow = DPsiData(a);
									//u equation
									a_u_1 += psi_flow*(itw_U*itw_U);
									a_v_1 += psi_flow*(itw_V*itw_U);
									a_w_1 += psi_flow*(itw_W*itw_U);
									
									//v equation
									a_u_2 += psi_flow*(itw_U*itw_V);
									a_v_2 += psi_flow*(itw_V*itw_V);
									a_w_2 += psi_flow*(itw_W*itw_V);
									
									//w equation
									a_u_3 += psi_flow*(itw_U*itw_W);
									a_v_3 += psi_flow*(itw_V*itw_W);
									a_w_3 += psi_flow*(itw_W*itw_W);
									
									// second term
									bu += - psi_flow*(itw_U*(it - i0));		   
									bv += - psi_flow*(itw_V*(it - i0));
									bw += - psi_flow*(itw_W*(it - i0));
								}
								
								if (pixel_doZt_views[view]) {
									double a = square(it + itw_U*du + itw_V*dv + itw_W*dw - 
													 (ilt + iltw_U*du + iltw_V*dv + iltw_W*dw ));
									double psi_stereo = DPsiData(a);
									//u equation
									a_u_1  += psi_stereo*((itw_U - iltw_U)*(itw_U - iltw_U));
									a_v_1  += psi_stereo*((itw_V - iltw_V)*(itw_U - iltw_U));
									a_w_1  += psi_stereo*((itw_W - iltw_W)*(itw_U - iltw_U));
									
									//v equation
									a_u_2  += psi_stereo*((itw_U - iltw_U)*(itw_V - iltw_V));
									a_v_2  += psi_stereo*((itw_V - iltw_V)*(itw_V - iltw_V));
									a_w_2  += psi_stereo*((itw_W - iltw_W)*(itw_V - iltw_V));
									
									//w equation
									a_u_3  += psi_stereo*((itw_U - iltw_U)*(itw_W - iltw_W));
									a_v_3  += psi_stereo*((itw_V - iltw_V)*(itw_W - iltw_W));
									a_w_3  += psi_stereo*((itw_W - iltw_W)*(itw_W - iltw_W));
									
									
									
									// second term
									bu += - psi_stereo*((it - ilt)*(itw_U - iltw_U));
									bv += - psi_stereo*((it - ilt)*(itw_V - iltw_V));
									bw += - psi_stereo*((it - ilt)*(itw_W - iltw_W));
									
								}
							} // for(view = 1; view < NumOfViews; view++){
							
							
							if (pixel_do_ref_flow) {
								//psi' coefficients computation
								//tali: the data term of the left camera OF
								double a = square(ilt + iltw_U*du + iltw_V*dv + iltw_W*dw  - il0);
								//divergence coefficients  
								double psi_flowL = DPsiData(a);

								// tali: the _1 refers to the the u equation, the _2 to the v equation the _3 to w and _4 z 
								//u equation								
								a_u_1 += psi_flowL*(iltw_U*iltw_U);
								a_v_1 += psi_flowL*(iltw_V*iltw_U);
								a_w_1 += psi_flowL*(iltw_W*iltw_U);

								//v equation
								a_u_2 += psi_flowL*(iltw_U*iltw_V);
								a_v_2 += psi_flowL*(iltw_V*iltw_V);
								a_w_2 += psi_flowL*(iltw_W*iltw_V);

								//w equation
								a_u_3 += psi_flowL*(iltw_U*iltw_W);
								a_v_3 += psi_flowL*(iltw_V*iltw_W);
								a_w_3 += psi_flowL*(iltw_W*iltw_W);

								// second term
								bu += - psi_flowL*(iltw_U*(ilt - il0));		   
								bv += - psi_flowL*(iltw_V*(ilt - il0));
								bw += - psi_flowL*(iltw_W*(ilt - il0));

							}
							// 3 lines corresponding to the 3 equations at (yl0,xl0) 
							cvmSet(M,i  ,0,a_u_1);
							cvmSet(M,i+1,0,a_u_2);
							cvmSet(M,i+2,0,a_u_3);
						
							cvmSet(M,i  ,1,a_v_1);
							cvmSet(M,i+1,1,a_v_2);
							cvmSet(M,i+2,1,a_v_3);
						
			
							cvmSet(M,i  ,2,a_w_1);
							cvmSet(M,i+1,2,a_w_2);
							cvmSet(M,i+2,2,a_w_3);
							
			   				cvmSet(M,i  ,3,(-alpha*mu_u/(4*h*h))*b_f_4);
							cvmSet(M,i+1,3,(-alpha*mu_v/(4*h*h))*b_f_4);
							cvmSet(M,i+2,3,(-alpha_w*mu_w/(4*h*h))*b_fw_4);
							
		 
							cvmSet(M,i  ,4,(-alpha*mu_u/(4*h*h))*b_f_3);
							cvmSet(M,i+1,4,(-alpha*mu_v/(4*h*h))*b_f_3);
							cvmSet(M,i+2,4,(-alpha_w*mu_w/(4*h*h))*b_fw_3);
							
									 
							cvmSet(M,i  ,5,(-alpha*mu_u/(4*h*h))*b_f_2);
							cvmSet(M,i+1,5,(-alpha*mu_v/(4*h*h))*b_f_2);
							cvmSet(M,i+2,5,(-alpha_w*mu_w/(4*h*h))*b_fw_2);
							
							
							cvmSet(M,i  ,6,(-alpha*mu_u/(4*h*h))*b_f_1);
							cvmSet(M,i+1,6,(-alpha*mu_v/(4*h*h))*b_f_1);
							cvmSet(M,i+2,6,(-alpha_w*mu_w/(4*h*h))*b_fw_1);
							
							
							cvmSet(B,i,0,bu);
							cvmSet(B,i+1,0,bv);
							cvmSet(B,i+2,0,bw);
				
						}

				
				} //Pixel inside the integration domain 
			} //for(i = inf; i <= sup; i+=dimension){


			//Initial energy computation
			if(logfile && iter_inner == 0 && iter_outer == 0){
				double E_Ref_flow;
				double E_Flow[MAX_NUM_OF_VIEWS], E_Stereo_0[MAX_NUM_OF_VIEWS], E_Stere0_t[MAX_NUM_OF_VIEWS];
				double Edata, Esmooth, Eintensity, Es_Motion, Es_Z;

				EnergyComputation3D((Images_0_pyramid[0])[level], 
					(const CvMat**)Images_0w,(const CvMat**)Images_tw,
					U, V, W, Z0,
					(const CvMat*)disactivated_pixels_ref_flow, (const CvMat**)disactivated_pixels_flow,
					(const CvMat**)disactivated_pixels_Z0, (const CvMat**)disactivated_pixels_Zt,
					N_x, N_y, &Edata, &Esmooth, mu_u, mu_v, mu_w, mu_z, alpha, alpha_w, alpha_z,opt_Z,
					&E_Ref_flow, E_Flow, E_Stereo_0, E_Stere0_t, &Es_Motion, &Es_Z, &Eintensity, level);

				fprintf(logfile, "** LEVEL %d resolution %d x %d\n", level+1, N_x, N_y);
				fprintf(logfile, "* final energy:\n");
				fprintf(logfile, "Edata = %g Esmooth = %g (alpha=%g) (alpha_w=%g) (alpha_z=%g)\n", Edata/(N_x*N_y), Esmooth/(N_x*N_y), alpha, alpha_w, alpha_z);
				for(view = 1; view < NumOfViews; view++){
					fprintf(logfile, "View = %d: E_flow = %g E_Stere0 = %g E_Stereo_t = %g\n",view, E_Flow[view]/(N_x*N_y), E_Stereo_0[view]/(N_x*N_y), E_Stere0_t[view]/(N_x*N_y));
					fflush(logfile);

				}
				fprintf(logfile, "E_ref_flow  = %g\n", E_Ref_flow);
				fprintf(logfile, "Edata breakdown (squared): Eintensity = %g \n", Eintensity);
				fprintf(logfile, "Esmooth breakdown (squared): flow = %g , surface = %g\n", Es_Motion, Es_Z);
				fflush(logfile);

			}
			if (max_iter_outer == 0) {
				level = pyramid_level_end-1;
				goto CLEANUP;
			}

			//SOR iterations
		    cvSetZero(SOR);
		    cvSetZero(SOR_prec);
		    first_iteration = 1;
		    iter_SOR = 0;
		    norm_SOR = 0.;
			
			CvMat *M2 = cvCreateMat(dimension * N_x * N_y, dimension + 4, CV_32FC1); 
			CvMat *B2 = cvCreateMat(dimension * N_x * N_y,1,CV_32FC1);	
			cvCopy(M,M2);
			cvCopy(B,B2);

			while ((norm_SOR >= epsilon_SOR && iter_SOR < max_iter_SOR) || first_iteration == 1){
				first_iteration = 0;

				for (int sub_iter_SOR = 0; sub_iter_SOR <4; sub_iter_SOR++) {
					int line, line_begin, line_end, line_incr;
					int col, col_begin, col_end, col_incr;
					switch(sub_iter_SOR){
							case 0:         // NE-SW
								line_begin = 0;
								line_end = N_y * dimension * N_x;
								line_incr = dimension * N_x;
								col_begin = 0;
								col_end = dimension * N_x;
								col_incr = 1;
								break;
							case 1:         // SW-NE
								line_begin = (N_y-1) * dimension * N_x;
								line_end = - dimension * N_x;
								line_incr = - dimension * N_x;
								col_begin = dimension * N_x - 1;
								col_end = -1;
								col_incr = -1;
								break;
							case 2:         // NW-SE
								line_begin = 0;
								line_end = N_y * dimension * N_x;
								line_incr = dimension * N_x;
								col_begin = dimension * N_x - 1;
								col_end = -1;
								col_incr = -1;
								break;
							default:         // SE-NW
								line_begin = (N_y-1) * dimension * N_x;
								line_end = - dimension * N_x;
								line_incr = - dimension * N_x;
								col_begin = 0;
								col_end = dimension * N_x;
								col_incr = 1;
					}
					for (line = line_begin; line != line_end; line+= line_incr) {
						for (col = col_begin; col != col_end; col+= col_incr) {
							int z_eq = 0;
							i = line + col;
							double sum = 0;
							double diag_term = 0.;

							// i refers to the relevant equations for line y=0
							if(i < dimension*N_x){
								diag_term = cvmGet(M2, i,0); 
								sum = cvmGet(M2, i,1)*cvmGet(SOR, i+dimension*N_x,0);
							}
							// i refers to the relevant equations for line y=Ny
							else if(i >= dimension*N_x*(N_y-1)){
								diag_term = cvmGet(M2, i,0); 
								sum = cvmGet(M2, i, 1)*cvmGet(SOR, i-dimension*N_x,0);
							}
							else{
								if(i%(dimension*N_x) == 0){
									diag_term = cvmGet(M2, i,0); 
									sum = cvmGet(M2, i,1)*cvmGet(SOR, i+dimension,0);
								}
								else if((i+dimension)%(dimension*N_x)==0){
									diag_term = cvmGet(M2, i,0); 
									sum = cvmGet(M2, i,1)*cvmGet(SOR, i-dimension,0);
								}
								else{
									int rest1 = i%(dimension*N_x);
									if((dimension == 3 && (rest1 == 1 || rest1 == 2)) ||
										(dimension == 4 && (rest1 == 1 || rest1 == 2 || rest1 == 3))){
											diag_term = cvmGet(M2, i,0); 
											sum = cvmGet(M2, i,1)*cvmGet(SOR, i+dimension,0);
									}
									else if((dimension == 4 && (rest1 == 4*N_x - 3 || rest1 == 4*N_x - 2 || rest1 == 4*N_x - 1)) ||
										(dimension == 3 && (rest1 == 3*N_x - 2 || rest1 == 3*N_x - 1))){
											diag_term = cvmGet(M2, i,0); 
											sum = cvmGet(M2, i,1)*cvmGet(SOR, i-dimension,0);
									}
									else{
										int rest2 = rest1 % dimension;

										if(dimension == 3){
											if(rest2 == 0){
												//u equation
												sum =
													cvmGet(M2, i,3)*cvmGet(SOR, i-3*N_x,0) +
													cvmGet(M2, i,5)*cvmGet(SOR, i-3    ,0) +
													cvmGet(M2, i,1)*cvmGet(SOR, i+1    ,0) +
													cvmGet(M2, i,2)*cvmGet(SOR, i+2    ,0) +
													cvmGet(M2, i,4)*cvmGet(SOR, i+3*N_x,0) +
													cvmGet(M2, i,6)*cvmGet(SOR, i+3    ,0);
												diag_term = cvmGet(M2, i,0);
											}
											else if(rest2 == 1){
												//v equation
												sum =
													cvmGet(M2, i,3)*cvmGet(SOR, i-3*N_x,0) +
													cvmGet(M2, i,5)*cvmGet(SOR, i-3    ,0) +
													cvmGet(M2, i,0)*cvmGet(SOR, i-1    ,0) +
													cvmGet(M2, i,2)*cvmGet(SOR, i+1    ,0) +
													cvmGet(M2, i,4)*cvmGet(SOR, i+3*N_x,0) +
													cvmGet(M2, i,6)*cvmGet(SOR, i+3    ,0) ;
												diag_term = cvmGet(M2, i,1);
											}
											else if(rest2 == 2){
												//w/z equation
												sum =
													cvmGet(M2, i,3)*cvmGet(SOR, i-3*N_x,0) +
													cvmGet(M2, i,5)*cvmGet(SOR, i-3    ,0) +
													cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) +
													cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) +
													cvmGet(M2, i,4)*cvmGet(SOR, i+3*N_x,0) +
													cvmGet(M2, i,6)*cvmGet(SOR, i+3    ,0);
												diag_term = cvmGet(M2, i,2);
											}
										}
										else if(dimension == 4){
											if(rest2 == 0){
												//u equation
												sum =
													// tali: the neighbour above
													cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
													// tali: the neighbour to the left
													cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
													// tali: the relevant v 
													cvmGet(M2, i,1)*cvmGet(SOR, i+1    ,0) +
													// tali: the relevant w
													cvmGet(M2, i,2)*cvmGet(SOR, i+2    ,0) +
													// tali: the relevant z
													cvmGet(M2, i,3)*cvmGet(SOR, i+3    ,0) +
													// tali: the neighbour below
													cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
													// tali: the neighbour to right
													cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0);
												diag_term = cvmGet(M2, i,0);
											}
											else if(rest2 == 1){
												//v equation
												sum =
													cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
													cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
													cvmGet(M2, i,0)*cvmGet(SOR, i-1    ,0) +
													cvmGet(M2, i,2)*cvmGet(SOR, i+1    ,0) +
													cvmGet(M2, i,3)*cvmGet(SOR, i+2    ,0) +
													cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
													cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0) ;
												diag_term = cvmGet(M2, i,1);
											}
											else if(rest2 == 2){
												double S1 = cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) ;
												double S2 = cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) ;
												double S3 = cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) ;
												double S4 = cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) ;
												double S5=  cvmGet(M2, i,3)*cvmGet(SOR, i+1    ,0) ;
												double S6 = cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) ;
												double S7 = cvmGet(M2, i,7)*cvmGet(SOR, i+4,0)  ;
												//w equation
												sum =
													cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
													cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
													cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) +
													cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) +
													cvmGet(M2, i,3)*cvmGet(SOR, i+1    ,0) +
													cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
													cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0);
												diag_term = cvmGet(M2, i,2);
											}
											else{
												//Z equation
												z_eq = 1;
												sum = 
													cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +  
													cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
													cvmGet(M2, i,0)*cvmGet(SOR, i-3    ,0) +
													cvmGet(M2, i,1)*cvmGet(SOR, i-2    ,0) +
													cvmGet(M2, i,2)*cvmGet(SOR, i-1    ,0) +
													cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
													cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0);
												diag_term = cvmGet(M2, i,3);	
											}
										}
									}
								}
							}

							double rold = cvmGet(SOR,i,0);
							double b = cvmGet(B2,i,0);
							double rnew = (1.-omega_SOR)*rold + (omega_SOR/diag_term) * (b - sum);

							if(methode == 0 ){
								if(z_eq) {
									cvmSet(SOR,i,0,Crop(rnew,MAX_INCR));
								}
								else {
									cvmSet(SOR,i,0,Crop(rnew,MAX_INCR_SF));
								}
							}
							else{
								cvmSet(SOR,i,0,Crop(rnew,MAX_INCR_SF));
							}


						} // for (col
					} // for (line
				} // for (sub_iter_SOR

				norm_SOR = cvNorm(SOR_prec, SOR, CV_RELATIVE_L2);
				cvCopy(SOR, SOR_prec);
				iter_SOR ++;

			} // while ((norm_SOR >= epsilon_SOR && iter_SOR < max_iter_SOR) || first_iteration == 1){
			//printf("iter_SOR=%d norm_SOR=%g\n",iter_SOR, norm_SOR);
			cvReleaseMat(&M2);
			cvReleaseMat(&B2);

			cvCopy(dU, dU_prec); 
			cvCopy(dV, dV_prec);
			cvCopy(dW, dW_prec);
			if (opt_Z) {
				cvCopy(dZ0, dZ0_prec);
			}
						                
			for(i = 0; i < dimension*N_x*N_y; i+=dimension){
				xl0 = (i%(dimension*N_x))/dimension;
				yl0 = i/(dimension*N_x);
				int d = 0;

				double du = cvmGet(SOR, i + d,0); d++;
				double dv = cvmGet(SOR, i + d,0); d++;
				double dw = cvmGet(SOR, i + d,0); d++;

				cvmSet(dU, yl0, xl0, Crop(du,MAX_INCR_SF));
				cvmSet(dV, yl0, xl0, Crop(dv,MAX_INCR_SF));
				cvmSet(dW, yl0, xl0, Crop(dw,MAX_INCR_SF));

				if(opt_Z){
					double dz = cvmGet(SOR, i + d,0); d++;
					cvmSet(dZ0, yl0, xl0, Crop(dz,MAX_INCR));
				}


			}
				
			norm_inner = 0;
			
			norm_inner = fmax(norm_inner, cvNorm(dU_prec, dU, CV_RELATIVE_L2));
			norm_inner = fmax(norm_inner, cvNorm(dV_prec, dV, CV_RELATIVE_L2));
			norm_inner = fmax(norm_inner, cvNorm(dW_prec, dW, CV_RELATIVE_L2));


			if (opt_Z) {
				norm_inner = fmax(norm_inner, cvNorm(dZ0_prec, dZ0, CV_RELATIVE_L2));
			}

			iter_inner++;

			/*cvSmooth(dW, dW, CV_GAUSSIAN,3,0.5);
			cvSmooth(dU, dU, CV_GAUSSIAN,3,0.5);
			cvSmooth(dV, dV, CV_GAUSSIAN,3,0.5);
			cvSmooth(dZ0, dZ0, CV_GAUSSIAN,3,0.5);*/


			}
/*
			    cvSmooth(dW, dW, CV_MEDIAN,1);
				cvSmooth(dU, dU, CV_MEDIAN,1);
				cvSmooth(dV, dV, CV_MEDIAN,1);
				cvSmooth(dZ0, dZ0, CV_MEDIAN,1);*/	
			
			cvCopy(U, U_prec);
			cvAdd(U, dU, U);

			cvCopy(V, V_prec);
			cvAdd(V, dV, V);

			cvCopy(W, W_prec);
			cvAdd(W, dW, W);				
		
			if (opt_Z) {
				cvCopy(Z0, Z0_prec);
				cvAdd(Z0, dZ0, Z0);
			}
			
			norm = 0;
			norm = fmax(norm, cvNorm(U_prec, U, CV_RELATIVE_L2));
			norm = fmax(norm, cvNorm(V_prec, V, CV_RELATIVE_L2));
			norm = fmax(norm, cvNorm(W_prec, W, CV_RELATIVE_L2));
	           
			if(opt_Z) {
				norm = fmax(norm, cvNorm(Z0_prec, Z0, CV_RELATIVE_L2));
			}
#ifdef DEBUG
				if (SF_U_FileName) {
					SaveResults(SF_U_FileName, U);
				}
				if (SF_V_FileName) {	     
					SaveResults(SF_V_FileName, V);
				}
				if (SF_W_FileName) {	     
					SaveResults(SF_W_FileName, W);
				}
		
				SaveResults(SF_Z0_FileName, Z0);
			
			
			if (SF_disactivated_ref_flow_FileName && disactivated_pixels_ref_flow) {				
				sprintf(file, "%s.%s",SF_disactivated_ref_flow_FileName, IMG_TYPE);
				SaveResults(file, disactivated_pixels_ref_flow, 0, 0., 255.);
			}
			for(view = 1; view < NumOfViews; view++) {
				if (SF_disactivated_flow_FileName && disactivated_pixels_flow) {
					sprintf(file, "%s%d.%s",SF_disactivated_flow_FileName, view, IMG_TYPE);
					SaveResults(file, disactivated_pixels_flow[view], 0, 0., 255.);
				}
				if (SF_disactivated_Surf_1_FileName && disactivated_pixels_Z0 && opt_Z) {
					sprintf(file, "%s%d.%s",SF_disactivated_Surf_1_FileName, view, IMG_TYPE);
					SaveResults(file, disactivated_pixels_Z0[view], 0, 0., 255.);
				}
				if (SF_disactivated_Surf_2_FileName && disactivated_pixels_Zt) {
					sprintf(file, "%s%d.%s",SF_disactivated_Surf_2_FileName, view, IMG_TYPE);
					SaveResults(file, disactivated_pixels_Zt[view], 0, 0., 255.);
				}
			}
			for(view = 1; view < NumOfViews; view++) {
				if(opt_Z && SF_disactivated_Surf_1_FileName){
					sprintf(file, "%sv%d_el%d.txt", SF_disactivated_Surf_1_FileName,view,level);
					f = fopen(file,"w");
					WriteRes8u(f,disactivated_pixels_Z0[view]);
					fclose(f);
				}

				if(SF_disactivated_flow_FileName){
					sprintf(file, "%sv%d_el%d.txt",SF_disactivated_flow_FileName,view,level);
					f = fopen(file,"w");
					WriteRes8u(f,disactivated_pixels_flow[view]);
					fclose(f);
				}

				if(SF_disactivated_Surf_2_FileName){
					sprintf(file, "%sv%d_el%d.txt",SF_disactivated_Surf_2_FileName,view,level);
					f = fopen(file,"w");
					WriteRes8u(f,disactivated_pixels_Zt[view]);
					fclose(f);
				}
			}

			if(SF_disactivated_ref_flow_FileName){
				sprintf(file, "%s_el%d.txt",SF_disactivated_ref_flow_FileName,level);
				f = fopen(file,"w");
				WriteRes8u	(f,disactivated_pixels_ref_flow);
				fclose(f);
			}
#endif
			
			iter_outer++;
		}

		double E_Ref_flow;
		double E_Flow[MAX_NUM_OF_VIEWS], E_Stereo_0[MAX_NUM_OF_VIEWS], E_Stere0_t[MAX_NUM_OF_VIEWS];
		double Edata, Esmooth, Eintensity, Es_Motion, Es_Z;


		EnergyComputation3D((Images_0_pyramid[0])[level], 
			(const CvMat**)Images_0w,(const CvMat**)Images_tw,
			U, V, W, Z0,
			(const CvMat*)disactivated_pixels_ref_flow, (const CvMat**)disactivated_pixels_flow,
			(const CvMat**)disactivated_pixels_Z0, (const CvMat**)disactivated_pixels_Zt,
			N_x, N_y, &Edata, &Esmooth, mu_u, mu_v, mu_w, mu_z, alpha, alpha_w, alpha_z,opt_Z,
			&E_Ref_flow, E_Flow, E_Stereo_0, E_Stere0_t, &Es_Motion, &Es_Z, &Eintensity, level);

		fprintf(logfile, "** LEVEL %d resolution %d x %d\n", level+1, N_x, N_y);
		fprintf(logfile, "* final energy:\n");
		fprintf(logfile, "Edata = %g Esmooth = %g (alpha=%g) (alpha_w=%g) (alpha_z=%g)\n", Edata/(N_x*N_y), Esmooth/(N_x*N_y), alpha, alpha_w, alpha_z);
		for(view = 1; view < NumOfViews; view++){
			fprintf(logfile, "View = %d: E_flow = %g E_Stere0 = %g E_Stereo_t = %g\n",view, E_Flow[view]/(N_x*N_y), E_Stereo_0[view]/(N_x*N_y), E_Stere0_t[view]/(N_x*N_y));
			fflush(logfile);

		}
		fprintf(logfile, "E_ref_flow  = %g\n", E_Ref_flow);
		fprintf(logfile, "Edata breakdown (squared): Eintensity = %g \n", Eintensity);
		fprintf(logfile, "Esmooth breakdown (squared): flow = %g , surface = %g\n", Es_Motion, Es_Z);
		fflush(logfile);

      CLEANUP:
	  //let's clean
        cvReleaseMat(&derivative_tmp);
        cvReleaseMat(&divergence_coefficients);
		cvReleaseMat(&MapIdentityX);
		cvReleaseMat(&MapIdentityY);
		for(view = 0; view < NumOfViews; view++) {
			cvReleaseMat(&Images_0_X[view]);
			cvReleaseMat(&Images_0_Y[view]);
			cvReleaseMat(&Images_tw[view]);
			cvReleaseMat(&Images_tw_X[view]);
			cvReleaseMat(&Images_tw_Y[view]);
			cvReleaseMat(&Proj_xt_x[view]);
			cvReleaseMat(&Proj_xt_y[view]);
			cvReleaseMat(&Proj_yt_x[view]);
			cvReleaseMat(&Proj_yt_y[view]);
			cvReleaseMat(&Proj_xt_u[view]);
			cvReleaseMat(&Proj_xt_v[view]);
			cvReleaseMat(&Proj_xt_w[view]);
			cvReleaseMat(&Proj_yt_u[view]);
			cvReleaseMat(&Proj_yt_v[view]);
			cvReleaseMat(&Proj_yt_w[view]);
			cvReleaseMat(&MapX_Img_t[view]);
			cvReleaseMat(&MapY_Img_t[view]);
			cvReleaseMat(&MapX_Img_t_inner[view]);
			cvReleaseMat(&MapY_Img_t_inner[view]);
			
			
				cvReleaseMat(&Images_0w[view]);
				cvReleaseMat(&MapX_Img_0[view]);
				cvReleaseMat(&MapY_Img_0[view]);
				cvReleaseMat(&MapX_Img_0_inner[view]);
				cvReleaseMat(&MapY_Img_0_inner[view]);
				if(view > 0) {
					cvReleaseMat(&Images_0w_X[view]);
					cvReleaseMat(&Images_0w_Y[view]);
				}
				cvReleaseMat(&Proj_x0_x[view]);
				cvReleaseMat(&Proj_x0_y[view]);
				cvReleaseMat(&Proj_y0_x[view]);
				cvReleaseMat(&Proj_y0_y[view]);
				if (opt_Z){
					cvReleaseMat(&Proj_x0_z[view]);
					cvReleaseMat(&Proj_y0_z[view]);
					cvReleaseMat(&Proj_xt_z[view]);
					cvReleaseMat(&Proj_yt_z[view]);
				}
			if(view > 0){ 
				cvReleaseMat(&disactivated_pixels_flow[view]);
				cvReleaseMat(&disactivated_pixels_Zt[view]);
				if(opt_Z){
					cvReleaseMat(&disactivated_pixels_Z0[view]);
				}
			}
		}
		cvReleaseMat(&disactivated_pixels_ref_flow);
		cvReleaseMat(&X);
		cvReleaseMat(&Y);
		cvReleaseMat(&X_x);
		cvReleaseMat(&X_y);
		cvReleaseMat(&Y_x);
		cvReleaseMat(&Y_y);
		cvReleaseMat(&X_z);
		cvReleaseMat(&Y_z);
	    cvReleaseMat(&M);
		cvReleaseMat(&B);
       
		cvReleaseMat(&Xt);
		cvReleaseMat(&Yt);
		cvReleaseMat(&Zt);
		cvReleaseMat(&Xt_x);
		cvReleaseMat(&Xt_y);
		cvReleaseMat(&Yt_x);
		cvReleaseMat(&Yt_y);
		
		cvReleaseMat(&U_prec); 
		cvReleaseMat(&V_prec);
		cvReleaseMat(&W_prec);
		cvReleaseMat(&dU);
		cvReleaseMat(&dV);
		cvReleaseMat(&dW);
		cvReleaseMat(&dU_X);
		cvReleaseMat(&dU_Y);
		cvReleaseMat(&dV_X);
		cvReleaseMat(&dV_Y);
		cvReleaseMat(&dW_X);
		cvReleaseMat(&dW_Y);
		cvReleaseMat(&U_X);
		cvReleaseMat(&U_Y);
		cvReleaseMat(&V_X);
		cvReleaseMat(&V_Y);
		cvReleaseMat(&W_X);
		cvReleaseMat(&W_Y);
		cvReleaseMat(&dU_prec);
		cvReleaseMat(&dV_prec);
		cvReleaseMat(&dW_prec);
	
		if (opt_Z){
			cvReleaseMat(&Xt_z);
			cvReleaseMat(&Yt_z);
			cvReleaseMat(&Z0_prec);
			cvReleaseMat(&Z0_X);
			cvReleaseMat(&Z0_Y);
			cvReleaseMat(&dZ0);
			cvReleaseMat(&dZ0_X);
			cvReleaseMat(&dZ0_Y);
			cvReleaseMat(&dZ0_prec);
		}
		cvReleaseMat(&SOR);
		cvReleaseMat(&SOR_prec);	

	//if this is not the last level, images must be upscaled before iterations
		if (level != (pyramid_level_end-1)) {
            const CvSize next_level_size = cvGetSize(Images_0_pyramid[0][level - 1]);
            CvMat *Interpolated_Data;

			Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
			cvResize(U, Interpolated_Data,CV_INTER_LINEAR);
			cvReleaseMat(&U);
			U = Interpolated_Data;
			//cvSmooth(U, U, CV_GAUSSIAN,3,0.5);

			Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
			cvResize(V, Interpolated_Data,CV_INTER_LINEAR);
			cvReleaseMat(&V);
			V = Interpolated_Data;
			//				cvSmooth(V, V, CV_GAUSSIAN,3,0.5);

			Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
			cvResize(W, Interpolated_Data,CV_INTER_LINEAR);
			cvReleaseMat(&W);
			W = Interpolated_Data;
			//				cvSmooth(W, W, CV_GAUSSIAN,3,0.5);
			
			Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
			cvResize(Z0, Interpolated_Data,CV_INTER_LINEAR);
			cvReleaseMat(&Z0);
			Z0 = Interpolated_Data;
	
				//cvSmooth(Z0, Z0, CV_GAUSSIAN,3,0.5);
		
			// tali: the projection matrix for the current pyramyid
			eta_curr = eta_curr/PyramidEta;
			if(ENABLE_ALPHA) { 
				alpha = alpha/PyramidEta;
				alpha_w = alpha_w/PyramidEta;
				alpha_z = alpha_z/PyramidEta;

			}

			for(view = 0; view < NumOfViews; view++) {
				ComputeProjMat(M_curr[view],R_Mat[view], T_Mat[view], C_Mat[view], eta_curr);
			}

#ifdef DEBUG
			sprintf(file, "%s/scaledProjMat%d.txt", TxtResultsDir,level-1);
				f = fopen(file,"w");
				WriteRes(f,M_curr[0]);
				fclose(f);
#endif
		}
#ifdef DEBUG	
		if (opt_Z) {
			sprintf(file, "%s/Z0_l%d.txt", TxtResultsDir,level);
			f = fopen(file,"w");
			WriteRes(f,Z0);
			fclose(f);
		}

		sprintf(file, "%s/W_l%d.txt", TxtResultsDir,level);
		f = fopen(file,"w");
		WriteRes(f,W);
		fclose(f);

		sprintf(file, "%s/U_l%d.txt", TxtResultsDir,level);
		f = fopen(file,"w");
		WriteRes(f,U);
		fclose(f);

		sprintf(file, "%s/V_l%d.txt", TxtResultsDir,level);
		f = fopen(file,"w");
		WriteRes(f,V);
		fclose(f);
		
#endif

	} // for(int level = pyramid_level_start - 1; level >= pyramid_level_end - 1; level--){
        
	if (U_end)
		cvCopy(U, U_end);
	if (V_end)
		cvCopy(V, V_end);
	if (W_end)
		cvCopy(W, W_end);
	cvReleaseMat(&U);
	cvReleaseMat(&V);
	cvReleaseMat(&W);

	
	if (Z0_end)
		cvCopy(Z0, Z0_end);
	cvReleaseMat(&Z0);
	
	for(view = 0; view < NumOfViews; view++){
		if(M_curr) {
			cvReleaseMat(&M_curr[view]);
		}
		
		if(COP && (view > 0)) {
			cvReleaseMat(&COP[view]);
		}

	}
}