#include <math.h>
#include <assert.h>
#include "Resolution.h"
#include "iniconfig.h"
#include "resize.h"
#include <stdlib.h>
#define snprintf _snprintf

#define MAX_HUE 150
//#define DIVERGENCE_8 0
//#define DISABLE_PIXEL_DISACTIVATION 0
#define GMRES 0
#define GMRES_KMAX 200
///#define OCCLUDED_PIXELS_LAPLACIAN 0
#define DISACTIVATION_THRESHOLD 1.7
#define DISACTIVATION_PRECENT  0.2
//#define VERBOSE_ENERGY 1
#define WARP_DETERMINANT_THRESHOLD 0.2// should be >= 0. and < 1.

#define EPSILON_PSI_SMOOTH 0.0001           // epsilon for the psi function
#define EPSILON_PSI_SMOOTH_W 0.00001           // epsilon for the psi function
#define EPSILON_PSI_DATA 1.               // epsilon for the psi function

#define MAX_INCR  105                  // maximum increment between two iterates in pixels
#define SCALE_B 1
#define MAX_NUM_OF_VIEWS 9

#define DEBUG_PIXEL 0
#define DEBUG_PIXEL_X 331
#define DEBUG_PIXEL_Y 87
#define DEBUG_PIXEL_COMPONENT 1
//tali 
#define ENABLE_ALPHA 1
#define OCCLUSION1 1

#define ASSERT  0
#define DEBUG  1
#define __BEGIN__       {

#define __END__         goto exit; exit: ; }

// Pixels that go beyond this border (in pixels) are disactivated. Must be at
// least 3, since gradient estimation for warped images is
// wrong 2 pixels from the border, and warping is wrong 1 pixel from the border
// but 4 (more than 2*sqrt(2)+1) is safer in
// case the transformed image border is at 45deg
#define DISACTIVATE_BORDER 4//3.9

#ifndef M_PI
# define M_PI               3.14159265358979323846
#endif

static void debug_me() {
}

static inline double square(double x) {
    return x*x;
}

static inline double fmax(double a, double b)
{
	return (a > b ? a : b);
}


CV_INLINE  unsigned char cvmGet8u( const CvMat* mat, int row, int col )
{
    assert( (unsigned)row < (unsigned)mat->rows &&
            (unsigned)col < (unsigned)mat->cols );
    assert( CV_MAT_TYPE(mat->type) == CV_8UC1 );
    return ((unsigned char*)(mat->data.ptr + (size_t)mat->step*row))[col];
}

CV_INLINE  static void  cvmSet8u( CvMat* mat, int row, int col, unsigned char value )
{
    assert( (unsigned)row < (unsigned)mat->rows &&
            (unsigned)col < (unsigned)mat->cols );
    assert( CV_MAT_TYPE(mat->type) == CV_8UC1 );

    ((unsigned char*)(mat->data.ptr + (size_t)mat->step*row))[col] = value;
}

// used to mark border pixels as disactivated (although they are not used anyway)
CV_INLINE  static void  cvmSet8u_borders( CvMat* mat, unsigned char value )
{
    int rows  = mat->rows;
    int cols = mat->cols;
    int i, j;

    // top border
    for(j = 0; j < cols; j++)
        cvmSet8u(mat,0,j,value);
    // bottom border
    for(j = 0; j < cols; j++)
        cvmSet8u(mat,rows-1,j,value);
    // left border
    for(i = 1; i < rows-1; i++)
        cvmSet8u(mat,i,0,value);
    // right border
    for(i = 1; i < rows-1; i++)
        cvmSet8u(mat,i,cols-1,value);
}

CV_INLINE  static void  cvmSet8u_col_borders( CvMat* mat, unsigned char value, int left, int col )
{
    int rows  = mat->rows;
    int cols = mat->cols;
    int i, j;

	if(left){
    // left border
		for(j=0; j < col; j++){
			for(i = 0; i < rows; i++)
				cvmSet8u(mat,i,j,value);
		}
	}else
	{
		for(j=cols; j > col; j--){
			for(i = 0; i < rows; i++)
				cvmSet8u(mat,i,j,value);
		}
	}
}



static inline double sign(double a)
{
    return (a > 0 ? 1.0 : -1.0);
}
static inline float sign(float a)
{
    return (a > 0.0 ? 1.0f : -1.0f);
}

/***************************************************Horizontal linear Interpolation**************************************************/
static inline void Interpolate_X(const CvMat* data, //data to be interpolated
				 const int x, //pixel x
				 const int y, //pixel y
				 const double alpha, //Interpolation coefficient
				 double *Interpolate, //Result
				 const int Nx){
    if(x < Nx - 1)
	*Interpolate = alpha * cvmGet(data, y, x + 1) + (1 - alpha) * cvmGet(data, y, x);
    else
	*Interpolate = alpha * cvmGet(data, y, x) + (1 - alpha) * cvmGet(data, y, x - 1); 
}

static inline void SmoothImage(CvArr *img) {
    cvSmooth(img,img, CV_GAUSSIAN, 3,3);
}

static inline void SmoothData(CvArr *img) {
#if 1
#ifndef _MSC_VER
#warning "not smoothing data between scale changes"
#endif
	(void)img;
#else
    cvSmooth(img,img, CV_GAUSSIAN, 3,3);
#endif
}

static inline double& Crop(double &val, double threshold) {
#if 1
    if (val > threshold)
        val = threshold;
    else if (val < -threshold)
        val = -threshold;
#endif
    return val;
}

static inline void CropArr(CvArr *arr, double threshold) {
#if 1
    cvMinS(arr, threshold, arr);
    cvMaxS(arr,-threshold, arr);
#endif
}


/***************************************************Initial Disparity d' estimation using left and right optical flows****************/
void DtEstimation(const CvMat *U, //Left optical flow
		  const CvMat *Ur, //Right optical flow
		  const CvMat *D0, //Disparity d at time t
		  CvMat*Dt)
{  
    double alpha;
    double xr0;
    double ur, u, d0;
    int xr0_int;
    int dimX = U->width;
    int dimY = U->height;
    //D' estimation
    for(int yl0 =0; yl0 < dimY; yl0++){
	for(int xl0 = 0; xl0 < dimX; xl0++){
	    //xr0 in the initial right image IR(x+d, yl0, t)
	    d0 =  cvmGet(D0,yl0,xl0);
	    xr0 = xl0 + d0;
	    //if D estimation is not correct, not changing Dt 
	    if(xr0 < 0 || xr0 > dimX)
		cvmSet(Dt,yl0,xl0, d0);
	    else{
		//Horizontal interpolation
		xr0_int = (int)cvFloor(xr0);
		alpha = xr0 - xr0_int;
		//Right flow estimation for the current point :U_r(x+d,yl0)
		Interpolate_X(Ur, xr0_int, yl0, alpha, &ur, dimX);
		//Left flow for the current point : U_l(x,yl0);
		u = cvmGet(U, yl0, xl0);
		//Disparity D' estimation  
		cvmSet(Dt, yl0, xl0, d0 + ur - u);
	    }
	}
    }
    //Results image save
    if (SF_Dt_Init_FileName)
        SaveResults(SF_Dt_Init_FileName, Dt);
}

/**************************************************Scene Flow Estimation*****************/
void ScaleChange (FILE* logfile, //the log file
                  const double eta, //reduction factor
                  int pyramid_levels, // total number of pyramid levels
                  const CvMat * const * const Il0_pyramid, //time t Image
                  const CvMat * const * const Ir0_pyramid, //time t Image
                  const CvMat * const * const Ilt_pyramid, //time t+1 Image
                  const CvMat * const * const Irt_pyramid, //time t+1 Image
                  int pyramid_level_start, // starting level
                  const CvMat* const U_start, //horizontal left optical flow initialization
                  const CvMat* const V_start, //vertical optical flow initialization
                  const CvMat* const D0_start, // disparity initialization
                  const CvMat* const Dt_start,
                  int pyramid_level_end,   // end pyramid level
                  CvMat* U_end, //final U
                  CvMat* V_end, //final V
                  CvMat* D0_end, //t disparity 
                  CvMat* Dt_end)
{
	(void)Il0_pyramid;
	(void)Ilt_pyramid;
	(void)Ir0_pyramid;
	(void)Irt_pyramid;
	(void)pyramid_levels;
	(void)eta;
	(void)logfile;
    if (U_start) {
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_start-1], U_start));
        assert(U_end);
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_end-1], U_end));
    }
    if (V_start) {
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_start-1], V_start));
        assert(V_end);
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_end-1], V_end));
    }
    if (D0_start) {
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_start-1], D0_start));
        assert(D0_end);
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_end-1], D0_end));
    }
    if (Dt_start) {
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_start-1], Dt_start));
        assert(Dt_end);
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_end-1], Dt_end));
    }
    if (pyramid_level_end == pyramid_level_start) {
        if (U_start)
            cvCopy(U_start, U_end);
        if (V_start)
            cvCopy(V_start, V_end);
        if (D0_start)
            cvCopy(D0_start, D0_end);
        if (Dt_start)
            cvCopy(Dt_start, Dt_end);
    }
   
    int interpolation = CV_INTER_LINEAR;
    double scale;
    if (pyramid_level_end > pyramid_level_start)
      interpolation = CV_INTER_AREA;  // downsizing is better with CV_INTER_AREA
    if (U_start) {
      cvResize(U_start, U_end, interpolation);
      scale = (double)U_end->width / U_start->width;
      cvConvertScale(U_end, U_end, scale);
    }
    if (V_start) {
      cvResize(V_start, V_end, interpolation);
      scale = (double)V_end->height / V_start->height;
      cvConvertScale(V_end, V_end, scale);
    }
    if (D0_start) {
      cvResize(D0_start, D0_end, interpolation);
      scale = (double)D0_end->width / D0_start->width;
      cvConvertScale(D0_end, D0_end, scale);
    }
   if (Dt_start) {
      cvResize(Dt_start, Dt_end, interpolation);
      scale = (double)Dt_end->width / Dt_start->width;
      cvConvertScale(Dt_end, Dt_end, scale);
    }
    
}

/*************************************** Gaussian density computation *******************/
static inline double NormalDensity(const double sigma, const double x){
    return (1./sqrt(2.*M_PI*fabs(sigma*sigma)))*exp((-1./(2.*sigma*sigma))*x*x);
}

/* lightweight convolution with 3x3 kernel */
static void GradientFast_32f( const CvMat* srcmat, CvMat* dstxmat, CvMat* dstymat )
{
    int  x, y;

    int src_type, dstx_type, dsty_type;
    src_type = CV_MAT_TYPE( srcmat->type );
    dstx_type = CV_MAT_TYPE( dstxmat->type );
    dsty_type = CV_MAT_TYPE( dstymat->type );
    assert(src_type == CV_32FC1 && dstx_type == CV_32FC1 && dsty_type == CV_32FC1);
    // assert(CV_ARE_SIZES_EQ( srcmat, dstxmat ));
//     assert(CV_ARE_SIZES_EQ( srcmat, dstymat ));
    int src_step = srcmat->step;
    int dst_step = dstxmat->step;
    assert(dst_step == dstymat->step);
    float* src = (float*)srcmat->data.ptr;
    float* dstx = (float*)dstxmat->data.ptr;
    float* dsty = (float*)dstymat->data.ptr;

    src_step /= sizeof(src[0]);
    dst_step /= sizeof(dstx[0]);

    for( y = 0; y < srcmat->height; y++, src += src_step, dstx += dst_step, dsty += dst_step)
    {
        float* srcprevline;
        float* srcnextline;
        float muly = 0.5;
        if (y == 0) {
            srcprevline = src;
            muly = 1;
        }
        else {
            srcprevline = src - src_step;
        }
        if (y == srcmat->height - 1) {
            srcnextline = src;
            muly = 1;
        }
        else {
            srcnextline = src + src_step;
        }
        for( x = 0; x < srcmat->width; x++ ) {
            float srcvalprevx;
            float srcvalnextx;
            float mulx = 0.5;
            if (x == 0) {
                srcvalprevx = src[x];
                mulx = 1;
            }
            else {
                srcvalprevx = src[x-1];
            }
            if (x == srcmat->width - 1) {
                srcvalnextx = src[x];
                mulx = 1;
            }
            else {
                srcvalnextx = src[x+1];
            }
            dstx[x] = (srcvalnextx-srcvalprevx)*mulx;
            dsty[x] = (srcnextline[x]-srcprevline[x])*muly;
        }
    }
}

/* lightweight convolution with 3x3 kernel */
static void ZBufMap_32f( const CvMat* src, CvMat* dst, const CvMat* mapx, const CvMat* mapy, float fillval )
{
    int  x, y;

    int src_type, dst_type, mapx_type, mapy_type;
    src_type = CV_MAT_TYPE( src->type );
    dst_type = CV_MAT_TYPE( dst->type );
    mapx_type = CV_MAT_TYPE( mapx->type );
    mapy_type = CV_MAT_TYPE( mapy->type );
    assert(src_type == CV_32FC1 && dst_type == CV_32FC1);
    assert(mapx_type == CV_32FC1 && mapy_type == CV_32FC1);
    assert(CV_ARE_SIZES_EQ( src, dst ));
    assert(CV_ARE_SIZES_EQ( src, mapx ));
    assert(CV_ARE_SIZES_EQ( src, mapy ));

    cvSet(dst, cvScalar(fillval));
    for( y = 0; y < src->height; y++)
    {
        for( x = 0; x < src->width; x++ ) {
            const double dstx = cvmGet(mapx,y,x);
            const double dsty = cvmGet(mapy,y,x);
            const double srcval = cvmGet(src,y,x);
            
            if (dstx < 0 || dstx > (dst->width-1) ||
                dsty < 0 || dsty > (dst->height-1))
                continue;
                
            for(int i=cvFloor(dsty); i <= cvCeil(dsty); i++) {
                for(int j=cvFloor(dstx); j <= cvCeil(dstx); j++) {
                    const double dstval = cvmGet(dst,i,j);
                    if (srcval < dstval)
                        cvmSet(dst,i,j,srcval);
                }
            }
        }
    }
}



static void DistBufMap_32f( CvMat* src, CvMat* dst, const CvMat* mapx, const CvMat* mapy, CvMat *X, CvMat *Y, CvMat *Z, const CvMat* COP)
{
    int  x, y;
    cvSet(dst, cvScalar(-1));
	double copx = cvmGet(COP, 0, 0);
	double copy = cvmGet(COP, 1, 0);
	double copz = cvmGet(COP, 2, 0);

    for( y = 0; y < dst->height; y++)
    {
        for( x = 0; x < dst->width; x++ ) {
            const double dstx = cvmGet(mapx,y,x);
            const double dsty = cvmGet(mapy,y,x);

            if (dstx < 0 || dstx > (dst->width-1) ||
                dsty < 0 || dsty > (dst->height-1))
                continue;
            
			
			const double X_val = cvmGet(X,y,x);
            const double Y_val = cvmGet(Y,y,x);
			const double Z_val = cvmGet(Z,y,x);
			const double curr_dist  = cvmGet(src,y,x);

            for(int i=cvFloor(dsty); i <= cvCeil(dsty); i++) {
                for(int j=cvFloor(dstx); j <= cvCeil(dstx); j++) {
                    const double prev_dist = cvmGet(dst,i,j);
                    if (curr_dist < prev_dist || prev_dist < 0)
                        cvmSet(dst,i,j,curr_dist);
                }
            }
        }
    }
}



/************** creates a distances map. for each of the pixels in the reference image it computes the distance of P from COP*****************/
static void Dist_map(CvMat* dst,  CvMat *X, CvMat *Y, CvMat *Z,CvMat *COP)
{
    int  x, y;
	double dist = 0.0;
	int dst_type;
    dst_type = CV_MAT_TYPE( dst->type );
    assert(dst_type == CV_32FC1);
	double copx = cvmGet(COP, 0, 0);
	double copy = cvmGet(COP, 1, 0);
	double copz = cvmGet(COP, 2, 0);


    for( y = 0; y < dst->height; y++)
    {
        for( x = 0; x < dst->width; x++ ) {
            const double X_val = cvmGet(X,y,x);
            const double Y_val = cvmGet(Y,y,x);
			const double Z_val = cvmGet(Z,y,x);
			dist  = sqrt((X_val - copx)*(X_val - copx) + (Y_val - copy)*(Y_val - copy) + (Z_val - copz)*(Z_val - copz));
            cvmSet(dst,y,x,dist);
        }
    }
}


/***************************Saving results in false color images********************/
void SaveImage(const char *filename, const CvArr* arr)
{
    int type = cvGetElemType(arr);
    if(type == CV_8UC1 || type == CV_8UC3) {
        cvSaveImage(filename, arr);
    }
    else {
        CvMat *charim = cvCreateMat(cvGetDimSize(arr,0), cvGetDimSize(arr,1), CV_8UC1);
        cvConvert(arr, charim);
        cvSaveImage(filename, charim);
        cvReleaseMat(&charim);
    }
}

/***************************Saving results in false color images********************/
void SaveResults(const char *filename, const CvArr* arr, int addscale, double min, double max)
{  
    double values[MAX_HUE+1];
    int i;
    char chaine[256];
    CvFont police;
    int ydiv = 1;
    CvSize size;

    if (CV_IS_SPARSE_MAT( arr ) && ((CvSparseMat*)arr)->dims == 2) {
        size.height = ((CvSparseMat*)arr)->size[0];
        size.width = ((CvSparseMat*)arr)->size[1];
    }
    else {
        size = cvGetSize(arr);
    }
    
    CvMat *Temp = cvCreateMat(size.height, size.width, CV_8UC1);

    if (min == 0. && max == -1.) {
        cvMinMaxLoc(arr, &min, &max);
    }
    //Data scaling
    if(max <= min) {
	cvSetZero(Temp);
    }
    else {
        double scale = (double)MAX_HUE/(max-min);
        double shift =  -(double)MAX_HUE*min/(max - min);
        
        if (!CV_IS_SPARSE_MAT( arr )) {
            cvConvertScale(arr, Temp, scale, shift);
        }
        else {
            cvSetZero(Temp);
            CvSparseMatIterator mat_iterator;
            const CvSparseMat *array = (CvSparseMat*) arr;
            CvSparseNode* node = cvInitSparseMatIterator( array, &mat_iterator );
            for( ; node != 0; node = cvGetNextSparseNode( &mat_iterator ))
            {
                int* idx = CV_NODE_IDX( array, node ); /* get pointer to the element indices */
                double val;
                switch(array->type) {
                    case CV_8UC1:
                        val = *(unsigned char*)CV_NODE_VAL( array, node );
                        break;
                    case CV_32FC1:
                        val = *(float*)CV_NODE_VAL( array, node );
                        break;
                    case CV_64FC1:
                        val = *(double*)CV_NODE_VAL( array, node );
                        break;
                    default:
                        val = 1.;
                }
                cvSetReal2D(Temp, idx[0], idx[1], val*scale+shift);
            }
        }
    }
  
    for(i = 0; i <= MAX_HUE; i++)
	values[i] = min + i*(max - min)/(double)MAX_HUE;
  
    //We use HSV images 
    IplImage *HSV_arr = cvCreateImage(size, IPL_DEPTH_8U, 3);
    //H channel update
    cvSetImageCOI(HSV_arr, 1);
    cvCopy(Temp, HSV_arr);
    //S and V channels update

    //HSV space use
    cvSet(Temp, cvScalar(255));
    cvSetImageCOI(HSV_arr, 2);
    cvCopy(Temp, HSV_arr);
    cvSetImageCOI(HSV_arr, 3);
    cvCopy(Temp, HSV_arr);
    cvSetImageCOI(HSV_arr, 0);
    cvReleaseMat(&Temp);
 
    //We put a value scale on the images
    if (size.width < 50 || size.height < 32)
        addscale = 0;
    if (size.height < MAX_HUE)
        ydiv = cvCeil((double)MAX_HUE/size.height);
    assert(ydiv > 0);
    if(addscale){
	for(i = 0; i <= MAX_HUE; i++){
	    cvLine(HSV_arr, cvPoint(0,i/ydiv), cvPoint(10,i/ydiv), cvScalar(i,255,255));
	}
    }

    //HSV => RGB
    IplImage *RGB_arr = cvCreateImage(size, IPL_DEPTH_8U, 3);
    cvCvtColor(HSV_arr, RGB_arr, 55);
    cvReleaseImage(&HSV_arr);

    if(addscale){
	cvInitFont(&police, CV_FONT_HERSHEY_SIMPLEX, 0.3, 0.3, 0.);
	snprintf(chaine, 256,"%g", values[0]);
	cvPutText(RGB_arr, chaine, cvPoint(11,10), &police, cvScalar(0,0,0));
	snprintf(chaine, 256,"%g", values[MAX_HUE/2]);
	cvPutText(RGB_arr, chaine, cvPoint(11,5+MAX_HUE/(2*ydiv)), &police, cvScalar(0,0,0));
	snprintf(chaine, 256,"%g", values[MAX_HUE]);
	cvPutText(RGB_arr, chaine, cvPoint(11,MAX_HUE/ydiv), &police, cvScalar(0,0,0));
    }
  
    //Writing the values scale 
    cvSaveImage(filename, RGB_arr);
    cvReleaseImage(&RGB_arr);
}

/************************************* Psi values estimation ********************/
static inline double PsiData(const double x){
    return sqrt(x + EPSILON_PSI_DATA*EPSILON_PSI_DATA);
}
static inline double PsiSmooth(const double x){
    return sqrt(x + EPSILON_PSI_SMOOTH*EPSILON_PSI_SMOOTH);
}
static inline double PsiSmoothW(const double x){
    return sqrt(x + EPSILON_PSI_SMOOTH_W*EPSILON_PSI_SMOOTH_W);
}

//**************************************Divergence operator coefficients computation***********************************************************
static inline double DPsiSmooth(const double grad2)
{
    //phi(s^2) = 2sqrt(1+s^2)-2
    return 1./sqrt((EPSILON_PSI_SMOOTH*EPSILON_PSI_SMOOTH + grad2));
}

static inline double DPsiSmoothW(const double grad2)
{
    //phi(s^2) = 2sqrt(1+s^2)-2
    return 1./sqrt((EPSILON_PSI_SMOOTH_W*EPSILON_PSI_SMOOTH_W + grad2));
}

static inline double DPsiData(const double intensity2)
{
    //phi(s^2) = 2sqrt(1+s^2)-2
    return 1./sqrt((EPSILON_PSI_DATA*EPSILON_PSI_DATA + intensity2));
}

//**************************************Half pixel interpolation for the divergence coefficients in the principal directions *****************
static inline void divergence_coefficients_interpolate(CvMat * coeffs, const int i, const int j, double *b1, double *b2, double *b3, double *b4)
{
    double valcentre = cvmGet(coeffs, i, j);
    //The function is called only on the inside points so no condition over the image side
    *b1 = (valcentre + cvmGet(coeffs, i, j + 1)) / 2;	
    *b2 = (valcentre + cvmGet(coeffs, i, j - 1)) / 2;
    *b3 = (valcentre + cvmGet(coeffs, i + 1, j)) / 2;	
    *b4 = (valcentre + cvmGet(coeffs, i - 1, j)) / 2;
}

//**************************************Half pixel interpolation for the divergence coefficients in the diagonal directions *****************
static inline void divergence_coefficients_interpolate_diag(CvMat * coeffs, const int i, const int j, double *b5, double *b6, double *b7, double *b8)
{
    double valcentre = cvmGet(coeffs, i, j);
    //The function is called only on the inside points so no condition over the image side
    *b5 = (valcentre + cvmGet(coeffs, i+1, j + 1)) / 2;	//en X
    *b6 = (valcentre + cvmGet(coeffs, i+1, j - 1)) / 2;
    *b7 = (valcentre + cvmGet(coeffs, i - 1, j+1)) / 2;	//en Y
    *b8 = (valcentre + cvmGet(coeffs, i - 1, j-1)) / 2;
}


//**************************************Images gradient computation********************************
static inline void Derivative(const CvArr * image,  int dx, int dy, CvArr * derivative_int, CvArr * derivative)
{
    if (derivative_int == 0)
        derivative_int = derivative;
    if ((dx == 1 && dy == 0) || (dx == 0 && dy == 1)) {
        cvSobel(image, derivative_int, dx, dy, CV_SCHARR);
        cvConvertScale(derivative_int, derivative, 1.0/32.0);
    }
    else {
        assert(dx+dy == 2);
        cvSobel(image, derivative_int, dx, dy, 3);
        cvConvertScale(derivative_int, derivative, 1.0/4.0);
    }
}

/*************************************Matrix rescaling for linear system computation ******************/
void MatrixNormalize(CvMat *M, CvMat*B, const int Nbcol,  const int inf, const int sup){
    double line_norm;
    int i,j;
    for(i=inf; i < sup; i++){
        line_norm = 0.;
        for(j=0; j < Nbcol; j++)
            line_norm += square(cvGetReal2D(M, i, j));
        line_norm = sqrt(line_norm);//euclidean norm of each matrix line
        for(j=0; j < Nbcol; j++)
            cvSetReal2D(M,i,j,cvGetReal2D(M, i, j)/line_norm); 
        cvSetReal2D(B,i,0,cvGetReal2D(B, i, 0)/line_norm); 
    
    }
}


/***************************Energy computation for each multiscale level with each part of the energy******************************************/
/***************************Take as input the warped images******************************************************/

void EnergyComputation(const CvMat *Il0, //Left t image
		       const CvMat *Ir0w, //Right t image
		       const CvMat *Iltw, //Left t+1 image
		       const CvMat *Irtw, //Right t+1 image 
		       const CvMat* U    , //Horizontal optical flow
		       const CvMat* V    , //Vertical optical flow
		       const CvMat* Dt       , //disparity t+1
		       const CvMat* D0          , //Disparity t
                       const CvMat *disactivated_pixels_leftflow,
                       const CvMat *disactivated_pixels_rightflow,
                       const CvMat *disactivated_pixels_stereoinit,
                       const CvMat *disactivated_pixels_stereo,
		       const int N_x, //Horizontal pixels number 
		       const int N_y, //Vertical pixels number
		       double *data, //Data term
		       double *smooth, //Smoothing term
		       const double gamma , //Images gradient weight
		       const double lambda, //Smooth parameter for disparity variation
		       const double mu, //Smooth parameter for t disparity 
		       const int doleftflow, //Booleans
		       const int dorightflow, 
		       const int dodisparity, 
		       const int dodisparity0, 
		       double *Efl, //Left optical flow contribution
		       double *Efr, //Right optical flow contribution
		       double *Est, //t+1 stereo contribution
		       double *Es0, //t stereo contribution
		       double *Esflp, 
		       double *Esfrp, 
		       double *Esdp, 
		       double *Esd0p, 
		       double *Eintensityp, 
		       double *Egradientp){

    double Esmooth, Esmoothtmp;
    double El,Er,Elr0, Elr;
    double Es, Esfl, Esfr, Esd, Esd0;
    double Ei, Eg, Eintensity, Egradient;
    
    CvMat *Iltw_X       = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Iltw_Y       = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Il0_X  = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Il0_Y  = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Irtw_X      = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Irtw_Y      = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Ir0w_X = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Ir0w_Y = cvCreateMat(N_y, N_x, CV_32FC1);
   
    CvMat *U_X = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *U_Y = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *V_X = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *V_Y = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Dt_X   = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Dt_Y   = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *D0_X  = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *D0_Y  = cvCreateMat(N_y, N_x, CV_32FC1);

    if(doleftflow) {
        assert(Iltw && Il0 && U && V);
        GradientFast_32f(U, U_X, U_Y);
        GradientFast_32f(V, V_X, V_Y);
    }
    if (dorightflow) {
        assert(Irtw && Ir0w);
        fprintf(stderr,"cannot compute regularization term for right flow\n");
        return;
    }
    if (dodisparity) {
        assert(Iltw && Irtw && Dt);
	GradientFast_32f(Dt, Dt_X, Dt_Y);
    }
    if (dodisparity0) {
        assert(Il0 && Ir0w && D0);
    }
    if (dodisparity || dodisparity0) {
        GradientFast_32f(D0, D0_X, D0_Y);
    }
    
    // gradients
    CvMat *derivative_tmp = 0;
    if (CV_MAT_TYPE(Il0->type) == CV_8UC1)
        derivative_tmp = cvCreateMat(N_y, N_x, CV_16SC1);
    if (Iltw) {
        assert(Iltw && Iltw_X && Iltw_Y);
        Derivative(Iltw, 1, 0, derivative_tmp, Iltw_X);
        Derivative(Iltw, 0, 1, derivative_tmp, Iltw_Y);
    }
    if (Irtw) {
        assert(Irtw && Irtw_X && Irtw_Y);
        Derivative(Irtw, 1, 0, derivative_tmp, Irtw_X);
        Derivative(Irtw, 0, 1, derivative_tmp, Irtw_Y);
    }
    if (Il0) {
        assert(Il0 && Il0_X && Il0_Y);
        Derivative(Il0, 1, 0, derivative_tmp, Il0_X);
        Derivative(Il0, 0, 1, derivative_tmp, Il0_Y);
    }
    if (Ir0w) {
        assert(Ir0w && Ir0w_X && Ir0w_Y);
        Derivative(Ir0w, 1, 0, derivative_tmp, Ir0w_X);
        Derivative(Ir0w, 0, 1, derivative_tmp, Ir0w_Y);
    }
    cvReleaseMat(&derivative_tmp);

    // energy
    El = 0.;
    Er = 0.;
    Elr = 0.;
    Elr0 = 0.;
    Esmooth = 0.;
    Esfl = 0.;
    Esfr = 0.;
    Esd = 0;
    Esd0 = 0;
    Eintensity = 0.;
    Egradient = 0.;
    for(int yl0=0; yl0 < N_y; yl0++){
	for(int xl0=0; xl0 < N_x; xl0++){
            double ilt = 0., irt = 0.;
            double iltw_X = 0, iltw_Y = 0.;
            double irtw_X = 0, irtw_Y = 0.;
            double il0 = 0., ir0 = 0.;
            double il0_X = 0., il0_Y = 0.;
            double ir0w_X = 0., ir0w_Y = 0.;
            int pixel_doleftflow = 1;
            int pixel_dorightflow = 1;
            int pixel_dostereoinit = 1;
            int pixel_dostereo = 1;
            if (disactivated_pixels_leftflow)
                pixel_doleftflow = !cvmGet8u(disactivated_pixels_leftflow,yl0,xl0);
            if (disactivated_pixels_rightflow)
                pixel_dorightflow = !cvmGet8u(disactivated_pixels_rightflow,yl0,xl0);
            if (disactivated_pixels_stereoinit)
                pixel_dostereoinit = !cvmGet8u(disactivated_pixels_stereoinit,yl0,xl0);
            if (disactivated_pixels_stereo)
                pixel_dostereo = !cvmGet8u(disactivated_pixels_stereo,yl0,xl0);
            if (Il0) {
                il0 = cvGetReal2D(Il0,yl0,xl0);
                il0_X = cvmGet(Il0_X,yl0,xl0);
                il0_Y = cvmGet(Il0_Y,yl0,xl0);
            }
            if (Ir0w) {
                ir0 = cvGetReal2D(Ir0w,yl0,xl0);
                ir0w_X = cvmGet(Ir0w_X,yl0,xl0);
                ir0w_Y = cvmGet(Ir0w_Y,yl0,xl0);
            }
            if (Iltw) {
                ilt = cvGetReal2D(Iltw,yl0,xl0);
                iltw_X = cvmGet(Iltw_X,yl0,xl0);
                iltw_Y = cvmGet(Iltw_Y,yl0,xl0);
            }
	    if (Irtw) {
                irt = cvGetReal2D(Irtw,yl0,xl0);
                irtw_X = cvmGet(Irtw_X,yl0,xl0);
                irtw_Y = cvmGet(Irtw_Y,yl0,xl0);
            }

	    
            // data terms
	    if (pixel_doleftflow && (doleftflow || (dorightflow && dodisparity && dodisparity0))) {
                Ei = square(ilt - il0);
                Eg = square(iltw_X - il0_X)  + square(iltw_Y - il0_Y);
		El += PsiData(Ei  + gamma*Eg);
                Eintensity += Ei;
                Egradient += Eg;
	    }
            if (pixel_dorightflow && (dorightflow || (doleftflow && dodisparity && dodisparity0))) {
                Ei = square(irt - ir0);
                Eg = square(irtw_X - ir0w_X)  + square(irtw_Y - ir0w_Y);
		Er += PsiData(Ei + gamma*Eg);
                Eintensity += Ei;
                Egradient += Eg;
            }
            if (pixel_dostereoinit && (dodisparity0 || (doleftflow && dorightflow && dodisparity))) {
                Ei = square(ir0 - il0);
                Eg = square(ir0w_X - il0_X) + square(ir0w_Y - il0_Y);
		Elr0 += PsiData(Ei + gamma*Eg);
                Eintensity += Ei;
                Egradient += Eg;
            }
            if (pixel_dostereo && (dodisparity || (doleftflow && dorightflow && dodisparity0))) {
                Ei = square(irt - ilt);
                Eg = square(irtw_X - iltw_X)   + square(irtw_Y - iltw_Y);
		Elr += PsiData(Ei  + gamma*Eg);
                Eintensity += Ei;
                Egradient += Eg;
            }
            // regularization terms
            Esmoothtmp = 0.;
	    if(doleftflow) {
                double u_X = cvmGet(U_X,yl0,xl0);
                double u_Y = cvmGet(U_Y,yl0,xl0);
                double v_X = cvmGet(V_X,yl0,xl0);
                double v_Y = cvmGet(V_Y,yl0,xl0);
                Es = square(u_X) + square(u_Y) + square(v_X) + square(v_Y);
		Esmoothtmp += Es;
                Esfl += Es;
	    }
            if (dorightflow) {
            }
            if (dodisparity0) {
                double d0_X = cvmGet(D0_X,yl0,xl0);
                double d0_Y = cvmGet(D0_Y,yl0,xl0);
                Es = mu*(square(d0_X) + square(d0_Y));
                Esmoothtmp += Es;
                Esd0 += Es;
            }
            if (dodisparity) {
                double d0_X = cvmGet(D0_X,yl0,xl0);
                double d0_Y = cvmGet(D0_Y,yl0,xl0);
                double dt_X = cvmGet(Dt_X,yl0,xl0);
                double dt_Y = cvmGet(Dt_Y,yl0,xl0);
                Es = lambda*(square(dt_X - d0_X) + square(dt_Y - d0_Y));
                Esmoothtmp += Es;
                Esd += Es;
            }
	    Esmooth += PsiSmooth(Esmoothtmp);
	    
	}
    }
    *data = El + Er + Elr + Elr0;
    *smooth = Esmooth;
    if (Efl)
        *Efl = El;
    if (Efr)
        *Efr = Er;
    if (Est)
        *Est = Elr;
    if (Es0)
        *Es0 = Elr0;
    if (Esflp)
        *Esflp = Esfl;
    if (Esfrp)
        *Esfrp = Esfr;
    if (Esdp)
        *Esdp = Esd;
    if (Esd0p)
        *Esd0p = Esd0;
    if (Eintensityp)
        *Eintensityp = Eintensity;
    if (Egradientp)
        *Egradientp = Egradient;
    cvReleaseMat(&U_X);
    cvReleaseMat(&U_Y);
    cvReleaseMat(&V_X);
    cvReleaseMat(&V_Y);
    cvReleaseMat(&Dt_X);
    cvReleaseMat(&Dt_Y);
    cvReleaseMat(&D0_X);
    cvReleaseMat(&D0_Y);
    cvReleaseMat(&Iltw_X);
    cvReleaseMat(&Iltw_Y);
    cvReleaseMat(&Irtw_X);
    cvReleaseMat(&Irtw_Y);
    cvReleaseMat(&Il0_X);
    cvReleaseMat(&Il0_Y);
    cvReleaseMat(&Ir0w_X);
    cvReleaseMat(&Ir0w_Y);
}


void EnergyComputation3D(const CvMat *I0, //Images at time t
					   const CvMat **I0w,   //Wraped images at time t 
					   const CvMat **Itw,   //Wraped images at time t+1
					   const CvMat* U    , //Horizontal scene flow
					   const CvMat* V    , //Vertical scene flow
					   const CvMat* W    , //Z scene flow t+1
					   const CvMat* Z0   , //surface at time t
                       const CvMat *disactivated_pixels_ref_flow,
                       const CvMat **disactivated_pixels_flow,
                       const CvMat **disactivated_pixels_Z0,
                       const CvMat **disactivated_pixels_Zt,
					   const int N_x, //Horizontal pixels number 
					   const int N_y, //Vertical pixels number
		               double *data, //Data term
		               double *smooth, //Smoothing term
		               const double mu, //Smooth parameter for t disparity 
		               const int do_ref_flow, //Booleans
		               const int do_not_ref_flow, 
		               const int do_Z0, 
		               const int do_Zt, 
		               double *E_Ref_flow, //Reference optical flow contribution
		               double *E_Flow, //Right optical flow contribution
		               double *E_Stereo_0, //t stereo contribution
		               double *E_Stereo_t, //t+1 stereo contribution
		               double *Esflp, 
		               double *Esz0p,
					   double *Eintensityp,
					   int level){

    double Esmooth, Esmoothtmp;
	double E_ref_flow = 0;
	double E_flow[MAX_NUM_OF_VIEWS] = {0}, E_stereo_0[MAX_NUM_OF_VIEWS] = {0}, E_stereo_t[MAX_NUM_OF_VIEWS] = {0};
    double Es, Esfl, Esz0;
    double Ei, Eintensity, Edata_pixel, Esmooth_pixel;
	int view = 1;
	CvMat *EnergyMapData = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *EnergyMapSmooth = cvCreateMat(N_y, N_x, CV_32FC1);
	cvZero(EnergyMapData);
	cvZero(EnergyMapSmooth);
 
    CvMat *U_x = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *U_y = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *V_x = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *V_y = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *W_x = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *W_y = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Z0_x = cvCreateMat(N_y, N_x, CV_32FC1);
    CvMat *Z0_y   = cvCreateMat(N_y, N_x, CV_32FC1);
    
	if(do_ref_flow || do_not_ref_flow || do_Zt) {
		assert(U && V && W);
		GradientFast_32f(U, U_x, U_y);
        GradientFast_32f(V, V_x, V_y);
		GradientFast_32f(W, W_x, W_y);
	}

    if (do_Z0) {
        GradientFast_32f(Z0, Z0_x, Z0_y);
    }
    
    // energy
    Esmooth = 0.;
    Esfl = 0.;
	Esz0 = 0.;
    Eintensity = 0.;
	
    for(int yl0=0; yl0 < N_y; yl0++){
		for(int xl0=0; xl0 < N_x; xl0++){
			Edata_pixel = 0.;
			Esmooth_pixel = 0.;
            double it = 0., itw = 0.;
			double i0 = 0., i0w = 0.;
            int pixel_do_ref_flow = 1;
            int pixel_do_flow = 1;
            int pixel_doZ0 = 1;
            int pixel_doZt = 1;
			for(view = 1; view < NumOfViews; view++){
				if (disactivated_pixels_ref_flow)
					pixel_do_ref_flow = !cvmGet8u(disactivated_pixels_ref_flow,yl0,xl0);
				if (disactivated_pixels_flow)
					pixel_do_flow = !cvmGet8u(disactivated_pixels_flow[view],yl0,xl0);
				if (disactivated_pixels_Z0)
					pixel_doZ0 = !cvmGet8u(disactivated_pixels_Z0[view],yl0,xl0);
				if (disactivated_pixels_Zt)
					pixel_doZt = !cvmGet8u(disactivated_pixels_Zt[view],yl0,xl0);
           
				i0 = cvGetReal2D(I0,yl0,xl0);

				if (I0w) {
					i0w = cvGetReal2D(I0w[view],yl0,xl0);
				}
				if (Itw) {
					it = cvGetReal2D(Itw[0],yl0,xl0);
					itw = cvGetReal2D(Itw[view],yl0,xl0);
				}

				// data terms
				if (pixel_do_ref_flow && do_ref_flow) {
					Ei = square(it - i0);
					Edata_pixel += PsiData(Ei);
					E_ref_flow += PsiData(Ei);
					Eintensity += Ei;
				}
				if (pixel_do_flow && do_not_ref_flow) {
					Ei = square(itw - i0w);
					Edata_pixel += PsiData(Ei);
					E_flow[view] += PsiData(Ei);
					Eintensity += Ei;
				}
				if (pixel_doZ0 && do_Z0) {
					Ei = square(i0w - i0);
					Edata_pixel += PsiData(Ei);
					E_stereo_0[view] += PsiData(Ei);
					Eintensity += Ei;
				 }
				if (pixel_doZt && do_Zt) {
					Ei = square(itw - it);
					Edata_pixel += PsiData(Ei);
					E_stereo_t[view] += PsiData(Ei);
					Eintensity += Ei;
				 }
			}
            // regularization terms
            Esmoothtmp = 0.;
			if(do_ref_flow || do_not_ref_flow || do_Zt) {
                double u_x = cvmGet(U_x,yl0,xl0);
                double u_y = cvmGet(U_y,yl0,xl0);
                double v_x = cvmGet(V_x,yl0,xl0);
                double v_y = cvmGet(V_y,yl0,xl0);
				double w_x = cvmGet(W_x,yl0,xl0);
                double w_y = cvmGet(W_y,yl0,xl0);
                Es = square(u_x) + square(u_y) + square(v_x) + square(v_y)+ square(w_x) + square(w_y);
				Esmoothtmp += Es;
                Esfl += Es;
				Esmooth += PsiSmooth(Esmoothtmp);
			}
            
            if (do_Z0) {
                double z0_x = cvmGet(Z0_x,yl0,xl0);
                double z0_y = cvmGet(Z0_y,yl0,xl0);
                Es = mu*(square(z0_x) + square(z0_y));
                Esmoothtmp = Es;
                Esz0 += Es;
            }
            
			Esmooth_pixel = PsiSmooth(Esmoothtmp);
			Esmooth += Esmooth_pixel;
			cvmSet(EnergyMapData, yl0, xl0, Edata_pixel);
			cvmSet(EnergyMapSmooth, yl0, xl0, Esmooth_pixel);
		}
    }
#ifdef DEBUG
	FILE *f;
	char file[256];
	sprintf(file,"%s/SmoothEnergyMap%d.txt", TxtResultsDir, level);
	f = fopen(file,"w");
	WriteRes(f,EnergyMapSmooth);
	fclose(f);
	sprintf(file,"%s/DataEnergyMap%d.txt", TxtResultsDir, level);
	f = fopen(file,"w");
	WriteRes(f,EnergyMapData);
	fclose(f);
#endif
	
	cvReleaseMat(&EnergyMapData);
	cvReleaseMat(&EnergyMapSmooth);
	*data = E_ref_flow;
	for(view = 1; view < NumOfViews; view++) {
		*data += E_flow[view] + E_stereo_0[view] + E_stereo_t[view];
		if (E_Flow)
			E_Flow[view] = E_flow[view];
		if (E_Stereo_0)
			E_Stereo_0[view] = E_stereo_0[view];
		if (E_Stereo_t)
			E_Stereo_t[view] = E_stereo_t[view];
	}

    *smooth = Esmooth;
	if(E_Ref_flow)
		*E_Ref_flow = E_ref_flow;
    if (Esflp)
        *Esflp = Esfl;
    if (Esz0p)
        *Esz0p = Esz0;
    if (Eintensityp)
        *Eintensityp = Eintensity;
    
    cvReleaseMat(&U_x);
    cvReleaseMat(&U_y);
    cvReleaseMat(&V_x);
    cvReleaseMat(&V_y);
    cvReleaseMat(&W_x);
    cvReleaseMat(&W_y);
    cvReleaseMat(&Z0_x);
    cvReleaseMat(&Z0_y);
    
}



//************************3 elements maxima computation*****************
static inline double max_3_arg(const double x, const double y, const double z)
{
    return fmax(fmax(x,y),z);
}

//**************************************Vector euclidean norm computation*******************************************************
static inline double norm_vect(const double x, const double y)
{
    return sqrt(x * x + y * y);
}

//**************************************Matrix euclidean norm computation*******************************************************
static inline double norm_mat(const double a, const double b, const double c, const double d)
{
    return sqrt(a * a + b * b + c * c + d * d);
}

//***************************************Gaussian data pyramid computation **********
static inline void GaussianPyramid(CvArr * Data, CvArr * Undersampled_Data)
{
    cvPyrDown(Data, Undersampled_Data);
}

//***************************************Angle of a vector computation with the X axis*********************
static inline double teta_vect(const double x, const double y){
    return acos(x/norm_vect(x,y));
}

//***********************************************Underpolate data to the k-ith resolution level with the reduction factor eta ***************
#if 1
static inline void Underpolate(const CvMat* data, //Data to be undersampled
			       CvMat* data_underpolated, //Undersampled data
			       const int k, //Resolution level
			       const double eta, //Reduction factor
			       const double total_scale)
{
	(void)eta;
	(void)k;
    cvResize(data, data_underpolated, CV_INTER_AREA);
    cvConvertScale(data_underpolated,data_underpolated, total_scale);
}
#else
static void Underpolate(CvMat* data, CvMat* data_unterpolated, const int k, const double eta, const double total_scale){
   
    int N_x = cvGetSize(data).width;
    int N_y = cvGetSize(data).height;
    CvMat* data_cur = cvCreateMat(N_y, N_x, CV_32FC1);
    cvCopy(data, data_cur);
    double eta_total = 1.;
    for(int i = 0; i < k; i++){
        eta_total *= eta;
        CvMat* data_next =  cvCreateMat(cvCeil(N_y*eta_total), cvCeil(N_x*eta_total) , CV_32FC1);
        cvResize(data_cur, data_next, CV_INTER_AREA);
        //SmoothData(data_next);
	cvReleaseMat(&data_cur);
	data_cur = data_next;
    }
    assert(CV_ARE_SIZES_EQ(data_cur,data_underpolated));
    cvConvertScale(data_cur,data_underpolated, total_scale);
    cvReleaseMat(&data_cur);
}
#endif

/***************************************************euclidean norm in R^n*****************************/
double n_norm(CvMat *x, const int Nx){
    double n = 0.;
    for(int k=0; k < Nx; k++)
        n += square(cvGetReal2D(x,k,0));
    return sqrt(n);
}

/***************************************************QR Factorization*********************************/
void QR(CvMat *M, CvMat *Q, CvMat *R, const int Nx, const int Ny){
    cvSetZero(Q);
    cvSetZero(R);
  
}

/**************************************************Matrix Transposition*****************************/
void Transpose(CvMat *Q1, CvMat *Q2, const int N){
    double tmp;
    cvSetZero(Q2);
    for(int i = 0; i < N; i++){
        for(int j = 0; j <= i; j++){
            if(i != j){
                tmp = cvGetReal2D(Q1,i,j);
                cvSetReal2D(Q2,j,i,tmp);
                tmp = cvGetReal2D(Q1,j,i);
                cvSetReal2D(Q2,i,j,tmp);
            }
            else{
                tmp = cvGetReal2D(Q1,i,j);
                cvSetReal2D(Q2,i,j,tmp);
            }
        }
    }
}

/********************************************Scene flow matrix with vector multiplication to spare memory******/
void SFMatrixProduct(CvMat *SFMatrix, CvMat *x, CvMat *x_end, const int Nx, const int Ny, const int dimension){
  
    int i;
    int N = dimension * Nx * Ny;
    for (i = 0; i < N; i++) {
		    
	if(i < dimension*Nx){
	    double x1 = cvGetReal2D(x,i,0);
	    double x2 = cvGetReal2D(x,i+dimension*Nx,0);
	    double m1 = cvGetReal2D(SFMatrix,i,0);
	    double m2 = cvGetReal2D(SFMatrix,i,1);
	    cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
	}
	else if(i >= dimension*Nx*(Ny-1)){
	    double x1 = cvGetReal2D(x,i,0);
	    double x2 = cvGetReal2D(x,i-dimension*Nx,0);
	    double m1 = cvGetReal2D(SFMatrix,i,0);
	    double m2 = cvGetReal2D(SFMatrix,i,1);
	    cvSetReal2D(x_end, i, 0,  m1*x1 + m2*x2);
	}
	else{
	    if(i%(dimension*Nx) == 0){
		double x1 = cvGetReal2D(x,i,0);
		double x2 = cvGetReal2D(x,i+dimension,0);
		double m1 = cvGetReal2D(SFMatrix,i,0);
		double m2 = cvGetReal2D(SFMatrix,i,1);
		cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
	    }
	    else if((i+dimension)%(dimension*Nx)==0){
		double x1 = cvGetReal2D(x,i,0);
		double x2 = cvGetReal2D(x,i-dimension,0);
		double m1 = cvGetReal2D(SFMatrix,i,0);
		double m2 = cvGetReal2D(SFMatrix,i,1);
		cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
	    }
	    else{

		int rest1 = i%(dimension*Nx);
		if(rest1 == 1 || rest1 == 2 || rest1 == 3){
                    double x1 = cvGetReal2D(x,i,0);
                    double x2 = cvGetReal2D(x,i+dimension,0);
                    double m1 = cvGetReal2D(SFMatrix,i,0);
                    double m2 = cvGetReal2D(SFMatrix,i,1);
                    cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2); 
		}
		else if(rest1 == 4*Nx - 3 || rest1 == 4*Nx - 2 || rest1 == 4*Nx - 1){
		    double x1 = cvGetReal2D(x,i,0);
		    double x2 = cvGetReal2D(x,i-dimension,0);
		    double m1 = cvGetReal2D(SFMatrix,i,0);
		    double m2 = cvGetReal2D(SFMatrix,i,1);
		    cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2);
		}
		else{
		    int rest2 = rest1 % dimension;
		    
		    if(rest2 == 0){

                        double x1 = cvGetReal2D(x,i,0);
                        double x2 = cvGetReal2D(x,i-dimension*Nx,0);
                        double x3 = cvGetReal2D(x,i-dimension,0);
                        double x4 = cvGetReal2D(x,i+1,0);
                        double x5 = cvGetReal2D(x,i+2,0);
                        double x6 = cvGetReal2D(x,i+3,0);
                        double x7 = cvGetReal2D(x,i+dimension*Nx,0);
                        double x8 = cvGetReal2D(x,i+dimension,0);
			
                        double m1 = cvGetReal2D(SFMatrix,i,0);
                        double m2 = cvGetReal2D(SFMatrix,i,4);
                        double m3 = cvGetReal2D(SFMatrix,i,6);
                        double m4 = cvGetReal2D(SFMatrix,i,1);
                        double m5 = cvGetReal2D(SFMatrix,i,2);
                        double m6 = cvGetReal2D(SFMatrix,i,3);
                        double m7 = cvGetReal2D(SFMatrix,i,5);
                        double m8 = cvGetReal2D(SFMatrix,i,7);
			 
                        cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);
					
                    }
		    else if(rest2 == 1){
			//v equation
                        double x1 = cvGetReal2D(x,i,0);
                        double x2 = cvGetReal2D(x,i-dimension*Nx,0);
                        double x3 = cvGetReal2D(x,i-dimension,0);
                        double x4 = cvGetReal2D(x,i-1,0);
                        double x5 = cvGetReal2D(x,i+1,0);
                        double x6 = cvGetReal2D(x,i+2,0);
                        double x7 = cvGetReal2D(x,i+dimension*Nx,0);
                        double x8 = cvGetReal2D(x,i+dimension,0);
			
                        double m1 = cvGetReal2D(SFMatrix,i,1);
                        double m2 = cvGetReal2D(SFMatrix,i,4);
                        double m3 = cvGetReal2D(SFMatrix,i,6);
                        double m4 = cvGetReal2D(SFMatrix,i,0);
                        double m5 = cvGetReal2D(SFMatrix,i,2);
                        double m6 = cvGetReal2D(SFMatrix,i,3);
                        double m7 = cvGetReal2D(SFMatrix,i,5);
                        double m8 = cvGetReal2D(SFMatrix,i,7);
			 
                        cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);

		    }

		    else if(rest2 == 2){
			//dprime equation
                        double x1 = cvGetReal2D(x,i,0);
                        double x2 = cvGetReal2D(x,i-dimension*Nx,0);
                        double x3 = cvGetReal2D(x,i-dimension,0);
                        double x4 = cvGetReal2D(x,i-2,0);
                        double x5 = cvGetReal2D(x,i-1,0);
                        double x6 = cvGetReal2D(x,i+1,0);
                        double x7 = cvGetReal2D(x,i+dimension*Nx,0);
                        double x8 = cvGetReal2D(x,i+dimension,0);
			
                        double m1 = cvGetReal2D(SFMatrix,i,2);
                        double m2 = cvGetReal2D(SFMatrix,i,4);
                        double m3 = cvGetReal2D(SFMatrix,i,6);
                        double m4 = cvGetReal2D(SFMatrix,i,0);
                        double m5 = cvGetReal2D(SFMatrix,i,1);
                        double m6 = cvGetReal2D(SFMatrix,i,3);
                        double m7 = cvGetReal2D(SFMatrix,i,5);
                        double m8 = cvGetReal2D(SFMatrix,i,7);
			 
                        cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);
			  
					    
		    }
		    else{
			//d equation
			double x1 = cvGetReal2D(x,i,0);
			double x2 = cvGetReal2D(x,i-dimension*Nx,0);
			double x3 = cvGetReal2D(x,i-dimension,0);
			double x4 = cvGetReal2D(x,i-3,0);
			double x5 = cvGetReal2D(x,i-2,0);
			double x6 = cvGetReal2D(x,i-1,0);
			double x7 = cvGetReal2D(x,i+dimension*Nx,0);
			double x8 = cvGetReal2D(x,i+dimension,0);
			
			double m1 = cvGetReal2D(SFMatrix,i,3);
			double m2 = cvGetReal2D(SFMatrix,i,4);
			double m3 = cvGetReal2D(SFMatrix,i,6);
			double m4 = cvGetReal2D(SFMatrix,i,0);
			double m5 = cvGetReal2D(SFMatrix,i,1);
			double m6 = cvGetReal2D(SFMatrix,i,2);
			double m7 = cvGetReal2D(SFMatrix,i,5);
			double m8 = cvGetReal2D(SFMatrix,i,7);
			 
			cvSetReal2D(x_end, i, 0, m1*x1 + m2*x2 + m3*x3 + m4*x4 + m5*x5 + m6*x6 + m7*x7 + m8*x8);
					
		    }
		    
		}
	    }
	}
    }

}






/*******************************************Scalar Product in R^k****************************/
double scalar_prod(CvMat *x, CvMat *y, const int k){
    double scalar = 0.;
    for(int p = 0; p < k; p++)
        scalar += cvGetReal2D(x,p,0)*cvGetReal2D(y,p,0);
    return scalar;
}

/*************************************Extract a matrix column ****************/
void ColumnExtract(CvMat* Q, CvMat* q1, const int j, const int N){
    for(int p = 0; p < N; p++)
        cvSetReal2D(q1,p,0,cvGetReal2D(Q,p,j));
}

/**************************************************GMRES Method to solve the scene flow final linear system ***********/
void GmRes(CvMat *M, CvMat *b, CvMat *x0, CvMat *xk, CvMat *rk, const int Kmax, const double epsilon, const int Nx, const int Ny, const int dimension){
    int i,j,k, p;
    double h;
    const int N = Nx*Ny*dimension;
    CvMat *r0 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *q1 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *q2 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *Q =  cvCreateMat(N, Kmax, CV_32FC1);
    CvMat *x1 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *H = cvCreateMat(Kmax + 1, Kmax, CV_32FC1);
    CvMat *H_end = cvCreateMat(Kmax, Kmax, CV_32FC1);
    CvMat *c = cvCreateMat(Kmax, 1, CV_32FC1);
    CvMat *s = cvCreateMat(Kmax, 1, CV_32FC1);
    CvMat *g = cvCreateMat(Kmax+1, 1, CV_32FC1);
    CvMat *g_end = cvCreateMat(Kmax, 1, CV_32FC1);
    CvMat *X = cvCreateMat(N, 1, CV_32FC1);
    CvMat *Y = cvCreateMat(Kmax, 1, CV_32FC1);
  
    cvSetZero(H);
    cvSetZero(Q);
    cvSetZero(H_end);
    cvSetZero(g);
    cvSetZero(g_end);
    //Initial residu r0 initialization
    SFMatrixProduct(M, x0, x1, Nx, Ny, dimension);
    cvSub(b, x1, r0);
    double beta = n_norm(r0,N);
    cvConvertScale(r0, q1, 1./beta); 
    //Q initialization
    for(p=0; p < N; p++)
        cvSetReal2D(Q, p, 0, cvGetReal2D(q1, p, 0));
  
    k = 1;
    double rho = beta;
    cvSetReal2D(g,0,0,beta);
    //Main loop
    while(k <= Kmax /*&& rho > epsilon * beta*/ ){
        SFMatrixProduct(M, q1, q2, Nx, Ny, dimension); //q2 = M*qk 
        //extracting vk
        ColumnExtract(Q, q1, k-1, N);
        for(j = 1; j <= k; j++){
            CvMat *gamma =  cvCreateMat(N, 1, CV_32FC1);
            ColumnExtract(Q,gamma, j-1, N); //gamma = qj
            h = scalar_prod(q2, gamma, N); //<M*qk,qj>
            cvSetReal2D(H, j-1, k-1,h); //Hjk = <M*qk,qj>
            cvConvertScale(gamma, gamma, h);
            cvSub(q2, gamma, q2);
            cvReleaseMat(&gamma);
        }
        h = n_norm(q2,N);
        cvSetReal2D(H, k, k-1,h);
        cvConvertScale(q2,q2,1./h);
        //Q update
        if(k < Kmax){
            for(p=0; p < N; p++)
                cvSetReal2D(Q, p, k, cvGetReal2D(q2, p, 0));
        }
     
        //Givens Rotations
        for(i = 1; i < k; i++){
            double t1 = cvGetReal2D(H,k,i-1);
            double t2 = cvGetReal2D(H,k,i);
            double ci = cvGetReal2D(c, i-1,0);
            double si = cvGetReal2D(s, i-1,0);
            cvSetReal2D(H,k,i-1,ci*t1+si*t2);
            cvSetReal2D(H,k,i,ci*t2-si*t1);
        }
        //Givens rotations computations
        double h1 = cvGetReal2D(H, k-1,k-1);
        double h2 = cvGetReal2D(H, k,k-1);
        double nu = sqrt(square(h1) + square(h2));
        double ck = h1/nu;
        double sk = h2/nu;
        cvSetReal2D(c,k-1,0,ck);
        cvSetReal2D(s,k-1,0,sk);
     
        //H update
        cvSetReal2D(H,k-1,k-1,ck*h1+sk*h2);
        cvSetReal2D(H,k,k-1,0.);
     
        double a1 = cvGetReal2D(g,k-1,0);
        double a2 = cvGetReal2D(g,k,0);
        cvSetReal2D(g,k-1,0,ck*a1+sk*a2);
        cvSetReal2D(g,k,0,ck*a2-sk*a1);

        rho = fabs(ck*a2-sk*a1);
        k++;
    }
    //Get out the Kmax*Kmax matrix from H
    for(i = 0; i < Kmax; i++){
        cvSetReal2D(g_end, i,0, cvGetReal2D(g,i,0));
        for(j = 0; j < Kmax; j++)
            cvSetReal2D(H_end, i,j, cvGetReal2D(H,i,j));
    }
    //Solving the linear system
    double det = cvDet(H_end);
    cvSolve(H_end, g_end, Y);
  
    //X=QY
    cvGEMM(Q, Y, 1, NULL, 0., X);
 
    //The actual solution
    cvAdd(x0, X, xk);

    //The residu
    SFMatrixProduct(M, X, rk, Nx, Ny, dimension);
    cvSub(r0, rk, rk);
		  
    cvReleaseMat(&r0);
    cvReleaseMat(&q1);
    cvReleaseMat(&q2);
    cvReleaseMat(&x1);
    cvReleaseMat(&H);
    cvReleaseMat(&H_end);
    cvReleaseMat(&Q);
    cvReleaseMat(&X);	  
    cvReleaseMat(&Y);	  
    cvReleaseMat(&c);
    cvReleaseMat(&s);
    cvReleaseMat(&g);
    cvReleaseMat(&g_end);
}

/**************************************************GMRES Method to solve the scene flow final linear system ***********/
void GmRes2(CvMat *M, CvMat *b, CvMat *x0, CvMat *xk, CvMat *rk, const int Kmax, const double epsilon, const int Nx, const int Ny, const int dimension){
    int i,j, p;
    double h;
    const int N = Nx*Ny*dimension;
    CvMat *r0 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *v1 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *v2 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *V =  cvCreateMat(N, Kmax, CV_32FC1);
    CvMat *x1 = cvCreateMat(N, 1, CV_32FC1);
    CvMat *H = cvCreateMat(Kmax + 1, Kmax, CV_32FC1);
    CvMat *H_end = cvCreateMat(Kmax, Kmax, CV_32FC1);
    CvMat *c = cvCreateMat(Kmax, 1, CV_32FC1);
    CvMat *s = cvCreateMat(Kmax, 1, CV_32FC1);
    CvMat *g = cvCreateMat(Kmax+1, 1, CV_32FC1);
    CvMat *g_end = cvCreateMat(Kmax, 1, CV_32FC1);
    CvMat *X = cvCreateMat(N, 1, CV_32FC1);
    CvMat *Y = cvCreateMat(Kmax, 1, CV_32FC1);
  
    cvSetZero(H);
    cvSetZero(V);
    cvSetZero(H_end);
    cvSetZero(g);
    cvSetZero(g_end);
    //Initial residu r0 initialization
    SFMatrixProduct(M, x0, x1, Nx, Ny, dimension);
    cvSub(b, x1, r0);
    double beta = n_norm(r0,N);
    cvConvertScale(r0, v1, 1./beta); 
    //V Krylov vectors matrix initialization
    for(p=0; p < N; p++)
        cvSetReal2D(V, p, 0, cvGetReal2D(v1, p, 0));
  
    double rho = beta;
    cvSetReal2D(g,0,0,beta);
    //Main loop
    for(j = 1; j <= Kmax; j++){
	//v1 = vj
	ColumnExtract(V,v1,j-1,N);
	CvMat *gamma =  cvCreateMat(N, 1, CV_32FC1);
	SFMatrixProduct(M, v1, gamma, Nx, Ny, dimension); //gamma = M*vj 
        for(i = 1; i <= j; i++){
	   ColumnExtract(V,v2,i-1,N); //v2 = Vi		
	   h = scalar_prod(gamma, v2, N); //h_ij = <M*vj,vi>	
	   cvSetReal2D(H, i-1, j-1,h); //Hij = <M*vj,vi>
	   cvConvertScale(v2, v2, h); //vi = h_ij * vi
           cvSub(gamma, v2,v1); //vj+1 = M*vj - h_ij*vi 	
        }
	h = n_norm(v1, N);
	cvSetReal2D(H, j, j-1,h);		
	//normalize the new krylov vector
	cvConvertScale(v1, v1, h);
	//V update
        if(j < Kmax){
            for(p=0; p < N; p++)
                cvSetReal2D(V, p, j, cvGetReal2D(v1, p, 0));
        }	

	cvReleaseMat(&gamma);		
     
        //Givens Rotations
       /* int k;
	   for(i = 1; i < k; i++){
            double t1 = cvGetReal2D(H,k,i-1);
            double t2 = cvGetReal2D(H,k,i);
            double ci = cvGetReal2D(c, i-1,0);
            double si = cvGetReal2D(s, i-1,0);
            cvSetReal2D(H,k,i-1,ci*t1+si*t2);
            cvSetReal2D(H,k,i,ci*t2-si*t1);
        }
        //Givens rotations computations
        double h1 = cvGetReal2D(H, k-1,k-1);
        double h2 = cvGetReal2D(H, k,k-1);
        double nu = sqrt(square(h1) + square(h2));
        double ck = h1/nu;
        double sk = h2/nu;
        cvSetReal2D(c,k-1,0,ck);
        cvSetReal2D(s,k-1,0,sk);
     
        //H update
        cvSetReal2D(H,k-1,k-1,ck*h1+sk*h2);
        cvSetReal2D(H,k,k-1,0.);
     
        double a1 = cvGetReal2D(g,k-1,0);
        double a2 = cvGetReal2D(g,k,0);
        cvSetReal2D(g,k-1,0,ck*a1+sk*a2);
        cvSetReal2D(g,k,0,ck*a2-sk*a1);

        rho = fabs(ck*a2-sk*a1);
        k++;*/
    }
    //Get out the Kmax*Kmax matrix from H
   /* for(i = 0; i < Kmax; i++){
        cvSetReal2D(g_end, i,0, cvGetReal2D(g,i,0));
        for(j = 0; j < Kmax; j++)
            cvSetReal2D(H_end, i,j, cvGetReal2D(H,i,j));
    }
    //Solving the linear system
    //double det = cvDet(H_end);
    cvSolve(H_end, g_end, Y);
  
    //X=QY
    cvGEMM(Q, Y, 1, NULL, 0., X);
 
    //The actual solution
    cvAdd(x0, X, xk);

    //The residu
    SFMatrixProduct(M, X, rk, Nx, Ny, dimension);
    cvSub(r0, rk, rk);*/
		  
    cvReleaseMat(&r0);
    cvReleaseMat(&v1);
    cvReleaseMat(&v2);
    cvReleaseMat(&x1);
    cvReleaseMat(&H);
    cvReleaseMat(&H_end);
    cvReleaseMat(&V);
    cvReleaseMat(&X);	  
    cvReleaseMat(&Y);	  
    cvReleaseMat(&c);
    cvReleaseMat(&s);
    cvReleaseMat(&g);
    cvReleaseMat(&g_end);
}



void PyramidBuild(CvMat *I,
                  double eta,
                  int pyramid_levels,
                  CvMat **I_pyramid)//in: array[pyramid_levels] of non-allocated pointers. out: pyramid
{
    //Images Gaussian Pyramid Computation
    double eta_total = 1.;
    int N_x = cvGetSize(I).width;
    int N_y = cvGetSize(I).height;

    I_pyramid[0] = cvCloneMat(I);
    
    for (int level = 1; level < pyramid_levels; level++) {
        eta_total *= eta;
        CvSize size_i = cvSize(cvCeil(N_x*eta_total),cvCeil(N_y*eta_total));
	//Images memory allocation
	I_pyramid[level] = cvCreateMat(size_i.height, size_i.width, CV_MAT_TYPE(I->type));       
	cvResize(I_pyramid[level-1], I_pyramid[level], CV_INTER_AREA);
    }
}

void PyramidSmooth(int pyramid_levels,
                   CvMat **I_pyramid)
{
    for (int level = 0; level < pyramid_levels; level++)
        SmoothImage(I_pyramid[level]);  // Il0_pyramid is pre-smoothed
}

void PyramidCopy(int pyramid_levels,
                 const CvMat * const * const In_pyramid,
                 CvMat ** Out_pyramid)
{
    for (int level = 0; level < pyramid_levels; level++) {
        Out_pyramid[level] = cvCloneMat(In_pyramid[level]);
    }
}

/**************************************************Optical Flow computation using Brox et al method*****************/
void OpticalFlow(FILE* logfile, //The log file
	       const double eta, //reduction factor
	       int pyramid_levels, // total number of pyramid levels
	       const CvMat * const * const Il0_pyramid, //time t Image
	       const CvMat * const * const Ilt_pyramid, //time t+1 Image
               int pyramid_level_start, // starting level
	       const CvMat* U_start, // in : X component of Optical Flow
	       const CvMat* V_start, // in : X component of Optical Flow
               int pyramid_level_end,   // end pyramid level
	       CvMat* U_end, //final U
	       CvMat* V_end, //final V
	       const double h, //grid step for finite elements
	       const double epsilon_outer, //stopping criteria for outer iterations 
	       const double epsilon_inner, //stopping criteria for inner iterations
	       const double epsilon_SOR, //stopping criteria for the linear system resolution
	       const int max_iter_outer, //maximal number of outer iterations
	       const int max_iter_inner, //maximal number of inner iterations
	       const int max_iter_SOR, //maximal number of linear system resolution iterations
	       const double omega_SOR, //SOR method weight
	       const double alpha, //smoothing parameter
	       const double gamma) //the gradients weight
{
    int N_x,N_y;
    int pyramidtype;
    CvMat *U, *V;

    const int dimension = 2;

    assert(pyramid_level_start >= pyramid_level_end);
    assert(pyramid_level_end >= 1);
    assert(pyramid_level_start <= pyramid_levels);
    pyramidtype = CV_MAT_TYPE(Il0_pyramid[pyramid_level_start-1]->type);
    assert(pyramidtype == CV_8UC1 || pyramidtype == CV_32FC1);
    assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_start-1], U_start));
    assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_start-1], V_start));
    assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_end-1], U_end));
    assert(CV_ARE_SIZES_EQ(Il0_pyramid[pyramid_level_end-1], V_end));
    assert(CV_ARE_TYPES_EQ(Il0_pyramid[pyramid_level_start-1],Ilt_pyramid[pyramid_level_start-1]));
    
    //intial optical flow is set to the lowest resolution
    {
        CvSize size_start = cvGetSize(Il0_pyramid[pyramid_level_start-1]);
        
        U = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
        V = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
    }
    //Underpolate(U_start, U, pyramid_level_start - 1, eta, (double)U->width / U_start->width);
    cvCopy(U_start, U);
    //Underpolate(V_start, V, pyramid_level_start - 1, eta, (double)V->height / V_start->height);
    cvCopy(V_start, V);
    if(logfile){
        double max_Fx, min_Fx, max_Fy, min_Fy;
        cvMinMaxLoc(U, &min_Fx, &max_Fx);
        cvMinMaxLoc(V, &min_Fy, &max_Fy);
        fprintf(logfile, "initial optical flow (reduced resolution): %g<u<%g %g<v<%g\n", min_Fx, max_Fx, min_Fy, max_Fy);
        fflush(logfile);
    }
  
    //Loop over the resolution levels
    for(int level = pyramid_level_start - 1; level >= pyramid_level_end - 1; level--){
        assert(Il0_pyramid[level] && Ilt_pyramid[level]);
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[level], Ilt_pyramid[level]));
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[level], U));
        assert(CV_ARE_SIZES_EQ(Il0_pyramid[level], V));

        CvSize level_size = cvGetSize(Il0_pyramid[level]);
        int xl0, yl0;

        if (OF_I0_FileName) {
            SaveImage(OF_I0_FileName,  Il0_pyramid[level]);
        }
        if (OF_It_FileName) {
            SaveImage(OF_It_FileName,  Ilt_pyramid[level]);
        }
	//Size of the current images
	N_x = level_size.width;
	N_y = level_size.height;


	//Maps used for the images warping : Initialization
	CvMat *MapIdentityX = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *MapIdentityY = cvCreateMat(N_y, N_x, CV_32FC1);
	for(yl0=0; yl0 < N_y; yl0++){
	    for(xl0=0; xl0 < N_x; xl0++){
		cvmSet(MapIdentityX, yl0, xl0, xl0);
		cvmSet(MapIdentityY, yl0, xl0, yl0);
	    }
	}
	//Creation and initialization of the warped image I_{l}(x+u, y+v) and the coordinates maps used to create it
	CvMat *Iltw = cvCreateMat(N_y, N_x, pyramidtype);
	CvMat *MapX_Ilt =  cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *MapY_Ilt =  cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *MapX_Ilt_inner =  cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *MapY_Ilt_inner =  cvCreateMat(N_y, N_x, CV_32FC1);
        
	//Images gradients computation
        CvMat *derivative_tmp = 0;
        if (pyramidtype == CV_8UC1)
            derivative_tmp = cvCreateMat(N_y, N_x, CV_16SC1);
	CvMat *Il0_X = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *Il0_Y = cvCreateMat(N_y, N_x, CV_32FC1);
        CvMat *Iltw_X = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *Iltw_Y = cvCreateMat(N_y, N_x, CV_32FC1);
	// Second derivatives
	CvMat *Iltw_XX = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *Iltw_XY = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *Iltw_YY = cvCreateMat(N_y, N_x, CV_32FC1);
	    
        // compute gradients of the only non-warped image
        assert(Il0_X && Il0_Y);
        Derivative(Il0_pyramid[level], 1, 0, derivative_tmp, Il0_X);
        Derivative(Il0_pyramid[level], 0, 1, derivative_tmp, Il0_Y);


	//M is the matrix of the final linear system and B the second member of this system
        //int M_size[2] = {2 * N_x * N_y, 6};
        int M_dim;
#if DIVERGENCE_8
        M_dim = 10;
#else  // !DIVERGENCE_8
        M_dim = 6;
#endif  // !DIVERGENCE_8
	CvMat *M = cvCreateMat(dimension*N_x*N_y, M_dim, CV_32FC1); 
	CvMat *B = cvCreateMat(dimension*N_x*N_y,1,CV_32FC1);

        CvMat *disactivated_pixels = cvCreateMat(N_y, N_x, CV_8UC1);
        cvSetZero(disactivated_pixels);
        cvmSet8u_borders(disactivated_pixels, 127);
        
        //divergence coefficients
	CvMat *divergence_coefficients = cvCreateMat(N_y, N_x, CV_32FC1);

	//Saving the initial flow images at beginning of each resolution level
        if (OF_U_FileName) {
            SaveResults(OF_U_FileName, U);
        }	     
        if (OF_V_FileName) {
            SaveResults(OF_V_FileName, V);
        }
     
	CvMat *U_prec = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *V_prec = cvCreateMat(N_y, N_x, CV_32FC1);

	//Flow increments and their derivatives	
	CvMat *dU = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *dV = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *dU_X = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *dU_Y = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *dV_X = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *dV_Y = cvCreateMat(N_y, N_x, CV_32FC1);

	 
	// flow gradients
	CvMat *U_X = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *U_Y = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *V_X = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *V_Y = cvCreateMat(N_y, N_x, CV_32FC1);
	 
	CvMat *dU_prec = cvCreateMat(N_y, N_x, CV_32FC1);
	CvMat *dV_prec = cvCreateMat(N_y, N_x, CV_32FC1);	     
       
	//For the final linear system and SOR iterations
	CvMat *SOR = cvCreateMat(dimension*N_x*N_y,1, CV_32FC1);		
	CvMat *SOR_prec = cvCreateMat(dimension*N_x*N_y,1,CV_32FC1);
        
	//Main loop over the optical flow values
        int iter_SOR = 0, iter_inner = 0, iter_outer = 0;
        int first_iteration = 1;
	double norm = 0., norm_inner = 0., norm_SOR = 0.;
	while((norm > epsilon_outer && iter_outer <= max_iter_outer) || iter_outer < 2 || first_iteration == 1){
	    first_iteration = 0;
	    
	    //Coordinate maps update
	    cvAdd(MapIdentityX, U, MapX_Ilt);
	    cvAdd(MapIdentityY, V, MapY_Ilt);
	
	    //Images warping
	    // Images warping + Warped images gradient computation
	    // (Warped images are smoothed)
	    cvRemap(Ilt_pyramid[level], Iltw, MapX_Ilt, MapY_Ilt);
	    SmoothImage(Iltw);
	    if (OF_It_Warped_FileName) {
		SaveImage(OF_It_Warped_FileName, Iltw);
	    }
	    //Warped images gradient computation
            assert(Iltw && Iltw_X && Iltw_Y && Iltw_XX && Iltw_XY && Iltw_YY);
	    Derivative(Iltw, 1, 0, derivative_tmp, Iltw_X);
	    Derivative(Iltw, 0, 1, derivative_tmp, Iltw_Y);
	    Derivative(Iltw, 2, 0, derivative_tmp, Iltw_XX);
	    Derivative(Iltw, 1, 1, derivative_tmp, Iltw_XY);
	    Derivative(Iltw, 0, 2, derivative_tmp, Iltw_YY);


	    //Flow gradients computation
	    GradientFast_32f(U, U_X, U_Y);
	    GradientFast_32f(V, V_X, V_Y);
	     
            //Increments initialization
	    cvSetZero(dU);
	    cvSetZero(dV);
	  
	    norm_inner = 0.;
	    first_iteration = 1;
	    iter_inner = 0;
	    //secondary fixed point iterations
	    while((norm_inner > epsilon_inner && iter_inner <= max_iter_inner) || first_iteration == 1){
                int i;
                
		first_iteration = 0;

		cvAdd(MapX_Ilt, dU, MapX_Ilt_inner);
		cvAdd(MapY_Ilt, dV, MapY_Ilt_inner);

		//flow increments gradients computation
		GradientFast_32f(dU, dU_X, dU_Y);
		GradientFast_32f(dV, dV_X, dV_Y);
		for(yl0 = 0; yl0 < N_y; yl0++){
		    for(xl0 = 0; xl0 < N_x; xl0++){
			double du_X = cvmGet(dU_X, yl0, xl0);
			double du_Y = cvmGet(dU_Y, yl0, xl0);
			double u_X = cvmGet(U_X, yl0, xl0);
			double u_Y = cvmGet(U_Y, yl0, xl0);
			double dv_X = cvmGet(dV_X, yl0, xl0);
			double dv_Y = cvmGet(dV_Y, yl0, xl0);
			double v_X = cvmGet(V_X, yl0, xl0);
			double v_Y = cvmGet(V_Y, yl0, xl0);
                        double s2 = (square(u_X + du_X) + square(u_Y + du_Y) +
                                     square(v_X + dv_X) + square(v_Y + dv_Y));
			double a = DPsiSmooth(s2) ;
			cvmSet(divergence_coefficients, yl0, xl0, a);
		    }
		}
		  
		//Computing M and B for the final linear system 
		cvSetZero(M);
		cvSetZero(B);

		//The dimension*N_x first lines in M correspond to the points on the y = 0 on the image
		int inf = 0;
                int sup = dimension*N_x;
		for (i = inf; i < sup; i++) {
		    cvmSet(M, i, 0, -1./h); 
		    cvmSet(M, i, 1,  1./h);			 
		}

		//Points on the lower horizontal side of the image      
		inf = dimension*N_x*(N_y-1);
		sup = dimension*N_x*N_y;
		for (i = inf; i < sup; i++) {
                    cvmSet(M, i, 0,  1./h); 
                    cvmSet(M, i, 1, -1./h);
		}

	
		inf = dimension*N_x;
		sup = dimension*N_x*(N_y-1) - 1;
		     
                //We fill in the lines of the matrix M corresponding to images pixels excepted points on the upper and lower horizontal side
		for(i = inf; i <= sup; i+=dimension){
                    yl0 = i/(dimension*N_x);
		    xl0 = (i % (inf))/dimension;
		    if(xl0 == 0){ //if point on the left side of image
			for(int d=0; d<dimension; d++) {
                            cvmSet(M,i+d,0,-1./h);
			    cvmSet(M,i+d,1, 1./h);
			}
		    }
		    else if(xl0 == (N_x-1)){//if point on the right side
                        for(int d=0; d<dimension; d++) {
			    cvmSet(M,i+d,0, 1./h);
			    cvmSet(M,i+d,1,-1./h);  
			}
		    }
		    else{
                        //Pixel inside the integration domain 
                        double ilt = 0., il0 = 0.;				//tali: il0=Il0(x,y) ilt = Iltw(x,y)
                        double ilt_X = 0., ilt_Y = 0.;
                        double iltw_X = 0., iltw_Y = 0.;
                        double il0_X = 0., il0_Y = 0.;
                        double iltw_XX = 0., iltw_XY = 0., iltw_YY = 0.;
                        double iltw_X_X = 0., iltw_X_Y = 0.;
                        double iltw_Y_X = 0., iltw_Y_Y = 0.;
                        //Interpolated divergence coefficients
                        double b_f_1, b_f_2, b_f_3, b_f_4;
			

                        // gradients of the functions used for image warping
                        double uw_X = cvmGet(U_X, yl0, xl0);
                        double uw_Y = cvmGet(U_Y, yl0, xl0);
                        double vw_X = cvmGet(V_X, yl0, xl0);
                        double vw_Y = cvmGet(V_Y, yl0, xl0);
   
                        // Half pixel interpolation for the divergence coefficients 
                        divergence_coefficients_interpolate(divergence_coefficients, yl0, xl0, &b_f_1, &b_f_2, &b_f_3, &b_f_4);
			//If using the 8 points scheme for the divergence
#if DIVERGENCE_8
                        divergence_coefficients_interpolate_diag(divergence_coefficients, yl0, xl0, &b_f_5, &b_f_6, &b_f_7, &b_f_8);
#endif  // DIVERGENCE_8

			//Current coordinates of the followed point 
			double xlt = cvmGet(MapX_Ilt_inner,yl0,xl0);
			double ylt = cvmGet(MapY_Ilt_inner,yl0,xl0);

		
			
#if DIVERGENCE_8
			double b_f_5, b_f_6, b_f_7, b_f_8;
                        //For the 8-divergence
                        double alpha_p_u, alpha_p_v, alpha_d_u, alpha_d_v;
    
			//U and V Gradients angle with the X axis
			double u_X = cvmGet(U_X, yl0, xl0) + cvmGet(dU_X, yl0, xl0);
			double u_Y = cvmGet(U_Y, yl0, xl0) + cvmGet(dU_Y, yl0, xl0);
			double v_X = cvmGet(V_X, yl0, xl0) + cvmGet(dV_X, yl0, xl0);
			double v_Y = cvmGet(V_Y, yl0, xl0) + cvmGet(dV_Y, yl0, xl0);

			//Coefficients computation for 8-divergence in u and v
                        // must be 1 when the gradient is aligned with axes, 0 when diagonal
			//We check if gradients are 0
			if(u_X == 0. && u_Y == 0.) //We take constant coefficients
                            //if(1)
                            alpha_p_u = 0.5;
			else{
                            if (u_X == 0)
                                alpha_p_u = 1;
                            else
                                alpha_p_u = fabs(fabs(atan(u_Y/u_X))/M_PI_4-1);
			}

			if(v_X == 0. && v_Y == 0.)
                            //if(1) 
			    alpha_p_v = 0.5;
			else{
                            if (v_X == 0)
                                alpha_p_v = 1;
                            else
                                alpha_p_v = fabs(fabs(atan(v_Y/v_X))/M_PI_4-1);
			}
			

			//The validity constraint for the directional weights for the divergence 
                        alpha_d_u = (1.-alpha_p_u)/2.;
			alpha_d_v = (1.-alpha_p_v)/2.;
#endif  // DIVERGENCE_8
                        //Pixels out must not be taken into account in the energy data term
                        int pixel_is_disactivated = cvmGet8u(disactivated_pixels,yl0,xl0);

			//is the point visible at time t+1 ?
			if(!pixel_is_disactivated &&                          
                           (xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h  ||
                            ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h)){
                            //printf("disactivate %d,%d\n", x, yl0);
                            cvmSet8u(disactivated_pixels,yl0,xl0,127);
                            pixel_is_disactivated = 1;
			}
		        
			double a_v_1 = 0;
                        double a_u_2 = 0;
			//If using the 8 points scheme for the divergence
#if DIVERGENCE_8
                        double a_u_1 = (alpha/(4.*h*h))*(alpha_p_u*(b_f_1 + b_f_2 + b_f_3 + b_f_4) + alpha_d_u*(b_f_5 + b_f_6 + b_f_7 + b_f_8));
                        double a_v_2 = (alpha/(4.*h*h))*(alpha_p_v*(b_f_1 + b_f_2 + b_f_3 + b_f_4) + alpha_d_v*(b_f_5 + b_f_6 + b_f_7 + b_f_8));
#else  // !DIVERGENCE_8
                        double a_u_1 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
                        double a_v_2 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);

#endif  // !DIVERGENCE_8
		        double psi_flowL = 0.;
                            //Data term estimation
                        if (!pixel_is_disactivated) {
                            il0 = cvGetReal2D(Il0_pyramid[level], yl0, xl0); 
                            il0_X = cvmGet(Il0_X,  yl0, xl0);
                            il0_Y = cvmGet(Il0_Y,  yl0, xl0);
			   
                            ilt = cvGetReal2D(Iltw,yl0,xl0);
                            iltw_X = cvmGet(Iltw_X, yl0, xl0);
                            iltw_Y = cvmGet(Iltw_Y, yl0, xl0);
			    iltw_XX = cvmGet(Iltw_XX,yl0,xl0);
			    iltw_XY = cvmGet(Iltw_XY,yl0,xl0);
			    iltw_YY = cvmGet(Iltw_YY,yl0,xl0);

                            // In Maple:
                            // Iltw(x,y):=Ilt(x+u,y+v);
                            // diff(Iltw(x,y),x);
                            // diff(Iltw(x,y),y);
                            // then, express D[1](Ilt) as a function of D[1](Iltw) and D[2](Iltw)
                            // and D[2](Ilt) as a function of D[1](Iltw) and D[2](Iltw)
							
							//tali: using chain rule and caremer to compute ilt_x ilt_y: Iltw(x,y) = Ilt(x+u(x,y),y+v(x,y))
                            double det = (1+uw_X)*(1+vw_Y) - uw_Y*vw_X;
                            if (det <= WARP_DETERMINANT_THRESHOLD) {
                                // image is reversed, set everything to 0
                                if (!pixel_is_disactivated) {
                                    cvmSet8u(disactivated_pixels,yl0,xl0,191);
                                    pixel_is_disactivated = 1;
                                }
                            }
                            else {              
                                ilt_X = ((1+vw_Y)*iltw_X - vw_X*iltw_Y)/det;
                                ilt_Y = ((1+uw_X)*iltw_Y - uw_Y*iltw_X)/det;
                            
                                // derivatives of warped image gradients with respect to non-warped coordinates
                                // 1-derivatives of the x coordinate of the gradient of the warped image
                                iltw_X_X = ((1+vw_Y)*iltw_XX - vw_X*iltw_XY)/det;
                                iltw_X_Y = ((1+uw_X)*iltw_XY - uw_Y*iltw_XX)/det;
                                // 2-derivatives of the y coordinate of the gradient of the warped image
                                iltw_Y_X = ((1+vw_Y)*iltw_XY - vw_X*iltw_YY)/det;
                                iltw_Y_Y = ((1+uw_X)*iltw_YY - uw_Y*iltw_XY)/det;

                                double du = cvmGet(dU, yl0, xl0);
                                double dv = cvmGet(dV, yl0, xl0);

                                // IN THE FOLLOWING:
                                // Il0 should match Ilt
                                // The first order Taylor expansion of ilt is:
                                // ilt(x+u+du,y+v+dv) = ilt + ilt_X*du + ilt_Y*dv
                                // iltw_X and iltw_Y should be interpolated similarly to ilt, since
                                // (Il0_X,Il0_Y) should match (Iltw_X,Iltw_Y), NOT (Ilt_X,Ilt_Y) !!!!
                                // see the expression of b below for an example

                                //psi' coefficients computation
                                double a = square(ilt + ilt_X*du + ilt_Y*dv - il0);
                                double b = gamma*(square(iltw_X + iltw_X_X*du + iltw_X_Y*dv - il0_X) +
                                                  square(iltw_Y + iltw_Y_X*du + iltw_Y_Y*dv - il0_Y));
                                //divergence coefficients
                                psi_flowL = DPsiData(a+b);
			       
                                //u equation
                                a_u_1 += psi_flowL*(ilt_X*ilt_X + gamma*(iltw_X_X*iltw_X_X + iltw_Y_X*iltw_Y_X));
                                a_v_1 += psi_flowL*(ilt_Y*ilt_X + gamma*(iltw_X_Y*iltw_X_X + iltw_Y_Y*iltw_Y_X));
                            
                                //v equation
                                a_u_2 += psi_flowL*(ilt_X*ilt_Y + gamma*(iltw_X_X*iltw_X_Y + iltw_Y_X*iltw_Y_Y));
                                a_v_2 += psi_flowL*(ilt_Y*ilt_Y + gamma*(iltw_X_Y*iltw_X_Y + iltw_Y_Y*iltw_Y_Y));
                            }
			}
			 
			//M estimation
			cvmSet(M,i,0,a_u_1);
			cvmSet(M,i+1,0,a_u_2);
			   
			cvmSet(M,i,1,a_v_1);
			cvmSet(M,i+1,1,a_v_2);
		
			 
			//If we use the 8 pts divergence scheme 
#if DIVERGENCE_8
                        cvmSet(M,i,2,(-alpha*alpha_p_u/(4*h*h))*b_f_4);
                        cvmSet(M,i+1,2,(-alpha*alpha_p_v/(4*h*h))*b_f_4);
			   
			   
                        cvmSet(M,i,3,(-alpha*alpha_p_u/(4*h*h))*b_f_3);
                        cvmSet(M,i+1,3,(-alpha*alpha_p_v/(4*h*h))*b_f_3);
				 

                        cvmSet(M,i,4,(-alpha*alpha_p_u/(4*h*h))*b_f_2);
                        cvmSet(M,i+1,4,(-alpha*alpha_p_v/(4*h*h))*b_f_2);
				

                        cvmSet(M,i,5,(-alpha*alpha_p_u/(4*h*h))*b_f_1);
                        cvmSet(M,i+1,5,(-alpha*alpha_p_v/(4*h*h))*b_f_1);
			
                        cvmSet(M,i,6,(-alpha*alpha_d_u/(4*h*h))*b_f_5);
                        cvmSet(M,i+1,6,(-alpha*alpha_d_v/(4*h*h))*b_f_5);
			   
			   
                        cvmSet(M,i,7,(-alpha*alpha_d_u/(4*h*h))*b_f_6);
                        cvmSet(M,i+1,7,(-alpha*alpha_d_v/(4*h*h))*b_f_6);
				 

                        cvmSet(M,i,8,(-alpha*alpha_d_u/(4*h*h))*b_f_7);
                        cvmSet(M,i+1,8,(-alpha*alpha_d_v/(4*h*h))*b_f_7);
				

                        cvmSet(M,i,9,(-alpha*alpha_d_u/(4*h*h))*b_f_8);
                        cvmSet(M,i+1,9,(-alpha*alpha_d_v/(4*h*h))*b_f_8);
#else  // !DIVERGENCE_8
                        cvmSet(M,i,2,(-alpha/(4*h*h))*b_f_4);
                        cvmSet(M,i+1,2,(-alpha/(4*h*h))*b_f_4);
			   
			   
                        cvmSet(M,i,3,(-alpha/(4*h*h))*b_f_3);
                        cvmSet(M,i+1,3,(-alpha/(4*h*h))*b_f_3);
				 

                        cvmSet(M,i,4,(-alpha/(4*h*h))*b_f_2);
                        cvmSet(M,i+1,4,(-alpha/(4*h*h))*b_f_2);
				

                        cvmSet(M,i,5,(-alpha/(4*h*h))*b_f_1);
                        cvmSet(M,i+1,5,(-alpha/(4*h*h))*b_f_1);   
#endif  // !DIVERGENCE_8
			   
                        //second member for the first equation
                        double bu0 = cvmGet(U, yl0, xl0);
                        double bu1 = cvmGet(U, yl0, xl0+1);
                        double bu2 = cvmGet(U, yl0, xl0-1);
                        double bu3 = cvmGet(U,yl0+1,xl0);
                        double bu4 = cvmGet(U,yl0-1,xl0);
                        //Crop(bu0, N_x);
#if DIVERGENCE_8
                        double bu5 = cvmGet(U, yl0+1, xl0+1);
                        double bu6 = cvmGet(U, yl0+1, xl0-1);
                        double bu7 = cvmGet(U,yl0-1,xl0+1);
                        double bu8 = cvmGet(U,yl0-1,xl0-1); 
                        double bu = (alpha/(4*h*h)) *
                            (alpha_p_u *
                             (b_f_1 * bu1 +
                              b_f_2 * bu2 +
                              b_f_3 * bu3 +
                              b_f_4 * bu4) +
                             alpha_d_u *
                             (b_f_5 * bu5 +
                              b_f_6 * bu6 +
                              b_f_7 * bu7 + 
                              b_f_8 * bu8)
                             - (alpha_p_u*(b_f_1 + b_f_2 + b_f_3 + b_f_4) + alpha_d_u*(b_f_5 + b_f_6 + b_f_7 + b_f_8)) * bu0);
#else  // !DIVERGENCE_8
                        double bu = (alpha/(4*h*h)) *
                            (b_f_1 * bu1 +
                             b_f_2 * bu2 +
                             b_f_3 * bu3 +
                             b_f_4 * bu4
                             - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bu0);
#endif  // !DIVERGENCE_8
			//second member for the second equation
                        double bv0 = cvmGet(V, yl0, xl0);
                        double bv1 = cvmGet(V, yl0, xl0+1);
                        double bv2 = cvmGet(V, yl0, xl0-1);
                        double bv3 = cvmGet(V,yl0+1,xl0);
                        double bv4 = cvmGet(V,yl0-1,xl0);
                        //Crop(bv0, N_y);
#if DIVERGENCE_8
                        double bv5 = cvmGet(V, yl0+1, xl0+1);
                        double bv6 = cvmGet(V, yl0+1, xl0-1);
                        double bv7 = cvmGet(V,yl0-1,xl0+1);
                        double bv8 = cvmGet(V,yl0-1,xl0-1); 
                        double bv = (alpha/(4*h*h)) *
                            (alpha_p_v * 
                             (b_f_1 * bv1 +
                              b_f_2 * bv2 +
                              b_f_3 * bv3 +
                              b_f_4 * bv4)+
                             alpha_d_v *
                             (b_f_5 * bv5 +
                              b_f_6 * bv6 +
                              b_f_7 * bv7 + 
                              b_f_8 * bv8)
                             - (alpha_p_v*(b_f_1 + b_f_2 + b_f_3 + b_f_4) + alpha_d_v*(b_f_5 + b_f_6 + b_f_7 + b_f_8)) * bv0);
#else  // !DIVERGENCE_8
                        double bv = (alpha/(4*h*h)) *
                            (b_f_1 * bv1 +
                             b_f_2 * bv2 +
                             b_f_3 * bv3 +
                             b_f_4 * bv4
                             - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bv0);
#endif  // !DIVERGENCE_8
                        if (!pixel_is_disactivated) {
                            //second member for the first equation
                            bu +=
                                - psi_flowL*(ilt_X*(ilt - il0) + gamma*(iltw_X_X*(iltw_X - il0_X) + iltw_Y_X*(iltw_Y - il0_Y)));
                            //second member for the second equation
                            bv +=
                                - psi_flowL*(ilt_Y*(ilt - il0) + gamma*(iltw_X_Y*(iltw_X - il0_X) + iltw_Y_Y*(iltw_Y - il0_Y)));
                        }
                        cvmSet(B,i,0,bu);
                        cvmSet(B,i+1,0,bv);
			   
                    }
		       
		}

		//Initial energy computation
		if(logfile && iter_inner == 0
#if VERBOSE_ENERGY
#else
                    && iter_outer == 0
#endif
                    )
                {
                    double Efl, Edata, Esmooth, Eintensity, Egradient;
                    EnergyComputation(Il0_pyramid[level], 0, Iltw, 0,
                                      U, V, 0, 0,
                                      disactivated_pixels, 0, 0, 0,
                                      N_x, N_y, &Edata, &Esmooth, gamma, 1.0, 1.0, 1, 0, 0, 0, &Efl, 0, 0, 0, 0, 0, 0, 0, &Eintensity, &Egradient);
		    fprintf(logfile, "** level %d resolution %d x %d\n", level+1, N_x, N_y);
		    fprintf(logfile, "* energy before outer iteration %d:\n", iter_outer+1);
		    fprintf(logfile, "Edata = %g Esmooth = %g Ed/Es = %g (alpha=%g)\n", Edata, Esmooth, Edata/Esmooth, alpha);
		    fprintf(logfile, "Edata breakdown (squared): Eintensity = %g, Egradient=%g Ei/Eg=%g (gamma=%g)\n", Eintensity, Egradient, Egradient != 0 ? Eintensity/Egradient : 0., gamma);
		    fflush(logfile);
		}


		//SOR iterations
		cvSetZero(SOR);
		cvSetZero(SOR_prec);
		first_iteration = 1;
		iter_SOR = 0;
		norm_SOR = 0.;
		while ((norm_SOR >= epsilon_SOR && iter_SOR < max_iter_SOR) || first_iteration == 1){
		    first_iteration = 0;
			 
			 
		    //Loop on the SOR vector
                    for (int sub_iter_SOR = 0; sub_iter_SOR <4; sub_iter_SOR++) {
                    int line, line_begin, line_end, line_incr;
                    int col, col_begin, col_end, col_incr;
                    switch (sub_iter_SOR) {
                        case 0:         // NE-SW
                            line_begin = 0;
                            line_end = N_y * dimension * N_x;
                            line_incr = dimension * N_x;
                            col_begin = 0;
                            col_end = dimension * N_x;
                            col_incr = 1;
                            break;
                        case 1:         // SW-NE
                            line_begin = (N_y-1) * dimension * N_x;
                            line_end = - dimension * N_x;
                            line_incr = - dimension * N_x;
                            col_begin = dimension * N_x - 1;
                            col_end = -1;
                            col_incr = -1;
                            break;
                        case 2:         // NW-SE
                            line_begin = 0;
                            line_end = N_y * dimension * N_x;
                            line_incr = dimension * N_x;
                            col_begin = dimension * N_x - 1;
                            col_end = -1;
                            col_incr = -1;
                            break;
                        default:         // SE-NW
                            line_begin = (N_y-1) * dimension * N_x;
                            line_end = - dimension * N_x;
                            line_incr = - dimension * N_x;
                            col_begin = 0;
                            col_end = dimension * N_x;
                            col_incr = 1;
                    }
                    for (line = line_begin; line != line_end; line+= line_incr) {
                    for (col = col_begin; col != col_end; col+= col_incr) {
                        i = line + col;
			double sum = 0;
                        double diag_term;
                        
			int rest = (i%(dimension*N_x))%dimension;
			// tali: 'i < 2*N_x' referes to the pixels in y=0 in the original image 
			if(i < dimension*N_x){
			    diag_term = cvmGet(M,i,0);
				// tali: the sum in that case the sum is only one term which is relevant to the pixel under the current pixel
			    sum = cvmGet(M,i,1)*cvmGet(SOR, i+dimension*N_x,0);
			}
				// tali: refers to pixel in the ;last row in the original image
			else if(i >= dimension*N_x*(N_y-1)){
			    diag_term = cvmGet(M, i,0);
			    sum = cvmGet(M, i,1)*cvmGet(SOR, i-dimension*N_x,0);
			}
			else{
				// tali: a new line in the original image -- u equation 
			    if(i%(dimension*N_x) == 0){
				diag_term = cvmGet(M, i,0);
				sum = cvmGet(M, i,1)*cvmGet(SOR, i+dimension,0);
			    }
			    else if((i+dimension)%(dimension*N_x)==0){
				diag_term = cvmGet(M, i,0);
				sum = cvmGet(M, i,1)*cvmGet(SOR, i-dimension,0);
			    }
			    else{
				int rest1 = i%(dimension*N_x);
				if(rest1 == 1){
				    diag_term = cvmGet(M, i,0);
				    sum = cvmGet(M, i,1)*cvmGet(SOR, i+dimension,0);
				}
				else if(rest1 == dimension*N_x - 1){
				    diag_term = cvmGet(M, i,0);
				    sum = cvmGet(M, i,1)*cvmGet(SOR, i-dimension,0);
				}
				else{
				    diag_term = cvmGet(M, i,rest);
				    int rest2 = rest1 % 2;
				    if(rest2 == 0){//u equation
#if DIVERGENCE_8
                                        sum =
                                            cvmGet(M, i,2)*cvmGet(SOR, i-dimension*N_x,0) +
                                            cvmGet(M, i,4)*cvmGet(SOR, i-dimension,0) +
                                            cvmGet(M, i,8)*cvmGet(SOR, i-dimension*(N_x-1),0) +
                                            cvmGet(M, i,9)*cvmGet(SOR, i-dimension*(N_x+1),0) +
                                            cvmGet(M, i,5)*cvmGet(SOR, i+dimension,0) +
                                            cvmGet(M, i,3)*cvmGet(SOR, i+dimension*N_x,0) +
                                            cvmGet(M, i,1)*cvmGet(SOR, i+1,0) +
                                            cvmGet(M, i,7)*cvmGet(SOR, i+dimension*(N_x-1),0) +
                                            cvmGet(M, i,6)*cvmGet(SOR, i+dimension*(N_x+1),0) ;
#else  // !DIVERGENCE_8
                                        sum =
											// tali: the neighbour above
                                            cvmGet(M, i,2)*cvmGet(SOR, i-dimension*N_x,0) +
											// tali: the neighbour to the left
                                            cvmGet(M, i,4)*cvmGet(SOR, i-dimension,0) +
											// tali: the neighbour to the right
                                            cvmGet(M, i,5)*cvmGet(SOR, i+dimension,0) +
											// tali: the neighbour below
                                            cvmGet(M, i,3)*cvmGet(SOR, i+dimension*N_x,0) +
											// tali: the relveant v
                                            cvmGet(M, i,1)*cvmGet(SOR, i+1,0);
#endif  // !DIVERGENCE_8
	                                diag_term = cvmGet(M, i,0);
				    }
				    else{//v equation
#if DIVERGENCE_8
                                        sum =
                                            cvmGet(M, i,2)*cvmGet(SOR, i-dimension*N_x,0) +
                                            cvmGet(M, i,4)*cvmGet(SOR, i-dimension,0) +
                                            cvmGet(M, i,0)*cvmGet(SOR, i-1,0) +
                                            cvmGet(M, i,8)*cvmGet(SOR, i-dimension*(N_x-1),0) +
                                            cvmGet(M, i,9)*cvmGet(SOR, i-dimension*(N_x+1),0) +
                                            cvmGet(M, i,5)*cvmGet(SOR, i+dimension,0) +
                                            cvmGet(M, i,3)*cvmGet(SOR, i+dimension*N_x,0) +
                                            cvmGet(M, i,7)*cvmGet(SOR, i+dimension*(N_x-1),0) +
                                            cvmGet(M, i,6)*cvmGet(SOR, i+dimension*(N_x+1),0);
#else  // !DIVERGENCE_8
                                        sum =
											// tali: the neighbour above
                                            cvmGet(M, i,2)*cvmGet(SOR, i-dimension*N_x,0) +
											// tali: the neighbour to the left
                                            cvmGet(M, i,4)*cvmGet(SOR, i-dimension,0) +
											// tali: the relevant u
                                            cvmGet(M, i,0)*cvmGet(SOR, i-1,0) +
											// tali: the neighbour to the right
                                            cvmGet(M, i,5)*cvmGet(SOR, i+dimension,0) +
											// tali: the neighbour below
											cvmGet(M, i,3)*cvmGet(SOR, i+dimension*N_x,0);
#endif  // !DIVERGENCE_8
				        diag_term = cvmGet(M, i,1);
				    }
				}
			    }
			}
			// tali: in sor: the solution in the current iteration is: (1-w)prev_sol + (w/a_ii)(b_i-sum)
                        double rold = cvmGet(SOR, i,0);
                        double b = cvmGet(B, i,0);
                        double rnew = (1.-omega_SOR)*rold + (omega_SOR/diag_term) * (b - sum);
                        //assert(fabs(rnew)<(N_x+N_y));
			cvmSet(SOR, i,0,Crop(rnew,MAX_INCR));
		    } // for (col
		    } // for (line
		    } // for (sub_iter_SOR
			 
		    norm_SOR = cvNorm(SOR_prec, SOR, CV_RELATIVE_L2);
		    cvCopy(SOR, SOR_prec);
		    iter_SOR ++;
		} // while ((norm_SOR ...
		cvCopy(dU, dU_prec);
		cvCopy(dV, dV_prec);
		for(i = 0; i < dimension*N_x*N_y; i+=dimension){
                    xl0 = (i%(dimension*N_x))/dimension;
                    yl0 = i/(dimension*N_x);
                    double du = cvmGet(SOR, i,0);
                    double dv = cvmGet(SOR, i + 1,0);
		    cvmSet(dU, yl0, xl0, Crop(du,MAX_INCR));
		    cvmSet(dV, yl0, xl0, Crop(dv,MAX_INCR));
		}
                if (OF_ZeroTopBorder) {
                    for(xl0 = 0; xl0 < N_x; xl0++){
                        cvmSet(dU, 0, xl0, 0);
                        cvmSet(dV, 0, xl0, 0);
                        cvmSet(dU, 1, xl0, 0);
                        cvmSet(dV, 1, xl0, 0);
                    }
                    for(yl0 = 0; yl0 < N_y*7/8; yl0++){
                        cvmSet(dU, yl0, 0, 0);
                        cvmSet(dV, yl0, 0, 0);
                        cvmSet(dU, yl0, N_x-1, 0);
                        cvmSet(dV, yl0, N_x-1, 0);
                        cvmSet(dU, yl0, 1, 0);
                        cvmSet(dV, yl0, 1, 0);
                        cvmSet(dU, yl0, N_x-2, 0);
                        cvmSet(dV, yl0, N_x-2, 0);
                        cvmSet(dU, yl0, 2, 0);
                        cvmSet(dV, yl0, 2, 0);
                        cvmSet(dU, yl0, N_x-3, 0);
                        cvmSet(dV, yl0, N_x-3, 0);
                    }
                }
		norm_inner = fmax(cvNorm(dU_prec, dU, CV_RELATIVE_L2), cvNorm(dV_prec, dV, CV_RELATIVE_L2));
		
		iter_inner++;
	    }
	      
	   
            if (OF_U_increment_FileName) {
                SaveResults(OF_U_increment_FileName, dU);
            }
	     
            if (OF_V_increment_FileName) {
                SaveResults(OF_V_increment_FileName, dV);
            }

	    cvCopy(U, U_prec);
	    cvCopy(V, V_prec);
            cvAdd(U, dU, U);
	    cvAdd(V, dV, V);
	     
	    norm = fmax(cvNorm(U_prec, U, CV_RELATIVE_L2), cvNorm(V_prec, V, CV_RELATIVE_L2));
	     
	    if (OF_U_FileName) {
                SaveResults(OF_U_FileName, U);
            }	     
	    if (OF_V_FileName) {
                SaveResults(OF_V_FileName, V);
            }
	    if (OF_disactivated_FileName) {
                SaveResults(OF_disactivated_FileName, disactivated_pixels, 0, 0., 255.);
            }	     
	 
	    iter_outer++;
	}

	if (logfile) {
	    //Coordinate maps update
	    cvAdd(MapIdentityX, U, MapX_Ilt);
	    cvAdd(MapIdentityY, V, MapY_Ilt);
	
	    //Images warping
	    // Images warping + Warped images gradient computation
	    // (Warped images are smoothed)
	    cvRemap(Ilt_pyramid[level], Iltw, MapX_Ilt, MapY_Ilt);
	    SmoothImage(Iltw);
            fprintf(logfile, "* iterations: %d\n", iter_outer);
            fprintf(logfile, "iter_SOR=%d<%d norm_SOR=%g<%g\n", iter_SOR, max_iter_SOR, norm_SOR, epsilon_SOR);
            fprintf(logfile, "iter_inner=%d<=%d norm_inner=%g<%g\n",iter_inner,max_iter_inner,norm_inner,epsilon_inner);
            fprintf(logfile, "iter_outer=%d<=%d norm=%g<%g\n", iter_outer, max_iter_outer, norm, epsilon_outer);
            fflush(logfile);
            for(yl0 = 0; yl0<N_y; yl0++) {
                for(xl0 = 0; xl0<N_x; xl0++) {
                    //Current coordinates of the followed point 
                    double xlt = cvmGet(MapX_Ilt,yl0,xl0);
                    double ylt = cvmGet(MapY_Ilt,yl0,xl0);
                    //Pixels out must not be taken into account in the energy data term
                    int pixel_is_disactivated = cvmGet8u(disactivated_pixels,yl0,xl0);

                    //is the point visible at time t+1 ?
                    if(!pixel_is_disactivated &&                          
                       (xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h  ||
                        ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h)){
                        //printf("disactivate %d,%d\n", x, yl0);
                        cvmSet8u(disactivated_pixels,yl0,xl0,127);
                        pixel_is_disactivated = 1;
                    }
                }
            }
            double Efl, Edata, Esmooth, Eintensity, Egradient;
            EnergyComputation(Il0_pyramid[level], 0, Iltw, 0,
                              U, V, 0, 0,
                              disactivated_pixels, 0, 0, 0,
                              N_x, N_y, &Edata, &Esmooth, gamma, 1.0, 1.0, 1, 0, 0, 0, &Efl, 0, 0, 0, 0, 0, 0, 0, &Eintensity, &Egradient);
            fprintf(logfile, "* final energy:\n");
            fprintf(logfile, "Edata = %g Esmooth = %g Ed/Es = %g (alpha=%g)\n", Edata, Esmooth, Edata/Esmooth, alpha);
            fprintf(logfile, "Edata breakdown (squared): Eintensity = %g, Egradient=%g Ei/Eg=%g (gamma=%g)\n", Eintensity, Egradient,  Egradient != 0 ? Eintensity/Egradient : 0., gamma);
            fflush(logfile);
        }

	//Let's Clean
	cvReleaseMat(&Iltw_X);
	cvReleaseMat(&Iltw_Y);
	cvReleaseMat(&Il0_X);
	cvReleaseMat(&Il0_Y);
	cvReleaseMat(&Iltw_XX);
	cvReleaseMat(&Iltw_XY);
	cvReleaseMat(&Iltw_YY);
	cvReleaseMat(&derivative_tmp);

        cvReleaseMat(&divergence_coefficients);

	cvReleaseMat(&Iltw);
	cvReleaseMat(&MapX_Ilt);
	cvReleaseMat(&MapY_Ilt);
	cvReleaseMat(&MapX_Ilt_inner);
	cvReleaseMat(&MapY_Ilt_inner);
	cvReleaseMat(&MapIdentityX);
	cvReleaseMat(&MapIdentityY);
	cvReleaseMat(&M);
	cvReleaseMat(&B);
        cvReleaseMat(&disactivated_pixels);
	cvReleaseMat(&U_prec); 
	cvReleaseMat(&V_prec);
	     

	cvReleaseMat(&dU);
	cvReleaseMat(&dV);
	cvReleaseMat(&dU_X);
	cvReleaseMat(&dU_Y);
	cvReleaseMat(&dV_X);
	cvReleaseMat(&dV_Y);

	 

	cvReleaseMat(&U_X);
	cvReleaseMat(&U_Y);
	cvReleaseMat(&V_X);
	cvReleaseMat(&V_Y);
	 
	cvReleaseMat(&dU_prec);
	cvReleaseMat(&dV_prec);
	 
	cvReleaseMat(&SOR);
	cvReleaseMat(&SOR_prec);

	//if this is not the last level, images must be upscaled before iterations
	if (level != (pyramid_level_end-1)) {
            const CvSize next_level_size = cvGetSize(Il0_pyramid[level - 1]);
	    CvMat *Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
	    cvResize(U, Interpolated_Data,CV_INTER_LINEAR);
            double uscale = (double)Interpolated_Data->width / U->width;
	    cvReleaseMat(&U);
	    U = Interpolated_Data;
            SmoothData(U);
	    cvConvertScale(U, U, uscale);
	   
	    Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
	    cvResize(V, Interpolated_Data,CV_INTER_LINEAR);
            double vscale = (double)Interpolated_Data->height / V->height;
	    cvReleaseMat(&V);
	    V = Interpolated_Data;
            SmoothData(V);
	    cvConvertScale(V, V, vscale);
	}
    }//  for(int level = pyramid_level_start - 1; level >= pyramid_level_end - 1; level--){
     
 
    //cvResize(U, U_end, CV_INTER_AREA);
    //cvResize(V, V_end, CV_INTER_AREA);
    cvCopy(U, U_end);
    cvCopy(V, V_end);
    cvReleaseMat(&U);
    cvReleaseMat(&V);
}

/**************************************************Scene Flow Estimation*****************/


/***************Computes 3D points with respect to the reference image******************/
/* tali: this function computes the 3D points given the reference image coordinates and 
   the depth. it scale the projection matrix to fit the current pyramid level*/

void ComputeXY(CvMat* X,CvMat* Y,CvMat* Z0,
			   CvMat* MapIdentityX,CvMat* MapIdentityY,
			   int N_x,int N_y,
			   const double* params,
			   double scale)
{
	CvMat* temp = cvCreateMat(X->rows, X->cols, CV_32FC1);

	// x-scale*u0
    cvSubS(MapIdentityX,cvScalar(params[2]*scale),temp);
	// Z*(x-u0)/(scale*fku)
	cvMul(temp,Z0,X,double(1/(params[0]*scale)));
	// y-scale*v0
	cvSubS(MapIdentityY,cvScalar(params[3]*scale),temp);
	// Z*(y-v0)/(scale*fkv)
	cvMul(temp,Z0,Y,double(1/(params[1]*scale)));

	cvReleaseMat(&temp);

}
/***************Converts between disparity and depth (rectified images)*****************/
void ConvertDZ(CvMat* D,CvMat* Z,double scale, double maxZ){
	int i=0;
	int j=0;
	cvZero(Z);
	double Val = 0;
	for (int i = 0; i < D->rows; ++i){
	  for (int j = 0; j < D->cols; ++j){
		  Val  = cvmGet(D,i,j);
		  if(abs(Val) > abs(scale/maxZ))
			  cvmSet(Z,i,j,scale/Val);
		  else
			cvmSet(Z,i,j,maxZ);
	  }
	}
}
/********Copy a specific row from the src mat to a specific row in the dst mat ***********/
void SetRow(CvMat* src, CvMat* dst, int src_row, int dst_row){
	int i=0;
	switch(CV_MAT_TYPE(src->type)){
		case CV_8UC1:  
			for(i=0; i<src->cols; i++){
				((unsigned char *)(dst->data.ptr + SIZE_T(dst_row*dst->step)))[i] = ((unsigned char *)(src->data.ptr + SIZE_T(src_row*src->step)))[i];				
			}
			break;
		case CV_32FC1:
			for(i=0; i<src->cols; i++){
				((float *)(dst->data.ptr + SIZE_T(dst_row*dst->step)))[i] = ((float *)(src->data.ptr + SIZE_T(src_row*src->step)))[i];				
			}
			
			break;
		case CV_64FC1:
			for(i=0; i<src->cols; i++){
				((double *)(dst->data.ptr + SIZE_T(dst_row*dst->step)))[i] = ((double *)(src->data.ptr + SIZE_T(src_row*src->step)))[i];				
			}
			
			break;
		default:
			break;
	}
}
/*****Computes projection matrix from the intrinsic and extrinsic parameters***********/
void ComputeProjMat(CvMat *P, CvMat *R, CvMat *T, CvMat *C,double eta_curr) {
	CvMat *RT = cvCreateMat(3, 4, CV_64FC1);
	cvSetZero(RT);
	double R11 = cvmGet(R,0,0);
	double R12 = cvmGet(R,0,1);
	double R13 = cvmGet(R,0,2);
	double R21 = cvmGet(R,1,0);
	double R22 = cvmGet(R,1,1);
	double R23 = cvmGet(R,1,2);
	double R31 = cvmGet(R,2,0);
	double R32 = cvmGet(R,2,1);
	double R33 = cvmGet(R,2,2);
	double T1 = cvmGet(T,0,0);
	double T2 = cvmGet(T,1,0);
	double T3 = cvmGet(T,2,0);
    // setting the elements in the matrix [R | T]
	cvmSet(RT,0,0,R11);
	cvmSet(RT,0,1,R12);
	cvmSet(RT,0,2,R13);
	cvmSet(RT,1,0,R21);
	cvmSet(RT,1,1,R22);
	cvmSet(RT,1,2,R23);
	cvmSet(RT,2,0,R31);
	cvmSet(RT,2,1,R32);
	cvmSet(RT,2,2,R33);
	cvmSet(RT,0,3,T1);
	cvmSet(RT,1,3,T2);
	cvmSet(RT,2,3,T3);
	cvmSet(C,2,2,1/eta_curr);
	cvMatMul(C,RT,P);
	cvReleaseMat(&RT);
}

void ComputeRefProjMat(CvMat *P, CvMat *R, CvMat *T, CvMat *C,double eta_curr) {
	CvMat *RT = cvCreateMat(3, 4, CV_64FC1);
	cvSetZero(RT);
	cvSetZero(C);

	 // setting the elements in the matrix [R | T]
	cvmSet(RT,0,0,1);
	cvmSet(RT,0,1,0);
	cvmSet(RT,0,2,0);
	cvmSet(RT,1,0,0);
	cvmSet(RT,1,1,1);
	cvmSet(RT,1,2,0);
	cvmSet(RT,2,0,0);
	cvmSet(RT,2,1,0);
	cvmSet(RT,2,2,1);
	cvmSet(RT,0,3,0);
	cvmSet(RT,1,3,0);
	cvmSet(RT,2,3,0);

	cvmSet(T,0,0,0);
	cvmSet(T,1,0,0);
	cvmSet(T,2,0,0);
	cvmSet(R,0,0,1);
	cvmSet(R,0,1,0);
	cvmSet(R,0,2,0);
	cvmSet(R,1,0,0);
	cvmSet(R,1,1,1);
	cvmSet(R,1,2,0);
	cvmSet(R,2,0,0);
	cvmSet(R,2,1,0);
	cvmSet(R,2,2,1);
    // setting the elements in the calibriation matrix
	cvmSet(C,0,0,f0ku);
	cvmSet(C,1,1,f0kv);
	//scaling the projection matrix to fit to the current pyramid size
	cvmSet(C,2,2,1/eta_curr);
	cvmSet(C,0,2,u0);
	cvmSet(C,1,2,v0);

	cvMatMul(C,RT,P);
    cvReleaseMat(&RT);

}


void ProjectPoints2( const CvMat* objectPoints,
                  const CvMat* r_vec,
                  const CvMat* t_vec,
                  const CvMat* A,
                  const CvMat* distCoeffs,
                  CvMat* imagePoints,
				  double scale,
				  CvMat* dpdr,
                  CvMat* dpdt, CvMat* dpdf,
                  CvMat* dpdc, CvMat* dpdk,
                  double aspectRatio )
{
    CvMat *_M = 0, *_m = 0;
    CvMat *_dpdr = 0, *_dpdt = 0, *_dpdc = 0, *_dpdf = 0, *_dpdk = 0;

    CV_FUNCNAME( "cvProjectPoints2" );

    __BEGIN__;

    int i, j, count;
    int calc_derivatives;
    const CvPoint3D64f* M;
    CvPoint2D64f* m;
    double r[3], R[9], dRdr[27], t[3], a[9], k[5] = {0,0,0,0,0}, fx, fy, cx, cy;
    CvMat _r, _t, _a = cvMat( 3, 3, CV_64F, a ), _k;
    CvMat _R = cvMat( 3, 3, CV_64F, R ), _dRdr = cvMat( 3, 9, CV_64F, dRdr );
    double *dpdr_p = 0, *dpdt_p = 0, *dpdk_p = 0, *dpdf_p = 0, *dpdc_p = 0;
    int dpdr_step = 0, dpdt_step = 0, dpdk_step = 0, dpdf_step = 0, dpdc_step = 0;
    bool fixedAspectRatio = aspectRatio > FLT_EPSILON;

    if( !CV_IS_MAT(objectPoints) || !CV_IS_MAT(r_vec) ||
        !CV_IS_MAT(t_vec) || !CV_IS_MAT(A) ||
        /*!CV_IS_MAT(distCoeffs) ||*/ !CV_IS_MAT(imagePoints) )
        CV_ERROR( CV_StsBadArg, "One of required arguments is not a valid matrix" );

    count = MAX(objectPoints->rows, objectPoints->cols);

    if( CV_IS_CONT_MAT(objectPoints->type) && CV_MAT_DEPTH(objectPoints->type) == CV_64F &&
        ((objectPoints->rows == 1 && CV_MAT_CN(objectPoints->type) == 3) ||
        (objectPoints->rows == count && CV_MAT_CN(objectPoints->type)*objectPoints->cols == 3)))
        _M = (CvMat*)objectPoints;
    else
    {
        CV_CALL( _M = cvCreateMat( 1, count, CV_64FC3 ));
        CV_CALL( cvConvertPointsHomogeneous( objectPoints, _M ));
    }

    if( CV_IS_CONT_MAT(imagePoints->type) && CV_MAT_DEPTH(imagePoints->type) == CV_64F &&
        ((imagePoints->rows == 1 && CV_MAT_CN(imagePoints->type) == 2) ||
        (imagePoints->rows == count && CV_MAT_CN(imagePoints->type)*imagePoints->cols == 2)))
        _m = imagePoints;
    else
        CV_CALL( _m = cvCreateMat( 1, count, CV_64FC2 ));

    M = (CvPoint3D64f*)_M->data.db;
    m = (CvPoint2D64f*)_m->data.db;

    if( (CV_MAT_DEPTH(r_vec->type) != CV_64F && CV_MAT_DEPTH(r_vec->type) != CV_32F) ||
        (((r_vec->rows != 1 && r_vec->cols != 1) ||
        r_vec->rows*r_vec->cols*CV_MAT_CN(r_vec->type) != 3) &&
        ((r_vec->rows != 3 && r_vec->cols != 3) || CV_MAT_CN(r_vec->type) != 1)))
        CV_ERROR( CV_StsBadArg, "Rotation must be represented by 1x3 or 3x1 "
                  "floating-point rotation vector, or 3x3 rotation matrix" );

    if( r_vec->rows == 3 && r_vec->cols == 3 )
    {
        _r = cvMat( 3, 1, CV_64FC1, r );
        CV_CALL( cvRodrigues2( r_vec, &_r ));
        CV_CALL( cvRodrigues2( &_r, &_R, &_dRdr ));
        cvCopy( r_vec, &_R );
    }
    else
    {
        _r = cvMat( r_vec->rows, r_vec->cols, CV_MAKETYPE(CV_64F,CV_MAT_CN(r_vec->type)), r );
        CV_CALL( cvConvert( r_vec, &_r ));
        CV_CALL( cvRodrigues2( &_r, &_R, &_dRdr ) );
    }

    if( (CV_MAT_DEPTH(t_vec->type) != CV_64F && CV_MAT_DEPTH(t_vec->type) != CV_32F) ||
        (t_vec->rows != 1 && t_vec->cols != 1) ||
        t_vec->rows*t_vec->cols*CV_MAT_CN(t_vec->type) != 3 )
        CV_ERROR( CV_StsBadArg,
            "Translation vector must be 1x3 or 3x1 floating-point vector" );

    _t = cvMat( t_vec->rows, t_vec->cols, CV_MAKETYPE(CV_64F,CV_MAT_CN(t_vec->type)), t );
    CV_CALL( cvConvert( t_vec, &_t ));

    if( (CV_MAT_TYPE(A->type) != CV_64FC1 && CV_MAT_TYPE(A->type) != CV_32FC1) ||
        A->rows != 3 || A->cols != 3 )
        CV_ERROR( CV_StsBadArg, "Instrinsic parameters must be 3x3 floating-point matrix" );

    CV_CALL( cvConvert( A, &_a ));
    fx = a[0]; fy = a[4];
    cx = a[2]; cy = a[5];

    if( fixedAspectRatio )
        fx = fy*aspectRatio;

    if( distCoeffs )
    {
        if( !CV_IS_MAT(distCoeffs) ||
            (CV_MAT_DEPTH(distCoeffs->type) != CV_64F &&
            CV_MAT_DEPTH(distCoeffs->type) != CV_32F) ||
            (distCoeffs->rows != 1 && distCoeffs->cols != 1) ||
            (distCoeffs->rows*distCoeffs->cols*CV_MAT_CN(distCoeffs->type) != 4 &&
            distCoeffs->rows*distCoeffs->cols*CV_MAT_CN(distCoeffs->type) != 5) )
            CV_ERROR( CV_StsBadArg,
                "Distortion coefficients must be 1x4, 4x1, 1x5 or 5x1 floating-point vector" );

        _k = cvMat( distCoeffs->rows, distCoeffs->cols,
                    CV_MAKETYPE(CV_64F,CV_MAT_CN(distCoeffs->type)), k );
        CV_CALL( cvConvert( distCoeffs, &_k ));
    }

    if( dpdr )
    {
        if( !CV_IS_MAT(dpdr) ||
            (CV_MAT_TYPE(dpdr->type) != CV_32FC1 &&
            CV_MAT_TYPE(dpdr->type) != CV_64FC1) ||
            dpdr->rows != count*2 || dpdr->cols != 3 )
            CV_ERROR( CV_StsBadArg, "dp/drot must be 2Nx3 floating-point matrix" );

        if( CV_MAT_TYPE(dpdr->type) == CV_64FC1 )
            _dpdr = dpdr;
        else
            CV_CALL( _dpdr = cvCreateMat( 2*count, 3, CV_64FC1 ));
        dpdr_p = _dpdr->data.db;
        dpdr_step = _dpdr->step/sizeof(dpdr_p[0]);
    }

    if( dpdt )
    {
        if( !CV_IS_MAT(dpdt) ||
            (CV_MAT_TYPE(dpdt->type) != CV_32FC1 &&
            CV_MAT_TYPE(dpdt->type) != CV_64FC1) ||
            dpdt->rows != count*2 || dpdt->cols != 3 )
            CV_ERROR( CV_StsBadArg, "dp/dT must be 2Nx3 floating-point matrix" );

        if( CV_MAT_TYPE(dpdt->type) == CV_64FC1 )
            _dpdt = dpdt;
        else
            CV_CALL( _dpdt = cvCreateMat( 2*count, 3, CV_64FC1 ));
        dpdt_p = _dpdt->data.db;
        dpdt_step = _dpdt->step/sizeof(dpdt_p[0]);
    }

    if( dpdf )
    {
        if( !CV_IS_MAT(dpdf) ||
            (CV_MAT_TYPE(dpdf->type) != CV_32FC1 && CV_MAT_TYPE(dpdf->type) != CV_64FC1) ||
            dpdf->rows != count*2 || dpdf->cols != 2 )
            CV_ERROR( CV_StsBadArg, "dp/df must be 2Nx2 floating-point matrix" );

        if( CV_MAT_TYPE(dpdf->type) == CV_64FC1 )
            _dpdf = dpdf;
        else
            CV_CALL( _dpdf = cvCreateMat( 2*count, 2, CV_64FC1 ));
        dpdf_p = _dpdf->data.db;
        dpdf_step = _dpdf->step/sizeof(dpdf_p[0]);
    }

    if( dpdc )
    {
        if( !CV_IS_MAT(dpdc) ||
            (CV_MAT_TYPE(dpdc->type) != CV_32FC1 && CV_MAT_TYPE(dpdc->type) != CV_64FC1) ||
            dpdc->rows != count*2 || dpdc->cols != 2 )
            CV_ERROR( CV_StsBadArg, "dp/dc must be 2Nx2 floating-point matrix" );

        if( CV_MAT_TYPE(dpdc->type) == CV_64FC1 )
            _dpdc = dpdc;
        else
            CV_CALL( _dpdc = cvCreateMat( 2*count, 2, CV_64FC1 ));
        dpdc_p = _dpdc->data.db;
        dpdc_step = _dpdc->step/sizeof(dpdc_p[0]);
    }

    if( dpdk )
    {
        if( !CV_IS_MAT(dpdk) ||
            (CV_MAT_TYPE(dpdk->type) != CV_32FC1 && CV_MAT_TYPE(dpdk->type) != CV_64FC1) ||
            dpdk->rows != count*2 || (dpdk->cols != 5 && dpdk->cols != 4 && dpdk->cols != 2) )
            CV_ERROR( CV_StsBadArg, "dp/df must be 2Nx5, 2Nx4 or 2Nx2 floating-point matrix" );

        if( !distCoeffs )
            CV_ERROR( CV_StsNullPtr, "distCoeffs is NULL while dpdk is not" );

        if( CV_MAT_TYPE(dpdk->type) == CV_64FC1 )
            _dpdk = dpdk;
        else
            CV_CALL( _dpdk = cvCreateMat( dpdk->rows, dpdk->cols, CV_64FC1 ));
        dpdk_p = _dpdk->data.db;
        dpdk_step = _dpdk->step/sizeof(dpdk_p[0]);
    }

    calc_derivatives = dpdr || dpdt || dpdf || dpdc || dpdk;

    for( i = 0; i < count; i++ )
    {
        double X = M[i].x, Y = M[i].y, Z = M[i].z;
        double x = R[0]*X + R[1]*Y + R[2]*Z + t[0];
        double y = R[3]*X + R[4]*Y + R[5]*Z + t[1];
        double z = R[6]*X + R[7]*Y + R[8]*Z + t[2];
        double r2, r4, r6, a1, a2, a3, cdist;
        double xd, yd;

        z = z ? scale/z : scale/1;
        x *= z; y *= z;

        r2 = x*x + y*y;
        r4 = r2*r2;
        r6 = r4*r2;
        a1 = 2*x*y;
        a2 = r2 + 2*x*x;
        a3 = r2 + 2*y*y;
        cdist = 1 + k[0]*r2 + k[1]*r4 + k[4]*r6;
        xd = x*cdist + k[2]*a1 + k[3]*a2;
        yd = y*cdist + k[2]*a3 + k[3]*a1;

        m[i].x = xd*fx + cx*scale;
        m[i].y = yd*fy + cy*scale;

        if( calc_derivatives )
        {
            if( dpdc_p )
            {
                dpdc_p[0] = 1; dpdc_p[1] = 0;
                dpdc_p[dpdc_step] = 0;
                dpdc_p[dpdc_step+1] = 1;
                dpdc_p += dpdc_step*2;
            }

            if( dpdf_p )
            {
                if( fixedAspectRatio )
                {
                    dpdf_p[0] = 0; dpdf_p[1] = xd*aspectRatio;
                    dpdf_p[dpdf_step] = 0;
                    dpdf_p[dpdf_step+1] = yd;
                }
                else
                {
                    dpdf_p[0] = xd; dpdf_p[1] = 0;
                    dpdf_p[dpdf_step] = 0;
                    dpdf_p[dpdf_step+1] = yd;
                }
                dpdf_p += dpdf_step*2;
            }

            if( dpdk_p )
            {
                dpdk_p[0] = fx*x*r2;
                dpdk_p[1] = fx*x*r4;
                dpdk_p[dpdk_step] = fy*y*r2;
                dpdk_p[dpdk_step+1] = fy*y*r4;
                if( _dpdk->cols > 2 )
                {
                    dpdk_p[2] = fx*a1;
                    dpdk_p[3] = fx*a2;
                    dpdk_p[dpdk_step+2] = fy*a3;
                    dpdk_p[dpdk_step+3] = fy*a1;
                    if( _dpdk->cols > 4 )
                    {
                        dpdk_p[4] = fx*x*r6;
                        dpdk_p[dpdk_step+4] = fy*y*r6;
                    }
                }
                dpdk_p += dpdk_step*2;
            }

            if( dpdt_p )
            {
                double dxdt[] = { z, 0, -x*z }, dydt[] = { 0, z, -y*z };
                for( j = 0; j < 3; j++ )
                {
                    double dr2dt = 2*x*dxdt[j] + 2*y*dydt[j];
                    double dcdist_dt = k[0]*dr2dt + 2*k[1]*r2*dr2dt + 3*k[4]*r4*dr2dt;
                    double da1dt = 2*(x*dydt[j] + y*dxdt[j]);
                    double dmxdt = fx*(dxdt[j]*cdist + x*dcdist_dt +
                                k[2]*da1dt + k[3]*(dr2dt + 2*x*dxdt[j]));
                    double dmydt = fy*(dydt[j]*cdist + y*dcdist_dt +
                                k[2]*(dr2dt + 2*y*dydt[j]) + k[3]*da1dt);
                    dpdt_p[j] = dmxdt;
                    dpdt_p[dpdt_step+j] = dmydt;
                }
                dpdt_p += dpdt_step*2;
            }

            if( dpdr_p )
            {
                double dx0dr[] =
                {
                    X*dRdr[0] + Y*dRdr[1] + Z*dRdr[2],
                    X*dRdr[9] + Y*dRdr[10] + Z*dRdr[11],
                    X*dRdr[18] + Y*dRdr[19] + Z*dRdr[20]
                };
                double dy0dr[] =
                {
                    X*dRdr[3] + Y*dRdr[4] + Z*dRdr[5],
                    X*dRdr[12] + Y*dRdr[13] + Z*dRdr[14],
                    X*dRdr[21] + Y*dRdr[22] + Z*dRdr[23]
                };
                double dz0dr[] =
                {
                    X*dRdr[6] + Y*dRdr[7] + Z*dRdr[8],
                    X*dRdr[15] + Y*dRdr[16] + Z*dRdr[17],
                    X*dRdr[24] + Y*dRdr[25] + Z*dRdr[26]
                };
                for( j = 0; j < 3; j++ )
                {
                    double dxdr = z*(dx0dr[j] - x*dz0dr[j]);
                    double dydr = z*(dy0dr[j] - y*dz0dr[j]);
                    double dr2dr = 2*x*dxdr + 2*y*dydr;
                    double dcdist_dr = k[0]*dr2dr + 2*k[1]*r2*dr2dr + 3*k[4]*r4*dr2dr;
                    double da1dr = 2*(x*dydr + y*dxdr);
                    double dmxdr = fx*(dxdr*cdist + x*dcdist_dr +
                                k[2]*da1dr + k[3]*(dr2dr + 2*x*dxdr));
                    double dmydr = fy*(dydr*cdist + y*dcdist_dr +
                                k[2]*(dr2dr + 2*y*dydr) + k[3]*da1dr);
                    dpdr_p[j] = dmxdr;
                    dpdr_p[dpdr_step+j] = dmydr;
                }
                dpdr_p += dpdr_step*2;
            }
        }
    }

    if( _m != imagePoints )
        cvConvertPointsHomogeneous( _m, imagePoints );
    if( _dpdr != dpdr )
        cvConvert( _dpdr, dpdr );
    if( _dpdt != dpdt )
        cvConvert( _dpdt, dpdt );
    if( _dpdf != dpdf )
        cvConvert( _dpdf, dpdf );
    if( _dpdc != dpdc )
        cvConvert( _dpdc, dpdc );
    if( _dpdk != dpdk )
        cvConvert( _dpdk, dpdk );

    __END__;

    if( _M != objectPoints )
        cvReleaseMat( &_M );
    if( _m != imagePoints )
        cvReleaseMat( &_m );
    if( _dpdr != dpdr )
        cvReleaseMat( &_dpdr );
    if( _dpdt != dpdt )
        cvReleaseMat( &_dpdt );
    if( _dpdf != dpdf )
        cvReleaseMat( &_dpdf );
    if( _dpdc != dpdc )
        cvReleaseMat( &_dpdc );
    if( _dpdk != dpdk )
        cvReleaseMat( &_dpdk );
}

void ComputeCOP(CvMat **COP, CvMat **R_Mat, CvMat **T_Mat) {
	CvMat *temp = cvCreateMat(3, 1, CV_64F);
	CvMat *R_inv = cvCreateMat(3, 3, CV_64F);
	cvZero(temp);
	COP[0] = cvCreateMat(3, 1, CV_64F);
	cvZero(COP[0]);
	int view = 1;
	for( view = 1; view < NumOfViews; view++) {
		cvmSet(temp, 0, 0, COPX); 
		cvmSet(temp, 1, 0, COPY);
		cvmSet(temp, 2, 0, COPZ);
		// FIX ME: check if this computation is correct
		cvmSub(temp, T_Mat[view], temp);
		cvInvert(R_Mat[view], R_inv);
		COP[view] = cvCreateMat(3, 1, CV_64F);
		cvMatMul(R_inv, temp, COP[view]);
	}
	cvReleaseMat(&temp);

}

void ComputeOcclusion(CvMat *MapX,CvMat *MapY, CvMat *X, CvMat *Y, CvMat *Z0, CvMat *disactivated_pixels_map, CvMat *cop) {
	int n_x = MapX->cols;
	int n_y = MapX->rows;
	double xk = 0; double yk = 0;
	int rxk = 0; int ryk = 0;
	int x_ref = 0; int y_ref = 0;
	double X_curr = 0; double Y_curr = 0; double Z_curr = 0;
	double X_prev = 0; double Y_prev = 0; double Z_prev = 0;
	double dist = 0;
	double curr_dist = 0;
	double prev_dist = 0;
	double copx = 0; double copy = 0; double copz = 0;
	BOOL is_occluded = 0;
	int disactive = 0;
	CvMat *RefMap_X = cvCreateMat(n_y,n_x,CV_32FC1);
	CvMat *RefMap_Y = cvCreateMat(n_y,n_x,CV_32FC1);
	cvZero(RefMap_X);
	cvZero(RefMap_Y);
	
	if(cop) {
		copx = cvmGet(cop, 0, 0);
		copy = cvmGet(cop, 1, 0);
		copz = cvmGet(cop, 2, 0);
	}
	//loops over all pixels 
	for(int i = 0; i < n_x; i++) {
		for(int j = 0; j < n_y; j++) {
			is_occluded = 0;
			//the coordinates in the second image which correspond to (j,i)
			xk = cvmGet(MapX, j, i);
			yk = cvmGet(MapY, j, i);
			for(int ryk=cvRound(yk); ryk <= cvRound(yk); ryk++) {
				for(int rxk=cvRound(xk); rxk <= cvRound(xk); rxk++) {
			
			if (rxk < 0 || rxk >= n_x || ryk < 0 || ryk >= n_y)
				continue;
			
			// the remap reference coordinates 
			x_ref = (int)cvmGet(RefMap_X, ryk, rxk);
			y_ref = (int)cvmGet(RefMap_Y, ryk, rxk);
			// if xk and yk where previously then ref_x and ref_y are non zero
			if(x_ref == 0 && y_ref == 0) {
				// its the first time there is a map of reference image pixel to (ryk, rxk)
				cvmSet(RefMap_X, ryk, rxk, i);
				cvmSet(RefMap_Y, ryk, rxk, j);
				continue;
			}
			is_occluded = (x_ref - i)*(x_ref - i) + (y_ref - j)*(y_ref - j) > 1 ? 1 : 0;
			X_curr = cvmGet(X, j, i);
			Y_curr = cvmGet(Y, j, i);
			Z_curr = cvmGet(Z0, j, i);
			X_prev = cvmGet(X, y_ref, x_ref);
			Y_prev = cvmGet(Y, y_ref, x_ref);
			Z_prev = cvmGet(Z0, y_ref, x_ref);
			disactive = cvmGet8u(disactivated_pixels_map, y_ref, x_ref);
				// computes the distances 
			curr_dist = (X_curr-copx)*(X_curr-copx) + (Y_curr-copy)*(Y_curr-copy) + (Z_curr-copz)*(Z_curr-copz);
			prev_dist = (X_prev-copx)*(X_prev-copx) + (Y_prev-copy)*(Y_prev-copy) + (Z_prev-copz)*(Z_prev-copz);

			if(is_occluded)
				is_occluded = (X_curr - X_prev)*(X_curr - X_prev) + (Y_curr - Y_prev)*(Y_curr - Y_prev) + (Z_curr - Z_prev)*(Z_curr - Z_prev) > DISACTIVATION_THRESHOLD_Z ? 1 : 0;
				//is_occluded =  Z_curr/Z_prev > 1+DISACTIVATION_PRECENT || Z_curr/Z_prev < 1-DISACTIVATION_PRECENT? 1 : 0;
				//is_occluded =  curr_dist/prev_dist > 1+DISACTIVATION_PRECENT || curr_dist/prev_dist < 1-DISACTIVATION_PRECENT? 1 : 0;
				
			
				// in case the previous point is occluded
			if(curr_dist < prev_dist) {
				cvmSet(RefMap_X, ryk, rxk, i);
				cvmSet(RefMap_Y, ryk, rxk, j);
				if(disactive==0 && is_occluded)
					cvmSet8u(disactivated_pixels_map, y_ref, x_ref, 100);
			}else {
					// the current point is occluded
					if(is_occluded)
						cvmSet8u(disactivated_pixels_map, j, i, 100);
				}
			}
		}
			}
		}

			
	cvReleaseMat(&RefMap_X);
	cvReleaseMat(&RefMap_Y);
}
/***************computes the projection of world points to the image*******************/
void ComputeProjMaps(CvMat* X,CvMat* Y,
					 CvMat* Z,CvMat *R, 
					 CvMat* T, CvMat* C,
					 CvMat* Proj_MapX_Ir0,
					 CvMat* Proj_MapY_Ir0,
					 double eta_curr){
	int i=0;
	int n_x = X->cols;
	int n_y = X->rows;

	CvMat *X_row = cvCreateMat(1,n_y, CV_32FC1); 
	CvMat *Y_row = cvCreateMat(1,n_y, CV_32FC1); 
	CvMat *Z_row = cvCreateMat(1,n_y, CV_32FC1); 
	CvMat *proj_points = cvCreateMat(2,n_x, CV_32FC1); 

	CvMat *Point_row = cvCreateMat(3,n_x, CV_32FC1); 
    
	for(i=0; i< X->rows; i++){
		SetRow(X,Point_row,i,0);
		SetRow(Y,Point_row,i,1);
		SetRow(Z,Point_row,i,2);
		
		ProjectPoints2(Point_row, R, T, C, NULL,proj_points,eta_curr);
		SetRow(proj_points, Proj_MapX_Ir0, 0, i);
		SetRow(proj_points, Proj_MapY_Ir0, 1, i);
		
	}
	
	// tali: release the allocated matrixes
    cvReleaseMat(&X_row);
	cvReleaseMat(&Y_row);
	cvReleaseMat(&Z_row);
	cvReleaseMat(&Point_row);
	cvReleaseMat(&proj_points);
	
}


/************************* Writes a matrix to txt file **********************************/
void WriteRes(FILE* filename,CvMat* M){
	int i=0;
	int j=0;
	double Val = 0;
	if(M == NULL){
		i=1;
		return;
	}
	
	if (filename){
		  for (int i = 0; i < M->rows; ++i){
			  for (int j = 0; j < M->cols; ++j){
				  Val  = cvmGet(M,i,j);
				  fprintf (filename, "%f\t",Val );
			  }
			  fprintf (filename, "\n");
		  }
	  }
}
/************************* Reads a matrix to txt file **********************************/
void ReadMat(char* filename,CvMat* M){
	int i=0;
	int j=0;
	double Val = 0;
	if(M == NULL){
		i=1;
		return;
	}
	
	if (filename){
		FILE *f = fopen(filename,"r");
		  for (int i = 0; i < M->rows; ++i){
			  for (int j = 0; j < M->cols; ++j){
				  fscanf(f,"%lf",&Val);
				  cvmSet(M,i,j,Val);	  
			  }
		  }
		 fclose(f);
	  }

}

void WriteRes8u(FILE* filename,const CvMat* M){
	int i=0;
	int j=0;
	double Val = 0;
	if(M == NULL){
		i=1;
		return;
	}
	
	if (filename){
		  for (int i = 0; i < M->rows; ++i){
			  for (int j = 0; j < M->cols; ++j){
				  Val  = cvmGet8u(M,i,j);
				  fprintf (filename, "%f\t",Val );
			  }
			  fprintf (filename, "\n");
		  }
	  }
}

/********************** Computes the world coordinates gradients ************************/
void Gradient3DPoints(CvMat* MapIdentityX,CvMat* MapIdentityY, CvMat* Z, CvMat* Z_X, CvMat* Z_Y, 
					  double eta_curr,
					  CvMat* Xx, CvMat* Xy,CvMat* Xz, CvMat* Yx,CvMat* Yy,CvMat* Yz)
{	
	CvMat *temp1 = cvCreateMat(Z->rows, Z->cols, CV_32FC1);
	CvMat *temp2 = cvCreateMat(Z->rows, Z->cols, CV_32FC1);
    //---------------- X derivatives -----------------------
	// temp1<- x - u0*eta_curr
	cvAddS(MapIdentityX,cvScalar(-u0*eta_curr),temp1);
	cvCvtScale(temp1,Xz,1/(f0ku*eta_curr),0);
	
	//computes Xx: [Z_X*temp1 + Z]/[f0ku*eta_curr]
	cvMul(temp1,Z_X,temp2,1/(f0ku*eta_curr));
	cvAddWeighted(temp2,1,Z,1/(f0ku*eta_curr),0,Xx);

	// computes Xy: [Z_Y*temp1]/[f0ku*eta_curr]
    cvMul(temp1,Z_Y,Xy,1/(f0ku*eta_curr));
	//------------------------------------------------------
	cvZero(temp1);
	cvZero(temp2);
	//---------------- Y derivatives -----------------------
	// temp1<- y - v0*eta_curr
	cvAddS(MapIdentityY,cvScalar(-v0*eta_curr),temp1);
	cvCvtScale(temp1,Yz,1/(f0kv*eta_curr),0);
	
	// computes Yx
	cvMul(temp1,Z_X,Yx,1/(f0kv*eta_curr));

	// computes Yy
	cvMul(temp1,Z_Y,temp2,1/(f0kv*eta_curr));
	cvAddWeighted(temp2,1,Z,1/(f0kv*eta_curr),0,Yy);

	cvReleaseMat(&temp1);
	cvReleaseMat(&temp2);
}


/******************** Computes the projection coordinates gradients ************************/
void GradientProjCoor(CvMat* X, CvMat* Y, CvMat* Z, CvMat *X_driv, CvMat *Y_driv,CvMat* Z_driv, 
					  CvMat* M, CvMat* Proj_driv, int grad_y){
    
	CvMat *denom = cvCreateMat(X->rows, X->cols, CV_32FC1);
	CvMat *numer = cvCreateMat(X->rows, X->cols, CV_32FC1);
	CvMat *temp1 = cvCreateMat(X->rows, X->cols, CV_32FC1);
	CvMat *temp2 = cvCreateMat(X->rows, X->cols, CV_32FC1);
	double Mn1; double Mn2; double Mn3; double Mn4;
	double Md1; double Md2; double Md3; double Md4;
		
	// the last line of the projection matrix
	Md1 = cvmGet(M, 2, 0);
	Md2 = cvmGet(M, 2, 1);
	Md3 = cvmGet(M, 2, 2);
	Md4 = cvmGet(M, 2, 3);
	
	// denom = M31*X + M32*Y + M33*Z + M34
	cvAddWeighted(X, Md1, Y, Md2, 0, temp1);
	cvAddWeighted(temp1, 1, Z ,Md3, Md4,denom);
	cvZero(temp1);
	
	// if computing the gradient of the horizintal projected coordinate
	if(grad_y==0){
		Mn1 = cvmGet(M, 0, 0);
		Mn2 = cvmGet(M, 0, 1);
		Mn3 = cvmGet(M, 0, 2);
		Mn4 = cvmGet(M, 0, 3);
	}
	else 
	{
	    Mn1 = cvmGet(M, 1, 0);
		Mn2 = cvmGet(M, 1, 1);
		Mn3 = cvmGet(M, 1, 2);
		Mn4 = cvmGet(M, 1, 3);
	}

	//numer = M11*X + M12*Y + M13*Z + M14
    cvAddWeighted(X, Mn1, Y, Mn2, 0, temp1);
	cvAddWeighted(temp1, 1, Z, Mn3, Mn4,numer);
	

	// Mn1*X_driv + Mn2*Y_driv + Mn3*Z_driv
	cvAddWeighted(X_driv, Mn1, Y_driv, Mn2, 0, temp1);
	cvAddWeighted(temp1,1, Z_driv, Mn3, 0, temp1);

	// Md1*X_driv + Md2*Y_driv + Md3*Z_driv
	cvAddWeighted(X_driv, Md1, Y_driv, Md2, 0,temp2);
	cvAddWeighted(temp2, 1, Z_driv, Md3, 0,temp2);
	//Proj_driv <- temp1*denom
	cvMul(temp1,denom,Proj_driv);
	cvZero(temp1);
	// temp1 <-- temp2*numer
	cvMul(temp2,numer,temp1);
	// Proj_driv <--denom(...)-numer(...)
	cvSub(Proj_driv,temp1, Proj_driv);
    cvZero(temp1);
	cvZero(temp2);
	
	cvMul(denom,denom,temp1);
	cvDiv(Proj_driv, temp1, Proj_driv);
	
    cvReleaseMat(&denom);
	cvReleaseMat(&numer);
	cvReleaseMat(&temp1);
	cvReleaseMat(&temp2);
}
void SceneFlow_MVZ(FILE* logfile, //the log file
	       const double eta, //reduction factor
	       int pyramid_levels, // total number of pyramid levels
	       const CvMat * const * const * const Images_0_pyramid, //time t Images
	       const CvMat * const * const * const Images_t_pyramid, //time t+1 Images
           int pyramid_level_start, // starting level
	       CvMat* U_start, //horizontal left scene flow initialization
	       CvMat* V_start, //vertical scene flow initialization
		   CvMat* W_start, //depth scene flow initialization
           CvMat* Z0_start, // Z(x,y) initialization
           int pyramid_level_end, // end pyramid level
	       CvMat* U_end, //final U
	       CvMat* V_end, //final V
		   CvMat* W_end, //final V
		   CvMat* Z0_end, //t Z 
		   const double h, //grid step for finite elements
	       const double epsilon_outer, //stopping criteria for outer iterations 
	       const double epsilon_inner, //stopping criteria for inner iterations
	       const double epsilon_SOR, //stopping criteria for the linear system resolution
	       const int max_iter_outer, //maximal number of outer iterations
	       const int max_iter_inner, //maximal number of inner iterations
	       const int max_iter_SOR, //maximal number of linear system resolution iterations
	       const double omega_SOR, //SOR method weight
           double alpha, // smooth term weight
		   double alpha_w, // smooth term weight
		   double alpha_wz, // smooth term weight
           const double gamma, // data term: gradient weight
           const double lambda, // smooth term: disparity flow weight
           const double mu, // smooth term: disparity weight
		   const double mu2, // smooth term: disparity weight
           const double mu_z,
		   const double mu_u,
		   const double mu_v,
		   const double mu_w,
		   const double f0ku, // x focal length of the reference camera
		   const double f0kv, // y focal length of the refernce camera
		   const double u0, 
		   const double v0,
		   CvMat** R_Mat, //the rotation matrixes
		   CvMat** T_Mat, //the translation matrixes
		   CvMat** C_Mat, //the calibration matrixes
		   const int methode) //the chosen method 
{
    int N_x,N_y;
    char file[256];
	FILE *f = 0;
    CvMat *Z0=0, *U=0, *V=0, *W=0;
	CvMat *X=0,  *Y=0, *Xt=0, *Yt=0, *Zt=0;
	CvMat **M_curr = 0; 
	
	int dimension = 0;
    int do_ref_flow = 0;
    int do_not_ref_flow = 0;
    int do_Z0 = 0;
    int do_Zt = 0;
    int xl0, yl0;
	const double ref_params[4] = {f0ku, f0kv, u0, v0};
	double eta_curr = 1;

	for (int i = 1; i < pyramid_level_start; i++) {
		eta_curr *= PyramidEta;
	}
	if(ENABLE_ALPHA) {
		alpha = alpha*eta_curr;
		alpha_w = alpha_w*eta_curr;
		alpha_wz = alpha_wz*eta_curr;
	}

	// computes the initial reference projection matrix
	int view = 0;
	M_curr = new CvMat*[NumOfViews];
	for(view = 0; view < NumOfViews; view++) {
		M_curr[view] = cvCreateMat(3,4,CV_64FC1);
		ComputeProjMat(M_curr[view],R_Mat[view], T_Mat[view], C_Mat[view], eta_curr);
		#ifdef DEBUG
			sprintf(file, "%s/ProjMat%d.txt", TxtResultsDir,view);
			f = fopen(file,"w");
			WriteRes(f,M_curr[view]);
			fclose(f);
		#endif
	}
    if(methode == 0){
        // computing all 4 data terms. U,V,W,Z optimization
        do_ref_flow = 1;
		do_not_ref_flow = 1;
        do_Z0 = 1;
        do_Zt = 1;
        dimension = 4;
    }
	if(methode == 1) {
		// U,V,W optimization 
		do_ref_flow = 1;
		do_not_ref_flow = 1;
        do_Z0 = 0;
        do_Zt = 1;
        dimension = 3;
	}
	if(methode == 2) {
		do_ref_flow = 1;
		do_not_ref_flow = 1;
		do_Z0 = 1;
		do_Zt = 1;
		dimension = 3;
	}
	else if (methode == 6) {
		//only Z optimization at time t0
		do_Z0 = 1;
		dimension = 1;
	}

	int pyramidtype = CV_MAT_TYPE(Images_0_pyramid[0][pyramid_level_start-1]->type);

    //intial scene flow is set to the lowest resolution
    {
        CvSize size_start = cvGetSize(Images_0_pyramid[0][pyramid_level_start-1]);
		Z0 = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
		cvCopy(Z0_start, Z0);
		
		if (do_ref_flow || do_not_ref_flow || do_Zt) {
            U = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
            V = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
			W = cvCreateMat(size_start.height, size_start.width, CV_32FC1);
			cvCopy(U_start, U);
            cvCopy(V_start, V);
			cvCopy(W_start, W);
        }
	}
	//Computes the center of projection for all views
	CvMat **COP = 0; 
	COP = new CvMat*[NumOfViews];
	ComputeCOP(COP, R_Mat, T_Mat);
    
    //Loop over the resolution levels
	for(int level = pyramid_level_start - 1; level >= pyramid_level_end - 1; level--){

        CvSize level_size = cvGetSize(Images_0_pyramid[0][level]);
		//Size of the current images
		N_x = level_size.width;
		N_y = level_size.height;

		//Maps used for the images warping : Initialization
		CvMat *MapIdentityX = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *MapIdentityY = cvCreateMat(N_y, N_x, CV_32FC1);

		if(do_ref_flow || do_not_ref_flow || do_Zt) {
			//3D points at t+1
			Xt = cvCreateMat(N_y, N_x, CV_32FC1);
			Yt = cvCreateMat(N_y, N_x, CV_32FC1);
			Zt = cvCreateMat(N_y, N_x, CV_32FC1);
		}
		//MeshGrid 
		for(yl0 = 0; yl0 < N_y; yl0++){
			for(xl0 = 0; xl0 < N_x; xl0++){
				cvmSet(MapIdentityX, yl0, xl0, xl0);
				cvmSet(MapIdentityY, yl0, xl0, yl0);
			}
		}
		////////////////////////////////////////////////////////
		//Warped Images
		CvMat **Images_0w = new CvMat*[(size_t)NumOfViews];
		CvMat **Images_tw = new CvMat*[NumOfViews];
		//Maps for warping
		CvMat **MapX_Img_0 = new CvMat*[NumOfViews];
		CvMat **MapY_Img_0 = new CvMat*[NumOfViews];
		CvMat **MapX_Img_0_inner = new CvMat*[NumOfViews];
		CvMat **MapY_Img_0_inner = new CvMat*[NumOfViews];
		CvMat **MapX_Img_t = new CvMat*[NumOfViews];
		CvMat **MapY_Img_t = new CvMat*[NumOfViews];
		CvMat **MapX_Img_t_inner = new CvMat*[NumOfViews];
		CvMat **MapY_Img_t_inner = new CvMat*[NumOfViews];

		for(view = 0; view < NumOfViews; view++) {
			Images_0w[view] = cvCreateMat(N_y, N_x, pyramidtype);
			MapX_Img_0[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_0[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapX_Img_0_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_0_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			Images_tw[view] = cvCreateMat(N_y, N_x, pyramidtype);
			MapX_Img_t[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_t[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapX_Img_t_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
			MapY_Img_t_inner[view] =  cvCreateMat(N_y, N_x, CV_32FC1);
		}
		CvMat *X = cvCreateMat(N_y, N_x, CV_32FC1);
        CvMat *Y = cvCreateMat(N_y, N_x, CV_32FC1);

		//Images gradients computation
		CvMat *derivative_tmp = 0;
		if (pyramidtype == CV_8UC1)
			derivative_tmp = cvCreateMat(N_y, N_x, CV_16SC1);
		CvMat **Images_0_X  = new CvMat*[NumOfViews];
		CvMat **Images_0_Y  = new CvMat*[NumOfViews];
		CvMat **Images_0w_X = new CvMat*[NumOfViews];
		CvMat **Images_0w_Y = new CvMat*[NumOfViews];
		CvMat **Images_t_X  = new CvMat*[NumOfViews];
		CvMat **Images_t_Y  = new CvMat*[NumOfViews];
		CvMat **Images_tw_X = new CvMat*[NumOfViews];
		CvMat **Images_tw_Y = new CvMat*[NumOfViews];
		CvMat **Proj_x0_x = 0, **Proj_x0_y = 0, **Proj_x0_z = 0;
		CvMat **Proj_x0_u = 0, **Proj_x0_v = 0, **Proj_x0_w = 0;
		CvMat **Proj_y0_x = 0, **Proj_y0_y = 0, **Proj_y0_z = 0;
		CvMat **Proj_y0_u = 0, **Proj_y0_v = 0, **Proj_y0_w = 0;
		CvMat **Proj_xt_x = 0, **Proj_xt_y = 0, **Proj_xt_z = 0;
		CvMat **Proj_xt_u = 0, **Proj_xt_v = 0, **Proj_xt_w = 0;
		CvMat **Proj_yt_x = 0, **Proj_yt_y = 0, **Proj_yt_z = 0;
		CvMat **Proj_yt_u = 0, **Proj_yt_v = 0, **Proj_yt_w = 0;

		CvMat *X_x = 0, *X_y = 0, *X_z = 0;
		CvMat *Y_x = 0, *Y_y = 0, *Y_z = 0;
		CvMat *Xt_x = 0, *Xt_y = 0, *Xt_z = 0;
		CvMat *Yt_x = 0, *Yt_y = 0, *Yt_z = 0;
		CvMat *Zt_X = 0, *Zt_Y = 0;

		// the world point X,Y derivatives with respect to the ref camera coordinates
		X_x = cvCreateMat(N_y, N_x, CV_32FC1);
		X_y = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_x = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_y = cvCreateMat(N_y, N_x, CV_32FC1);
		// if Z optimization 
		X_z = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_z = cvCreateMat(N_y, N_x, CV_32FC1);
		// if U,V,W
		if(do_ref_flow || do_not_ref_flow || do_Zt){
			Xt_x = cvCreateMat(N_y, N_x, CV_32FC1);
			Xt_y = cvCreateMat(N_y, N_x, CV_32FC1);
			Yt_x = cvCreateMat(N_y, N_x, CV_32FC1);
			Yt_y = cvCreateMat(N_y, N_x, CV_32FC1);
			Zt_X = cvCreateMat(N_y, N_x, CV_32FC1);
			Zt_Y = cvCreateMat(N_y, N_x, CV_32FC1);
			if(do_Z0)
			{
				Xt_z = cvCreateMat(N_y, N_x, CV_32FC1);
				Yt_z = cvCreateMat(N_y, N_x, CV_32FC1);
			}
		}
		if (do_ref_flow ||do_not_ref_flow || do_Zt) { 
			Proj_xt_x = new CvMat*[NumOfViews];
			Proj_xt_y = new CvMat*[NumOfViews];
			Proj_yt_x = new CvMat*[NumOfViews];
			Proj_yt_y = new CvMat*[NumOfViews];
			Proj_xt_u = new CvMat*[NumOfViews]; 
			Proj_xt_v = new CvMat*[NumOfViews];
			Proj_xt_w = new CvMat*[NumOfViews];
			Proj_yt_u = new CvMat*[NumOfViews]; 
			Proj_yt_v = new CvMat*[NumOfViews];
			Proj_yt_w = new CvMat*[NumOfViews];
			if(do_Z0){
				Proj_xt_z = new CvMat*[NumOfViews];
				Proj_yt_z = new CvMat*[NumOfViews];
			}
			for(view = 0; view < NumOfViews; view++){
				Images_t_X[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
				Images_t_Y[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_xt_x[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_xt_y[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_yt_x[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_yt_y[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_xt_u[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_xt_v[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_xt_w[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_yt_u[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_yt_v[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_yt_w[view]   = cvCreateMat(N_y, N_x, CV_32FC1);
				if(do_Z0){
					Proj_xt_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_yt_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				}
			}
		}
        if (do_Z0 || do_not_ref_flow) {
			Proj_x0_x = new CvMat*[NumOfViews];
			Proj_x0_y = new CvMat*[NumOfViews];
			Proj_y0_x = new CvMat*[NumOfViews];
			Proj_y0_y = new CvMat*[NumOfViews];
			Proj_x0_u = new CvMat*[NumOfViews]; 
			Proj_x0_v = new CvMat*[NumOfViews];
			Proj_x0_w = new CvMat*[NumOfViews];
			Proj_y0_u = new CvMat*[NumOfViews]; 
			Proj_y0_v = new CvMat*[NumOfViews];
			Proj_y0_w = new CvMat*[NumOfViews];
			if(do_Z0){
				Proj_x0_z = new CvMat*[NumOfViews];
				Proj_y0_z = new CvMat*[NumOfViews];
			}
			for(view = 0; view < NumOfViews; view++){
				Images_0_X[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Images_0_Y[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_x0_x[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_x0_y[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_y0_x[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
				Proj_y0_y[view]  = cvCreateMat(N_y, N_x, CV_32FC1);
				if(do_not_ref_flow) { 
					Proj_x0_u[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_x0_v[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_x0_w[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_y0_u[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_y0_v[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_y0_w[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				}
				if(do_Z0) {
					Proj_x0_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Proj_y0_z[view] = cvCreateMat(N_y, N_x, CV_32FC1);
				}
			}
		}
	  
        //Compute gradients of the only non-warped image
		for(view = 0; view < NumOfViews; view++) {
			Derivative(Images_0_pyramid[view][level], 1, 0, derivative_tmp, Images_0_X[view]);
			Derivative(Images_0_pyramid[view][level], 0, 1, derivative_tmp, Images_0_Y[view]);
		}
        
		//M is the matrix of the final linear system and B the second member of this system
		CvMat *M = cvCreateMat(dimension * N_x * N_y, dimension + 4, CV_32FC1); 
	    CvMat *B = cvCreateMat(dimension * N_x * N_y,1,CV_32FC1);
		
        CvMat *disactivated_pixels_ref_flow = 0;
		CvMat **disactivated_pixels_flow = 0;
        CvMat **disactivated_pixels_Z0 = 0;
		CvMat **disactivated_pixels_Zt = 0;
		//tali: CHECK ME
		// occlution map for Images_tw[0]: between the reference images at time t and t+1
		if (do_ref_flow || do_Zt) {
            disactivated_pixels_ref_flow = cvCreateMat(N_y, N_x, CV_8UC1);
        }
		// occlution map for I0w: between the reference image and other wraped images at time t
        if (do_Z0 || do_not_ref_flow ) {
            disactivated_pixels_Z0 = new CvMat*[NumOfViews];
        }
		// occlution map for Itw: between the reference image and other wraped images at time t+1
        if (do_not_ref_flow || do_Zt) {
            disactivated_pixels_flow = new CvMat*[NumOfViews];
        }
	
		// occlution map for Itw: between the reference image and other wraped images at time t+1
		if (do_Zt || do_not_ref_flow) {
            disactivated_pixels_Zt = new CvMat*[NumOfViews];
        }
        
        //divergence coefficients
		CvMat *divergence_coefficients = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *divergence_coefficients_Z = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *divergence_coefficients_W = cvCreateMat(N_y, N_x, CV_32FC1);
		CvMat *divergence_coefficients_WZ = cvCreateMat(N_y, N_x, CV_32FC1);
        CvMat *Z0_X,*Z0_Y,*U_X, *U_Y, *V_X, *V_Y, *W_X, *W_Y;
        CvMat *dZ0, *dU, *dW, *dV;
        CvMat *dZ0_prec,  *dU_prec,  *dV_prec, *dW_prec;
        CvMat *dZ0_X,*dZ0_Y, *dU_X, *dU_Y, *dV_X, *dV_Y,*dW_X, *dW_Y;
        CvMat *Z0_prec,*U_prec, *V_prec, *W_prec;
		CvMat *Z_inner=0, *X_inner=0, *Y_inner=0;
		CvMat *Zt_inner=0, *Xt_inner=0, *Yt_inner=0;
		
		//flow
		if(do_ref_flow || do_not_ref_flow || do_Zt ){
#ifdef DEBUG
            //saving the initial flow images
            if (SF_U_FileName) {
                SaveResults(SF_U_FileName, U);
            }
            if (SF_V_FileName) {
                SaveResults(SF_V_FileName, V);
            }
			if (SF_W_FileName) {
                SaveResults(SF_W_FileName, W);
            }
#endif

	
            U_prec = cvCreateMat(N_y, N_x, CV_32FC1);
            V_prec = cvCreateMat(N_y, N_x, CV_32FC1);
			W_prec = cvCreateMat(N_y, N_x, CV_32FC1);
	   
            dU = cvCreateMat(N_y, N_x, CV_32FC1);
            dV = cvCreateMat(N_y, N_x, CV_32FC1);
			dW = cvCreateMat(N_y, N_x, CV_32FC1);
            dU_X = cvCreateMat(N_y, N_x, CV_32FC1);
            dU_Y = cvCreateMat(N_y, N_x, CV_32FC1);
            dV_X = cvCreateMat(N_y, N_x, CV_32FC1);
            dV_Y = cvCreateMat(N_y, N_x, CV_32FC1);
			dW_X = cvCreateMat(N_y, N_x, CV_32FC1);
            dW_Y = cvCreateMat(N_y, N_x, CV_32FC1);
	     
            U_X = cvCreateMat(N_y, N_x, CV_32FC1);
            U_Y = cvCreateMat(N_y, N_x, CV_32FC1);
            V_X = cvCreateMat(N_y, N_x, CV_32FC1);
            V_Y = cvCreateMat(N_y, N_x, CV_32FC1);
			W_X = cvCreateMat(N_y, N_x, CV_32FC1);
            W_Y = cvCreateMat(N_y, N_x, CV_32FC1);

            dU_prec = cvCreateMat(N_y, N_x, CV_32FC1);
            dV_prec = cvCreateMat(N_y, N_x, CV_32FC1);
			dW_prec = cvCreateMat(N_y, N_x, CV_32FC1);

			Zt_inner = cvCreateMat(N_y, N_x, CV_32FC1);
			Xt_inner = cvCreateMat(N_y, N_x, CV_32FC1);
			Yt_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		}
		
		Z0_X = cvCreateMat(N_y, N_x, CV_32FC1);
		Z0_Y = cvCreateMat(N_y, N_x, CV_32FC1);
		Z_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		X_inner = cvCreateMat(N_y, N_x, CV_32FC1);
		Y_inner = cvCreateMat(N_y, N_x, CV_32FC1);
			
		if(do_Z0) {
			Z0_prec = cvCreateMat(N_y, N_x, CV_32FC1);
		}

		if(do_Z0) {
			if (SF_D0_FileName) {
				SaveResults(SF_D0_FileName, Z0);
			}	   
			dZ0 = cvCreateMat(N_y, N_x, CV_32FC1);
			dZ0_X = cvCreateMat(N_y, N_x, CV_32FC1);
			dZ0_Y = cvCreateMat(N_y, N_x, CV_32FC1);
			dZ0_prec = cvCreateMat(N_y, N_x, CV_32FC1);
		}
		//For the final linear system and SOR iterations
        CvMat *SOR = cvCreateMat(dimension*N_x*N_y,1, CV_32FC1);	
        CvMat *SOR_prec = cvCreateMat(dimension*N_x*N_y,1,CV_32FC1);

		//Main fixed point iterations, loop over optical flow/disparity values
        double norm = 0., norm_inner = 0., norm_SOR = 0.;
        int iter_SOR = 0, iter_inner = 0, iter_outer = 0;
        int first_iteration = 1;
		
		while((norm > epsilon_outer && iter_outer <= max_iter_outer) || iter_outer < 2 || first_iteration == 1){
			first_iteration = 0;
		    
			//Maps update
			ComputeXY(X, Y, Z0, MapIdentityX, MapIdentityY, N_x, N_y, ref_params, eta_curr);
			//Images_0w
			if(do_Z0 || do_not_ref_flow ) {
				for(view = 1; view < NumOfViews; view++) {
					ComputeProjMaps(X, Y, Z0, R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_0[view], MapY_Img_0[view], eta_curr);
				}
			}
			if (do_ref_flow || do_not_ref_flow || do_Zt) {
				cvAdd(X, U, Xt);
				cvAdd(Y, V, Yt);
				cvAdd(Z0,W, Zt);
			}

			// Images_tw - ref image
			if(do_ref_flow || do_Zt) {
				ComputeProjMaps(Xt, Yt, Zt, R_Mat[0], T_Mat[0], C_Mat[0], MapX_Img_t[0], MapY_Img_t[0], eta_curr);
			}
			//Images_tw
			if(do_not_ref_flow || do_Zt) {
				for(view = 1; view < NumOfViews; view++) {
					ComputeProjMaps(Xt, Yt, Zt, R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_t[view], MapY_Img_t[view], eta_curr);
				}
			}	

	#if OCCLUSION1 
			if (do_ref_flow || do_Zt) {
					cvZero(disactivated_pixels_ref_flow);
					ComputeOcclusion(MapX_Img_t[0], MapY_Img_t[0], Xt, Yt, Zt, disactivated_pixels_ref_flow);
					cvmSet8u_borders(disactivated_pixels_ref_flow, 127);
			}
			if (do_Z0 || do_not_ref_flow) {
				for(view = 1; view < NumOfViews; view++){
					disactivated_pixels_Z0[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					cvZero(disactivated_pixels_Z0[view]);
					ComputeOcclusion(MapX_Img_0[view], MapY_Img_0[view], X, Y, Z0, disactivated_pixels_Z0[view], COP[view]);
					cvmSet8u_borders(disactivated_pixels_Z0[view], 127);
				}
			}
			// not so sure about how to compute right flow occlusion...
			//CHECK ME: notice that the computation is done with respect reference coordinates:
			// Namely : the occloution is computed between the left and right image (Irtw). 
			if (do_not_ref_flow || do_Zt) {
				for(view = 1; view < NumOfViews; view++){
					disactivated_pixels_flow[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					disactivated_pixels_Zt[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					cvZero(disactivated_pixels_flow[view]);
					if (BLACK_LINE && view == 1){
						CvMat *img = cvLoadImageM(Zt_initFile, CV_LOAD_IMAGE_GRAYSCALE);
						cvResize(img, disactivated_pixels_flow[view],CV_INTER_AREA);
						cvReleaseMat(&img);
					}
					ComputeOcclusion(MapX_Img_t[view], MapY_Img_t[view], Xt, Yt, Zt, disactivated_pixels_flow[view], COP[view]);
					cvmSet8u_borders(disactivated_pixels_flow[view], 127);
					cvCopy(disactivated_pixels_flow[view],disactivated_pixels_Zt[view]);
				}
			}
	#else
			CvMat *Dist = cvCreateMat(N_y, N_x, CV_32FC1);
			CvMat *Dist_warped = cvCreateMat(N_y, N_x, CV_32FC1);
			CvMat *Dist_backwarped = cvCreateMat(N_y, N_x, CV_32FC1);
			CvMat *temp = cvCreateMat(N_y, N_x, CV_32FC1);
		

			if (do_ref_flow || do_Zt) {
				cvZero(disactivated_pixels_ref_flow);
				Dist_map(Dist,Xt, Yt, Zt, COP[0]);
				DistBufMap_32f(Dist, Dist_warped, MapX_Img_t[0], MapY_Img_t[0], Xt, Yt, Zt,COP[0]);
				cvRemap(Dist_warped, Dist_backwarped,MapX_Img_t[0], MapY_Img_t[0], CV_INTER_NN+CV_WARP_FILL_OUTLIERS, cvScalar(N_x));
				cvDiv(Dist, Dist_backwarped, temp);
				cvCmpS(temp, 1+DISACTIVATION_PRECENT, disactivated_pixels_ref_flow, CV_CMP_GT);
				//cvAddS(Dist_backwarped, cvScalar(DISACTIVATION_THRESHOLD), Dist_backwarped);
				//cvCmp(Dist, Dist_backwarped, disactivated_pixels_ref_flow, CV_CMP_GE);
				cvmSet8u_borders(disactivated_pixels_ref_flow, 127);
				

			}
			if (do_Z0 || do_not_ref_flow) {
				for(view = 1; view < NumOfViews; view++){
					disactivated_pixels_Z0[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					cvZero(disactivated_pixels_Z0[view]);
					Dist_map(Dist, X, Y, Zt, COP[view]);
					DistBufMap_32f(Dist, Dist_warped, MapX_Img_0[view], MapY_Img_0[view], X, Y, Z0, COP[view]);
					cvRemap(Dist_warped, Dist_backwarped, MapX_Img_0[view], MapY_Img_0[view], CV_INTER_NN+CV_WARP_FILL_OUTLIERS, cvScalar(N_x));
					cvDiv(Dist, Dist_backwarped, temp);
					cvCmpS(temp, 1+DISACTIVATION_PRECENT, disactivated_pixels_Z0[view], CV_CMP_GT);
					//cvAddS(Dist_backwarped, cvScalar(DISACTIVATION_THRESHOLD), Dist_backwarped);
					//cvCmp(Dist, Dist_backwarped, disactivated_pixels_Z0[view], CV_CMP_GE);
					cvmSet8u_borders(disactivated_pixels_Z0[view], 127);
				}
			}

			if (do_not_ref_flow || do_Zt) {
				for(view = 1; view < NumOfViews; view++){
					disactivated_pixels_flow[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					disactivated_pixels_Zt[view] = cvCreateMat(N_y, N_x, CV_8UC1); 
					cvZero(disactivated_pixels_flow[view]);
					Dist_map(Dist, Xt, Yt, Zt, COP[view]);
					DistBufMap_32f(Dist, Dist_warped, MapX_Img_t[view], MapY_Img_t[view], Xt, Yt, Zt, COP[view]);
					cvRemap(Dist_warped, Dist_backwarped, MapX_Img_t[view], MapY_Img_t[view], CV_INTER_NN+CV_WARP_FILL_OUTLIERS, cvScalar(N_x));
					cvDiv(Dist, Dist_backwarped, temp);
					cvCmpS(temp, 1+DISACTIVATION_PRECENT, disactivated_pixels_flow[view], CV_CMP_GT);
					//cvAddS(Dist_backwarped, cvScalar(DISACTIVATION_THRESHOLD), Dist_backwarped);
					//cvCmp(Dist, Dist_backwarped, disactivated_pixels_flow[view], CV_CMP_GE);
					if (BLACK_LINE && view == 1){
						CvMat *img_temp1 = cvLoadImageM(Zt_initFile, CV_LOAD_IMAGE_GRAYSCALE);
						CvMat *img_temp2 = cvCreateMat(N_y, N_x, CV_8UC1);
						cvResize(img_temp1, img_temp2, CV_INTER_AREA);
						cvAdd(img_temp2, disactivated_pixels_flow[view], disactivated_pixels_flow[view]);
						cvReleaseMat(&img_temp1);
						cvReleaseMat(&img_temp2);
					}
					cvmSet8u_borders(disactivated_pixels_flow[view], 127);
					cvCopy(disactivated_pixels_flow[view],disactivated_pixels_Zt[view]);
				}
			}
			cvReleaseMat(&Dist);
			cvReleaseMat(&Dist_warped);
			cvReleaseMat(&Dist_backwarped);
			cvReleaseMat(&temp);
			
	#endif
#ifdef DEBUG
			for(view = 1; view < NumOfViews; view++) {
				sprintf(file, "%sv%d_l%d.txt", SF_disactivated_Surf_1_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_Z0[view]);
				fclose(f);

				sprintf(file, "%sv%d_l%d.txt",SF_disactivated_flow_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_flow[view]);
				fclose(f);

				sprintf(file, "%sv%d_l%d.txt",SF_disactivated_Surf_2_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_Zt[view]);
				fclose(f);

			}
			sprintf(file, "%s_l%d.txt",SF_disactivated_ref_flow_FileName,level);
			f = fopen(file,"w");
			WriteRes8u	(f,disactivated_pixels_ref_flow);
			fclose(f);
#endif

			if (do_ref_flow || do_Zt) {
				cvRemap(Images_t_pyramid[0][level], Images_tw[0], MapX_Img_t[0], MapY_Img_t[0]);
				SmoothImage(Images_tw[0]);

				Images_tw_X[0] = cvCreateMat(N_y, N_x, CV_32FC1);
				Images_tw_Y[0] = cvCreateMat(N_y, N_x, CV_32FC1);
				Derivative(Images_tw[0] , 1, 0, derivative_tmp, Images_tw_X[0]);
				Derivative(Images_tw[0] , 0, 1, derivative_tmp, Images_tw_Y[0]);
			}
			if (do_Z0 || do_not_ref_flow) {
				for(view = 1; view < NumOfViews; view++) {
					cvRemap(Images_0_pyramid[view][level], Images_0w[view], MapX_Img_0[view], MapY_Img_0[view]);
					SmoothImage(Images_0w[view]);
					if (SF_I0_Warped_FileName) {
						sprintf(file, "%s%d.%s", SF_I0_Warped_FileName, view, IMG_TYPE);
						SaveImage(file, Images_0w[view]);
					}
					
					assert(Images_0w[view] && Images_0w_X[view] && Images_0w_Y[view]);
					Images_0w_X[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Images_0w_Y[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Derivative(Images_0w[view], 1, 0, derivative_tmp, Images_0w_X[view]);
					Derivative(Images_0w[view], 0, 1, derivative_tmp, Images_0w_Y[view]);
				}
			}
			if (do_Zt || do_not_ref_flow) {
				for(view = 1; view < NumOfViews; view++) {
					cvRemap(Images_t_pyramid[view][level], Images_tw[view], MapX_Img_t[view], MapY_Img_t[view]);
					SmoothImage(Images_tw[view]);
					if (SF_It_Warped_FileName) {
						sprintf(file, "%s%d.%s", SF_It_Warped_FileName, view, IMG_TYPE);
						SaveImage(file, Images_tw[view]);
					}
					assert(Images_tw[view] && Images_tw_X[view] && Images_tw_Y[view]);
					Images_tw_X[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Images_tw_Y[view] = cvCreateMat(N_y, N_x, CV_32FC1);
					Derivative(Images_tw[view], 1, 0, derivative_tmp, Images_tw_X[view]);
					Derivative(Images_tw[view], 0, 1, derivative_tmp, Images_tw_Y[view]);
				}
			}
			//Flow gradients computation
			GradientFast_32f(Z0, Z0_X, Z0_Y);
			Gradient3DPoints(MapIdentityX, MapIdentityY, Z0, Z0_X, Z0_Y, eta_curr, X_x, X_y, X_z, Y_x, Y_y, Y_z);
			
			if(do_ref_flow || do_not_ref_flow || do_Zt){
				GradientFast_32f(U, U_X, U_Y);
				GradientFast_32f(V, V_X, V_Y);
				GradientFast_32f(W, W_X, W_Y);
				cvAdd(U_X,X_x,Xt_x);
				cvAdd(U_Y,X_y,Xt_y);
				cvAdd(V_X,Y_x,Yt_x);
				cvAdd(V_Y,Y_y,Yt_y);
				cvAdd(W_X,Z0_X,Zt_X);
				cvAdd(W_Y,Z0_Y,Zt_Y);
			}
			if(do_Z0 || do_not_ref_flow){
				//computes the projected coordinates derivatives at time t
				for(view = 1; view < NumOfViews; view++) {
					GradientProjCoor(X, Y, Z0, X_x, Y_x, Z0_X, M_curr[view], Proj_x0_x[view], 0);
					GradientProjCoor(X, Y, Z0, X_y, Y_y, Z0_Y, M_curr[view], Proj_x0_y[view], 0);
					GradientProjCoor(X, Y, Z0, X_x, Y_x, Z0_X, M_curr[view], Proj_y0_x[view], 1);
					GradientProjCoor(X, Y, Z0, X_y, Y_y, Z0_Y, M_curr[view], Proj_y0_y[view], 1);
					if(smooth_flag) {
						SmoothImage(Proj_x0_x[view]);
						SmoothImage(Proj_x0_y[view]);
						SmoothImage(Proj_y0_x[view]);
						SmoothImage(Proj_y0_y[view]);
					}
					if(do_Z0) {
						CvMat *temp = cvCreateMat(N_y, N_x, CV_32FC1); 
						cvSet(temp, cvScalarAll(1),0);
						GradientProjCoor(X, Y, Z0, X_z, Y_z, temp, M_curr[view], Proj_x0_z[view], 0);
						GradientProjCoor(X, Y, Z0, X_z, Y_z, temp, M_curr[view], Proj_y0_z[view], 1);
						cvReleaseMat(&temp);
						if(smooth_flag) {
							SmoothImage(Proj_x0_z[view]);
							SmoothImage(Proj_y0_z[view]);
						}
					}
				}
			}
			if(do_not_ref_flow || do_Zt) {
				CvMat *ZerosDerive = cvCreateMat(N_y, N_x, CV_32FC1); 
				CvMat *OnesDerive = cvCreateMat(N_y, N_x, CV_32FC1); 
				cvSet(OnesDerive, cvScalarAll(1),0);
				cvZero(ZerosDerive);
				for(view = 1; view < NumOfViews; view++) {
					GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[view], Proj_xt_u[view], 0);
					GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[view], Proj_xt_v[view], 0);
					GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[view], Proj_xt_w[view], 0);
					
					GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[view], Proj_yt_u[view], 1);
					GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[view], Proj_yt_v[view], 1);
					GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[view], Proj_yt_w[view], 1);

					GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X, M_curr[view], Proj_xt_x[view], 0);
					GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y, M_curr[view], Proj_xt_y[view], 0);
					GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X, M_curr[view], Proj_yt_x[view], 1);
					GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y, M_curr[view], Proj_yt_y[view], 1);
					if(smooth_flag) {
						SmoothImage(Proj_xt_u[view]);
						SmoothImage(Proj_xt_v[view]);
						SmoothImage(Proj_xt_w[view]);
						SmoothImage(Proj_yt_u[view]);
						SmoothImage(Proj_yt_v[view]);
						SmoothImage(Proj_yt_w[view]);
						SmoothImage(Proj_xt_x[view]);
						SmoothImage(Proj_xt_y[view]);
						SmoothImage(Proj_yt_x[view]);
						SmoothImage(Proj_yt_y[view]);
					}
					if(do_Z0) {
						CvMat *temp = cvCreateMat(N_y, N_x, CV_32FC1); 
						cvSet(temp, cvScalarAll(1),0);
						GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp, M_curr[view], Proj_xt_z[view], 0);
						GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp, M_curr[view], Proj_yt_z[view], 1);
						cvReleaseMat(&temp);
						if(smooth_flag) {
							SmoothImage(Proj_xt_z[view]);
							SmoothImage(Proj_yt_z[view]);
						}
					}
				}
				cvReleaseMat(&ZerosDerive);
				cvReleaseMat(&OnesDerive);
			}
			
			if(do_ref_flow || do_Zt) {
				CvMat *ZerosDerive = cvCreateMat(N_y, N_x, CV_32FC1); 
				CvMat *OnesDerive = cvCreateMat(N_y, N_x, CV_32FC1); 
				cvSet(OnesDerive, cvScalarAll(1),0);
				cvZero(ZerosDerive);
				GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[0], Proj_xt_u[0], 0);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[0], Proj_xt_v[0], 0);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[0], Proj_xt_w[0], 0);

				GradientProjCoor(Xt, Yt, Zt, OnesDerive, ZerosDerive, ZerosDerive, M_curr[0], Proj_yt_u[0], 1);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, OnesDerive, ZerosDerive, M_curr[0], Proj_yt_v[0], 1);
				GradientProjCoor(Xt, Yt, Zt, ZerosDerive, ZerosDerive, OnesDerive, M_curr[0], Proj_yt_w[0], 1);
				
			    GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X,  M_curr[0], Proj_xt_x[0], 0);
				GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y,  M_curr[0], Proj_xt_y[0], 0);
				GradientProjCoor(Xt, Yt, Zt, Xt_x, Yt_x, Zt_X,  M_curr[0], Proj_yt_x[0], 1);
				GradientProjCoor(Xt, Yt, Zt, Xt_y, Yt_y, Zt_Y,  M_curr[0], Proj_yt_y[0], 1);
				if(smooth_flag) {
					SmoothImage(Proj_xt_x[0]);
					SmoothImage(Proj_xt_y[0]);
					SmoothImage(Proj_yt_x[0]);
					SmoothImage(Proj_yt_y[0]);
					SmoothImage(Proj_xt_u[0]);
					SmoothImage(Proj_xt_v[0]);
					SmoothImage(Proj_xt_w[0]);
					SmoothImage(Proj_yt_u[0]);
					SmoothImage(Proj_yt_v[0]);
					SmoothImage(Proj_yt_w[0]);
				}
				if(do_Z0) {
					CvMat *temp = cvCreateMat(N_y, N_x, CV_32FC1); 
					cvSet(temp, cvScalarAll(1),0);
					GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp,  M_curr[0], Proj_xt_z[0], 0);
					GradientProjCoor(Xt, Yt, Zt, X_z, Y_z, temp,  M_curr[0], Proj_yt_z[0], 1);
					cvReleaseMat(&temp);
					if(smooth_flag) {
						SmoothImage(Proj_xt_z[0]);
						SmoothImage(Proj_yt_z[0]);
					}
				}
				cvReleaseMat(&ZerosDerive);
				cvReleaseMat(&OnesDerive);
			}
			//Increments initialization
			if(do_ref_flow || do_not_ref_flow || do_Zt){
				cvSetZero(dU);
				cvSetZero(dV);
				cvSetZero(dW);
			}
			if(do_Z0){
				cvSetZero(dZ0);
			}

		  	norm_inner = 0.;
			first_iteration = 1;
			iter_inner = 0;
			//secondary fixed point iterations
			while((norm_inner > epsilon_inner && iter_inner <= max_iter_inner) || first_iteration == 1){
				first_iteration = 0;
	                
				//Coordinates maps update
				cvCopy(X, X_inner);
				cvCopy(Y, Y_inner);
				cvCopy(Z0,Z_inner);
				// need to update Z0 and re-compute the world coordinates at time t
				if (do_Z0 || do_not_ref_flow){
					if(do_Z0)
						cvAdd(Z_inner, dZ0, Z_inner);
					ComputeXY(X_inner, Y_inner, Z_inner, MapIdentityX, MapIdentityY, N_x, N_y, ref_params, eta_curr);
					for(view = 1; view < NumOfViews; view++){
						ComputeProjMaps(X_inner, Y_inner, Z_inner, R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_0_inner[view], MapY_Img_0_inner[view], eta_curr);
					}
				}
				// need to update the world coordinates at time t+1
				if(do_ref_flow || do_not_ref_flow || do_Zt) {
					// Xt_inner = X_inner + U + dU
					cvAdd(X_inner, U, Xt_inner);
					cvAdd(Xt_inner, dU, Xt_inner);
					// Yt_inner = Y_inner + V + dV
					cvAdd(Y_inner, V, Yt_inner);
					cvAdd(Yt_inner, dV, Yt_inner);
					// Zt_inner = (Z0+dZ0)+W+dW
					cvAdd(Z_inner, W, Zt_inner);
					cvAdd(Zt_inner, dW, Zt_inner);
				}
				if (do_ref_flow || do_Zt ) {
					ComputeProjMaps(Xt_inner,Yt_inner,Zt_inner,R_Mat[0], T_Mat[0], C_Mat[0], MapX_Img_t_inner[0], MapY_Img_t_inner[0], eta_curr);
				}
				if (do_not_ref_flow || do_Zt) {
					for(view = 1; view < NumOfViews; view++) {
						ComputeProjMaps(Xt_inner,Yt_inner,Zt_inner,R_Mat[view], T_Mat[view], C_Mat[view], MapX_Img_t_inner[view], MapY_Img_t_inner[view], eta_curr);
					}
				}
								
				//Flow increments gradients computation
				if(do_ref_flow || do_not_ref_flow || do_Zt){
					GradientFast_32f(dU, dU_X, dU_Y);
					GradientFast_32f(dV, dV_X, dV_Y);
					GradientFast_32f(dW, dW_X, dW_Y);
				}
				if(do_Z0){
					GradientFast_32f(dZ0, dZ0_X, dZ0_Y);
				}
				for(yl0 = 0; yl0 < N_y; yl0++){
					for(xl0 = 0; xl0 < N_x; xl0++){
				    double z0 = 0.;
					double z0_X = 0.;
					double z0_Y = 0.;
					double dz0_X = 0.;
					double dz0_Y = 0.;
					double s2 = 0.;
					double s3 = 0.;
					double s4 = 0.;
					double sZ = 0.;
					
					if (do_ref_flow || do_not_ref_flow || do_Zt) {
						double du_X = cvmGet(dU_X, yl0, xl0);
						double du_Y = cvmGet(dU_Y, yl0, xl0);
						double u_X = cvmGet(U_X, yl0, xl0);
						double u_Y = cvmGet(U_Y, yl0, xl0);
						double dv_X = cvmGet(dV_X, yl0, xl0);
						double dv_Y = cvmGet(dV_Y, yl0, xl0);
						double v_X = cvmGet(V_X, yl0, xl0);
						double v_Y = cvmGet(V_Y, yl0, xl0);
						double dw_X = cvmGet(dW_X, yl0, xl0);
						double dw_Y = cvmGet(dW_Y, yl0, xl0);
						double w_X = cvmGet(W_X, yl0, xl0);
						double w_Y = cvmGet(W_Y, yl0, xl0);
						if(DIV_UVW){
							s2 += mu_u*(square(u_X + du_X) + square(u_Y + du_Y)) +
								  mu_v*(square(v_X + dv_X) + square(v_Y + dv_Y)) + 
								  mu_w*(square(w_X + dw_X) + square(w_Y + dw_Y));
						}
						else{
							s2 += (square(u_X + du_X) + square(u_Y + du_Y) +
								square(v_X + dv_X) + square(v_Y + dv_Y)) ;
							
							s3 +=  square(w_X + dw_X) + square(w_Y + dw_Y) ;
						}
						if(DIV_ZW & do_Z0)
						{
							z0_X = cvmGet(Z0_X, yl0, xl0);
							z0_Y = cvmGet(Z0_Y, yl0, xl0);
							dz0_X = cvmGet(dZ0_X, yl0, xl0);
							dz0_Y = cvmGet(dZ0_Y, yl0, xl0);
							s4 +=  square(w_X + dw_X) + mu2*square(z0_X + dz0_X) + square(w_Y + dw_Y) + mu2*square(z0_Y + dz0_Y);
						}
					}
				           
					if (do_Z0) {
						z0_X = cvmGet(Z0_X, yl0, xl0);
						z0_Y = cvmGet(Z0_Y, yl0, xl0);
						dz0_X = cvmGet(dZ0_X, yl0, xl0);
						dz0_Y = cvmGet(dZ0_Y, yl0, xl0);
						sZ += mu_z*(square(z0_X + dz0_X) + square(z0_Y + dz0_Y));
					}
					double a = DPsiSmooth(s2);
					cvmSet(divergence_coefficients, yl0, xl0, a);
					a = DPsiSmooth(sZ);
					cvmSet(divergence_coefficients_Z, yl0, xl0, a);
					if(!DIV_UVW){
						a = DPsiSmoothW(s3);
						cvmSet(divergence_coefficients_W, yl0, xl0, a);
					}
					if(DIV_ZW){
						a = DPsiSmooth(s4);
						cvmSet(divergence_coefficients_WZ, yl0, xl0, a);
					}

					}
				}

				//Computing M and B for the final linear system 
				cvSetZero(M);
				cvSetZero(B);
				
				int i;
				
				//The dimension*N_x first lines in M correspond to the points on the y = 0 on the image
				//tali: The matrix indexes are a row-stack indexes
				int inf = 0;
				int sup = dimension*N_x;
				for (i = inf; i < sup; i++) {
					cvmSet(M, i, 0, -1./h); 
					cvmSet(M, i, 1,  1./h);	
				}
				//Points on the lower horizontal side of the image      
				inf = dimension*N_x*(N_y-1);
				sup = dimension*N_x*N_y;
			
				for (i = inf; i < sup; i++) {
					cvmSet(M, i, 0,  1./h); 
					cvmSet(M, i, 1, -1./h);

				}

				inf = dimension*N_x;
				sup = dimension*N_x*(N_y-1)-1;
			
				//We fill in the lines of the matrix M corresponding to images pixels except points on the upper and lower horizontal side
				for(i = inf; i <= sup; i+=dimension){
					yl0 = i/(dimension*N_x);
					xl0 = (i % (inf))/dimension;
		           
					if(xl0 == 0){ //if point on the left side of image
						for(int d=0; d<dimension; d++) {
							cvmSet(M,i+d,0,-1./h);
							cvmSet(M,i+d,1, 1./h);
							
						}
					}
					else if(xl0 == (N_x-1)){//if point on the right side
						for(int d=0; d<dimension; d++) {
							cvmSet(M,i+d,0, 1./h);
							cvmSet(M,i+d,1,-1./h);  
						}
					}
					else{
						//Pixel inside the integration domain 
						if(methode == 0){
							double il0 = 0.;
							double il0_X = 0., il0_Y = 0., il0w_Z = 0.;
							double i0 = 0.;
							double i0_X = 0., i0_Y = 0.;
							double i0w_X = 0., i0w_Y = 0., i0w_Z = 0.; 
							double proj_x0_x = 0., proj_x0_y = 0;
							double proj_y0_x = 0., proj_y0_y = 0;
							double proj_x0_z = 0., proj_y0_z = 0;
							double ilt = 0.;
							double ilt_X = 0., ilt_Y = 0.;
							double iltw_X = 0., iltw_Y = 0., iltw_Z = 0.;
							double iltw_U = 0., iltw_V = 0., iltw_W = 0.;
							double proj_xlt_x = 0., proj_xlt_y = 0., proj_xlt_z = 0.;
							double proj_xlt_u = 0., proj_xlt_v = 0., proj_xlt_w = 0.;
							double proj_ylt_x = 0., proj_ylt_y = 0., proj_ylt_z = 0.;
							double proj_ylt_u = 0., proj_ylt_v = 0., proj_ylt_w = 0.;
							double it = 0.;
							double it_X = 0., it_Y = 0.;
							double itw_X = 0., itw_Y = 0., itw_Z = 0.;
							double itw_U = 0., itw_V = 0., itw_W = 0.;
							double proj_xt_x = 0., proj_xt_y = 0., proj_xt_z = 0.;
							double proj_xt_u = 0., proj_xt_v = 0., proj_xt_w = 0.;
							double proj_yt_x = 0., proj_yt_y = 0., proj_yt_z = 0.;
							double proj_yt_u = 0., proj_yt_v = 0., proj_yt_w = 0.;
							double du = 0, dv = 0, dw = 0, dz = 0;

							//interpolated divergence coefficients
							double b_f_1, b_f_2, b_f_3, b_f_4;
							double b_fz_1, b_fz_2, b_fz_3, b_fz_4;
							double b_fw_1, b_fw_2, b_fw_3, b_fw_4;
							double b_fwz_1, b_fwz_2, b_fwz_3, b_fwz_4;
							if(DIV_ZW){
								divergence_coefficients_interpolate(divergence_coefficients_WZ, yl0, xl0, &b_fwz_1, &b_fwz_2, &b_fwz_3, &b_fwz_4);
							}

							// Half pixel interpolation for the divergence coefficients 
							divergence_coefficients_interpolate(divergence_coefficients, yl0, xl0, &b_f_1, &b_f_2, &b_f_3, &b_f_4);
							divergence_coefficients_interpolate(divergence_coefficients_Z, yl0, xl0, &b_fz_1, &b_fz_2, &b_fz_3, &b_fz_4);
							if(!DIV_UVW)
								divergence_coefficients_interpolate(divergence_coefficients_W, yl0, xl0, &b_fw_1, &b_fw_2, &b_fw_3, &b_fw_4);
							else {
								b_fw_1 = b_f_1;
								b_fw_2 = b_f_2;
								b_fw_3 = b_f_3;
								b_fw_4 = b_f_4;
								alpha_w = alpha;
							}

	                        	//Matrix terms
							double a_u_1, a_u_2, a_u_3, a_u_4, a_v_1, a_v_2, a_v_3, a_v_4; 
							double a_w_1, a_w_2, a_w_3, a_w_4,a_z_1, a_z_2, a_z_3, a_z_4;
		    
							a_u_1 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_v_1 = 0;
							a_w_1 = 0;	  
							a_z_1 = 0;
		    
							a_u_2 = 0;
							a_v_2 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_w_2 = 0;	  
							a_z_2 = 0;
		    
							a_u_3 = 0;
							a_v_3 = 0;
							a_w_3 = (alpha_w/(4.*h*h))*(b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4);	  
							a_z_3 = 0; 

							a_u_4 = 0;
							a_v_4 = 0;
							a_w_4 = 0; 
							a_z_4 = (alpha_w*(mu)/(4.*h*h))*(b_fz_1 + b_fz_2 + b_fz_3 + b_fz_4);
							if(DIV_ZW) {
								a_z_4 += (alpha_wz*mu2/(4.*h*h))*(b_fwz_1 + b_fwz_2 + b_fwz_3 + b_fwz_4);	 
								a_w_3 += (alpha_wz/(4.*h*h))*(b_fwz_1 + b_fwz_2 + b_fwz_3 + b_fwz_4);	 

								}
	
							//second member for the first equation
							double bu0 = cvmGet(U, yl0, xl0);
							double bu1 = cvmGet(U, yl0, xl0+1);
							double bu2 = cvmGet(U, yl0, xl0-1);
							double bu3 = cvmGet(U,yl0+1,xl0);
							double bu4 = cvmGet(U,yl0-1,xl0);
							//Crop(bu0, N_x);
							double bv0 = cvmGet(V, yl0, xl0);
							double bv1 = cvmGet(V, yl0, xl0+1);
							double bv2 = cvmGet(V, yl0, xl0-1);
							double bv3 = cvmGet(V,yl0+1,xl0);
							double bv4 = cvmGet(V,yl0-1,xl0);
							//Crop(bw0, N_x);
							double bw0 = cvmGet(W, yl0, xl0);
							double bw1 = cvmGet(W, yl0, xl0+1);
							double bw2 = cvmGet(W, yl0, xl0-1);
							double bw3 = cvmGet(W,yl0+1,xl0);
							double bw4 = cvmGet(W,yl0-1,xl0);

							double bzi0 = cvmGet(Z0, yl0, xl0);
							double bzi1 = cvmGet(Z0, yl0, xl0+1);
							double bzi2 = cvmGet(Z0, yl0, xl0-1);
							double bzi3 = cvmGet(Z0,yl0+1,xl0);
							double bzi4 = cvmGet(Z0,yl0-1,xl0);

							//tali: bu is the discretization of div(Psi()*grad(u))
							double bu = (alpha/(4*h*h)) *
								(b_f_1 * bu1 +
								 b_f_2 * bu2 +
								 b_f_3 * bu3 +
								 b_f_4 * bu4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bu0);

							//Crop(bv0, N_y);
							double bv = (alpha/(4*h*h)) *
								(b_f_1 * bv1 +
								 b_f_2 * bv2 +
								 b_f_3 * bv3 +
								 b_f_4 * bv4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bv0);
							//Crop(bv0, N_y);
							double bw = (alpha_w/(4*h*h)) *
								(b_fw_1 * bw1 +
								 b_fw_2 * bw2 +
								 b_fw_3 * bw3 +
								 b_fw_4 * bw4
								 - (b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4) * bw0);
							double bz = (alpha_w*(mu)/(4*h*h)) *
								(b_fz_1 * bzi1 +
								 b_fz_2 * bzi2 +
								 b_fz_3 * bzi3 +
								 b_fz_4 * bzi4
								 - (b_fz_1 + b_fz_2 + b_fz_3 + b_fz_4)* bzi0);
							if(DIV_ZW) {
									bz += (alpha_wz*mu2/(4*h*h)) *
									(b_fwz_1 * bzi1 +
									b_fwz_2 * bzi2 +
									b_fwz_3 * bzi3 +
									b_fwz_4 * bzi4
									- (b_fwz_1 + b_fwz_2 + b_fwz_3 + b_fwz_4)* bzi0);

									bw += (alpha_wz/(4*h*h)) *
									(b_fwz_1 * bw1 +
									b_fwz_2 * bw2 +
									b_fwz_3 * bw3 +
									b_fwz_4 * bw4
									- (b_fwz_1 + b_fwz_2 + b_fwz_3 + b_fwz_4)* bw0);
								}
						
							//Pixels out must not be taken into account in the energy data term
							int pixel_do_ref_flow = 0;
							int pixel_do_flow_views[MAX_NUM_OF_VIEWS] = {0};
							int pixel_doZ0_views[MAX_NUM_OF_VIEWS] = {0};
							int pixel_doZt_views[MAX_NUM_OF_VIEWS] = {0};
							if (disactivated_pixels_ref_flow)
								pixel_do_ref_flow = !cvmGet8u(disactivated_pixels_ref_flow,yl0,xl0);
							
							for(view=1; view < NumOfViews; view++) {
								if (disactivated_pixels_flow)
									pixel_do_flow_views[view] = !cvmGet8u(disactivated_pixels_flow[view],yl0,xl0);
								if (disactivated_pixels_Z0)
									pixel_doZ0_views[view] = !cvmGet8u(disactivated_pixels_Z0[view],yl0,xl0);
								if (disactivated_pixels_Zt)
									pixel_doZt_views[view] = !cvmGet8u(disactivated_pixels_Zt[view],yl0,xl0);
								//is the point visible at time t+1 ?
								if (pixel_do_ref_flow || pixel_doZt_views[view]) {
									double xlt = cvmGet(MapX_Img_t_inner[0], yl0, xl0);
									double ylt = cvmGet(MapY_Img_t_inner[0], yl0, xl0);
									if(xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h ||
									   ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h){
										   if(pixel_do_ref_flow){
												cvmSet8u(disactivated_pixels_ref_flow,yl0,xl0,127);
												pixel_do_ref_flow = 0;
										   }
										   if(pixel_doZt_views[view]){
												cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,127);
												pixel_doZt_views[view] = 0;
										   }
									}
								}
								//On the image at time t
								if (pixel_doZ0_views[view] || pixel_do_flow_views[view]) {
									double x0 = cvmGet(MapX_Img_0_inner[view], yl0, xl0);
									double y0 = cvmGet(MapY_Img_0_inner[view], yl0, xl0);
									if(x0 >= (N_x - 1 - DISACTIVATE_BORDER)*h || x0 <= DISACTIVATE_BORDER*h ||
									   y0 >= (N_y - 1 - DISACTIVATE_BORDER)*h || y0 <= DISACTIVATE_BORDER*h) {
										if (pixel_doZ0_views[view]) {
											cvmSet8u(disactivated_pixels_Z0[view],yl0,xl0,127);
											pixel_doZ0_views[view] = 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,127);
											pixel_do_flow_views[view] = 0;
										}
									}
								}

								//On the right image at time t+1
								if (pixel_doZt_views[view] || pixel_do_flow_views[view]) {
									double xt = cvmGet(MapX_Img_t_inner[view], yl0, xl0);
									double yt = cvmGet(MapY_Img_t_inner[view], yl0, xl0);
									if(xt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xt <= DISACTIVATE_BORDER*h ||
									   yt >= (N_y - 1 - DISACTIVATE_BORDER)*h || yt < DISACTIVATE_BORDER*h) {
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,127);
											pixel_doZt_views[view] = 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,127);
											pixel_do_flow_views[view] = 0;
										}
									}
								}
							}
							int pixel_doZ0 = 0;
							int pixel_doZt = 0;
							int pixel_do_flow = 0;
							for(view = 1; view < NumOfViews; view++){
								pixel_doZ0 += pixel_doZ0_views[view];
								pixel_doZt += pixel_doZt_views[view];
								pixel_do_flow += pixel_do_flow_views[view];
							}
							if (pixel_do_ref_flow || pixel_do_flow || pixel_doZt) {
								du = cvmGet(dU, yl0, xl0);
								dv = cvmGet(dV, yl0, xl0);
								dw = cvmGet(dW, yl0, xl0);
							}
							if (pixel_doZ0) {
								dz = cvmGet(dZ0, yl0, xl0);
							}
							//Data term estimation
							if (pixel_do_ref_flow || pixel_doZ0) {
								il0 = cvGetReal2D(Images_0_pyramid[0][level], yl0, xl0); 
								il0_X = cvmGet(Images_0_X[0],  yl0, xl0);
								il0_Y = cvmGet(Images_0_Y[0],  yl0, xl0);
							}
							for(view = 1; view < NumOfViews; view++){
								if (pixel_do_flow_views[view] || pixel_doZ0_views[view]) {
									i0 = cvGetReal2D(Images_0w[view], yl0, xl0);
									i0w_X = cvmGet(Images_0w_X[view],  yl0, xl0);
									i0w_Y = cvmGet(Images_0w_Y[view],  yl0, xl0);
									proj_x0_x = cvmGet(Proj_x0_x[view], yl0, xl0);
									proj_x0_y = cvmGet(Proj_x0_y[view], yl0, xl0);
									proj_y0_x = cvmGet(Proj_y0_x[view], yl0, xl0);
									proj_y0_y = cvmGet(Proj_y0_y[view], yl0, xl0);
									if(pixel_doZ0) {
										proj_x0_z = cvmGet(Proj_x0_z[view], yl0, xl0);
										proj_y0_z = cvmGet(Proj_y0_z[view], yl0, xl0);
									}
									double det = proj_x0_x*proj_y0_y - proj_x0_y*proj_y0_x;
									if (det <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_doZ0_views[view]) {
											cvmSet8u(disactivated_pixels_Z0[view],yl0,xl0,191);
											pixel_doZ0_views[view]= 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,191);
											pixel_do_flow_views[view] = 0;
										}
									}
									else {
										i0_X = (i0w_X*proj_y0_y - i0w_Y*proj_y0_x)/det;
										i0_Y = (i0w_Y*proj_x0_x - i0w_X*proj_x0_y)/det;
										if(do_Z0) {
											i0w_Z = i0_X*proj_x0_z + i0_Y*proj_y0_z;
											if(deriv_norm) 
												if(proj_x0_z || proj_y0_z)
													i0w_Z = i0w_Z/(proj_x0_z*proj_x0_z + proj_y0_z*proj_y0_z);
										}
									}
								}
								if (pixel_do_ref_flow || pixel_doZt_views[view]) {
									// Wraped image values
									ilt = cvGetReal2D(Images_tw[0], yl0, xl0);
									iltw_X = cvmGet(Images_tw_X[0], yl0, xl0);
									iltw_Y = cvmGet(Images_tw_Y[0],  yl0, xl0);
									// projected coordinated derivatives
									proj_xlt_x = cvmGet(Proj_xt_x[0], yl0, xl0);
									proj_xlt_y = cvmGet(Proj_xt_y[0], yl0, xl0);
									proj_ylt_x = cvmGet(Proj_yt_x[0], yl0, xl0);
									proj_ylt_y = cvmGet(Proj_yt_y[0], yl0, xl0);
									proj_xlt_z = cvmGet(Proj_xt_z[0], yl0, xl0);
									proj_ylt_z = cvmGet(Proj_yt_z[0], yl0, xl0);
									proj_xlt_u = cvmGet(Proj_xt_u[0], yl0, xl0);
									proj_ylt_u = cvmGet(Proj_yt_u[0], yl0, xl0);
									proj_xlt_v = cvmGet(Proj_xt_v[0], yl0, xl0);
									proj_ylt_v = cvmGet(Proj_yt_v[0], yl0, xl0);
									proj_xlt_w = cvmGet(Proj_xt_w[0], yl0, xl0);
									proj_ylt_w = cvmGet(Proj_yt_w[0], yl0, xl0);
								
									double det = proj_xlt_x*proj_ylt_y - proj_xlt_y*proj_ylt_x;
									if ((det) <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_do_ref_flow) {
											cvmSet8u(disactivated_pixels_ref_flow,yl0,xl0,191);
											pixel_do_ref_flow = 0;
										}
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,191);
											pixel_doZt_views[view] = 0;
										}
									}
									else {
										ilt_X = (iltw_X*proj_ylt_y - iltw_Y*proj_ylt_x)/det ;
										ilt_Y = (iltw_Y*proj_xlt_x - iltw_X*proj_xlt_y)/det ;
										iltw_Z = ilt_X*proj_xlt_z + ilt_Y*proj_ylt_z ;
										iltw_U = ilt_X*proj_xlt_u + ilt_Y*proj_ylt_u ; 
										iltw_V = ilt_X*proj_xlt_v + ilt_Y*proj_ylt_v ; 
										iltw_W = ilt_X*proj_xlt_w + ilt_Y*proj_ylt_w ;
										if(deriv_norm) {
											if(proj_xlt_z || proj_ylt_z)
													iltw_Z = iltw_Z/(proj_xlt_z*proj_xlt_z + proj_ylt_z*proj_ylt_z);
											if(proj_xlt_u || proj_ylt_u)
													iltw_U = iltw_U/(proj_xlt_u*proj_xlt_u + proj_ylt_u*proj_ylt_u);
											if(proj_xlt_v || proj_ylt_v)
													iltw_V = iltw_V/(proj_xlt_v*proj_xlt_v + proj_ylt_v*proj_ylt_v);
											if(proj_xlt_w || proj_ylt_w)
													iltw_W = iltw_W/(proj_xlt_w*proj_xlt_w  + proj_ylt_w*proj_ylt_w);
										}
									}
								}
								if (pixel_doZt_views[view] || pixel_doZt_views[view]) {
									it = cvGetReal2D(Images_tw[view],yl0,xl0);
									itw_X = cvmGet(Images_tw_X[view],  yl0, xl0);
									itw_Y = cvmGet(Images_tw_Y[view],  yl0, xl0);
									
									proj_xt_x = cvmGet(Proj_xt_x[view], yl0, xl0);
									proj_xt_y = cvmGet(Proj_xt_y[view], yl0, xl0);
									proj_yt_x = cvmGet(Proj_yt_x[view], yl0, xl0);
									proj_yt_y = cvmGet(Proj_yt_y[view], yl0, xl0);
									
									proj_xt_u = cvmGet(Proj_xt_u[view], yl0, xl0);
									proj_xt_v = cvmGet(Proj_xt_v[view], yl0, xl0);
									proj_xt_w = cvmGet(Proj_xt_w[view], yl0, xl0);
									proj_yt_u = cvmGet(Proj_yt_u[view], yl0, xl0);
									proj_yt_v = cvmGet(Proj_yt_v[view], yl0, xl0);
									proj_yt_w = cvmGet(Proj_yt_w[view], yl0, xl0);
									
									if(pixel_doZ0_views[view]) {
										proj_xt_z = cvmGet(Proj_xt_z[view], yl0, xl0);
										proj_yt_z = cvmGet(Proj_yt_z[view], yl0, xl0);
									}
									double det = proj_xt_x*proj_yt_y - proj_xt_y*proj_yt_x;
									if ((det) <= WARP_DETERMINANT_THRESHOLD) {
										// image is reversed, set everything to 0
										if (pixel_doZt_views[view]) {
											cvmSet8u(disactivated_pixels_Zt[view],yl0,xl0,191);
											pixel_doZt_views[view] = 0;
										}
										if (pixel_do_flow_views[view]) {
											cvmSet8u(disactivated_pixels_flow[view],yl0,xl0,191);
											pixel_do_flow_views[view] = 0;
										}
									}
									else {
										it_X = (itw_X*proj_yt_y - itw_Y*proj_yt_x)/det;
										it_Y = (itw_Y*proj_xt_x - itw_X*proj_xt_y)/det;
										if(pixel_doZ0_views[view]) {
											itw_Z = it_X*proj_xt_z + it_Y*proj_yt_z;
											if(deriv_norm)
												if(proj_xt_z || proj_yt_z)
													itw_Z = itw_Z/(proj_xt_z*proj_xt_z + proj_yt_z*proj_yt_z);
										}
										itw_U = it_X*proj_xt_u + it_Y*proj_yt_u ; 
										itw_V = it_X*proj_xt_v + it_Y*proj_yt_v ; 
										itw_W = it_X*proj_xt_w + it_Y*proj_yt_w ;
										if(deriv_norm) {
											if(proj_xt_u || proj_yt_u)
													itw_U = itw_U/(proj_xt_u*proj_xt_u + proj_yt_u*proj_yt_u);
											if(proj_xt_v || proj_yt_v)
													itw_V = itw_V/(proj_xt_v*proj_xt_v + proj_yt_v*proj_yt_v);
											if(proj_xt_w || proj_yt_w)
													itw_W = itw_W/(proj_xt_w*proj_xt_w  + proj_yt_w*proj_yt_w);
										}
									}
								}
								if (pixel_do_flow_views[view]) {
									double a = square(it + itw_U*du + itw_V*dv + itw_W*dw + itw_Z*dz - 
													  (i0 + i0w_Z*dz));
									double psi_flow = DPsiData(a);
									//u equation
									a_u_1 += psi_flow*(itw_U*itw_U);
									a_v_1 += psi_flow*(itw_V*itw_U);
									a_w_1 += psi_flow*(itw_W*itw_U);
									a_z_1 += psi_flow*((itw_Z - i0w_Z)*itw_U);
									//v equation
									a_u_2 += psi_flow*(itw_U*itw_V);
									a_v_2 += psi_flow*(itw_V*itw_V);
									a_w_2 += psi_flow*(itw_W*itw_V);
									a_z_2 += psi_flow*((itw_Z - i0w_Z)*itw_V);
									//w equation
									a_u_3 += psi_flow*(itw_U*itw_W);
									a_v_3 += psi_flow*(itw_V*itw_W);
									a_w_3 += psi_flow*(itw_W*itw_W);
									a_z_3 += psi_flow*((itw_Z - i0w_Z)*itw_W);
									//z equation
									a_u_4 += psi_flow*(itw_U*(itw_Z - i0w_Z));
									a_v_4 += psi_flow*(itw_V*(itw_Z - i0w_Z));
									a_w_4 += psi_flow*(itw_W*(itw_Z - i0w_Z));
									a_z_4 += psi_flow*((itw_Z - i0w_Z)*(itw_Z - i0w_Z));
									// second term
									bu += - psi_flow*(itw_U*(it - i0));		   
									bv += - psi_flow*(itw_V*(it - i0));
									bw += - psi_flow*(itw_W*(it - i0));
									bz += - psi_flow*((itw_Z - i0w_Z)*(i0 - it));
								}
								if (pixel_doZ0_views[view]) {
									double a = square(i0 + i0w_Z*dz - il0);
									double psi_stereo_init = DPsiData(a);
			 
									//z equation
									a_z_4 += psi_stereo_init*(i0w_Z*i0w_Z); 

									// second term
									bz += -psi_stereo_init*((i0 - il0)*i0w_Z);
								}
								if (pixel_doZt_views[view]) {
									double a = square(it + itw_U*du + itw_V*dv + itw_W*dw +itw_Z*dz - 
													 (ilt + iltw_U*du + iltw_V*dv + iltw_W*dw +iltw_Z*dz));
									double psi_stereo = DPsiData(a);
									//u equation
									a_u_1  += psi_stereo*((itw_U - iltw_U)*(itw_U - iltw_U));
									a_v_1  += psi_stereo*((itw_V - iltw_V)*(itw_U - iltw_U));
									a_w_1  += psi_stereo*((itw_W - iltw_W)*(itw_U - iltw_U));
									a_z_1  += psi_stereo*((itw_Z - iltw_Z)*(itw_U - iltw_U));
									//v equation
									a_u_2  += psi_stereo*((itw_U - iltw_U)*(itw_V - iltw_V));
									a_v_2  += psi_stereo*((itw_V - iltw_V)*(itw_V - iltw_V));
									a_w_2  += psi_stereo*((itw_W - iltw_W)*(itw_V - iltw_V));
									a_z_2  += psi_stereo*((itw_Z - iltw_Z)*(itw_V - iltw_V));
									//w equation
									a_u_3  += psi_stereo*((itw_U - iltw_U)*(itw_W - iltw_W));
									a_v_3  += psi_stereo*((itw_V - iltw_V)*(itw_W - iltw_W));
									a_w_3  += psi_stereo*((itw_W - iltw_W)*(itw_W - iltw_W));
									a_z_3  += psi_stereo*((itw_Z - iltw_Z)*(itw_W - iltw_W));
									//z equation
									a_u_4  += psi_stereo*((itw_U - iltw_U)*(itw_Z - iltw_Z));
									a_v_4  += psi_stereo*((itw_V - iltw_V)*(itw_Z - iltw_Z));
									a_w_4  += psi_stereo*((itw_W - iltw_W)*(itw_Z - iltw_Z));
									a_z_4  += psi_stereo*((itw_Z - iltw_Z)*(itw_Z - iltw_Z));

									// second term
									bu += - psi_stereo*((it - ilt)*(itw_U - iltw_U));
									bv += - psi_stereo*((it - ilt)*(itw_V - iltw_V));
									bw += - psi_stereo*((it - ilt)*(itw_W - iltw_W));
									bz += - psi_stereo*((it - ilt)*(itw_Z - iltw_Z));
								}
							} // for(view = 1; view < NumOfViews; view++){
							
							
							if (pixel_do_ref_flow) {
								//psi' coefficients computation
								//tali: the data term of the left camera OF
								double a = square(ilt + iltw_U*du + iltw_V*dv + iltw_W*dw + iltw_Z*dz - il0);
								//divergence coefficients  
								double psi_flowL = DPsiData(a);

								// tali: the _1 refers to the the u equation, the _2 to the v equation the _3 to w and _4 z 
								//u equation								
								a_u_1 += psi_flowL*(iltw_U*iltw_U);
								a_v_1 += psi_flowL*(iltw_V*iltw_U);
								a_w_1 += psi_flowL*(iltw_W*iltw_U);
								a_z_1 += psi_flowL*(iltw_Z*iltw_U);
								//v equation
								a_u_2 += psi_flowL*(iltw_U*iltw_V);
								a_v_2 += psi_flowL*(iltw_V*iltw_V);
								a_w_2 += psi_flowL*(iltw_W*iltw_V);
								a_z_2 += psi_flowL*(iltw_Z*iltw_V);
								//w equation
								a_u_3 += psi_flowL*(iltw_U*iltw_W);
								a_v_3 += psi_flowL*(iltw_V*iltw_W);
								a_w_3 += psi_flowL*(iltw_W*iltw_W);
								a_z_3 += psi_flowL*(iltw_Z*iltw_W);
								// second term
								bu += - psi_flowL*(iltw_U*(ilt - il0));		   
								bv += - psi_flowL*(iltw_V*(ilt - il0));
								bw += - psi_flowL*(iltw_W*(ilt - il0));
								bz += - psi_flowL*(iltw_Z*(ilt - il0));
							}
							// 3 lines corresponding to the 3 equations at (yl0,xl0) 
							cvmSet(M,i  ,0,a_u_1);
							cvmSet(M,i+1,0,a_u_2);
							cvmSet(M,i+2,0,a_u_3);
							cvmSet(M,i+3,0,a_u_4);
			   
							cvmSet(M,i  ,1,a_v_1);
							cvmSet(M,i+1,1,a_v_2);
							cvmSet(M,i+2,1,a_v_3);
							cvmSet(M,i+3,1,a_v_4);
			
							cvmSet(M,i  ,2,a_w_1);
							cvmSet(M,i+1,2,a_w_2);
							cvmSet(M,i+2,2,a_w_3);
							cvmSet(M,i+3,2,a_w_4);
			   
							cvmSet(M,i  ,3,a_z_1);
							cvmSet(M,i+1,3,a_z_2);
							cvmSet(M,i+2,3,a_z_3);
							cvmSet(M,i+3,3,a_z_4);

							cvmSet(M,i  ,4,(-alpha/(4*h*h))*b_f_4);
							cvmSet(M,i+1,4,(-alpha/(4*h*h))*b_f_4);
							
							if(DIV_ZW){
								cvmSet(M,i+2,4,(-alpha_w/(4*h*h))*b_fw_4 + (-alpha_wz/(4*h*h))*b_fwz_4);
								cvmSet(M,i+3,4,(-alpha_w*(mu)/(4*h*h))*b_fz_4+ (-alpha_wz*mu2/(4*h*h))*b_fwz_4);
							}
							else {
								cvmSet(M,i+2,4,(-alpha_w/(4*h*h))*b_fw_4);
								cvmSet(M,i+3,4,(-alpha_w*(mu)/(4*h*h))*b_fz_4);
							}
		 
							cvmSet(M,i  ,5,(-alpha/(4*h*h))*b_f_3);
							cvmSet(M,i+1,5,(-alpha/(4*h*h))*b_f_3);
							
							if(DIV_ZW){
								cvmSet(M,i+2,5,(-alpha_w/(4*h*h))*b_fw_3 + (-alpha_wz/(4*h*h))*b_fwz_3);
								cvmSet(M,i+3,5,(-alpha_w*(mu)/(4*h*h))*b_fz_3+(-alpha_wz*mu2/(4*h*h))*b_fwz_3);
							}
							else {
								cvmSet(M,i+2,5,(-alpha_w/(4*h*h))*b_fw_3);
								cvmSet(M,i+3,5,(-alpha_w*(mu)/(4*h*h))*b_fz_3);
							}
		 
							cvmSet(M,i  ,6,(-alpha/(4*h*h))*b_f_2);
							cvmSet(M,i+1,6,(-alpha/(4*h*h))*b_f_2);
							
							if(DIV_ZW){
								cvmSet(M,i+2,6,(-alpha_w/(4*h*h))*b_fw_2 + (-alpha_wz/(4*h*h))*b_fwz_2);
								cvmSet(M,i+3,6,(-alpha_w*(mu)/(4*h*h))*b_fz_2+(-alpha_wz*mu2/(4*h*h))*b_fwz_2);
							}
							else{
								cvmSet(M,i+2,6,(-alpha_w/(4*h*h))*b_fw_2);
								cvmSet(M,i+3,6,(-alpha_w*(mu)/(4*h*h))*b_fz_2);
							}
		 
							cvmSet(M,i  ,7,(-alpha/(4*h*h))*b_f_1);
							cvmSet(M,i+1,7,(-alpha/(4*h*h))*b_f_1);
							
							if(DIV_ZW){
								cvmSet(M,i+2,7,(-alpha_w/(4*h*h))*b_fw_1+(-alpha_wz/(4*h*h))*b_fwz_1);
								cvmSet(M,i+3,7,(-alpha_w*(mu)/(4*h*h))*b_fz_1+(-alpha_wz*mu2/(4*h*h))*b_fwz_1);
							}
							else{
								cvmSet(M,i+2,7,(-alpha_w/(4*h*h))*b_fw_1);
								cvmSet(M,i+3,7,(-alpha_w*(mu)/(4*h*h))*b_fz_1);
							}
				
							cvmSet(B,i,0,bu);
							cvmSet(B,i+1,0,bv);
							cvmSet(B,i+2,0,bw);
							cvmSet(B,i+3,0,bz);
				}
				/*else if(methode == 1) {
							double il0 = 0.;
							double il0_X = 0., il0_Y = 0., il0w_Z = 0.;
							double ir0 = 0.;
							double ir0_X = 0., ir0_Y = 0.;
							double ir0w_X = 0., ir0w_Y = 0., ir0w_Z = 0.; 
							double ir0w_XX = 0., ir0w_XY = 0., ir0w_YY = 0.;
							double ir0w_X_X = 0., ir0w_X_Y = 0.;
							double ir0w_Y_X = 0., ir0w_Y_Y = 0.;
							double proj_xr0_x = 0., proj_xr0_y = 0;
							double proj_yr0_x = 0., proj_yr0_y = 0;
							double ilt = 0.;
							double ilt_X = 0., ilt_Y = 0.;
							double iltw_X = 0., iltw_Y = 0., iltw_Z = 0.;
							double iltw_U = 0., iltw_V = 0., iltw_W = 0.;
							double iltw_XX = 0., iltw_XY = 0., iltw_YY = 0.;
							double iltw_X_X = 0., iltw_X_Y = 0.;
							double iltw_Y_X = 0., iltw_Y_Y = 0.;
							double proj_xlt_x = 0., proj_xlt_y = 0.;
							double proj_xlt_u = 0., proj_xlt_v = 0., proj_xlt_w = 0.;
							double proj_ylt_x = 0., proj_ylt_y = 0.;
							double proj_ylt_u = 0., proj_ylt_v = 0., proj_ylt_w = 0.;
							double irt = 0.;
							double irt_X = 0., irt_Y = 0.;
							double irtw_X = 0., irtw_Y = 0.;
							double irtw_U = 0., irtw_V = 0., irtw_W = 0.;
							double irtw_XX = 0., irtw_XY = 0., irtw_YY = 0.;
							double irtw_X_X = 0., irtw_X_Y = 0.;
							double irtw_Y_X = 0., irtw_Y_Y = 0.;
							double proj_xrt_x = 0., proj_xrt_y = 0., proj_xrt_z = 0.;
							double proj_xrt_u = 0., proj_xrt_v = 0., proj_xrt_w = 0.;
							double proj_yrt_x = 0., proj_yrt_y = 0., proj_yrt_z = 0.;
							double proj_yrt_u = 0., proj_yrt_v = 0., proj_yrt_w = 0.;
							double du = 0, dv = 0, dw = 0;

							//interpolated divergence coefficients
							double b_f_1, b_f_2, b_f_3, b_f_4;
							double b_fw_1, b_fw_2, b_fw_3, b_fw_4;
							
							// Half pixel interpolation for the divergence coefficients 
							divergence_coefficients_interpolate(divergence_coefficients, yl0, xl0, &b_f_1, &b_f_2, &b_f_3, &b_f_4);
							divergence_coefficients_interpolate(divergence_coefficients_W, yl0, xl0, &b_fw_1, &b_fw_2, &b_fw_3, &b_fw_4);
							//Pixels out must not be taken into account in the energy data term
							int pixel_doleftflow = 0;
							int pixel_dorightflow = 0;
							int pixel_doZt = 0;
							
							if (disactivated_pixels_leftflow)
								pixel_doleftflow = !cvmGet8u(disactivated_pixels_leftflow,yl0,xl0);
							if (disactivated_pixels_rightflow)
								pixel_dorightflow = !cvmGet8u(disactivated_pixels_rightflow,yl0,xl0);
							if (disactivated_pixels_Zt)
								pixel_doZt = !cvmGet8u(disactivated_pixels_Zt,yl0,xl0);

							//is the point visible at time t+1 ?
							if (pixel_doleftflow || pixel_doZt) {
								double xlt = cvmGet(MapX_Ilt_inner, yl0, xl0);
								double ylt = cvmGet(MapY_Ilt_inner, yl0, xl0);
								if(xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h ||
								   ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h){
									cvmSet8u(disactivated_pixels_leftflow,yl0,xl0,127);
									pixel_doleftflow = 0;
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,127);
										pixel_doZt = 0;
									}
								}
							}
							if (pixel_doZt || pixel_dorightflow) {
								double xrt = cvmGet(MapX_Irt_inner, yl0, xl0);
								double yrt = cvmGet(MapY_Irt_inner, yl0, xl0);
								if(xrt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xrt < DISACTIVATE_BORDER*h ||
								   yrt >= (N_y - 1 - DISACTIVATE_BORDER)*h || yrt < DISACTIVATE_BORDER*h) {
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,127);
										pixel_doZt = 0;
									}
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,127);
										pixel_dorightflow = 0;
									}
								}
							}
							
							//On the right image at time t
							if (pixel_dorightflow) {
								double xr0 = cvmGet(MapX_Ir0_inner, yl0, xl0);
								double yr0 = cvmGet(MapY_Ir0_inner, yl0, xl0);
								if(xr0 >= (N_x - 1 - DISACTIVATE_BORDER)*h || xr0 <= DISACTIVATE_BORDER*h ||
								   yr0 >= (N_y - 1 - DISACTIVATE_BORDER)*h || yr0 <= DISACTIVATE_BORDER*h) {
								
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,127);
										pixel_dorightflow = 0;
									}
								}
							}
						
							double a_u_1, a_u_2, a_u_3, a_v_1, a_v_2, a_v_3;
							double a_w_1, a_w_2, a_w_3;
							a_u_1 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_v_1 = 0;
							a_w_1 = 0;	  
							
							a_u_2 = 0;
							a_v_2 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_w_2 = 0;	  
							
							a_u_3 = 0;
							a_v_3 = 0;
							a_w_3 = (alpha/(4.*h*h))*(b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4);	  
							//second member for the first equation
							double bu0 = cvmGet(U, yl0, xl0);
							double bu1 = cvmGet(U, yl0, xl0+1);
							double bu2 = cvmGet(U, yl0, xl0-1);
							double bu3 = cvmGet(U,yl0+1,xl0);
							double bu4 = cvmGet(U,yl0-1,xl0);
							//Crop(bu0, N_x);
							double bv0 = cvmGet(V, yl0, xl0);
							double bv1 = cvmGet(V, yl0, xl0+1);
							double bv2 = cvmGet(V, yl0, xl0-1);
							double bv3 = cvmGet(V,yl0+1,xl0);
							double bv4 = cvmGet(V,yl0-1,xl0);
							//Crop(bw0, N_x);
							double bw0 = cvmGet(W, yl0, xl0);
							double bw1 = cvmGet(W, yl0, xl0+1);
							double bw2 = cvmGet(W, yl0, xl0-1);
							double bw3 = cvmGet(W,yl0+1,xl0);
							double bw4 = cvmGet(W,yl0-1,xl0);
							
							double bu = (alpha/(4*h*h)) *
								(b_f_1 * bu1 +
								 b_f_2 * bu2 +
								 b_f_3 * bu3 +
								 b_f_4 * bu4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bu0);

							//Crop(bv0, N_y);
							double bv = (alpha/(4*h*h)) *
								(b_f_1 * bv1 +
								 b_f_2 * bv2 +
								 b_f_3 * bv3 +
								 b_f_4 * bv4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bv0);
							//Crop(bv0, N_y);
							double bw = (alpha/(4*h*h)) *
								(b_fw_1 * bw1 +
								 b_fw_2 * bw2 +
								 b_fw_3 * bw3 +
								 b_fw_4 * bw4
								 - (b_fw_1 + b_fw_2 + b_fw_3 + b_fw_4) * bw0);
							
							if (pixel_doleftflow ) {
								il0 = cvGetReal2D(Il0_pyramid[level], yl0, xl0); 
								il0_X = cvmGet(Il0_X,  yl0, xl0);
								il0_Y = cvmGet(Il0_Y,  yl0, xl0);
							}
							if (pixel_dorightflow || do_Zt) {
								ir0 = cvGetReal2D(Ir0w, yl0, xl0);
								ir0w_X = cvmGet(Ir0w_X,  yl0, xl0);
								ir0w_Y = cvmGet(Ir0w_Y,  yl0, xl0);
								ir0w_XX = cvmGet(Ir0w_XX,  yl0, xl0);
								ir0w_XY = cvmGet(Ir0w_XY,  yl0, xl0);
								ir0w_YY = cvmGet(Ir0w_YY,  yl0, xl0);
								
								irt = cvGetReal2D(Irtw, yl0, xl0);
								irtw_X = cvmGet(Irtw_X,  yl0, xl0);
								irtw_Y = cvmGet(Irtw_Y,  yl0, xl0);

								proj_xr0_x = cvmGet(Proj_xr0_x, yl0, xl0);
								proj_xr0_y = cvmGet(Proj_xr0_y, yl0, xl0);
								proj_yr0_x = cvmGet(Proj_yr0_x, yl0, xl0);
								proj_yr0_y = cvmGet(Proj_yr0_y, yl0, xl0);

								proj_xrt_x = cvmGet(Proj_xrt_x, yl0, xl0);
								proj_xrt_y = cvmGet(Proj_xrt_y, yl0, xl0);
								proj_yrt_x = cvmGet(Proj_yrt_x, yl0, xl0);
								proj_yrt_y = cvmGet(Proj_yrt_y, yl0, xl0);
								
								proj_xrt_u = cvmGet(Proj_xrt_u, yl0, xl0);
								proj_yrt_u = cvmGet(Proj_yrt_u, yl0, xl0);
								proj_xrt_v = cvmGet(Proj_xrt_v, yl0, xl0);
								proj_yrt_v = cvmGet(Proj_yrt_v, yl0, xl0);
								proj_xrt_w = cvmGet(Proj_xrt_w, yl0, xl0);
								//proj_xrt_u = proj_xrt_w <= -0.3 ? -(proj_xrt_u) :(proj_xrt_u);
								proj_yrt_w = cvmGet(Proj_yrt_w, yl0, xl0);
								//proj_yrt_u = proj_xrt_w <= -0.3 ? -(proj_yrt_u) :(proj_yrt_u);
			
								double det = proj_xrt_x*proj_yrt_y - proj_xrt_y*proj_yrt_x;
								if ((det) <= WARP_DETERMINANT_THRESHOLD) {
									// image is reversed, set everything to 0
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,191);
										pixel_dorightflow = 0;
									}
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,191);
										pixel_doZt = 0;
									}
								}
								else {
									irt_X = (irtw_X*proj_yrt_y - irtw_Y*proj_yrt_x)/det;
									irt_Y = (irtw_Y*proj_xrt_x - irtw_X*proj_xrt_y)/det;
									irtw_U = irt_X*proj_xrt_u + irt_Y*proj_yrt_u ; 
									irtw_V = irt_X*proj_xrt_v + irt_Y*proj_yrt_v ; 
									irtw_W = irt_X*proj_xrt_w + irt_Y*proj_yrt_w ;
									if(proj_xrt_u || proj_yrt_u)
										irtw_U = irtw_U/sqrt(proj_xrt_u*proj_xrt_u + proj_yrt_u*proj_yrt_u);
									if(proj_xrt_y || proj_yrt_y)
										irtw_V = irtw_V/sqrt(proj_xrt_v*proj_xrt_v + proj_yrt_v*proj_yrt_v);
									if(proj_xrt_w || proj_yrt_w)
										irtw_W = irtw_W/sqrt(proj_xrt_w*proj_xrt_w + proj_yrt_w*proj_yrt_w);
										
								}
							}
							if (pixel_doleftflow || pixel_doZt) {
								// Wraped image values
								ilt = cvGetReal2D(Iltw, yl0, xl0);
								iltw_X = cvmGet(Iltw_X,  yl0, xl0);
								iltw_Y = cvmGet(Iltw_Y,  yl0, xl0);
								iltw_XX = cvmGet(Iltw_XX,  yl0, xl0);
								iltw_XY = cvmGet(Iltw_XY,  yl0, xl0);
								iltw_YY = cvmGet(Iltw_YY,  yl0, xl0);
								// projected coordinated derivatives
								proj_xlt_x = cvmGet(Proj_xlt_x, yl0, xl0);
								proj_xlt_y = cvmGet(Proj_xlt_y, yl0, xl0);
								proj_ylt_x = cvmGet(Proj_ylt_x, yl0, xl0);
								proj_ylt_y = cvmGet(Proj_ylt_y, yl0, xl0);
								proj_xlt_u = cvmGet(Proj_xlt_u, yl0, xl0);
								proj_ylt_u = cvmGet(Proj_ylt_u, yl0, xl0);
								proj_xlt_v = cvmGet(Proj_xlt_v, yl0, xl0);
								proj_ylt_v = cvmGet(Proj_ylt_v, yl0, xl0);
								proj_xlt_w = cvmGet(Proj_xlt_w, yl0, xl0);
								proj_ylt_w = cvmGet(Proj_ylt_w, yl0, xl0);
							
								double det = proj_xlt_x*proj_ylt_y - proj_xlt_y*proj_ylt_x;
								if ((det) <= WARP_DETERMINANT_THRESHOLD) {
									// image is reversed, set everything to 0
									if (pixel_doleftflow) {
										cvmSet8u(disactivated_pixels_leftflow,yl0,xl0,191);
										pixel_doleftflow = 0;
									}
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,191);
										pixel_doZt = 0;
									}
								}
								else {
									ilt_X = (iltw_X*proj_ylt_y - iltw_Y*proj_ylt_x)/det ;
									ilt_Y = (iltw_Y*proj_xlt_x - iltw_X*proj_xlt_y)/det ;
									iltw_U = ilt_X*proj_xlt_u + ilt_Y*proj_ylt_u ; 
									iltw_V = ilt_X*proj_xlt_v + ilt_Y*proj_ylt_v ; 
									iltw_W = ilt_X*proj_xlt_w + ilt_Y*proj_ylt_w ;
									if(proj_xlt_u || proj_ylt_u)
										iltw_U = iltw_U/sqrt(proj_xlt_u*proj_xlt_u + proj_ylt_u*proj_ylt_u);
									if(proj_xlt_v || proj_ylt_v)
										iltw_V = iltw_V/sqrt(proj_xlt_v*proj_xlt_v + proj_ylt_v*proj_ylt_v);
									if(proj_xlt_w || proj_ylt_w)
										iltw_W = iltw_W/sqrt(proj_xlt_w*proj_xlt_w + proj_ylt_w*proj_ylt_w);
								}
							}
						
							if (pixel_doleftflow || pixel_dorightflow || pixel_doZt) {
								du = cvmGet(dU, yl0, xl0);
								dv = cvmGet(dV, yl0, xl0);
								dw = cvmGet(dW, yl0, xl0);
							}
							if (pixel_doleftflow) {
								//psi' coefficients computation
								//tali: the data term of the left camera OF
								double a = square(ilt + iltw_U*du + iltw_V*dv + iltw_W*dw  - il0);
								//divergence coefficients
								double psi_flowL = DPsiData(a);

								// tali: the _1 refers to the the u equation, the _2 to the v equation the _3 to w and _4 z 
								//u equation								
								a_u_1 += psi_flowL*(iltw_U*iltw_U);
								a_v_1 += psi_flowL*(iltw_V*iltw_U);
								a_w_1 += psi_flowL*(iltw_W*iltw_U);
								
								//v equation
								a_u_2 += psi_flowL*(iltw_U*iltw_V);
								a_v_2 += psi_flowL*(iltw_V*iltw_V);
								a_w_2 += psi_flowL*(iltw_W*iltw_V);
								
								//w equation
								a_u_3 += psi_flowL*(iltw_U*iltw_W);
								a_v_3 += psi_flowL*(iltw_V*iltw_W);
								a_w_3 += psi_flowL*(iltw_W*iltw_W);
								
								// second term
								bu += - psi_flowL*(iltw_U*(ilt - il0));		   
								bv += - psi_flowL*(iltw_V*(ilt - il0));
								bw += - psi_flowL*(iltw_W*(ilt - il0));
								
							}
							if (pixel_dorightflow) {
								double a = square(irt + irtw_U*du + irtw_V*dv + irtw_W*dw - ir0);
								double psi_flowR = DPsiData(a);
								//u equation
								a_u_1 += psi_flowR*(irtw_U*irtw_U);
								a_v_1 += psi_flowR*(irtw_V*irtw_U);
								a_w_1 += psi_flowR*(irtw_W*irtw_U);
						
								//v equation
								a_u_2 += psi_flowR*(irtw_U*irtw_V);
								a_v_2 += psi_flowR*(irtw_V*irtw_V);
								a_w_2 += psi_flowR*(irtw_W*irtw_V);
								
								//w equation
								a_u_3 += psi_flowR*(irtw_U*irtw_W);
								a_v_3 += psi_flowR*(irtw_V*irtw_W);
								a_w_3 += psi_flowR*(irtw_W*irtw_W);
							
								// second term
								bu += - psi_flowR*(irtw_U*(irt - ir0));		   
								bv += - psi_flowR*(irtw_V*(irt - ir0));
								bw += - psi_flowR*(irtw_W*(irt - ir0));
							}
							if (pixel_doZt) {
								double a = square(irt + irtw_U*du + irtw_V*dv + irtw_W*dw  - 
												 (ilt + iltw_U*du + iltw_V*dv + iltw_W*dw));
								double psi_stereo = DPsiData(a);
								//u equation
								a_u_1  += psi_stereo*((irtw_U - iltw_U)*(irtw_U - iltw_U));
								a_v_1  += psi_stereo*((irtw_V - iltw_V)*(irtw_U - iltw_U));
								a_w_1  += psi_stereo*((irtw_W - iltw_W)*(irtw_U - iltw_U));
								
								//v equation
								a_u_2  += psi_stereo*((irtw_U - iltw_U)*(irtw_V - iltw_V));
								a_v_2  += psi_stereo*((irtw_V - iltw_V)*(irtw_V - iltw_V));
								a_w_2  += psi_stereo*((irtw_W - iltw_W)*(irtw_V - iltw_V));
								
								//w equation
								a_u_3  += psi_stereo*((irtw_U - iltw_U)*(irtw_W - iltw_W));
								a_v_3  += psi_stereo*((irtw_V - iltw_V)*(irtw_W - iltw_W));
								a_w_3  += psi_stereo*((irtw_W - iltw_W)*(irtw_W - iltw_W));
								
								// second term
								bu +=
									- psi_stereo*((irt - ilt)*(irtw_U - iltw_U));
								bv +=
									- psi_stereo*((irt - ilt)*(irtw_V - iltw_V));
								bw +=
									- psi_stereo*((irt - ilt)*(irtw_W - iltw_W));
							}
							// 3 lines corresponding to the 3 equations at (yl0,xl0) 
							cvmSet(M,i  ,0,a_u_1);
							cvmSet(M,i+1,0,a_u_2);
							cvmSet(M,i+2,0,a_u_3);
			   
							cvmSet(M,i  ,1,a_v_1);
							cvmSet(M,i+1,1,a_v_2);
							cvmSet(M,i+2,1,a_v_3);
						
							cvmSet(M,i  ,2,a_w_1);
							cvmSet(M,i+1,2,a_w_2);
							cvmSet(M,i+2,2,a_w_3);

							cvmSet(M,i  ,3,(-alpha/(4*h*h))*b_f_4);
							cvmSet(M,i+1,3,(-alpha/(4*h*h))*b_f_4);
							cvmSet(M,i+2,3,(-alpha*(mu)/(4*h*h))*b_fw_4);

							cvmSet(M,i  ,4,(-alpha/(4*h*h))*b_f_3);
							cvmSet(M,i+1,4,(-alpha/(4*h*h))*b_f_3);
							cvmSet(M,i+2,4,(-alpha*(mu)/(4*h*h))*b_fw_3);

							cvmSet(M,i  ,5,(-alpha/(4*h*h))*b_f_2);
							cvmSet(M,i+1,5,(-alpha/(4*h*h))*b_f_2);
							cvmSet(M,i+2,5,(-alpha*(mu)/(4*h*h))*b_fw_2);
		 
							cvmSet(M,i  ,6,(-alpha/(4*h*h))*b_f_1);
							cvmSet(M,i+1,6,(-alpha/(4*h*h))*b_f_1);
							cvmSet(M,i+2,6,(-alpha/(4*h*h))*b_fw_1);
							
							cvmSet(B,i,0,bu);
							cvmSet(B,i+1,0,bv);
							cvmSet(B,i+2,0,bw);
						}
				else if (methode == 2) {
							double il0 = 0.;
							double il0_X = 0., il0_Y = 0., il0w_Z = 0.;
							double ir0 = 0.;
							double ir0_X = 0., ir0_Y = 0.;
							double ir0w_X = 0., ir0w_Y = 0., ir0w_Z = 0.; 
							double ir0w_XX = 0., ir0w_XY = 0., ir0w_YY = 0.;
							double ir0w_X_X = 0., ir0w_X_Y = 0.;
							double ir0w_Y_X = 0., ir0w_Y_Y = 0.;
							double proj_xr0_x = 0., proj_xr0_y = 0;
							double proj_yr0_x = 0., proj_yr0_y = 0;
							double proj_xr0_z = 0., proj_yr0_z = 0;
							double ilt = 0.;
							double ilt_X = 0., ilt_Y = 0.;
							double iltw_X = 0., iltw_Y = 0., iltw_Z = 0.;
							double iltw_U = 0., iltw_V = 0.;
							double iltw_XX = 0., iltw_XY = 0., iltw_YY = 0.;
							double iltw_X_X = 0., iltw_X_Y = 0.;
							double iltw_Y_X = 0., iltw_Y_Y = 0.;
							double proj_xlt_x = 0., proj_xlt_y = 0., proj_xlt_z = 0.;
							double proj_xlt_u = 0., proj_xlt_v = 0.; 
							double proj_ylt_x = 0., proj_ylt_y = 0., proj_ylt_z = 0.;
							double proj_ylt_u = 0., proj_ylt_v = 0.; 
							double irt = 0.;
							double irt_X = 0., irt_Y = 0.;
							double irtw_X = 0., irtw_Y = 0., irtw_Z = 0.;
							double irtw_U = 0., irtw_V = 0., irtw_W = 0.;
							double irtw_XX = 0., irtw_XY = 0., irtw_YY = 0.;
							double irtw_X_X = 0., irtw_X_Y = 0.;
							double irtw_Y_X = 0., irtw_Y_Y = 0.;
							double proj_xrt_x = 0., proj_xrt_y = 0., proj_xrt_z = 0.;
							double proj_xrt_u = 0., proj_xrt_v = 0.; 
							double proj_yrt_x = 0., proj_yrt_y = 0., proj_yrt_z = 0.;
							double proj_yrt_u = 0., proj_yrt_v = 0.; 
							double du = 0, dv = 0, dz = 0;

							//interpolated divergence coefficients
							double b_f_1, b_f_2, b_f_3, b_f_4;
							double b_fz_1, b_fz_2, b_fz_3, b_fz_4;
							
							// Half pixel interpolation for the divergence coefficients 
							divergence_coefficients_interpolate(divergence_coefficients, yl0, xl0, &b_f_1, &b_f_2, &b_f_3, &b_f_4);
							divergence_coefficients_interpolate(divergence_coefficients_Z, yl0, xl0, &b_fz_1, &b_fz_2, &b_fz_3, &b_fz_4);
							//Pixels out must not be taken into account in the energy data term
							int pixel_doleftflow = 0;
							int pixel_dorightflow = 0;
							int pixel_doZ0 = 0;
							int pixel_doZt = 0;
							
							if (disactivated_pixels_leftflow)
								pixel_doleftflow = !cvmGet8u(disactivated_pixels_leftflow,yl0,xl0);
							if (disactivated_pixels_rightflow)
								pixel_dorightflow = !cvmGet8u(disactivated_pixels_rightflow,yl0,xl0);
							if (disactivated_pixels_Z0)
								pixel_doZ0 = !cvmGet8u(disactivated_pixels_Zt,yl0,xl0);
							if (disactivated_pixels_Zt)
								pixel_doZt = !cvmGet8u(disactivated_pixels_Zt,yl0,xl0);
							
							//is the point visible at time t+1 ?
							if (pixel_doleftflow || pixel_doZt) {
								double xlt = cvmGet(MapX_Ilt_inner, yl0, xl0);
								double ylt = cvmGet(MapY_Ilt_inner, yl0, xl0);
								if(xlt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xlt < DISACTIVATE_BORDER*h ||
								   ylt >= (N_y - 1 - DISACTIVATE_BORDER)*h || ylt < DISACTIVATE_BORDER*h){
									cvmSet8u(disactivated_pixels_leftflow,yl0,xl0,127);
									pixel_doleftflow = 0;
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,127);
										pixel_doZt = 0;
									}
								}
							}
							if (pixel_doZt || pixel_dorightflow) {
								double xrt = cvmGet(MapX_Irt_inner, yl0, xl0);
								double yrt = cvmGet(MapY_Irt_inner, yl0, xl0);
								if(xrt >= (N_x - 1 - DISACTIVATE_BORDER)*h || xrt < DISACTIVATE_BORDER*h ||
								   yrt >= (N_y - 1 - DISACTIVATE_BORDER)*h || yrt < DISACTIVATE_BORDER*h) {
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,127);
										pixel_doZt = 0;
									}
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,127);
										pixel_dorightflow = 0;
									}
								}
							}
							
							//On the right image at time t
							if (pixel_dorightflow || pixel_doZ0) {
								double xr0 = cvmGet(MapX_Ir0_inner, yl0, xl0);
								double yr0 = yl0;
								if(xr0 >= (N_x - 1 - DISACTIVATE_BORDER)*h || xr0 <= DISACTIVATE_BORDER*h ||
								   yr0 >= (N_y - 1 - DISACTIVATE_BORDER)*h || yr0 <= DISACTIVATE_BORDER*h) {
								
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,127);
										pixel_dorightflow = 0;
									}
									if (pixel_doZ0) {
										cvmSet8u(disactivated_pixels_Z0,yl0,xl0,127);
										pixel_doZ0 = 0;
									}
								}
							}

						
							double a_u_1, a_u_2, a_u_3, a_v_1, a_v_2, a_v_3;
							double a_z_1, a_z_2, a_z_3;
							a_u_1 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_v_1 = 0;
							a_z_1 = 0;	  
							
							a_u_2 = 0;
							a_v_2 = (alpha/(4.*h*h))*(b_f_1 + b_f_2 + b_f_3 + b_f_4);
							a_z_2 = 0;	  
							
							a_u_3 = 0;
							a_v_3 = 0;
							a_z_3 = (alpha*(mu)/(4.*h*h))*(b_fz_1 + b_fz_2 + b_fz_3 + b_fz_4);	  
							//second member for the first equation
							double bu0 = cvmGet(U, yl0, xl0);
							double bu1 = cvmGet(U, yl0, xl0+1);
							double bu2 = cvmGet(U, yl0, xl0-1);
							double bu3 = cvmGet(U,yl0+1,xl0);
							double bu4 = cvmGet(U,yl0-1,xl0);
							//Crop(bu0, N_x);
							double bv0 = cvmGet(V, yl0, xl0);
							double bv1 = cvmGet(V, yl0, xl0+1);
							double bv2 = cvmGet(V, yl0, xl0-1);
							double bv3 = cvmGet(V,yl0+1,xl0);
							double bv4 = cvmGet(V,yl0-1,xl0);
							//Crop(bw0, N_x);
							double bz0 = cvmGet(Z0, yl0, xl0);
							double bz1 = cvmGet(Z0, yl0, xl0+1);
							double bz2 = cvmGet(Z0, yl0, xl0-1);
							double bz3 = cvmGet(Z0,yl0+1,xl0);
							double bz4 = cvmGet(Z0,yl0-1,xl0);
							
							double bu = (alpha/(4*h*h)) *
								(b_f_1 * bu1 +
								 b_f_2 * bu2 +
								 b_f_3 * bu3 +
								 b_f_4 * bu4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bu0);

							//Crop(bv0, N_y);
							double bv = (alpha/(4*h*h)) *
								(b_f_1 * bv1 +
								 b_f_2 * bv2 +
								 b_f_3 * bv3 +
								 b_f_4 * bv4
								 - (b_f_1 + b_f_2 + b_f_3 + b_f_4) * bv0);
							//Crop(bv0, N_y);
							double bz = (alpha*(mu)/(4*h*h)) *
								(b_fz_1 * bz1 +
								 b_fz_2 * bz2 +
								 b_fz_3 * bz3 +
								 b_fz_4 * bz4
								 - (b_fz_1 + b_fz_2 + b_fz_3 + b_fz_4) * bz0);
							
							if (pixel_doleftflow || pixel_doZ0) {
								il0 = cvGetReal2D(Il0_pyramid[level], yl0, xl0); 
								il0_X = cvmGet(Il0_X,  yl0, xl0);
								il0_Y = cvmGet(Il0_Y,  yl0, xl0);
							}
							if(pixel_doZ0 || pixel_dorightflow) {
								ir0 = cvGetReal2D(Ir0w, yl0, xl0);
								ir0w_X = cvmGet(Ir0w_X,  yl0, xl0);
								ir0w_Y = cvmGet(Ir0w_Y,  yl0, xl0);
								ir0w_XX = cvmGet(Ir0w_XX,  yl0, xl0);
								ir0w_XY = cvmGet(Ir0w_XY,  yl0, xl0);
								ir0w_YY = cvmGet(Ir0w_YY,  yl0, xl0);
								proj_xr0_x = cvmGet(Proj_xr0_x, yl0, xl0);
								proj_xr0_y = cvmGet(Proj_xr0_y, yl0, xl0);
								proj_yr0_x = cvmGet(Proj_yr0_x, yl0, xl0);
								proj_yr0_y = cvmGet(Proj_yr0_y, yl0, xl0);
								proj_xr0_z = cvmGet(Proj_xr0_z, yl0, xl0);
								proj_yr0_z = cvmGet(Proj_yr0_z, yl0, xl0);
								
								double det = proj_xr0_x*proj_yr0_y - proj_xr0_y*proj_yr0_x;
								if ((det) <= WARP_DETERMINANT_THRESHOLD) {
									// image is reversed, set everything to 0
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,191);
										pixel_dorightflow = 0;
									}
									if (pixel_doZ0) {
										cvmSet8u(disactivated_pixels_Z0,yl0,xl0,191);
										pixel_doZ0 = 0;
									}
								}
								else {
									ir0_X = (ir0w_X*proj_yr0_y - irtw_Y*proj_yr0_x)/det;
									ir0_Y = (irtw_Y*proj_xr0_x - irtw_X*proj_xr0_y)/det;
									ir0w_Z = ir0_X*proj_xr0_z + ir0_Y*proj_yr0_z;
									if(deriv_norm) {
										if(proj_xr0_z || proj_yr0_z)
											ir0w_Z = ir0w_Z/sqrt(proj_xr0_z*proj_xr0_z + proj_yr0_z*proj_yr0_z);
									}
								}
							}
							
							if (pixel_dorightflow || do_Zt) {
								irt = cvGetReal2D(Irtw, yl0, xl0);
								irtw_X = cvmGet(Irtw_X,  yl0, xl0);
								irtw_Y = cvmGet(Irtw_Y,  yl0, xl0);
								proj_xrt_x = cvmGet(Proj_xrt_x, yl0, xl0);
								proj_xrt_y = cvmGet(Proj_xrt_y, yl0, xl0);
								proj_yrt_x = cvmGet(Proj_yrt_x, yl0, xl0);
								proj_yrt_y = cvmGet(Proj_yrt_y, yl0, xl0);
								
								proj_xrt_u = cvmGet(Proj_xrt_u, yl0, xl0);
								proj_yrt_u = cvmGet(Proj_yrt_u, yl0, xl0);
								proj_xrt_v = cvmGet(Proj_xrt_v, yl0, xl0);
								proj_yrt_v = cvmGet(Proj_yrt_v, yl0, xl0);
								proj_xrt_z = cvmGet(Proj_xrt_z, yl0, xl0);
								proj_yrt_z = cvmGet(Proj_yrt_z, yl0, xl0);
								
								double det = proj_xrt_x*proj_yrt_y - proj_xrt_y*proj_yrt_x;
								if ((det) <= WARP_DETERMINANT_THRESHOLD) {
									// image is reversed, set everything to 0
									if (pixel_dorightflow) {
										cvmSet8u(disactivated_pixels_rightflow,yl0,xl0,191);
										pixel_dorightflow = 0;
									}
								}
								else {
									irt_X = (irtw_X*proj_yrt_y - irtw_Y*proj_yrt_x)/det;
									irt_Y = (irtw_Y*proj_xrt_x - irtw_X*proj_xrt_y)/det;
									irtw_U = irt_X*proj_xrt_u + irt_Y*proj_yrt_u ; 
									irtw_V = irt_X*proj_xrt_v + irt_Y*proj_yrt_v ; 
									irtw_Z = irt_X*proj_xrt_z + irt_Y*proj_yrt_z ; 
									if(deriv_norm) {
										if(proj_xrt_u || proj_yrt_u)
											irtw_U = irtw_U/sqrt(proj_xrt_u*proj_xrt_u + proj_yrt_u*proj_yrt_u);
										if(proj_xrt_v || proj_yrt_v)
											irtw_V = irtw_V/sqrt(proj_xrt_v*proj_xrt_v + proj_yrt_v*proj_yrt_v);
										if(proj_xrt_z || proj_yrt_z)
											irtw_Z = irtw_Z/sqrt(proj_xrt_z*proj_xrt_z + proj_yrt_z*proj_yrt_z);
									}
								}
							}
							if (pixel_doleftflow || pixel_doZt) {
								// Wraped image values
								ilt = cvGetReal2D(Iltw, yl0, xl0);
								iltw_X = cvmGet(Iltw_X,  yl0, xl0);
								iltw_Y = cvmGet(Iltw_Y,  yl0, xl0);
								iltw_XX = cvmGet(Iltw_XX,  yl0, xl0);
								iltw_XY = cvmGet(Iltw_XY,  yl0, xl0);
								iltw_YY = cvmGet(Iltw_YY,  yl0, xl0);
								// projected coordinated derivatives
								proj_xlt_x = cvmGet(Proj_xlt_x, yl0, xl0);
								proj_xlt_y = cvmGet(Proj_xlt_y, yl0, xl0);
								proj_ylt_x = cvmGet(Proj_ylt_x, yl0, xl0);
								proj_ylt_y = cvmGet(Proj_ylt_y, yl0, xl0);
								proj_xlt_u = cvmGet(Proj_xlt_u, yl0, xl0);
								proj_ylt_u = cvmGet(Proj_ylt_u, yl0, xl0);
								proj_xlt_v = cvmGet(Proj_xlt_v, yl0, xl0);
								proj_ylt_v = cvmGet(Proj_ylt_v, yl0, xl0);
								proj_xlt_z = cvmGet(Proj_xlt_z, yl0, xl0);
								proj_ylt_z = cvmGet(Proj_ylt_z, yl0, xl0);
							
								double det = proj_xlt_x*proj_ylt_y - proj_xlt_y*proj_ylt_x;
								if ((det) <= WARP_DETERMINANT_THRESHOLD) {
									// image is reversed, set everything to 0
									if (pixel_doleftflow) {
										cvmSet8u(disactivated_pixels_leftflow,yl0,xl0,191);
										pixel_doleftflow = 0;
									}
									if (pixel_doZt) {
										cvmSet8u(disactivated_pixels_Zt,yl0,xl0,191);
										pixel_doZt = 0;
									}
								}
								else {
									ilt_X = (iltw_X*proj_ylt_y - iltw_Y*proj_ylt_x)/det ;
									ilt_Y = (iltw_Y*proj_xlt_x - iltw_X*proj_xlt_y)/det ;
									iltw_U = ilt_X*proj_xlt_u + ilt_Y*proj_ylt_u ; 
									iltw_V = ilt_X*proj_xlt_v + ilt_Y*proj_ylt_v ; 
									iltw_Z = ilt_X*proj_xlt_z + ilt_Y*proj_ylt_z ;
									if(deriv_norm) {
										if(proj_xlt_u || proj_ylt_u)
											iltw_U = iltw_U/sqrt(proj_xlt_u*proj_xlt_u + proj_yrt_u*proj_ylt_u);
										if(proj_xlt_v || proj_ylt_v)
											iltw_V = iltw_V/sqrt(proj_xlt_v*proj_xlt_v + proj_ylt_v*proj_ylt_v);
										if(proj_xlt_z || proj_ylt_z)
											iltw_Z = iltw_Z/sqrt(proj_xlt_z*proj_xlt_z + proj_ylt_z*proj_ylt_z);
									}
								}
							}
						
							if (pixel_doleftflow || pixel_dorightflow || pixel_doZt) {
								du = cvmGet(dU, yl0, xl0);
								dv = cvmGet(dV, yl0, xl0);
								dz = cvmGet(dZ0, yl0, xl0);
							}
							if (pixel_doleftflow) {
								//psi' coefficients computation
								//tali: the data term of the left camera OF
								double a = square(ilt + iltw_U*du + iltw_V*dv + iltw_Z*dz  - il0);
								//divergence coefficients
								double psi_flowL = DPsiData(a);

								// tali: the _1 refers to the the u equation, the _2 to the v equation the _3 to w and _4 z 
								//u equation								
								a_u_1 += psi_flowL*(iltw_U*iltw_U);
								a_v_1 += psi_flowL*(iltw_V*iltw_U);
								a_z_1 += psi_flowL*(iltw_Z*iltw_U);
								
								//v equation
								a_u_2 += psi_flowL*(iltw_U*iltw_V);
								a_v_2 += psi_flowL*(iltw_V*iltw_V);
								a_z_2 += psi_flowL*(iltw_Z*iltw_V);
								
								//w equation
								a_u_3 += psi_flowL*(iltw_U*iltw_Z);
								a_v_3 += psi_flowL*(iltw_V*iltw_Z);
								a_z_3 += psi_flowL*(iltw_Z*iltw_Z);
								
								// second term
								bu += - psi_flowL*(iltw_U*(ilt - il0));		   
								bv += - psi_flowL*(iltw_V*(ilt - il0));
								bz += - psi_flowL*(iltw_Z*(ilt - il0));
								
							}

							if (pixel_dorightflow) {
								double a = square(irt + irtw_U*du + irtw_V*dv + irtw_Z*dz - (ir0 + ir0w_Z*dz));
								double psi_flowR = DPsiData(a);
								//u equation
								a_u_1 += psi_flowR*(irtw_U*irtw_U);
								a_v_1 += psi_flowR*(irtw_V*irtw_U);
								a_z_1 += psi_flowR*((irtw_Z - ir0w_Z)*irtw_U);
						
								//v equation
								a_u_2 += psi_flowR*(irtw_U*irtw_V);
								a_v_2 += psi_flowR*(irtw_V*irtw_V);
								a_z_2 += psi_flowR*((irtw_Z - ir0w_Z)*irtw_V);
								
								//w equation
								a_u_3 += psi_flowR*(irtw_U*irtw_W);
								a_v_3 += psi_flowR*(irtw_V*irtw_W);
								a_z_3 += psi_flowR*((irtw_Z - ir0w_Z)*irtw_Z);
							
								// second term
								bu += - psi_flowR*(irtw_U*(irt - ir0));		   
								bv += - psi_flowR*(irtw_V*(irt - ir0));
								bz += - psi_flowR*((irtw_Z - ir0w_Z)*(irt - ir0));
							}
							if(pixel_doZ0) {
								double a = square(ir0 + ir0w_Z*dz - il0);
								double psi_Z0 = DPsiData(a);
								a_z_3 += psi_Z0*(ir0w_Z*ir0w_Z);
								bz    += - psi_Z0*((ir0 - il0)*ir0w_Z);
							}
							
							if (pixel_doZt) {
								double a = square(irt + irtw_U*du + irtw_V*dv + irtw_Z*dz  - 
												 (ilt + iltw_U*du + iltw_V*dv + iltw_Z*dz));
								double psi_stereo = DPsiData(a);
								//u equation
								a_u_1  += psi_stereo*((irtw_U - iltw_U)*(irtw_U - iltw_U));
								a_v_1  += psi_stereo*((irtw_V - iltw_V)*(irtw_U - iltw_U));
								a_z_1  += psi_stereo*((irtw_Z - iltw_Z)*(irtw_U - iltw_U));
								
								//v equation
								a_u_2  += psi_stereo*((irtw_U - iltw_U)*(irtw_V - iltw_V));
								a_v_2  += psi_stereo*((irtw_V - iltw_V)*(irtw_V - iltw_V));
								a_z_2  += psi_stereo*((irtw_Z - iltw_Z)*(irtw_V - iltw_V));
								
								//w equation
								a_u_3  += psi_stereo*((irtw_U - iltw_U)*(irtw_Z - iltw_Z));
								a_v_3  += psi_stereo*((irtw_V - iltw_V)*(irtw_Z - iltw_Z));
								a_z_3  += psi_stereo*((irtw_Z - iltw_Z)*(irtw_Z - iltw_Z));
								
								// second term
								bu +=
									- psi_stereo*((irt - ilt)*(irtw_U - iltw_U));
								bv +=
									- psi_stereo*((irt - ilt)*(irtw_V - iltw_V));
								bz +=
									- psi_stereo*((irt - ilt)*(irtw_Z - iltw_Z));
							}
							// 3 lines corresponding to the 3 equations at (yl0,xl0) 
							cvmSet(M,i  ,0,a_u_1);
							cvmSet(M,i+1,0,a_u_2);
							cvmSet(M,i+2,0,a_u_3);
			   
							cvmSet(M,i  ,1,a_v_1);
							cvmSet(M,i+1,1,a_v_2);
							cvmSet(M,i+2,1,a_v_3);
						
							cvmSet(M,i  ,2,a_z_1);
							cvmSet(M,i+1,2,a_z_2);
							cvmSet(M,i+2,2,a_z_3);

							cvmSet(M,i  ,3,(-alpha/(4*h*h))*b_f_4);
							cvmSet(M,i+1,3,(-alpha/(4*h*h))*b_f_4);
							cvmSet(M,i+2,3,(-alpha*(mu)/(4*h*h))*b_fz_4);

							cvmSet(M,i  ,4,(-alpha/(4*h*h))*b_f_3);
							cvmSet(M,i+1,4,(-alpha/(4*h*h))*b_f_3);
							cvmSet(M,i+2,4,(-alpha*(mu)/(4*h*h))*b_fz_3);

							cvmSet(M,i  ,5,(-alpha/(4*h*h))*b_f_2);
							cvmSet(M,i+1,5,(-alpha/(4*h*h))*b_f_2);
							cvmSet(M,i+2,5,(-alpha*(mu)/(4*h*h))*b_fz_2);
		 
							cvmSet(M,i  ,6,(-alpha/(4*h*h))*b_f_1);
							cvmSet(M,i+1,6,(-alpha/(4*h*h))*b_f_1);
							cvmSet(M,i+2,6,(-alpha*(mu)/(4*h*h))*b_fz_1);
							
							cvmSet(B,i  ,0,bu/SCALE_B);
							cvmSet(B,i+1,0,bv/SCALE_B);
							cvmSet(B,i+2,0,bz/SCALE_B);
						
				}
					else if(methode == 6) {
						double ir0 = 0., il0 = 0.;
						double il0_X = 0., il0_Y = 0.;
						double ir0_X = 0., ir0_Y = 0.;
						double ir0w_Z = 0;
						double ir0w_X = 0.,   ir0w_Y = 0.;
						double proj_x2_x = 0.,  proj_x2_y = 0., proj_x2_z = 0.;
						double proj_y2_x = 0.,  proj_y2_y = 0., proj_y2_z = 0.;
						double ir0w_XX = 0.,  ir0w_XY = 0., ir0w_YY = 0.;
						double ir0w_X_X = 0., ir0w_X_Y = 0.;
						double ir0w_Y_X = 0., ir0w_Y_Y = 0.;
						double psi_Z_init = 0.;
						double Z0w_X = cvmGet(Z0_X, yl0, xl0) ;
						double Z0w_Y = cvmGet(Z0_Y, yl0, xl0) ;
						double pb_f_1, pb_f_2, pb_f_3, pb_f_4;

						divergence_coefficients_interpolate(divergence_coefficients_Z, yl0, xl0, &pb_f_1, &pb_f_2, &pb_f_3, &pb_f_4);

						int pixel_is_disactivated_Z = cvmGet8u(disactivated_pixels_Z0,yl0,xl0);

						double xr0 = cvmGet(MapX_Ir0_inner, yl0, xl0);
						double yr0 = cvmGet(MapY_Ir0_inner, yl0, xl0);
						if(!pixel_is_disactivated_Z &&
						   (xr0 >= (N_x - 1 - DISACTIVATE_BORDER)*h || xr0 < DISACTIVATE_BORDER*h ||
							yr0 >= (N_y - 1 - DISACTIVATE_BORDER)*h || yr0 < DISACTIVATE_BORDER*h)) {
							cvmSet8u(disactivated_pixels_Z0,yl0,xl0,127);
							pixel_is_disactivated_Z = 1;
						}
						
						double a_Z0_4 = (alpha*mu/(4*h*h))*(pb_f_1 + pb_f_2 + pb_f_3 + pb_f_4);

						if(!pixel_is_disactivated_Z){
							//Data term estimation
							il0 = cvGetReal2D(Il0_pyramid[level], yl0, xl0);
							il0_X = cvmGet(Il0_X, yl0, xl0);
							il0_Y = cvmGet(Il0_Y, yl0, xl0);	 

							// tali: these are relevant for the Z optimization
							ir0 = cvGetReal2D(Ir0w, yl0, xl0);
							ir0w_X = cvmGet(Ir0w_X, yl0, xl0);
							ir0w_Y = cvmGet(Ir0w_Y, yl0, xl0);
							ir0w_XX = cvmGet(Ir0w_XX, yl0, xl0);
							ir0w_XY = cvmGet(Ir0w_XY, yl0, xl0);
							proj_x2_x = cvmGet(Proj_xr0_x, yl0, xl0);
							proj_x2_y = cvmGet(Proj_xr0_y, yl0, xl0);
							proj_x2_z = cvmGet(Proj_xr0_z, yl0, xl0);
							proj_y2_x = cvmGet(Proj_yr0_x, yl0, xl0);
							proj_y2_y = cvmGet(Proj_yr0_y, yl0, xl0);
							proj_y2_z = cvmGet(Proj_yr0_z, yl0, xl0);
							
							double det = proj_x2_x*proj_y2_y - proj_x2_y*proj_y2_x;
							// tali: i added the abs. need to check this condition
							if ((det) <= WARP_DETERMINANT_THRESHOLD) {
								// image is reversed, set everything to 0
								if (!pixel_is_disactivated_Z) {
									cvmSet8u(disactivated_pixels_Z0,yl0,xl0,191);
									pixel_is_disactivated_Z = 1;
								}
							}
							else {				
								ir0_X = (ir0w_X*proj_y2_y - ir0w_Y*proj_y2_x)/det;
								ir0_Y = (ir0w_Y*proj_x2_x - ir0w_X*proj_x2_y)/det;
								ir0w_Z = ir0_X*proj_x2_z + ir0_Y*proj_y2_z;
							}
							double dz0 = cvmGet(dZ0, yl0, xl0);
							double a = square(ir0 + ir0w_Z*dz0 - il0);
							
							// tali: might consider to add the gradient constancy condition
							psi_Z_init = DPsiData(a);
							a_Z0_4    += psi_Z_init*(ir0w_Z*ir0w_Z); 

						} // if(!pixel_is_disactivated_Z)
						
						cvmSet(M,i,0,a_Z0_4);
						  
						cvmSet(M,i,1,(-alpha*mu/(4*h*h))*pb_f_4);
					     
						cvmSet(M,i,2,(-alpha*mu/(4*h*h))*pb_f_3);
					   
						cvmSet(M,i,3,(-alpha*mu/(4*h*h))*pb_f_2);
		                            
						cvmSet(M,i,4,(-alpha*mu/(4*h*h))*pb_f_1);

						double bZi0 = cvmGet(Z0, yl0, xl0);
						double bZi1 = cvmGet(Z0, yl0, xl0+1);
						double bZi2 = cvmGet(Z0, yl0, xl0-1);
						double bZi3 = cvmGet(Z0,yl0+1,xl0);
						double bZi4 = cvmGet(Z0,yl0-1,xl0);
						double bZi = (alpha*mu/(4*h*h)) *
							(pb_f_1 * bZi1 +
							 pb_f_2 * bZi2 +
							 pb_f_3 * bZi3 +
							 pb_f_4 * bZi4
							 - (pb_f_1 + pb_f_2 + pb_f_3 + pb_f_4) * bZi0);
						if (!pixel_is_disactivated_Z) {
							// second term
							bZi +=
								-psi_Z_init*((ir0 - il0)*ir0w_Z );
						}
						cvmSet(B,i,0,bZi);
					}// else if(methode == 6) */
				} //Pixel inside the integration domain 
			} //for(i = inf; i <= sup; i+=dimension){
				//FIXME
                

				//Initial energy computation
				if(logfile && iter_inner == 0 && iter_outer == 0){
					double E_Ref_flow;
					double E_Flow[MAX_NUM_OF_VIEWS] = {0}, E_Stereo_0[MAX_NUM_OF_VIEWS] = {0}, E_Stere0_t[MAX_NUM_OF_VIEWS] = {0};
					double Esfl, Esz0;
					double Edata, Esmooth, Eintensity;
					
					EnergyComputation3D((Images_0_pyramid[0])[level], 
										(const CvMat**)Images_0w,(const CvMat**)Images_tw,
										U, V, W, Z0,
										(const CvMat*)disactivated_pixels_ref_flow, (const CvMat**)disactivated_pixels_flow,
										(const CvMat**)disactivated_pixels_Z0, (const CvMat**)disactivated_pixels_Zt,
										N_x, N_y, &Edata, &Esmooth, mu,
										do_ref_flow, do_not_ref_flow, do_Z0, do_Zt,
										&E_Ref_flow, E_Flow, E_Stereo_0, E_Stere0_t, &Esfl, &Esz0, &Eintensity, level);
					
					fprintf(logfile, "** LEVEL %d resolution %d x %d\n", level+1, N_x, N_y);
					fprintf(logfile, "* initial energy:\n");
					fprintf(logfile, "Edata = %g Esmooth = %g (alpha=%g)\n", Edata/(N_x*N_y), Esmooth/(N_x*N_y), alpha);
					for(view = 1; view < NumOfViews; view++){
						fprintf(logfile, "View = %d: E_flow = %g E_Stere0 = %g E_Stereo_t = %g\n",view, E_Flow[view]/(N_x*N_y), E_Stereo_0[view]/(N_x*N_y), E_Stere0_t[view]/(N_x*N_y));
					}
					fprintf(logfile, "E_ref_flow  = %g\n", E_Ref_flow/(N_x*N_y));
					fprintf(logfile, "Edata breakdown (squared): Eintensity = %g \n", Eintensity/(N_x*N_y));
					fprintf(logfile, "Esmooth breakdown (squared): flow = %g , surface = %g\n", Esfl/(N_x*N_y), Esz0/(N_x*N_y));
					fflush(logfile);
				}
				if (max_iter_outer == 0) {
					level = pyramid_level_end-1;
					goto CLEANUP;
				}

				//SOR iterations
			    cvSetZero(SOR);
			    cvSetZero(SOR_prec);
			    first_iteration = 1;
			    iter_SOR = 0;
			    norm_SOR = 0.;

	
			CvMat *M2 = cvCreateMat(dimension * N_x * N_y, dimension + 4, CV_32FC1); 
			CvMat *B2 = cvCreateMat(dimension * N_x * N_y,1,CV_32FC1);	
			cvCopy(M,M2);
			cvCopy(B,B2);

				while ((norm_SOR >= epsilon_SOR && iter_SOR < max_iter_SOR) || first_iteration == 1){
					first_iteration = 0;
					int Res = 0;
					for (int sub_iter_SOR = 0; sub_iter_SOR <4; sub_iter_SOR++) {
						int line, line_begin, line_end, line_incr;
						int col, col_begin, col_end, col_incr;
						switch (sub_iter_SOR) {
							case 0:         // NE-SW
								line_begin = 0;
								line_end = N_y * dimension * N_x;
								line_incr = dimension * N_x;
								col_begin = 0;
								col_end = dimension * N_x;
								col_incr = 1;
								break;
							case 1:         // SW-NE
								line_begin = (N_y-1) * dimension * N_x;
								line_end = - dimension * N_x;
								line_incr = - dimension * N_x;
								col_begin = dimension * N_x - 1;
								col_end = -1;
								col_incr = -1;
								break;
							case 2:         // NW-SE
								line_begin = 0;
								line_end = N_y * dimension * N_x;
								line_incr = dimension * N_x;
								col_begin = dimension * N_x - 1;
								col_end = -1;
								col_incr = -1;
								break;
							default:         // SE-NW
								line_begin = (N_y-1) * dimension * N_x;
								line_end = - dimension * N_x;
								line_incr = - dimension * N_x;
								col_begin = 0;
								col_end = dimension * N_x;
								col_incr = 1;
						}
						for (line = line_begin; line != line_end; line+= line_incr) {
							for (col = col_begin; col != col_end; col+= col_incr) {
								i = line + col;
								double sum = 0;
								double diag_term = 0.;
		                        
								//int rest = (i%(dimension*N_x))%dimension;
						  
								//tali: i refers to the relevant equations for y=0
								if(i < dimension*N_x){
									diag_term = cvmGet(M2, i,0); 
									sum = cvmGet(M2, i,1)*cvmGet(SOR, i+dimension*N_x,0);
								}
								//tali: i refers to the relevant equations for y=Ny
								else if(i >= dimension*N_x*(N_y-1)){
									diag_term = cvmGet(M2, i,0); 
									sum = cvmGet(M2, i, 1)*cvmGet(SOR, i-dimension*N_x,0);
								}
								else{
									if(i%(dimension*N_x) == 0){
								
									diag_term = cvmGet(M2, i,0); 
									sum = cvmGet(M2, i,1)*cvmGet(SOR, i+dimension,0);
									}
								else if((i+dimension)%(dimension*N_x)==0){
									diag_term = cvmGet(M2, i,0); 
									sum = cvmGet(M2, i,1)*cvmGet(SOR, i-dimension,0);
								}
								else{
									if(dimension == 1){
									sum = cvmGet(M2, i,1)*cvmGet(SOR, i-N_x,0) + cvmGet(M2, i,3)*cvmGet(SOR, i-1,0)
														+ cvmGet(M2, i,2)*cvmGet(SOR, i+N_x,0) + cvmGet(M2, i,4)*cvmGet(SOR, i+1,0);
									diag_term = cvmGet(M2, i,0);
								}
								else{
									int rest1 = i%(dimension*N_x);
						    
									if((dimension == 2 && (rest1 == 1)) ||
													   (dimension == 3 && (rest1 == 1 || rest1 == 2)) ||
													   (dimension == 4 && (rest1 == 1 || rest1 == 2 || rest1 == 3))){
										diag_term = cvmGet(M2, i,0); 
										sum = cvmGet(M2, i,1)*cvmGet(SOR, i+dimension,0);
									}
									else if((dimension == 4 && (rest1 == 4*N_x - 3 || rest1 == 4*N_x - 2 || rest1 == 4*N_x - 1)) ||
													(dimension == 3 && (rest1 == 3*N_x - 2 || rest1 == 3*N_x - 1)) ||
													(dimension == 2 && (rest1 == 2*N_x - 1))){
										diag_term = cvmGet(M2, i,0); 
										sum = cvmGet(M2, i,1)*cvmGet(SOR, i-dimension,0);
									}
							else{
							int rest2 = rest1 % dimension;
							Res = rest2;
							
							if(dimension == 2){
								if(rest2 == 0){
								  //u equation
								  sum =
								cvmGet(M2, i,2)*cvmGet(SOR, i-2*N_x,0) +
								cvmGet(M2, i,4)*cvmGet(SOR, i-2    ,0) +
								cvmGet(M2, i,1)*cvmGet(SOR, i+1    ,0) +
								cvmGet(M2, i,3)*cvmGet(SOR, i+2*N_x,0) +
								cvmGet(M2, i,5)*cvmGet(SOR, i+2    ,0);
								  diag_term = cvmGet(M2, i,0);
								}
								else if(rest2 == 1){
								  //v equation
								sum =
								cvmGet(M2, i,2)*cvmGet(SOR, i-2*N_x,0) +
								cvmGet(M2, i,4)*cvmGet(SOR, i-2    ,0) +
								cvmGet(M2, i,0)*cvmGet(SOR, i-1    ,0) +
								cvmGet(M2, i,3)*cvmGet(SOR, i+2*N_x,0) +
								cvmGet(M2, i,5)*cvmGet(SOR, i+2    ,0) ;
								diag_term = cvmGet(M2, i,1);
								}
							}
							else if(dimension == 3){
								if(methode == 1 || methode == 2){
								if(rest2 == 0){
									//u equation
									sum =
									cvmGet(M2, i,3)*cvmGet(SOR, i-3*N_x,0) +
									cvmGet(M2, i,5)*cvmGet(SOR, i-3    ,0) +
									cvmGet(M2, i,1)*cvmGet(SOR, i+1    ,0) +
									cvmGet(M2, i,2)*cvmGet(SOR, i+2    ,0) +
									cvmGet(M2, i,4)*cvmGet(SOR, i+3*N_x,0) +
									cvmGet(M2, i,6)*cvmGet(SOR, i+3    ,0);
									diag_term = cvmGet(M2, i,0);
								}
								else if(rest2 == 1){
									//v equation
									sum =
									cvmGet(M2, i,3)*cvmGet(SOR, i-3*N_x,0) +
									cvmGet(M2, i,5)*cvmGet(SOR, i-3    ,0) +
									cvmGet(M2, i,0)*cvmGet(SOR, i-1    ,0) +
									cvmGet(M2, i,2)*cvmGet(SOR, i+1    ,0) +
									cvmGet(M2, i,4)*cvmGet(SOR, i+3*N_x,0) +
									cvmGet(M2, i,6)*cvmGet(SOR, i+3    ,0) ;
									diag_term = cvmGet(M2, i,1);
								}
								else if(rest2 == 2){
									//w/z equation
									sum =
									cvmGet(M2, i,3)*cvmGet(SOR, i-3*N_x,0) +
									cvmGet(M2, i,5)*cvmGet(SOR, i-3    ,0) +
									cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) +
									cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) +
									cvmGet(M2, i,4)*cvmGet(SOR, i+3*N_x,0) +
									cvmGet(M2, i,6)*cvmGet(SOR, i+3    ,0);
									diag_term = cvmGet(M2, i,2);
									}
								}
							}
							else if(dimension == 4){
								if(rest2 == 0){
									//u equation
		#if DIVERGENCE_8                              
														sum =
															cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
															cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
															cvmGet(M2, i,11)*cvmGet(SOR, i-4*(N_x+1),0) +
															cvmGet(M2, i,10)*cvmGet(SOR, i-4*(N_x-1),0) +
															cvmGet(M2, i,1)*cvmGet(SOR, i+1    ,0) +
															cvmGet(M2, i,2)*cvmGet(SOR, i+2    ,0) +
															cvmGet(M2, i,3)*cvmGet(SOR, i+3    ,0) +
															cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
															cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0) +
															cvmGet(M2, i,9)*cvmGet(SOR, i+4*(N_x-1),0) +
															cvmGet(M2, i,8)*cvmGet(SOR, i+4*(N_x+1),0);
							     
		#else  // !DIVERGENCE_8
									sum =
										// tali: the neighbour above
										cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
										// tali: the neighbour to the left
										cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
										// tali: the relevant v 
										cvmGet(M2, i,1)*cvmGet(SOR, i+1    ,0) +
										// tali: the relevant w
										cvmGet(M2, i,2)*cvmGet(SOR, i+2    ,0) +
										// tali: the relevant z
										cvmGet(M2, i,3)*cvmGet(SOR, i+3    ,0) +
										// tali: the neighbour below
										cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
										// tali: the neighbour to right
										cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0);

									
		#endif  // !DIVERGENCE_8
										diag_term = cvmGet(M2, i,0);
								}
								else if(rest2 == 1){
									//v equation
		#if DIVERGENCE_8
														sum =
															cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
															cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
															cvmGet(M2, i,0)*cvmGet(SOR, i-1    ,0) +
															cvmGet(M2, i,11)*cvmGet(SOR, i-4*(N_x+1),0)+
															cvmGet(M2, i,10)*cvmGet(SOR, i-4*(N_x-1),0) +
															cvmGet(M2, i,2)*cvmGet(SOR, i+1    ,0) +
															cvmGet(M2, i,3)*cvmGet(SOR, i+2    ,0) +
															cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
															cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0) +
															cvmGet(M2, i,9)*cvmGet(SOR, i+4*(N_x-1),0) +
															cvmGet(M2, i,8)*cvmGet(SOR, i+4*(N_x+1),0);
							      
		#else  // !DIVERGENCE_8
										sum =
										cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
										cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
										cvmGet(M2, i,0)*cvmGet(SOR, i-1    ,0) +
										cvmGet(M2, i,2)*cvmGet(SOR, i+1    ,0) +
										cvmGet(M2, i,3)*cvmGet(SOR, i+2    ,0) +
										cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
										cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0) ;
		#endif  // !DIVERGENCE_8
										diag_term = cvmGet(M2, i,1);
								}
								else if(rest2 == 2){
									double S1 = cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) ;
									double S2 = cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) ;
									double S3 = cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) ;
									double S4 = cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) ;
									double S5=  cvmGet(M2, i,3)*cvmGet(SOR, i+1    ,0) ;
									double S6 = cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) ;
									double S7 = cvmGet(M2, i,7)*cvmGet(SOR, i+4,0)  ;
									//w equation
		#if DIVERGENCE_8	
														sum =
															cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) - cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x+1,0) +
															cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) - cvmGet(M2, i,6)*cvmGet(SOR, i-4    +1,0) +
															cvmGet(M2, i,11)*cvmGet(SOR, i-4*(N_x+1),0) - cvmGet(M2, i,11)*cvmGet(SOR, i-4*N_x-3,0)+
															cvmGet(M2, i,10)*cvmGet(SOR, i-4*(N_x-1),0) - cvmGet(M2, i,10)*cvmGet(SOR, i-4*(N_x-1)+1,0)+
															cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) +
															cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) +
															cvmGet(M2, i,3)*cvmGet(SOR, i+1    ,0) +
															cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) - cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x+1,0) +
															cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0) - cvmGet(M2, i,7)*cvmGet(SOR, i+4    +1,0) +
															cvmGet(M2, i,9)*cvmGet(SOR, i+4*(N_x-1),0) - cvmGet(M2, i,9)*cvmGet(SOR, i+4*(N_x-1)+1,0) +
															cvmGet(M2, i,8)*cvmGet(SOR, i+4*(N_x+1),0) - cvmGet(M2, i,8)*cvmGet(SOR, i+4*(N_x+1)+1,0);
		#else  // !DIVERGENCE_8
										/*sum =
										cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) - cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x+1,0) +
										cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) - cvmGet(M2, i,6)*cvmGet(SOR, i-4    +1,0) +
										cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) +
										cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) +
										cvmGet(M2, i,3)*cvmGet(SOR, i+1    ,0) +
										cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) - cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x+1,0) +
										cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0) - cvmGet(M2, i,7)*cvmGet(SOR, i+4    +1,0);
										*/
									sum =
										cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +
										cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
										cvmGet(M2, i,0)*cvmGet(SOR, i-2    ,0) +
										cvmGet(M2, i,1)*cvmGet(SOR, i-1    ,0) +
										cvmGet(M2, i,3)*cvmGet(SOR, i+1    ,0) +
										cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
										cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0);
		#endif  // !DIVERGENCE_8
										diag_term = cvmGet(M2, i,2);
								}
								else{
									//Z equation
														assert(rest2 == 3);
		#if DIVERGENCE_8						
								sum =
															cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) - cvmGet(M2, i-1 ,4)*cvmGet(SOR, i-4*N_x-1,0) +
															cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) - cvmGet(M2, i-1 ,6)*cvmGet(SOR, i-4    -1,0) +
															cvmGet(M2, i,11)*cvmGet(SOR, i-4*(N_x+1),0) - cvmGet(M2, i-1,11)*cvmGet(SOR, i-4*N_x-5,0)+
									cvmGet(M2, i,10)*cvmGet(SOR, i-4*(N_x-1),0) - cvmGet(M2, i-1,10)*cvmGet(SOR, i-4*(N_x-1)-1,0)+    
									cvmGet(M2, i,0)*cvmGet(SOR, i-3    ,0) +
															cvmGet(M2, i,1)*cvmGet(SOR, i-2    ,0) +
															cvmGet(M2, i,2)*cvmGet(SOR, i-1    ,0) +
															cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0)- cvmGet(M2, i-1,5)*cvmGet(SOR, i+4*N_x-1,0) +
															cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0)- cvmGet(M2, i-1,7)*cvmGet(SOR, i+4    -1,0) +
									cvmGet(M2, i,9)*cvmGet(SOR, i+4*(N_x-1),0) - cvmGet(M2, i-1,9)*cvmGet(SOR, i+4*(N_x-1)-1,0) +
									cvmGet(M2, i,8)*cvmGet(SOR, i+4*(N_x+1),0) - cvmGet(M2, i-1,8)*cvmGet(SOR, i+4*(N_x+1)-1,0);
								
		#else  // !DIVERGENCE_8
							/*	sum =
										cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) - cvmGet(M2, i-1 ,4)*cvmGet(SOR, i-4*N_x-1,0) +
										cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) - cvmGet(M2, i-1 ,6)*cvmGet(SOR, i-4    -1,0) +
										cvmGet(M2, i,0)*cvmGet(SOR, i-3    ,0) +
										cvmGet(M2, i,1)*cvmGet(SOR, i-2    ,0) +
										cvmGet(M2, i,2)*cvmGet(SOR, i-1    ,0) +
										cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0)- cvmGet(M2, i-1,5)*cvmGet(SOR, i+4*N_x-1,0) +
										cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0)- cvmGet(M2, i-1,7)*cvmGet(SOR, i+4    -1,0);*/
		#endif  // !DIVERGENCE_8
								sum =
										cvmGet(M2, i,4)*cvmGet(SOR, i-4*N_x,0) +  
										cvmGet(M2, i,6)*cvmGet(SOR, i-4    ,0) +
										cvmGet(M2, i,0)*cvmGet(SOR, i-3    ,0) +
										cvmGet(M2, i,1)*cvmGet(SOR, i-2    ,0) +
										cvmGet(M2, i,2)*cvmGet(SOR, i-1    ,0) +
										cvmGet(M2, i,5)*cvmGet(SOR, i+4*N_x,0) +
										cvmGet(M2, i,7)*cvmGet(SOR, i+4    ,0);
										diag_term = cvmGet(M2, i,3);	
								}
							}
							}
						}
						}
					}
					
					double rold = cvmGet(SOR,i,0);
					double b = cvmGet(B2,i,0);
					double rnew = (1.-omega_SOR)*rold + (omega_SOR/diag_term) * (b - sum);
					
		#ifdef DEBUG_PIXEL
								if(i == ((DEBUG_PIXEL_X+DEBUG_PIXEL_Y*N_x)*dimension+DEBUG_PIXEL_COMPONENT)) {
									debug_me();
								}
		#endif
								//assert(fabs(rnew)<(N_x+N_y));
					if(methode == 0 ){
						if(Res == 4 ) {
							cvmSet(SOR,i,0,0);
						}
						else {
						cvmSet(SOR,i,0,Crop(rnew,2*MAX_INCR));
						}
					}
					else{
						cvmSet(SOR,i,0,Crop(rnew,2*MAX_INCR));
					}

								
					} // for (col
					} // for (line
					} // for (sub_iter_SOR
					
				
				
					norm_SOR = cvNorm(SOR_prec, SOR, CV_RELATIVE_L2);
					cvCopy(SOR, SOR_prec);
					iter_SOR ++;
				    
		#if 1
		#ifndef _MSC_VER
		#warning  "preconditioning disactivated"
		#endif
		#else
					//Condition the system by using the increments norm every 4 steps
							if(iter_SOR < 10){
								const double eps_sor = 0.001;
					CvMat *Precond =  cvCreateMat(dimension*N_y*N_x,1,CV_32FC1);
					SFMatrixProduct(M, SOR, Precond, N_x, N_y, dimension);
								//printf("precond!\n");
								for (i = 0; i < dimension * N_y * N_x; i++) {
									//int xl0 = (i%(dimension*N_x))/dimension;
									//int yl0 = i/(dimension*N_x);
									double value = cvGetReal2D(B,i,0);
									double norm_inc_sor = fabs(cvGetReal2D(Precond,i,0));
									//if(norm_inc_sor > eps_sor)
									//    printf("inc(%d,%d,%d)=%g\n",xl0,yl0,i%dimension,norm_inc_sor);
									cvSetReal2D(B2,i,0,value/(norm_inc_sor+eps_sor));
									for(int p = 0; p < dimension + 4; p++){
										value = cvGetReal2D(M,i,p);
										cvSetReal2D(M2,i,p,value/(norm_inc_sor+eps_sor));
									}
								}
					cvReleaseMat(&Precond);
					}
					else
					{
								cvCopy(M,M2);
								cvCopy(B,B2);
					}
#endif
			    } // while ((norm_SOR >= epsilon_SOR && iter_SOR < max_iter_SOR) || first_iteration == 1){
					//printf("iter_SOR=%d norm_SOR=%g\n",iter_SOR, norm_SOR);
				cvReleaseMat(&M2);
				cvReleaseMat(&B2);
	#if GMRES
				//Solving the system with a Krylov spaces methods, linear GRMS method
				if(dimension == 4){
					int N = dimension*N_x*N_y;
					CvMat *x0 = cvCreateMat(N,1,CV_32FC1);
					CvMat *xk = cvCreateMat(N,1,CV_32FC1);
					CvMat *rk = cvCreateMat(N,1,CV_32FC1);
					cvSetZero(xk);
					cvSetZero(rk);
				    
					double norm_GMRES = n_norm(rk, N);
					int flag = 1;
					cvCopy(SOR, x0);
					int iter_GMRES = 0;
					while((norm_GMRES > epsilon_SOR || flag == 1) && iter_GMRES < 10){
					flag = 0;
					//GmRes Algorithm
					GmRes(M, B, x0, xk, rk, GMRES_KMAX, epsilon_SOR, N_x, N_y, dimension);
					norm_GMRES = n_norm(rk, N);
					cvCopy(xk, x0);  
					iter_GMRES++;
					}
					cvCopy(xk, SOR);
					cvReleaseMat(&x0);
					cvReleaseMat(&xk);
					cvReleaseMat(&rk);
				}
		#endif //GRMS
				if (do_ref_flow || do_not_ref_flow || do_Zt) {
					cvCopy(dU, dU_prec); 
					cvCopy(dV, dV_prec);
					if(methode != 2)
						cvCopy(dW, dW_prec);
				}
				if (do_Z0) {
					cvCopy(dZ0, dZ0_prec);
				}
						                
				for(i = 0; i < dimension*N_x*N_y; i+=dimension){
					xl0 = (i%(dimension*N_x))/dimension;
					yl0 = i/(dimension*N_x);
					int d = 0;

					if (do_ref_flow || do_not_ref_flow || do_Zt) {
						double du = cvmGet(SOR, i + d,0); d++;
						double dv = cvmGet(SOR, i + d,0); d++;
						double dw = 0;
						if(methode != 2) {
							dw = cvmGet(SOR, i + d,0); d++;
						}
						if(abs(du) > MAX_INCR || abs(dv) > MAX_INCR || abs(dw) > MAX_INCR)
						{
							double unw_norm = sqrt(du*du + dv*dv + dw*dw);
							du = MAX_INCR*du/unw_norm;
							dv = MAX_INCR*dv/unw_norm;
							dw = MAX_INCR*dw/unw_norm;
						}
						cvmSet(dU, yl0, xl0,du*SCALE_B);
						cvmSet(dV, yl0, xl0,dv*SCALE_B);/*
						cvmSet(dU, yl0, xl0, Crop(du,MAX_INCR));
						cvmSet(dV, yl0, xl0, Crop(dv,MAX_INCR));*/
						
						if(methode != 2)
							cvmSet(dW, yl0, xl0,dw*SCALE_B);
					}
	               
					if (do_Z0) {
						double dz = cvmGet(SOR, i + d,0); d++;
						cvmSet(dZ0, yl0, xl0, Crop(dz,MAX_INCR));
						cvmSet(dZ0, yl0, xl0,dz*SCALE_B);
					}

					assert(d == dimension);
				}
				//FIXME
				/*
				if (OF_ZeroTopBorder) {
					if (doleftflow) {
						for(xl0 = 0; xl0 < N_x; xl0++){
							cvmSet(dU, 0, xl0, 0);
							cvmSet(dV, 0, xl0, 0);
							cvmSet(dU, 1, xl0, 0);
							cvmSet(dV, 1, xl0, 0);
						}
						for(yl0 = 0; yl0 < N_y*7/8; yl0++){
							cvmSet(dU, yl0, 0, 0);
							cvmSet(dV, yl0, 0, 0);
							cvmSet(dU, yl0, N_x-1, 0);
							cvmSet(dV, yl0, N_x-1, 0);
							cvmSet(dU, yl0, 1, 0);
							cvmSet(dV, yl0, 1, 0);
							cvmSet(dU, yl0, N_x-2, 0);
							cvmSet(dV, yl0, N_x-2, 0);
							cvmSet(dU, yl0, 2, 0);
							cvmSet(dV, yl0, 2, 0);
							cvmSet(dU, yl0, N_x-3, 0);
							cvmSet(dV, yl0, N_x-3, 0);
						}
					}
					if (dodisparity) {
						for(xl0 = 0; xl0 < N_x; xl0++){
							cvmSet(dDt, 0, xl0, -40.*N_x/854 - cvmGet(Dt,0,xl0));
							cvmSet(dDt, 1, xl0, -40.*N_x/854 - cvmGet(Dt,1,xl0));
						}
						for(yl0 = 0; yl0 < N_y*7/8; yl0++){
							cvmSet(dDt, yl0, 0, -40.*N_x/854 - cvmGet(Dt,yl0,0));
							cvmSet(dDt, yl0, N_x-1, -40.*N_x/854 - cvmGet(Dt,yl0,N_x-1));
							cvmSet(dDt, yl0, 1, -40.*N_x/854 - cvmGet(Dt,yl0,1));
							cvmSet(dDt, yl0, N_x-2, -40.*N_x/854 - cvmGet(Dt,yl0,N_x-2));
							cvmSet(dDt, yl0, 2, -40.*N_x/854 - cvmGet(Dt,yl0,2));
							cvmSet(dDt, yl0, N_x-3, -40.*N_x/854 - cvmGet(Dt,yl0,N_x-3));
						}
					}
					if (dodisparity0) {
						for(xl0 = 0; xl0 < N_x; xl0++){
							cvmSet(dD0, 0, xl0, -40.*N_x/854 - cvmGet(D0,0,xl0));
							cvmSet(dD0, 1, xl0, -40.*N_x/854 - cvmGet(D0,1,xl0));
						}
						for(yl0 = 0; yl0 < N_y*7/8; yl0++){
							cvmSet(dD0, yl0, 0, -40.*N_x/854 - cvmGet(D0,yl0,0));
							cvmSet(dD0, yl0, N_x-1, -40.*N_x/854 - cvmGet(D0,yl0,N_x-1));
							cvmSet(dD0, yl0, 1, -40.*N_x/854 - cvmGet(D0,yl0,1));
							cvmSet(dD0, yl0, N_x-2, -40.*N_x/854 - cvmGet(D0,yl0,N_x-2));
							cvmSet(dD0, yl0, 2, -40.*N_x/854 - cvmGet(D0,yl0,2));
							cvmSet(dD0, yl0, N_x-3, -40.*N_x/854 - cvmGet(D0,yl0,N_x-3));
						}
					}
				}*/

				norm_inner = 0;
				if (do_ref_flow || do_not_ref_flow || do_Zt) {
					norm_inner = fmax(norm_inner, cvNorm(dU_prec, dU, CV_RELATIVE_L2));
					norm_inner = fmax(norm_inner, cvNorm(dV_prec, dV, CV_RELATIVE_L2));
					if(methode != 2)
						norm_inner = fmax(norm_inner, cvNorm(dW_prec, dW, CV_RELATIVE_L2));
				}
	           
				if (do_Z0) {
					norm_inner = fmax(norm_inner, cvNorm(dZ0_prec, dZ0, CV_RELATIVE_L2));
				}
				
				iter_inner++;
			}
			
			if (do_ref_flow || do_not_ref_flow || do_Zt) {
				cvCopy(U, U_prec);
				cvAdd(U, dU, U);
				//CropArr(U, N_x);
				cvCopy(V, V_prec);
				cvAdd(V, dV, V);
				//CropArr(V, N_y);
				if(methode != 2) {
					cvCopy(W, W_prec);
					cvAdd(W, dW, W);
					//CropArr(W, N_y);
				}
				
			}
			if (do_Z0) {
				cvCopy(Z0, Z0_prec);
				cvAdd(Z0, dZ0, Z0);
			//	CropArr(Z0, maxZ);
			}

			norm = 0;
			if(do_ref_flow || do_not_ref_flow || do_Zt){
				norm = fmax(norm, cvNorm(U_prec, U, CV_RELATIVE_L2));
				norm = fmax(norm, cvNorm(V_prec, V, CV_RELATIVE_L2));
				if(methode != 2)
					norm = fmax(norm, cvNorm(W_prec, W, CV_RELATIVE_L2));
			}
	           
			if(do_Z0) {
				norm = fmax(norm, cvNorm(Z0_prec, Z0, CV_RELATIVE_L2));
			}
#ifdef DEBUG
			if (do_ref_flow) {
				if (SF_U_FileName) {
					SaveResults(SF_U_FileName, U);
				}
				if (SF_V_FileName) {	     
					SaveResults(SF_V_FileName, V);
				}
				if (SF_W_FileName) {	     
					SaveResults(SF_W_FileName, W);
				}

			}
			if(do_Z0){
				SaveResults(SF_Z0_FileName, Z0);
			}
			
			if (SF_disactivated_ref_flow_FileName && disactivated_pixels_ref_flow) {				
				sprintf(file, "%s.%s",SF_disactivated_ref_flow_FileName, IMG_TYPE);
				SaveResults(file, disactivated_pixels_ref_flow, 0, 0., 255.);
			}
			for(view = 1; view < NumOfViews; view++) {
				if (SF_disactivated_flow_FileName && disactivated_pixels_flow) {
					sprintf(file, "%s%d.%s",SF_disactivated_flow_FileName, view, IMG_TYPE);
					SaveResults(file, disactivated_pixels_flow[view], 0, 0., 255.);
				}
				if (SF_disactivated_Surf_1_FileName && disactivated_pixels_Z0) {
					sprintf(file, "%s%d.%s",SF_disactivated_Surf_1_FileName, view, IMG_TYPE);
					SaveResults(file, disactivated_pixels_Z0[view], 0, 0., 255.);
				}
				if (SF_disactivated_Surf_2_FileName && disactivated_pixels_Zt) {
					sprintf(file, "%s%d.%s",SF_disactivated_Surf_2_FileName, view, IMG_TYPE);
					SaveResults(file, disactivated_pixels_Zt[view], 0, 0., 255.);
				}
			}
			for(view = 1; view < NumOfViews; view++) {
				sprintf(file, "%sv%d_el%d.txt", SF_disactivated_Surf_1_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_Z0[view]);
				fclose(f);

				sprintf(file, "%sv%d_el%d.txt",SF_disactivated_flow_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_flow[view]);
				fclose(f);

				sprintf(file, "%sv%d_el%d.txt",SF_disactivated_Surf_2_FileName,view,level);
				f = fopen(file,"w");
				WriteRes8u(f,disactivated_pixels_Zt[view]);
				fclose(f);

			}
			sprintf(file, "%s_el%d.txt",SF_disactivated_ref_flow_FileName,level);
			f = fopen(file,"w");
			WriteRes8u	(f,disactivated_pixels_ref_flow);
			fclose(f);
#endif
			
			iter_outer++;
		}
			



					double E_Ref_flow;
					double E_Flow[MAX_NUM_OF_VIEWS], E_Stereo_0[MAX_NUM_OF_VIEWS], E_Stere0_t[MAX_NUM_OF_VIEWS];
					double Esfl, Esz0;
					double Edata, Esmooth, Eintensity;
					
					EnergyComputation3D((Images_0_pyramid[0])[level], 
										(const CvMat**)Images_0w,(const CvMat**)Images_tw,
										U, V, W, Z0,
										(const CvMat*)disactivated_pixels_ref_flow, (const CvMat**)disactivated_pixels_flow,
										(const CvMat**)disactivated_pixels_Z0, (const CvMat**)disactivated_pixels_Zt,
										N_x, N_y, &Edata, &Esmooth, gamma,
										do_ref_flow, do_not_ref_flow, do_Z0, do_Zt,
										&E_Ref_flow, E_Flow, E_Stereo_0, E_Stere0_t, &Esfl, &Esz0, &Eintensity, level);
					
					fprintf(logfile, "** LEVEL %d resolution %d x %d\n", level+1, N_x, N_y);
					fprintf(logfile, "* final energy:\n");
					fprintf(logfile, "Edata = %g Esmooth = %g (alpha=%g)\n", Edata/(N_x*N_y), Esmooth/(N_x*N_y), alpha);
					for(view = 1; view < NumOfViews; view++){
						fprintf(logfile, "View = %d: E_flow = %g E_Stere0 = %g E_Stereo_t = %g\n",view, E_Flow[view]/(N_x*N_y), E_Stereo_0[view]/(N_x*N_y), E_Stere0_t[view]/(N_x*N_y));
							fprintf(logfile, "View = %d: E_flow = %g E_Stere0 = %g E_Stereo_t = %g\n",view, E_Flow[view]/(N_x*N_y), E_Stereo_0[view]/(N_x*N_y), E_Stere0_t[view]/(N_x*N_y));
							fflush(logfile);

					}
					fprintf(logfile, "E_ref_flow  = %g\n", E_Ref_flow);
					fprintf(logfile, "Edata breakdown (squared): Eintensity = %g \n", Eintensity);
					fprintf(logfile, "Esmooth breakdown (squared): flow = %g , surface = %g\n", Esfl, Esz0);
					fflush(logfile);
				
      CLEANUP:
	  //let's clean
        cvReleaseMat(&derivative_tmp);
        cvReleaseMat(&divergence_coefficients);
		cvReleaseMat(&MapIdentityX);
		cvReleaseMat(&MapIdentityY);
		for(view = 0; view < NumOfViews; view++) {
			cvReleaseMat(&Images_0_X[view]);
			cvReleaseMat(&Images_0_Y[view]);
			if (do_ref_flow || do_not_ref_flow || do_Zt) {
				cvReleaseMat(&Images_tw[view]);
				cvReleaseMat(&Images_tw_X[view]);
				cvReleaseMat(&Images_tw_Y[view]);
				cvReleaseMat(&Proj_xt_x[view]);
				cvReleaseMat(&Proj_xt_y[view]);
				cvReleaseMat(&Proj_yt_x[view]);
				cvReleaseMat(&Proj_yt_y[view]);
				cvReleaseMat(&Proj_xt_u[view]);
				cvReleaseMat(&Proj_xt_v[view]);
				cvReleaseMat(&Proj_xt_w[view]);
				cvReleaseMat(&Proj_yt_u[view]);
				cvReleaseMat(&Proj_yt_v[view]);
				cvReleaseMat(&Proj_yt_w[view]);
				cvReleaseMat(&MapX_Img_t[view]);
				cvReleaseMat(&MapY_Img_t[view]);
				cvReleaseMat(&MapX_Img_t_inner[view]);
				cvReleaseMat(&MapY_Img_t_inner[view]);
				if(do_Z0){
					cvReleaseMat(&Proj_xt_z[view]);
					cvReleaseMat(&Proj_yt_z[view]);
				}
			}
			if (do_Z0 || do_not_ref_flow) {
				cvReleaseMat(&Images_0w[view]);
				cvReleaseMat(&MapX_Img_0[view]);
				cvReleaseMat(&MapY_Img_0[view]);
				cvReleaseMat(&MapX_Img_0_inner[view]);
				cvReleaseMat(&MapY_Img_0_inner[view]);
				if(view > 0) {
					cvReleaseMat(&Images_0w_X[view]);
					cvReleaseMat(&Images_0w_Y[view]);
				}
				cvReleaseMat(&Proj_x0_x[view]);
				cvReleaseMat(&Proj_x0_y[view]);
				cvReleaseMat(&Proj_y0_x[view]);
				cvReleaseMat(&Proj_y0_y[view]);
				cvReleaseMat(&Proj_x0_u[view]);
				cvReleaseMat(&Proj_x0_v[view]);
				cvReleaseMat(&Proj_x0_w[view]);

				if(do_Z0) {
					cvReleaseMat(&Proj_x0_z[view]);
					cvReleaseMat(&Proj_y0_z[view]);
				}
			}
			if(view > 0){ 
				if (do_not_ref_flow || do_Zt) {
					 cvReleaseMat(&disactivated_pixels_flow[view]);
				}
				if(do_Z0) {
					cvReleaseMat(&disactivated_pixels_Z0[view]);
				}
				if (do_Zt) {
					cvReleaseMat(&disactivated_pixels_Zt[view]);
				}
			}
			if(do_ref_flow)
				cvReleaseMat(&disactivated_pixels_ref_flow);
		}
		cvReleaseMat(&X);
		cvReleaseMat(&Y);
		cvReleaseMat(&X_x);
		cvReleaseMat(&X_y);
		cvReleaseMat(&Y_x);
		cvReleaseMat(&Y_y);
		cvReleaseMat(&X_z);
		cvReleaseMat(&Y_z);
	    cvReleaseMat(&M);
		cvReleaseMat(&B);
        
        
#ifdef M_IMAGE
	if(doleftflow && dodisparity0 && dodisparity)
            cvReleaseMat(&M_image);
#endif
		if (do_ref_flow || do_Zt) {
			cvReleaseMat(&disactivated_pixels_ref_flow);
		}
		
		if(do_ref_flow || do_not_ref_flow || do_Zt){
			cvReleaseMat(&Xt);
			cvReleaseMat(&Yt);
			cvReleaseMat(&Zt);
			cvReleaseMat(&Xt_x);
			cvReleaseMat(&Xt_y);
			cvReleaseMat(&Yt_x);
			cvReleaseMat(&Yt_y);
			if(do_Z0) {
				cvReleaseMat(&Xt_z);
				cvReleaseMat(&Yt_z);
			}
			cvReleaseMat(&U_prec); 
			cvReleaseMat(&V_prec);
			cvReleaseMat(&W_prec);
			cvReleaseMat(&dU);
			cvReleaseMat(&dV);
			cvReleaseMat(&dW);
			cvReleaseMat(&dU_X);
			cvReleaseMat(&dU_Y);
			cvReleaseMat(&dV_X);
			cvReleaseMat(&dV_Y);
			cvReleaseMat(&dW_X);
			cvReleaseMat(&dW_Y);
			cvReleaseMat(&U_X);
			cvReleaseMat(&U_Y);
			cvReleaseMat(&V_X);
			cvReleaseMat(&V_Y);
			cvReleaseMat(&W_X);
			cvReleaseMat(&W_Y);
			cvReleaseMat(&dU_prec);
			cvReleaseMat(&dV_prec);
			cvReleaseMat(&dW_prec);
		}
	 
		if (do_Z0) {
			cvReleaseMat(&Z0_prec);
			cvReleaseMat(&Z0_X);
			cvReleaseMat(&Z0_Y);
			cvReleaseMat(&dZ0);
			cvReleaseMat(&dZ0_X);
			cvReleaseMat(&dZ0_Y);
			cvReleaseMat(&dZ0_prec);
		}
		cvReleaseMat(&SOR);
		cvReleaseMat(&SOR_prec);	

	//if this is not the last level, images must be upscaled before iterations
		if (level != (pyramid_level_end-1)) {
            const CvSize next_level_size = cvGetSize(Images_0_pyramid[0][level - 1]);
            CvMat *Interpolated_Data;
			if(do_ref_flow || do_not_ref_flow || do_Zt){
                Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
                cvResize(U, Interpolated_Data,CV_INTER_LINEAR);
                double uscale = 1 ; //(double)Interpolated_Data->width / U->width;
                cvReleaseMat(&U);
                U = Interpolated_Data;
//                SmoothData(U);
                cvConvertScale(U, U, uscale);
                
                Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
                cvResize(V, Interpolated_Data,CV_INTER_LINEAR);
				// FIXME: check the scale
                double vscale = 1 ; //= (double)Interpolated_Data->height / V->height;
                cvReleaseMat(&V);
                V = Interpolated_Data;
//                SmoothData(V);
                cvConvertScale(V, V, vscale);

				Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
                cvResize(W, Interpolated_Data,CV_INTER_LINEAR);
                double wscale = 1; //= (double)Interpolated_Data->height / W->height;
                cvReleaseMat(&W);
                W = Interpolated_Data;
//                SmoothData(W);
                cvConvertScale(W, W, vscale);
			}
			if(do_Z0) {
				Interpolated_Data =  cvCreateMat(next_level_size.height, next_level_size.width, CV_32FC1);
				cvResize(Z0, Interpolated_Data,CV_INTER_LINEAR);
				double dscale = 1;//(double)Interpolated_Data->width / Z0->width;
				cvReleaseMat(&Z0);
				Z0 = Interpolated_Data;
//				SmoothData(Z0);
				cvConvertScale(Z0, Z0, dscale);
			}
			// tali: the projection matrix for the current pyramyid
			eta_curr = eta_curr/PyramidEta;
			if(ENABLE_ALPHA) { 
				alpha = alpha/PyramidEta;
				alpha_w = alpha_w/PyramidEta;
				alpha_wz = alpha_wz/PyramidEta;

			}

			
			for(view = 0; view < NumOfViews; view++) {
				ComputeProjMat(M_curr[view],R_Mat[view], T_Mat[view], C_Mat[view], eta_curr);
			}
			

#ifdef DEBUG
			sprintf(file, "%s/scaledProjMat%d.txt", TxtResultsDir,level-1);
				f = fopen(file,"w");
				WriteRes(f,M_curr[0]);
				fclose(f);
#endif
		}
#ifdef DEBUG	
		if (do_Z0) {
			sprintf(file, "%s/Z0_l%d.txt", TxtResultsDir,level);
			f = fopen(file,"w");
			WriteRes(f,Z0);
			fclose(f);
		}
		if(do_ref_flow || do_not_ref_flow || do_Zt) { 
			sprintf(file, "%s/W_l%d.txt", TxtResultsDir,level);
			f = fopen(file,"w");
			WriteRes(f,W);
			fclose(f);

			sprintf(file, "%s/U_l%d.txt", TxtResultsDir,level);
			f = fopen(file,"w");
			WriteRes(f,U);
			fclose(f);

			sprintf(file, "%s/V_l%d.txt", TxtResultsDir,level);
			f = fopen(file,"w");
			WriteRes(f,V);
			fclose(f);
		}
#endif

	} // for(int level = pyramid_level_start - 1; level >= pyramid_level_end - 1; level--){
        
	if(do_ref_flow || do_Zt || do_not_ref_flow){
        if (U_end)
            cvCopy(U, U_end);
        if (V_end)
            cvCopy(V, V_end);
		if (W_end)
            cvCopy(W, W_end);
        cvReleaseMat(&U);
        cvReleaseMat(&V);
		cvReleaseMat(&W);
    }
	if(do_Z0){
        if (Z0_end)
            cvCopy(Z0, Z0_end);
        cvReleaseMat(&Z0);
    }
	for(view = 0; view < NumOfViews; view++){
		if(M_curr) {
			cvReleaseMat(&M_curr[view]);
		}
		if(R_Mat){
			cvReleaseMat(&R_Mat[view]);
		}
		if(T_Mat){
			cvReleaseMat(&T_Mat[view]);
		}
		if(C_Mat){
			cvReleaseMat(&C_Mat[view]);
		}
		if(COP && (view > 0)) {
			cvReleaseMat(&COP[view]);
		}
			
	}
}



